

#
# ChangeLayers
#

package require tablelist

set LayersNoRedraw 0

proc SetSelection { tbl } {
    global LayersSelection
    if { ![winfo exists $tbl] } { return }
    set LayersSelection ""
    set n [$tbl size]
    if { $n == 1 } { 
	$tbl selection set 0 
    }
    for { set i 0 } { $i < $n } { incr i } {
	if { [$tbl selection includes $i] } {
	    lappend LayersSelection [lindex [$tbl get $i] 1]
	}
    }
}

proc SetLayerAsSelection { tbl layer } {
    if { ![winfo exists $tbl] } { return }
    $tbl selection clear 0 end
    for { set i 0 } { $i < [$tbl size] } { incr i } {
	if { [lindex [$tbl get $i] 1] == $layer } {
	    $tbl selection set $i
	    $tbl see $i
	    break
	}
    }
    SetSelection $tbl
}

proc SelectLayersByRegexpPressOk { tbl } {
    global wChangeLay LayerSelRegexp LayerSelAdd LayerSelCase
    if { $LayerSelAdd == "substitute" } {
	$tbl sel clear 0 end
    }
    set idx 0
    foreach i [$tbl get 0 end] {
	set i [lindex $i 1]
	set change ""
	if { $LayerSelRegexp != "regex" } {
	    if { $LayerSelCase != "case" } {
		if { [string match [string tolower [$wChangeLay.sel.up.e get]] \
			[string tolower $i]] } { set change yes }
	    } else {
		if { [string match [$wChangeLay.sel.up.e get] $i] } {
		    set change yes
		}
	    }
	} else {
	    if { $LayerSelCase != "case" } {
		if { [regexp -nocase -- [$wChangeLay.sel.up.e get] $i res] } {
		    if { $i == $res } { set change yes }
		}
	    } else {
		if { [regexp -- [$wChangeLay.sel.up.e get] $i res] } {
		    if { $i == $res } { set change yes }
		}
	    }
	}
	if { $change == "yes" } {
	    if { $LayerSelAdd != "substract" } {
		$tbl sel set $idx
	    } else {
		$tbl sel clear $idx
	    }
	}
	incr idx
    }        
    SetSelection $tbl
    destroy $wChangeLay.sel
    focus $wChangeLay.ent.e   
}

proc SelectLayersByRegexp { tbl } {
    global wChangeLay LayerSelRegexp LayerSelAdd LayerSelCase

    if { [winfo exists $wChangeLay.sel] } {
	 destroy $wChangeLay.sel
    }
    frame $wChangeLay.sel -relief ridge -bd 4
    frame $wChangeLay.sel.up
    entry $wChangeLay.sel.up.e -relief sunken
    focus $wChangeLay.sel.up.e
    button $wChangeLay.sel.up.ok -text [_ "OK"] \
	    -command [list SelectLayersByRegexpPressOk $tbl] -padx 5
    pack $wChangeLay.sel.up.e $wChangeLay.sel.up.ok -side left
    
    frame $wChangeLay.sel.down
    checkbutton $wChangeLay.sel.down.reg -text [_ "Regexp"] -onvalue regex \
	    -variable LayerSelRegexp
    checkbutton $wChangeLay.sel.down.case -text [_ "Case"] -onvalue case \
	    -variable LayerSelCase

    frame $wChangeLay.sel.rad
    radiobutton $wChangeLay.sel.rad.substitute -text [_ "Substitute"] -value substitute \
	    -variable LayerSelAdd
    radiobutton $wChangeLay.sel.rad.add -text [_ "Add"] -value add \
	    -variable LayerSelAdd
    radiobutton $wChangeLay.sel.rad.substract -text [_ "Substract"] -value substract \
	    -variable LayerSelAdd

    set LayerSelRegexp ""
    set LayerSelAdd substitute
    set LayerSelCase ""
    pack $wChangeLay.sel.down.reg $wChangeLay.sel.down.case \
	    -side left -ipadx 3
    pack $wChangeLay.sel.rad.substitute $wChangeLay.sel.rad.add \
	    $wChangeLay.sel.rad.substract -side left
    pack $wChangeLay.sel.up $wChangeLay.sel.down $wChangeLay.sel.rad -side top

    place $wChangeLay.sel -x 0 -rely 0.6

    bind $wChangeLay.sel.up.e <Return> [list $wChangeLay.sel.up.ok invoke]
}

proc SendAndFill { layerList col } {
    global LayersNoRedraw

    if { $LayersNoRedraw != 2 } { GiD_Process Layers }
    foreach i $layerList {
	GiD_Process color $i $col
    }

    #FillInfoLayers ;#automatically called
    if { $LayersNoRedraw == 0 } { GiD_Process escape }
    if { $LayersNoRedraw == 1 } { set LayersNoRedraw 2 }
}

proc FillInfoLayers { } {
    global wChangeLay GidPriv
    if { [info exists wChangeLay] } {
	FillTableInfoLayers $wChangeLay.middle.tbl
	if { [winfo exists $wChangeLay.touse.l] } {
	    $wChangeLay.touse.l conf -text [GiD_Info Project LayerToUse]
	}
    }
    if { [info exists GidPriv(ComboLayers,table)] } {
	FillTableInfoLayers $GidPriv(ComboLayers,table)
    }
    if { [info exists GidPriv(ComboLayers,entry)] && [winfo exists $GidPriv(ComboLayers,entry)] } {
	$GidPriv(ComboLayers,entry) delete 0 end
	$GidPriv(ComboLayers,entry) insert end [GiD_Info Project LayerToUse]
    }
}

proc FillTableInfoLayers { tbl } {
    global LayersSelection GidPriv
    if { ![winfo exists $tbl] } { return }
  
    set LayersSelection ""
    set LayerNames [GiD_Info layers]
    set BackLayerNames [GiD_Info back_layers]
    
    if { $BackLayerNames != "" } {
	regsub -all {_\*back\*} $BackLayerNames {} pp
	set BackLayerNames $pp
    }
    
    set sellayers ""
    foreach i [$tbl curselection] {
	lappend sellayers [lindex [$tbl get $i] 1]
    }
    
    $tbl delete 0 end
    
    set LayerToUse [GiD_Info Project LayerToUse]
    
    if { [info exists GidPriv(LayersAlphabetic)] && $GidPriv(LayersAlphabetic) } {
	set LayerNames [lsort -dictionary $LayerNames]
    }
    foreach i $LayerNames {
	set LayersProp [lindex [GiD_Info layers $i] 0]
	$tbl insert end [list "" $i "" "" ""]
	
	set cur_col [ format "#%02x%02x%02x" [lindex $LayersProp 2] \
		[lindex $LayersProp 3] [lindex $LayersProp 4]]
	
	$tbl cellconfigure end,0 -background $cur_col -text {} -selectbackground $cur_col
	if { $i == $LayerToUse } {
	    $tbl cellconfigure end,1 -image [GetImage layer_touse.gif]
	} else {
	    $tbl cellconfigure end,1 -image [GetImage layer_nottouse.gif] ;#void image
	}
	
	if { [lindex $LayersProp 0] } {
	    $tbl cellconfigure end,2 -image [GetImage layer_on.gif]
	} else {
	    $tbl cellconfigure end,2 -image [GetImage layer_off.gif]
	}
	
	if { [lindex $LayersProp 1] } {
	    $tbl cellconfigure end,3 -image [GetImage layer_freeze.gif]
	} else {
	    $tbl cellconfigure end,3 -image [GetImage layer_unfreeze.gif]
	}
	if { [lindex $LayersProp 5] == 255 } {
	    $tbl cellconfigure end,4 -image [GetImage layer_transparent_off.gif]
	} else {
	    $tbl cellconfigure end,4 -image [GetImage layer_transparent_on.gif]
	}

	if { [ lsearch $BackLayerNames $i] != -1 } {
	    $tbl cellconfigure end,5 -image [GetImage layer_bringtofront.gif]
	} else {
	    $tbl cellconfigure end,5 -image [GetImage layer_sendtoback.gif] ;#void image
	}
	
	if { [lsearch -exact $sellayers $i] != -1 } {
	    $tbl selection set end
	}
    }
    
    $tbl activate [lsearch $LayerNames $LayerToUse]

    if { [llength [$tbl curselection]] == 1 } {
	$tbl see [$tbl curselection]
    }
}

set LayersSelection ""

proc SetLayersToDelete { tbl } {
    global LayersSelection wChangeLay LayersNoRedraw 
    SetSelection $tbl
    if { $LayersSelection == "" } {
	tk_dialogRAM  $wChangeLay.modal [_ "error"] \
		[_ "before applying operation delete it is necessary to select layers"] \
		error 0 [_ "OK"]
	return
    }

    if { $LayersNoRedraw != 2 } { GiD_Process Layers }
    set warnmessage 1
    foreach i $LayersSelection {
	GiD_Process delete $i
	
	if { $warnmessage && [catch { GiD_Info layers $i } ] == 0 } {
	    set ret [tk_dialogRAM  $wChangeLay.modal [_ "error"] \
		         [_ "Layer '%s' can't be erased.\n(see the message bar for details)" $i] \
		         error 0 [_ "OK"] [_ "OK to all"]]
	    if { $ret == 1 } { set warnmessage 0 }
	}
    }
    
    if { $LayersNoRedraw == 0 } { GiD_Process escape }
    if { $LayersNoRedraw == 1 } { set LayersNoRedraw 2 }

    #FillInfoLayers ;#automatically called
}

proc SetLayersToColor { tbl } {
    global LayersSelection wChangeLay LayersNoRedraw 
    set parent [winfo parent $tbl]
    SetSelection $tbl
    if { $LayersSelection == "" } {
	tk_dialogRAM  $parent.modal [_ "error"] \
	[_ "before applying operation color it is necessary to select layers"] \
	error 0 [_ "OK"]
	return
    }
    set sel $LayersSelection ;#trick because GiD_Process Layers can change LayersSelection

    if { $LayersNoRedraw != 2 } { 
	GiD_Process Layers 
	set LayersSelection $sel
    }

    set LayersProp [lindex [GiD_Info layers [lindex $LayersSelection 0]] 0]

    set cur_col [ format "#%02x%02x%02x%02x" [lindex $LayersProp 2] \
	    [lindex $LayersProp 3] [lindex $LayersProp 4] [lindex $LayersProp 5]]

    #set retcol [tk_chooseColor -parent $parent -title [_ "Select color"] -initialcolor $cur_col]
    set retcol [GIDChooseColor $parent.selectcolor -title [_ "Select color"] -color $cur_col]

    if { $retcol != "" } {
	set retcol [string range $retcol 1 end]
	if { [string length $retcol] == 8 } {
	    scan $retcol %2x%2x%2x%2x col1 col2 col3 alpha
	    set col [format %03i%03i%03i $col1 $col2 $col3]
	    foreach i $LayersSelection {
		GiD_Process color $i $col transparent $i $alpha
	    }
	} else {
	    scan $retcol %2x%2x%2x col1 col2 col3
	    set col [format %03i%03i%03i $col1 $col2 $col3]
	    foreach i $LayersSelection {
		GiD_Process color $i $col
	    }
	}
	#    RamColor .ramcolor "SendAndFill [list $LayersSelection]"
    }
    if { $LayersNoRedraw == 0 } { GiD_Process escape }
    if { $LayersNoRedraw == 1 } { set LayersNoRedraw 2 }

    #FillInfoLayers ;#automatically called

}

proc SetLayersToRename { tbl } {
    global LayersSelection wChangeLay LayersNoRedraw 
    SetSelection $tbl
    if { [llength $LayersSelection] !=1 } {
	tk_dialogRAM $wChangeLay.m [_ "error"] \
		[_ "Before applying rename it is necessary to select 1 layer"] \
		error 0 [_ "OK"]
	return
    }
    if { [winfo exists $wChangeLay.ent.e] } {
	set name [string trim [$wChangeLay.ent.e get]]
    } else {
	set name ""
    }
    if { $name == "" } {
	#tk_dialogRAM $wChangeLay.m [_ "error"] \
		#[_ "To rename one layer select it and write a new name before pressing the button"] \
		#error 0 [_ "OK"]
	#return
	set name [tk_dialogEntryRAM .gid.tempwinw [_ "Layers"] \
		[_ "Enter new layer name"]  gidquestionhead any "" ]
	if { $name == "--CANCEL--" ||  $name == "" } { return  }
    }
 
    foreach i [GiD_Info layers] {
	if { $i == $name } {
	    tk_dialogRAM $wChangeLay.m [_ "error"] [_ "Layer '%s' already exists" $i] error 0 [_ "OK"]
	    return
	}
    }

    if { [winfo exists $wChangeLay.ent.e] } {
	$wChangeLay.ent.e delete 0 end
    }
    SetLayersTo RENAME $tbl $name
}

proc SetLayersTo { type tbl {newname ""} } {
    global LayersSelection wChangeLay LayersNoRedraw 
    
    SetSelection $tbl
    if { $LayersSelection == "" } {
	tk_dialogRAM  $wChangeLay.modal [_ "error"] \
		[_ "before applying operation %s it is necessary to select layers" $type] \
		error 0 [_ "OK"]
	return
    }

    set VarToSend GiD_Process
    if { $LayersNoRedraw != 2 } { 
	lappend VarToSend Layers 
    }
    foreach i $LayersSelection {
	switch $type {
	    ON {
		lappend VarToSend On $i
	    }
	    OFF {
		lappend VarToSend Off $i
	    }
	    FREEZE {
		lappend VarToSend Freeze $i
	    }
	    UNFREEZE {
		lappend VarToSend UnFreeze $i
	    }
	    TRANSPARENT {
		lappend VarToSend Transparent $i 127
	    }
	    OPAQUE {
		lappend VarToSend Opaque $i
	    }
	    TOUSE {
		lappend VarToSend ToUse $i
	    }
	    RENAME {
		lappend VarToSend ChangeName $i $newname
	    }
	    BRINGTOFRONT {
		lappend VarToSend BringToFront $i
	    }
	}
    }    
    if { $LayersNoRedraw == 0 } { 
	lappend VarToSend escape 
    }
    
    if { $type == "RENAME" } {
	#Trick to have selected the new name after FillInfoLayers will be automatically called
	set row [$tbl curselection]
	$tbl cellconfigure $row,1 -text $newname
    }

    eval $VarToSend
    if { $LayersNoRedraw == 1 } { set LayersNoRedraw 2 }

    #FillInfoLayers ;#automatically called
}

proc SendLayerEntitiesAA { what } {
    global LayersSelection wChangeLay GidPriv

    if { [llength $LayersSelection] !=1 } {
	tk_dialogRAM  $wChangeLay.modal [_ "error"] \
       [_ "before applying 'Entities' it is necessary to select 1 layer"] error \
		0 [_ "OK"]
	return
    }
    set layer [lindex $LayersSelection 0]

    if { $GidPriv(LayersAlsoLower) == 0 } {
	GiD_Process Layers Entities $layer $what
    } else {
	GiD_Process Layers Entities $layer LowerEntities $what
    }

    set SmallWinSelecting [GiD_Info variables SmallWinSelecting]
    FinishButton $wChangeLay $wChangeLay.buts \
	    [_ "Press 'Finish' to end selection"] \
	    "" disableall $SmallWinSelecting
}


proc SendLayerEntities { tbl } {                 
    global wChangeLay
    SetSelection $tbl

    AddEntitiesToMenuAfter 0 $wChangeLay.buts.act.m "SendLayerEntitiesAA " \
	    [list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] [_ "Dimensions"] [_ "All"]] \
	    {Points Lines Surfaces Volumes Dimensions All} \
	    [list [_ "Nodes"] [_ "Elements"] [_ "Dimensions"] [_ "All"]] \
	    {Nodes Elements Dimensions All}

}

proc EndLayerToUseEntitiesAA { what } {
    global GidPriv
    foreach i [trace vinfo GidPriv(selection)] {
	trace vdelete GidPriv(selection) [lindex $i 0] [lindex $i 1] 
    }

    set num [lindex [lindex $GidPriv(selection) 0] 1]
    set nnum ""
    regexp {[0-9]*} $num nnum
    if { $nnum != "" } {
	set retval [GiD_Info list_entities $what $nnum]
	set layername ""
	regexp {LAYER:([^\n\r]+)} $retval {} layername
	set layername [string trim $layername]
	if { $layername != "" } {
	    GiD_Process Layers ToUse $layername
	    FillInfoLayers

	    global wChangeLay
	    $wChangeLay.middle.tbl selection clear 0 end
	    for { set i 0 } { $i < [$wChangeLay.middle.tbl size] } { incr i } {
		if { [lindex [$wChangeLay.middle.tbl get $i] 1] == $layername } {
		    $wChangeLay.middle.tbl selection set $i
		    $wChangeLay.middle.tbl see $i
		    break
		}
	    }
	} else {
	    WarnWin [_ "Selected entity does not belong to one layer"]
	}
    }
}

proc LayerToUseEntitiesAA { what } {
    global GidPriv
    
    if { [trace vinfo GidPriv(selection)] != "" } {
	foreach i [trace vinfo GidPriv(selection)] {
	    trace vdelete GidPriv(selection) [lindex $i 0] [lindex $i 1] 
	}
	GiD_Process escape
    }

    GiD_Process SelectEntities JustOne $what
    trace var GidPriv(selection) w "EndLayerToUseEntitiesAA $what ;#"
}

proc LayerToUseEntities { menu } {
    AddEntitiesToMenu $menu "LayerToUseEntitiesAA " \
	    [list [_ "Point"] [_ "Line"] [_ "Surface"] [_ "Volume"] ] {Points Lines Surfaces Volumes} \
	    [list [_ "Node"] [_ "Element"] ] {Nodes Elements}
}

proc SendLayerEntitiesToBackAA { what } {
    global wChangeLay GidPriv

    if { $GidPriv(LayersBackOpposite) } {
	if { $GidPriv(LayersAlsoLower) == 0 } {
	    GiD_Process Layers SendToBackOpposite $what
	} else {
	    GiD_Process Layers SendToBackOpposite LowerEntities $what
	}
    } else {
	if { $GidPriv(LayersAlsoLower) == 0 } {
	    GiD_Process Layers SendToBack $what
	} else {
	    GiD_Process Layers SendToBack LowerEntities $what
	}
    }
    set SmallWinSelecting [GiD_Info variables SmallWinSelecting]
    FinishButton $wChangeLay $wChangeLay.dr.bdel \
	    [_ "Press 'Finish' to end selection"] \
	    "FillInfoLayers" disableall $SmallWinSelecting
    
}


proc SendLayerEntitiesToBack { tbl } {                 
    global  wChangeLay
    SetSelection $tbl

    AddEntitiesToMenuAfter 1 $wChangeLay.buts.back.m "SendLayerEntitiesToBackAA" \
	    [list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] [_ "All"]] \
	    {Points Lines Surfaces Volumes All} \
	    [list [_ "Nodes"] [_ "Elements"] [_ "All"]] \
	    {Nodes Elements All}

    $wChangeLay.buts.back.m add separator
    if { [GiD_Info layers -hasbacklayers] } {
	$wChangeLay.buts.back.m add command -label [_ "Bring to front"] -command \
		"SetLayersTo BRINGTOFRONT $tbl"
	$wChangeLay.buts.back.m add command -label [_ "Bring all to front"] -command \
		"GiD_Process Layers BringToFrontAll escape ; FillInfoLayers"
    } else {
	$wChangeLay.buts.back.m add command -label [_ "Bring to front"] -state disabled
	$wChangeLay.buts.back.m add command -label [_ "Bring all to front"] -state disabled
    }

}

proc GetAutomaticLayerName { } {
    set LayerNames [GiD_Info layers]
    set name ""
    for {set i 0} {$i<10000} {incr i} {
	set name Layer$i
	if { [lsearch -exact $LayerNames $name] == -1 } { break } 
    } 
    return $name
}


proc CreateNewLayer { {name ""} } {
    if { $name == "" } {
	set name [GetAutomaticLayerName]
    }
    GiD_Process Layers New $name escape
    return $name
}


proc NewLayerFromEntry { w } {
    global wChangeLay

    set name [string trim [$w.ent.e get]]
    #regsub -all {[ ]+} $name {_} name
   
    if { $name != "" } {
	set LayerNames [GiD_Info layers]
	set ipos [lsearch -exact $LayerNames $name]
	if { $ipos != -1 } {
	    tk_dialogRAM  $w.modal [_ "error"] [_ "Layer %s already exists" $name] error 0 [_ "OK"]
	    return
	}
    }
    set name [CreateNewLayer $name]
 
    #FillInfoLayers ;#automatically called
    $w.ent.e delete 0 end

    # seleccionamos la que acabamos de crear
    SetLayerAsSelection $wChangeLay.middle.tbl $name
}

proc ConfigureListScrollbars { listbox sx sy } {
    foreach i "x y" {
	if { ![info exists s${i}] || ![winfo exists [set s${i}]] } { continue }
	foreach "${i}1 ${i}2" [$listbox ${i}view] break
	if { [set ${i}1] == 0 && [set ${i}2] == 1 } {
	    after idle grid remove [set s${i}]
	} else {
	    after idle grid [set s${i}]
	}
    }
}   

proc MotionTableList { x y w tbl } {
    set tablelist::x [expr {$x + [winfo x $w]}]
    set tablelist::y [expr {$y + [winfo y $w]}]
    set cell [$tbl nearestcell $tablelist::x $tablelist::y]
    foreach {row col} [ split $cell ,] break
    if { $col != 1 } {
	return -code break
    }
}

proc ClickTableList { x y w tbl } { 
    focus $w
    set cell [$tbl containingcell $x $y] 
    foreach {row col} [ split $cell ,] break
    if { $row < 0 || $col < 0 } return
    #if click a unselected row then select only this row (but not if column == 1)
    if { $col != 1 } {
	if { ![$tbl selection includes $row] } {
	    $tbl selection clear 0 end
	    $tbl selection set $row
	}
    }

    if { $col == 0 } {
	SetLayersToColor $tbl
    } elseif { $col == 1 } {
	#SetLayersTo TOUSE $tbl
    } elseif { $col == 2 } {
	if { [lindex [$tbl cellconfigure $row,$col -image] 4] == [GetImage layer_off.gif] } {
	    SetLayersTo ON $tbl
	    #$tbl cellconfigure $row,$col -image [GetImage layer_on.gif]
	} else {
	    SetLayersTo OFF $tbl
	    #$tbl cellconfigure $row,$col -image [GetImage layer_off.gif]
	}
    } elseif { $col == 3 } {
	if { [lindex [$tbl cellconfigure $row,$col -image] 4] == [GetImage layer_freeze.gif] } {
	    SetLayersTo UNFREEZE $tbl
	    #$tbl cellconfigure $row,$col -image [GetImage layer_unfreeze.gif]
	} else {
	    SetLayersTo FREEZE $tbl
	    #$tbl cellconfigure $row,$col -image [GetImage layer_freeze.gif]
	}
    } elseif { $col == 4 } {
	if { [lindex [$tbl cellconfigure $row,$col -image] 4] == [GetImage layer_transparent_off.gif] } {
	    SetLayersTo TRANSPARENT $tbl
	    #$tbl cellconfigure $row,$col -image layer_transparent_on.gif
	} else {
	    SetLayersTo OPAQUE $tbl
	    #$tbl cellconfigure $row,$col -image layer_transparent_off.gif
	}
    } elseif { $col == 5 } {
	if { [lindex [$tbl cellconfigure $row,$col -image] 4] == [GetImage layer_bringtofront.gif] } {
	   SetLayersTo BRINGTOFRONT $tbl
	    #$tbl cellconfigure $row,$col -image layer_bringtofront.gif
	} else {
	    #$tbl cellconfigure $row,$col -image layer_sendtoback.gif
	}
    }


    if { $col != 1 } {
	return -code break
    }
}

proc InvertSelectionTableList { tbl } { 
    set newselection {}
    for { set i 0 } { $i < [$tbl size] } { incr i } {
	if { ![$tbl selection includes $i] } {
	    lappend newselection $i
	}
    }
    $tbl selection clear 0 end
    foreach i $newselection {
	$tbl selection set $i
    }
    #SetSelection $tbl
}


proc DoubleClickTableList { x y w tbl } { 
    set cell [$tbl containingcell $x $y] 
    foreach {row col} [ split $cell ,] break
    if { $row < 0 || $col < 0 } return
    $tbl selection set $row
    if { $col == 1 } {
	SetLayersTo TOUSE $tbl
    }
}

proc LayersCreateListWin { w} {
    global GidPriv

    set vsb $w.vsb1
    set hsb $w.hsb1
    set tbl [tablelist::tablelist $w.tbl \
	    -exportselection 0 \
	    -columns [list \
	    0 [_ "C#C#Color"] left \
	    5 [_ "Name"] left \
	    0 [_ "I/O#C#On/Off"] left \
	    0 [_ "F/U#C#Freeze/Unfreeze"] left \
	    0 [_ "Tr#C#transparent"] left \
	    0 [_ "B#C#back"] left \
	    ] \
	    -selectbackground black -selectforeground white \
	    -background $GidPriv(Color,BackgroundListbox) \
	    -stretch 1 \
	    -selectmode extended \
	    -highlightthickness 0 \
	    -labelborderwidth 1 \
	    -labelcommand tablelist::sortByColumn \
	    -labelpady 0 \
	    -showarrow 0 \
	    -showseparators 0 \
	    -xscrollcommand [list $hsb set] \
	    -yscrollcommand [list $vsb set] \
	    -stripebackground [CCColorSombra $GidPriv(Color,BackgroundListbox)] \
	    -resizablecolumns 0]
	    #-labelcommand tablelist::sortByColumn -showarrow 1

    scrollbar $vsb -orient vertical   -command [list $tbl yview]
    scrollbar $hsb -orient horizontal -command [list $tbl xview]

    # bind [$tbl bodypath] <Button-1> [list ClickTableList %x %y %W $tbl]
    # correcion for tablelist 4.1, else the binding of the image inside the label is lost 
    bind [$tbl bodytag] <Button-1> "\
	    foreach {tablelist::W tablelist::x tablelist::y} \
	    \[tablelist::convEventFields %W %x %y] {}
    ClickTableList \$tablelist::x \$tablelist::y %W $tbl"
    
    #bind [$tbl bodypath] <Double-Button-1> [list DoubleClickTableList %x %y %W $tbl]
    bind [$tbl bodytag] <Double-Button-1> "\
	    foreach {tablelist::W tablelist::x tablelist::y} \
	    \[tablelist::convEventFields %W %x %y] {}
    DoubleClickTableList \$tablelist::x \$tablelist::y %W $tbl"

    bind [$tbl bodytag] <Return> [list SetLayersTo TOUSE $tbl]
    bind [$tbl bodytag] <Key-Delete> [list SetLayersToDelete $tbl]
    

    bind [$tbl bodypath] <Motion> [list MotionTableList %x %y %W $tbl]

    bind [$tbl bodypath] <Alt_L> [list InvertSelectionTableList $tbl]
    bind [$tbl bodypath] <Alt_R> [list InvertSelectionTableList $tbl]
    bind [$tbl bodypath] <Meta_L> [list InvertSelectionTableList $tbl]
    bind [$tbl bodypath] <Meta_R> [list InvertSelectionTableList $tbl]

    bind [$tbl bodypath] <F2> [list SetLayersToRename $tbl]

    grid $tbl $vsb -sticky nsew
    grid configure $vsb -sticky ns
    grid $hsb -sticky ew
    grid remove $vsb $hsb
    bind $tbl <Configure> [list ConfigureListScrollbars $tbl $hsb $vsb]

    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1
    return $tbl
}

proc ChangeLayers { { w .gid.wlay } } {
    global wChangeLay LayersNoRedraw GidPriv 
    set wChangeLay $w
    set LayersNoRedraw 0
    InitWindow $w [_ "Layers"] LayersWindowGeom ChangeLayers
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0

    frame $w.middle -relief ridge -bd 2

    set tbl [LayersCreateListWin $w.middle]
    GidHelp $w.middle \
	[_ "Layer properties:\n Color, Name, Visible/Hidden, Frozen/Unfrozen, Exists back entities."]

    pack $w.middle -expand yes -fill both

    frame $w.options -bd 1 -relief sunken
    button $w.options.col -text [_ "Color"]... -command [list SetLayersToColor $tbl] -bd 1 -padx 5
    GidHelp $w.options.col \
	[_ "Changes the color used to draw the entities in render mode."]
    button $w.options.on -text [_ "On#C#verb"] -command [list SetLayersTo ON $tbl] -bd 1 -padx 2 -underline 0
    GidHelp $w.options.on \
	[_ "Entities belonging to a layer ON will be drawn."]
    button $w.options.off -text [_ "Off#C#verb"] -command [list SetLayersTo OFF $tbl] -bd 1 -padx 2 -underline 1
    GidHelp $w.options.off \
	[_ "Entities belonging to a layer OFF will not be drawn."]
    grid $w.options.on -sticky e -row 0 -column 0 -padx 2 -pady 1
    grid $w.options.off -sticky w -row 0 -column 1 -padx 2 -pady 1
    grid $w.options.col -sticky e -row 0 -column 2 -padx 2 -pady 1
    grid rowconfigure $w.options 0 -weight 1
    grid columnconfigure $w.options 0 -weight 1
    grid columnconfigure $w.options 1 -weight 1
    grid columnconfigure $w.options 2 -weight 1

    frame $w.options2 -bd 1 -relief sunken

    set def_back [ $w.options2 cget -background]
    BotonPlano $w.options2.fr "" [list SetLayersTo FREEZE $tbl] $def_back [ CCColorActivo $def_back]
    $w.options2.fr conf -image [GetImage Lock-on.gif] -takefocus 0 -padx 4 -pady 2 -width 26 -height 26
    GidHelp $w.options2.fr \
	[_ "Entities belonging to a freeze layer will be drawn but cannot be selected or modified."]
    BotonPlano $w.options2.ufr  "" [list SetLayersTo UNFREEZE $tbl] $def_back [ CCColorActivo $def_back]
    $w.options2.ufr conf -image [GetImage Lock-off.gif] -takefocus 0 -padx 4 -pady 2 -width 26 -height 26
    GidHelp "$w.options2.ufr " \
	[_ "Entities belonging to a unfreeze layer will be selected and modified."]

    BotonPlano $w.options2.opaque "" [list SetLayersTo OPAQUE $tbl] $def_back [ CCColorActivo $def_back]
    $w.options2.opaque conf -image [GetImage opaco.gif] -takefocus 0 -padx 4 -pady 2 -width 26 -height 26
    GidHelp $w.options2.opaque [_ "Turns the selected layers opaque."]
    BotonPlano $w.options2.transparent "" [list SetLayersTo TRANSPARENT $tbl] $def_back [ CCColorActivo $def_back]
    $w.options2.transparent conf -image [GetImage transparente.gif] -takefocus 0 -padx 4 -pady 2 -width 26 -height 26
    GidHelp $w.options2.opaque [_ "Turns the selected layers transparent."]

    button $w.options2.sel -text [_ "Sel"]... -command [list SelectLayersByRegexp $tbl] -bd 1 -padx 5 -width 6
    GidHelp $w.options2.sel \
	[_ "Select layers with filters (for example: 'b*' select all the layers with name beginning by 'b')"]
    grid $w.options2.fr $w.options2.ufr $w.options2.opaque $w.options2.transparent $w.options2.sel \
	    -sticky w -padx 2 -pady 1
    grid configure $w.options2.sel -sticky e
    grid rowconfigure $w.options2 0 -weight 1
    grid columnconfigure $w.options2 {4} -weight 1

    bind $w <Alt-o> "$w.options.on invoke; break"
    bind $w <Alt-f> "$w.options.off invoke; break"
    bind $w <Alt-e> "$w.options2.fr invoke; break"
    bind $w <Alt-u> "$w.options2.ufr invoke; break"

    bind $w <Control-o> "$w.options.on invoke; break"
    bind $w <Control-f> "$w.options.off invoke; break"
    bind $w <Control-e> "$w.options2.fr invoke; break"
    bind $w <Control-u> "$w.options2.ufr invoke; break"

    frame $w.touse -bd 0
    button $w.touse.b -text [_ "Layer To use"]: -command [list SetLayersTo TOUSE $tbl] -bd 1 -padx 5 
    GidHelp $w.touse.b \
	[_ "Set the current layer (the new entities are created in this layer)."]

    menubutton $w.touse.m -image [GetImage arrow_down_vsmall.gif] -menu $w.touse.m.m \
	    -bd 1 -direction flush
    GidHelp $w.touse.m [_ "This option permmits to select one entity and retrieve its layer."]
    menu $w.touse.m.m -postcommand "LayerToUseEntities $w.touse.m.m"
    bind $w.touse.m <1> {
	if {$tkPriv(inMenubutton) != ""} {
	    tkMbPost_ram $tkPriv(inMenubutton) $wChangeLay.touse.b 
	}
	break
    }

    label $w.touse.l -foreground blue

    set LayerToUse [GiD_Info Project LayerToUse]
    $w.touse.l conf -text $LayerToUse
    grid $w.touse.b -row 0 -column 0 -sticky e -padx 0 -pady 1
    grid $w.touse.m -row 0 -column 1 -sticky wns -padx 0 -pady 1
    grid $w.touse.l -row 0 -column 2 -sticky w -padx 3 -pady 1

    grid rowconfigure $w.touse 0 -weight 1
    grid columnconfigure $w.touse 2 -weight 1

    frame $w.ent -bd 0
    entry $w.ent.e -relief sunken -bd 1
    button $w.ent.b -text [_ "New#C#layer"] -command [list NewLayerFromEntry $w] -bd 1 -padx 5 -width 6
    GidHelp $w.ent.b [_ "Create a new layer."]

    SetWidgetsWidthFromText $w.options.col $w.options.on $w.options.off $w.options2.sel $w.touse.b $w.ent.b

    bind $w.ent.e <Return> { 
	$wChangeLay.ent.b config -relief sunken
	update idletasks
	$wChangeLay.ent.b invoke
	$wChangeLay.ent.b config -relief raised
    }
 
    grid $w.ent.e $w.ent.b -sticky ew -padx 2 -pady 1
    grid rowconfigure $w.ent 0 -weight 1
    grid columnconfigure $w.ent 0 -weight 1


    frame $w.dr -bd 0
    button $w.dr.bdel -text [_ "Delete"] -command [list SetLayersToDelete $tbl] -bd 1 -padx 5    
    GidHelp $w.dr.bdel [_ "Delete a layer."]
    button $w.dr.bren -text [_ "Rename"] -bd 1 -padx 5 -command [list SetLayersToRename $tbl] 
    GidHelp $w.dr.bren [_ "Rename a layer."]

    SetWidgetsWidthFromText $w.dr.bdel $w.dr.bren

    checkbutton $w.dr.bred -text [_ "alphabetic"] -variable GidPriv(LayersAlphabetic) \
	    -command FillInfoLayers
    GidHelp $w.dr.bred [_ "Sort the list of layers by name."]
    
    grid $w.dr.bdel $w.dr.bren $w.dr.bred -sticky we -padx 1 -pady 1
    
    #frame $w.buts -bd 0
    set def_back [$w cget -background]
    frame $w.buts -bg [CCColorActivo $def_back]

    if { ![info exists GidPriv(LayersAlsoLower)] } {
	set GidPriv(LayersAlsoLower) 0
    }
    if { ![info exists GidPriv(LayersBackOpposite)] } {
	set GidPriv(LayersBackOpposite) 0
    }

    menubutton $w.buts.back -text [_ "To back"] \
	    -image [GetImage arrow_down_vsmall.gif] -compound right \
	    -menu $w.buts.back.m 
   
    GidHelp $w.buts.back \
	[_ "Send entities to back/front of a layer (entities in back are not drawn and frozen)"]
    menu $w.buts.back.m -postcommand "SendLayerEntitiesToBack $tbl"
    $w.buts.back.m add check -label [_ "Also lower entities"] \
	    -var GidPriv(LayersAlsoLower)
    $w.buts.back.m add check -label [_ "Opposite"] \
	    -var GidPriv(LayersBackOpposite)

    $w.buts.back conf -bd 1 -highlightthickness 1
    $w.buts.back.m conf -bd 1 -activeborderwidth 1

    menubutton $w.buts.act -text [_ "Send To"] \
	    -image [GetImage arrow_down_vsmall.gif] -compound right \
	    -menu $w.buts.act.m
    GidHelp $w.buts.act \
	[_ "Send the selected entities to a layer.\
	   If 'Also lower entities' is checked, the subentities are also send to the layer"]
    menu $w.buts.act.m -postcommand "SendLayerEntities $tbl"
    $w.buts.act.m add check -label [_ "Also lower entities"] \
	    -var GidPriv(LayersAlsoLower)
    
    $w.buts.act conf -bd 1 -highlightthickness 1
    $w.buts.act.m conf -bd 1 -activeborderwidth 1
    
    button $w.buts.can -text [_ "Close"] -command [list destroy $w]

    SetWidgetsWidthFromText $w.buts.back $w.buts.act $w.buts.can

    grid $w.buts.back $w.buts.act $w.buts.can -sticky ew -padx 5 -pady 6

    grid $w.middle -sticky wens -pady 0
    grid $w.touse -sticky we -pady 0
    grid $w.options -sticky we -pady 0
    grid $w.options2 -sticky we -pady 0
    grid $w.ent -sticky we -pady 0
    grid $w.dr -sticky ew -pady 0
    grid $w.buts -sticky ews
    if { $::tcl_version >= 8.5 } { grid anchor $w.buts center }
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    FillInfoLayers

    bind $w <FocusIn> "focus $w.ent.e ; bind $w <FocusIn> {}"
    bind $w <Alt-c> "tkButtonInvoke $w.buts.can"
    bind $w <Escape> "tkButtonInvoke $w.buts.can"
}

proc tkMbPost_ram {w wbut {x {}} {y {}}} {
    global tkPriv errorInfo
    global tcl_platform

    if {([$w cget -state] == "disabled") || ($w == $tkPriv(postedMb))} {
	return
    }
    set menu [$w cget -menu]
    if {$menu == ""} {
	return
    }
    set tearoff [expr {($tcl_platform(platform) == "unix") \
		     || ([$menu cget -type] == "tearoff")}]
    if {[string first $w $menu] != 0} {
	error "can't post $menu:  it isn't a descendant of $w (this is a new requirement in Tk versions 3.0 and later)"
    }
    set cur $tkPriv(postedMb)
    if {$cur != ""} {
	tkMenuUnpost {}
    }
    set tkPriv(cursor) [$w cget -cursor]
    set tkPriv(relief) [$w cget -relief]
    $w configure -cursor arrow

    # the one by default:
    set new_relief raised
    if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
	if { [$w cget -relief] eq "round_raised"} {
	    set new_relief round_sunken
	}
    }
    $w configure -relief $new_relief

    set tkPriv(postedMb) $w
    set tkPriv(focus) [focus]
    $menu activate none
    tkGenerateMenuSelect $menu

    # If this looks like an option menubutton then post the menu so
    # that the current entry is on top of the mouse.  Otherwise post
    # the menu just below the menubutton, as for a pull-down.

    update idletasks
    if {[catch {
	     switch [$w cget -direction] {
		above {
		        set x [winfo rootx $wbut]
		        set y [expr {[winfo rooty $wbut] - [winfo reqheight $menu]}]
		        $menu post $x $y
		}
		below {
		        set x [winfo rootx $wbut]
		        set y [expr {[winfo rooty $wbut] + [winfo height $wbut]}]
		        $menu post $x $y
		}
		left {
		        set x [expr {[winfo rootx $wbut] - [winfo reqwidth $menu]}]
		        set y [expr {(2 * [winfo rooty $wbut] + [winfo height $wbut]) / 2}]
		        set entry [tkMenuFindName $menu [$w cget -text]]
		        if {[$w cget -indicatoron]} {
		    if {$entry == [$menu index last]} {
		            incr y [expr {-([$menu yposition $entry] \
		                    + [winfo reqheight $menu])/2}]
		    } else {
		            incr y [expr {-([$menu yposition $entry] \
		                + [$menu yposition [expr {$entry+1}]])/2}]
		    }
		        }
		        $menu post $x $y
		        if {($entry != {}) && ([$menu entrycget $entry -state] != "disabled")} {
		            $menu activate $entry
		    tkGenerateMenuSelect $menu
		        }
		}
		right {
		        set x [expr {[winfo rootx $wbut] + [winfo width $wbut]}]
		        set y [expr {(2 * [winfo rooty $wbut] + [winfo height $wbut]) / 2}]
		        set entry [tkMenuFindName $menu [$w cget -text]]
		        if {[$w cget -indicatoron]} {
		    if {$entry == [$menu index last]} {
		            incr y [expr {-([$menu yposition $entry] \
		                    + [winfo reqheight $menu])/2}]
		    } else {
		            incr y [expr {-([$menu yposition $entry] \
		                + [$menu yposition [expr {$entry+1}]])/2}]
		    }
		        }
		        $menu post $x $y
		        if {($entry != {}) && ([$menu entrycget $entry -state] != "disabled")} {
		            $menu activate $entry
		    tkGenerateMenuSelect $menu
		        }
		}
		default {
		        if {[$w cget -indicatoron]} {
		        if {$y == ""} {
		        set x [expr {[winfo rootx $wbut] + [winfo width $wbut]/2}]
		        set y [expr {[winfo rooty $wbut] + [winfo height $wbut]/2}]
		        }
		    tkPostOverPoint $menu $x $y [tkMenuFindName $menu [$w cget -text]]
		} else {
		        $menu post [winfo rootx $wbut] [expr {[winfo rooty $wbut]+[winfo height $wbut]}]
		        }  
		}
	     }
     } msg]} {
	# Error posting menu (e.g. bogus -postcommand). Unpost it and
	# reflect the error.
	
	set savedInfo $errorInfo
	tkMenuUnpost {}
	error $msg $savedInfo

    }

    set tkPriv(tearoff) $tearoff
    if {$tearoff != 0} {
	    focus $menu
	    tkSaveGrabInfo $w
	    grab -global $w
    }
}

