# -*- mode: Tcl; tcl-indent-level :  4 -*-

set ::MSGCAT_GID 1

set ::PYRAMID_GID 1

set _DebugDefaultValues 0

if { $::MSGCAT_GID } {
    # forzamos el autoload de msgcat
    msgcat::mc ""

    # ahora sobreescribimos el procedimiento mcunknown
    proc msgcat::mcunknown {locale src args} {
	# para encontrar en la fase de desarrollo cadenas por traducir
	# Retornamos el string con la primera letra en minuscula y el resto en mayusculas
	# Evitamos modificar los argumentos al printf para evitar errores
	if {[llength $args]} {
	    set src [eval [ list format $src] $args]
	}
	return $src
#         set ret [string tolower [string index $src 0]
#         set prevchar ""
#         for {set i 1} {$i < [string length $src]} {incr i} {
#             set char [string index $src $i]
#             if { $prevchar == "%" } {
#                 # to avoid errors as %d format translated as %D
#                 append ret $char
#             } else {
#                 append ret [string toupper $char]
#             }
#             set prevchar $char
#         }
#         return $ret
    }
}

proc GidTkBackCompatibility {} {
    #tcl/tk 8.4 hide private tk[a-Z]* procedures 
    set comms { tkSaveGrabInfo tkButtonInvoke tkTextSelectTo tkEntryInsert tkEntryBackspace \
	    tk_textCut tk_textCopy tk_textPaste tk_focusNext tk_focusPrev tkTextClosestGap \
	    tkTextAutoScan tkCancelRepeat tkTextInsert tkTabToWindow \
	    tkGenerateMenuSelect tk_popup tk_menuSetFocus tkMbPost }
    set hay_expose [ info command tk::unsupported::ExposePrivateCommand]
    foreach i $comms {
	auto_load $i
	if { ![llength [info commands $i]] && ( "$hay_expose" != "")} {
	    tk::unsupported::ExposePrivateCommand $i
	}
    }
    set vars { tkPriv }
    foreach i $vars {
	::tk::unsupported::ExposePrivateVariable $i
    }
}

GidTkBackCompatibility

#avoid wrong menu images with Windows animation effects
if { $::tcl_platform(platform) == "windows" } {
    proc GiD_PopupMenu {menu x y} {
	# temporarily disable menu animations because they trigger a bug in
	# Tk causing menu images to be incorrectly displayed
	set enabled [GiD_SetMenuAnimation]
	if {$enabled} {
	    GiD_SetMenuAnimation 0
	}
	tk_popup $menu $x $y
	if {$enabled} {
	    GiD_SetMenuAnimation $enabled
	}
    }
} else {
    proc GiD_PopupMenu {menu x y} {
	tk_popup $menu $x $y
    }
}

proc .central.s { args } {
    return [eval .gid.central.s $args]
}

proc GiD_Process { args } {
    return [eval .gid.central.s processlist $args]
}

proc GiD_Info { args } {
    return [eval .gid.central.s info $args]
}

proc GiveGidDefaultsFile { } {
    global tcl_platform env
    if { $tcl_platform(platform) == "windows" } {
	#return [file join $env(windir) gid.ini]
	return [ gidgetdefaultsfile]
    } else {
	return [file join $env(HOME) .gidDefaults]
    }
}

global GidPriv

if { ![ info exist GidPriv(VerifImage)]} {
    #set GidPriv(VerifImage) [ image create photo -file $GIDDEFAULT/bitmaps/verif.gif]
    set GidPriv(VerifImage) [ image create photo -data {
	R0lGODlhEAAQAKUAANqwAKWFAO/BAP/OAPLCAOO3AOK1APzPAM6lAP7QAOW4AN+zAP//AP/2
	ALyXAPnJAPvLAFxKAMihAP/fAP/4AKuKAO/AAP/RANasAH1kANetAP/eAP/3AJZ4AP/QAP3L
	AHtjANesAP/hANitANWrAIVqANmuAP/gAPfGAPPDAIBnAP/PAP7OANGnAIFoANarAP7ZAPLD
	AHJcAK6MAJx9AP///////////////////////////////////////////yH5BAEKAD8ALAAA
	AAAQABAAAAZYwJ9wSCwaj8jkD6AkBgTN4SAqJBSohgMVkVAQFz9Gg+h4QIgRyYTBoAgrlgum
	mNFs2JxOwfM5gkIibCMfFyRIJSYnDCgeKUoqJiIrLC1NLi8wMVQyMzRHQQA7
    }
    ]
}
if { ![ info exist GidPriv(VerifNoImage)]} {
    #set GidPriv(VerifNoImage) [ image create photo -file $GIDDEFAULT/bitmaps/verif_no.gif]
    set GidPriv(VerifNoImage) [ image create photo -data {
	R0lGODlhEAAQAIAAAP///////yH5BAEKAAEALAAAAAAQABAAAAIOjI+py+0Po5y02ouzPgUA
	Ow==
    }
    ]
}

# Author:       RAMSAN
# Date:         7 Feb 1997
# Function:     ReadDefaultValues
# Description:  Reads in file ~/.gidDefaults or window\gid.init  and finds there
#               some GUI configuration variables
#
# Note: ReadDefaultValues now is called before with the best estimated size file gidDefaults1024x768
# and after with gid.init or .gidDefaults, ant the second time overwrite previous values, 
# with the old - inverse system if call again ReadDefaultValues with other file, previous values don't change

proc ReadDefaultValues { file } {
    global GidPriv
    global MainWindowGeom DefaultSearchDirectory DefaultSearchDirectoryP
    global CalcAngleUnits RunProcInfo GidProcWin
    global tcl_platform
    if { $::_DebugDefaultValues} {
	DebugText ReadDefaultValues file= $file
    }
  
    if { ![info exists ::GiDTBLayout] } {
	array set ::GiDTBLayout {
	    INSIDE,CURRENT ""
	    INSIDE,INI     ""
	    INSIDE,PRE     ""
	    INSIDE,POST    ""
	}
    }
    if { ![info exists GidProcWin(remoteserver)] } { 
	set GidProcWin(remoteserver) "" 
    }
    if { ![info exists GidProcWin(remotegroup)] } {
	set GidProcWin(remotegroup) ""
    }
    if { ![info exists GidProcWin(remoteuser)] } { 
	set GidProcWin(remoteuser) "" 
    }

    if { [catch { set fileid [open $file r] }] } {
	return
    }
    fconfigure $fileid -encoding "utf-8"
    
    lappend ::GiDTBLayout(inifile) $file
    set ::GiDTBLayout($file,INSIDE,INI) [list]
    #set ::GiDTBLayout($file,INSIDE,cmd)  [list]

    while { ![eof $fileid] } {
	set aa [gets $fileid]
	if { [catch { set varname [lindex $aa 0] } ] } { continue }
	switch -- $varname {
	    MainWindowGeom {
		set MainWindowGeom [lrange $aa 1 end]
	    }
	    BitmapsGeom {
		set varname PrePostBitmapsWindowGeom
		if { [lindex $aa 1] != "NONE" } {
		    set GidPriv(PrePostBitmapsWindowGeom) \
		            "OPEN [lindex $aa 1] 1 CreateBitmaps"
		    #why ? after idle CreateBitmaps
		} else {
		    set GidPriv(PrePostBitmapsWindowGeom) "NONE {} 1 \
		            CreateBitmaps"
		}
	    }
	    DefaultSearchDirectory {
		set DefaultSearchDirectory [lrange $aa 1 end]
	    }
	    DefaultSearchDirectoryP {
		set DefaultSearchDirectoryP [lrange $aa 1 end]
	    }
	    RightButtonsGeom {
		set varname PrePostRightButWindowGeom
		if { [lindex $aa 1] != "NONE" } {
		    set GidPriv(PrePostRightButWindowGeom) "OPEN \
		            [lindex $aa 1]  1 RightButtons"
		    #why ? after idle RightButtons
		} else {
		    set GidPriv(PrePostRightButWindowGeom) "NONE {} 1 \
		            RightButtons"
		}
	    }
	    BottomEntryGeom {
		set varname PrePostEntryWindowGeom
		if { [lindex $aa 1] != "NONE" } {
		    set GidPriv(PrePostEntryWindowGeom) "OPEN [lindex $aa 1] \
		            1 BottomEntryFrame"
		    # why? after idle CreateBitmaps
		} else {
		    set GidPriv(PrePostEntryWindowGeom) "NONE {} 1 \
		            BottomEntryFrame"
		}
	    }
	    CalcAngleUnits {
		set CalcAngleUnits [lindex $aa 1]
	    }
	    ListEntitiesGeom {
		if { [lindex $aa 1] != "NONE" } {
		    set GidPriv(ListEntitiesWindowGeom) "OPEN [lindex $aa 1]"
		} else {
		    set GidPriv(ListEntitiesWindowGeom) "NONE +0+176 "
		}
	    }
	    RunProcInfo {
		set RunProcInfo [lrange $aa 1 end]
		# 3 seconds to permmit the GiD window to appear
		after 3000 ActualizeProcInfoBegin
	    }
	    remoteserver {
		set GidProcWin(remoteserver) [lindex $aa 1]
	    }
	    remotegroup {
		set GidProcWin(remotegroup) [lindex $aa 1]
	    }
	    remoteuser {
		set GidProcWin(remoteuser) [lindex $aa 1]
	    }
	    remoteauto {
		set GidProcWin(remoteauto) [lindex $aa 1]
	    }
	    remotetime {
		set GidProcWin(remotetime) [lindex $aa 1]
	    }
	    remoteport  {
		set GidProcWin(remoteport) [lindex $aa 1]
	    }
	    remotetimeout {
		set GidProcWin(remotetimeout) [lindex $aa 1]
	    }
	    remoteresultsdownload {
		set GidProcWin(remoteresultsdownload) [lindex $aa 1]
	    }
	    remoteresultsdelete {
		set GidProcWin(remoteresultsdelete) [lindex $aa 1]
	    }
	    NormalFont {
		foreach "fn family size weight slant" $aa ""
		font conf $fn -family $family -size $size -weight $weight \
		    -slant $slant
		font conf BoldFont -family $family -size $size -weight bold \
		    -slant $slant
		font conf SmallFont -family $family -size [expr $size-2] -weight $weight \
		    -slant $slant
	    }
	    FixedFont -
	    BigFont {
		foreach "fn family size weight slant" $aa ""
		font conf $fn -family $family -size $size -weight $weight \
		    -slant $slant
	    }
	    LayersAlsoLower {
		set GidPriv(LayersAlsoLower) [lindex $aa 1]
	    }
	    LayersBackOpposite {
		set GidPriv(LayersBackOpposite) [lindex $aa 1]
	    }
	    DuplicateEntities {
		set GidPriv(DuplicateEntities) [lindex $aa 1]
	    }
	    GiDTkFBoxPreview {
		set GidPriv(GiDTkFBoxPreview) [lindex $aa 1]
	    }
	    LayersAlphabetic {
		set GidPriv(LayersAlphabetic) [lindex $aa 1]
	    }
	    SoftwareOpenGL {
		set GidPriv(SoftwareOpenGL) [lindex $aa 1]
	    }
	    HaveViewMenu {        
		#obsolete option
	    }
	    GUIAppearance {
		set GidPriv(GUIAppearance) [lindex $aa 1]
	    }
	    DisplayCoordinates {
		set GidPriv(Displaycoordinates) [lindex $aa 1]
	    }
	    ToglWinMode {
		set GidPriv(ToglWinMode) [lindex $aa 1]
	    }
	    Language {
		GidUtils::SetCurrentLanguage [lindex $aa 1] 0              
	    }
	    RealFormatDefault {
		set GidPriv(RealFormat,Default) [lindex $aa 1]   
	    }
	    default {
		if { [string match *WindowGeom $varname] } {
		    #set name [lindex $aa 0]
		    set GidPriv($varname) [lrange $aa 1 end]
		    if { [lindex $GidPriv($varname) 2] != "0" && \
		        [lindex $GidPriv($varname) 2] != "1" } {
		        set GidPriv($varname) [linsert $GidPriv($varname) 2 1]
		    }
		}
	    }
	}
	if {[regexp ^INSIDE [lindex $aa 1]]} {
	    if {[lsearch $::GiDTBLayout($file,INSIDE,INI) $varname] == -1} {
		lappend ::GiDTBLayout($file,INSIDE,INI) $varname
#                 set cmd [lindex $aa 4 end]
#                 if {$cmd eq ""} {
#                     # PrePostRightButWindowGeom INSIDE {} RightButtons
#                     set cmd [lindex $aa end]
#                 }
#                 lappend ::GiDTBLayout($file,INSIDE,cmd)  $cmd
	    }
	}
    }
    if {[llength $::GiDTBLayout($file,INSIDE,INI)]} {
	set ::GiDTBLayout(INSIDE,INI) $::GiDTBLayout($file,INSIDE,INI)
    }
    close $fileid
}

# GiDTBLayout -- Data Structure
# 
#  INSIDE,CURRENT : PRE | POST | {}
#
#  INSIDE,INI : names from ini file 
#
#  INSIDE,PRE : last order of toolbars in PRE, initially it's empty
#
#  INSISE,POST : last order of toolbars in POST, initially it's empty
#
#  .gid.$frame : the list of toolbars gridded in order
#
#  

# where must be PRE or POST
#
proc SetGiDTBLayout { where } {
    set current $::GiDTBLayout(INSIDE,CURRENT)
    if {$current eq ""} {
	set ::GiDTBLayout(INSIDE,PRE) $::GiDTBLayout(INSIDE,INI)
	set ::GiDTBLayout(INSIDE,POST) $::GiDTBLayout(INSIDE,INI)
    } else {
	# save current status
	set ::GiDTBLayout(INSIDE,$current) {}
	foreach f {top left right bottom} {
	    foreach tb $::GiDTBLayout(.gid.$f) {
		lappend ::GiDTBLayout(INSIDE,$current) $::GiDTBLayout(tb,$tb)
		destroy $tb
	    }
	    set ::GiDTBLayout(.gid.$f) {}
	    grid remove .gid.$f
	}       
    }
    set ::GiDTBLayout(INSIDE,CURRENT) $where
}

#
# what can be: PRE or POST
#

proc OpenOutsideWindows { what } {
    global GidPriv
    if { [winfo exists .gid.central.s] && ( [.gid.central.s disable windows] || \
	    [.gid.central.s info variables usemorewindows] == 0 ) } return

    # hace falta? update idletasks
    set visited {}

    SetGiDTBLayout $what
    
    # create the INSIDE windows just in the order they where read
    # from ini or saved
    #WarnWinText "=========== Visitando $what [clock format [clock seconds]] ==========="
    set ::tb_errors {}
    foreach name $::GiDTBLayout(INSIDE,$what) {
	lappend visited $name
	if {[catch {set val $GidPriv($name)}]} continue
	if {[catch { eval [lrange $val 3 end] } msg]} {
	    lappend ::tb_errors $msg
	    # remove that entry GidPriv and ::GiDTBLayout
	    set GidPriv($name) [lindex $GidPriv($name) 0]
	} 
	#WarnWinText "$name -> $val"
	#if {[catch { eval [lrange $val 3 end] } msg]} {
	#    WarnWinText "failed: $msg"
	#    WarnWinText "$::errorInfo"
	#}
    }
    #WarnWinText "=========== Saliendo $what [clock format [clock seconds]] ==========="

    foreach {name val} [array get GidPriv PrePost*Geom] {
	if {[lsearch $visited $name] != -1} continue
	lappend visited $name
	if { ([lindex $val 0] == "OPEN" || [string match INSIDE* [lindex $val 0]]) \
		 || $name == "PrePostTopMenuWindowGeom" } {
	    if {[catch { eval [lrange $val 3 end] } msg]} {
		lappend ::tb_errors "$name \n\t $msg"
	    }
	}
    }
    foreach {name val} [array get GidPriv *Geom] {
      if {[lsearch $visited $name] != -1} continue
	regsub {Geom$} $name {} namegeom
	if { [string match *PrePost* $name] } { 
	    continue
	} elseif { $what == "PRE" && ![string match *Post* $name] && \
		([lindex $val 0] == "OPEN" || \
		[string match INSIDE* [lindex $val 0]])} {
	  catch { eval [lrange $val 3 end] }
	} elseif { $what == "PRE" && [string match *Post* $name] } {
	    set aa [lindex $val 0]
	    catch {
		destroy $GidPriv($namegeom)
	    }
	    set GidPriv($name) "$aa [lrange $val 1 end]"
	} elseif { $what == "POST" && [string match *Post* $name] && \
		[lindex $val 0] == "OPEN"  || \
		[string match INSIDE* [lindex $val 0]]} {
	    catch { eval [lrange $val 3 end] }
	} elseif { $what == "POST" && ![string match *Post* $name] } {
	    set aa [lindex $val 0]
	    catch {
		destroy $GidPriv($namegeom)
	    }
	    set GidPriv($name) "$aa [lrange $val 1 end]"
	}
    }
    ###puts " . fin"
    if {[info exists GidPriv(ToglWinMode)]} {
	after idle after idle GiD_Process NewWin $GidPriv(ToglWinMode)
    }
}

proc IsSplashWindowOnInDefaults {} {

    set file [GiveGidDefaultsFile]

    if { [catch { set fileid [open $file r] }] } {
	return 1
    }
    fconfigure $fileid -encoding "utf-8"
    while { ![eof $fileid] } {
	set aa [gets $fileid]
	if { [regexp {SplashWindow[ ]*([01])} $aa {} val] } {
	    close $fileid
	    return $val
	}
    }
    close $fileid
    return 1
}

proc WriteTCLDefaults { } {
    WriteTCLDefaultsInFile [GiveGidDefaultsFile]
    

    #to know the more important messages to be translated
    if { [info exists ::COUNTHITSMSGS] && $::COUNTHITSMSGS } {
	set filename [file join [file dirname [GiveGidDefaultsFile]] "MsgsHits.xml"]
	if { [file exists $filename] } {
	    #append the previous hits counter
	    set fp [open $filename r]
	    if { $fp != "" } {
		set map [list "&gt;" > "&lt;" < "&amp;" & "&quot;" \" "&apos;" ']
		while { ![eof $fp] } {
		    gets $fp aa
		    if { [regexp {<string>(.*)</string><hits>([0-9]*)</hits>} $aa dummy name frequency] } {
		        if { [string is integer $frequency] } {
		            set name [string map $map $name]
		            if { [info exists ::MsgsHits($name)] } {
		                incr ::MsgsHits($name) $frequency
		            } else {
		                set ::MsgsHits($name) $frequency                                
		            }
		        }
		    }
		}
		close $fp
	    }
	}
	#save accumulated hits counter
	set fp [open $filename w]
	if { $fp != "" } {
	    set map [list > "&gt;" < "&lt;" & "&amp;" \" "&quot;" ' "&apos;"]
	    puts $fout <frequencies>
	    set orderedlist [lsort -integer -decreasing [array names ::MsgsHits]]
	    foreach str $orderedlist {
		puts -nonewline $fp "<string>[string map $map $str]</string>"
		puts $fp "<hits>$::MsgsHits($str)</hits>"
	    }
	    puts $fout </frequencies>
	    close $fp
	}
    }

}

proc WriteTCLDefaultsInFile { file } {
    global GidPriv GiDPrivToolBar
    global MainWindowGeom DefaultSearchDirectory DefaultSearchDirectoryP
    global CalcAngleUnits RunProcInfo GidProcWin
    global tcl_platform

    if { $::GidPriv(OffScreen)} { 
	return
    }
    
    if { $::_DebugDefaultValues} {
	DebugText "WriteTCLDefaultsInFile  file = $file"
    }

    if { $tcl_platform(platform) == "windows" } {
	if { [wm state .gid] == "zoomed" } {
	    if { ![info exists MainWindowGeom] } {
		set MainWindowGeom "800x600+100+100"
	    }
	    set geom $MainWindowGeom
	    if { [lindex $geom 0] ne "ZOOMED" && \
		[regexp {([^x]*)x([^+-]*)([+-]*[0-9]*)([+-]*[0-9]*)} $geom] } {
		set MainWindowGeom "ZOOMED $geom"
	    } elseif { [regexp {([^x]*)x([^+-]*)([+-]*[0-9]*)([+-]*[0-9]*)} \
		    [lindex $geom 1]] } {
		set MainWindowGeom "$geom"
	    } else {
		set MainWindowGeom "ZOOMED 800x600+100+100"
	    }
	} else {
	    set MainWindowGeom [wm geom .gid]
	    regexp {([^x]*)x([^+-]*)([+-]*[0-9]*)([+-]*[0-9]*)} $MainWindowGeom \
		    trash w h x y
	    if { [.gid cget -menu] != "" } {
		set menuheight [GiD_Info menuheight]
		incr h $menuheight
	    }
	    set windowcaptionheight [GiD_Info windowcaptionheight]
	    if { $x <= 0 } { set x +-4 }
	    if { $y <= 0 } { set y +-4 }
	    if { $w >= [ winfo screenwidth .gid] && \
		    [expr $h+$windowcaptionheight] >= [ winfo screenheight .gid] } {
		set w [ winfo screenwidth .gid]
		set h [ winfo screenheight .gid]
		set x +-4
		set y +-4
	    }
	    set MainWindowGeom ${w}x${h}${x}${y}
	}
    } else {
	set MainWindowGeom [wm geom .gid]
    }

    if { ![info exists CalcAngleUnits] } {
	set CalcAngleUnits "rad"
    }
    set varslist { MainWindowGeom DefaultSearchDirectory \
	    DefaultSearchDirectoryP CalcAngleUnits}

    set fileid [open $file a]
    fconfigure $fileid -encoding "utf-8"
    foreach i $varslist {
	if { [info exists $i] } {
	    set text "puts $fileid \"$i $$i\""
	    eval $text
	}
    }

    if { [info exists GidPriv(LayersAlsoLower)] } {
	puts $fileid "LayersAlsoLower $GidPriv(LayersAlsoLower)"
    }
    if { [info exists GidPriv(LayersBackOpposite)] } {
	puts $fileid "LayersBackOpposite $GidPriv(LayersBackOpposite)"
    }
    if { [info exists GidPriv(DuplicateEntities)] } {
	puts $fileid "DuplicateEntities $GidPriv(DuplicateEntities)"
    }
    if { [info exists GidPriv(GiDTkFBoxPreview)] } {
	puts $fileid "GiDTkFBoxPreview $GidPriv(GiDTkFBoxPreview)"
    }
    if { [info exists GidPriv(LayersAlphabetic)] } {
	puts $fileid "LayersAlphabetic $GidPriv(LayersAlphabetic)"
    }
    if { [info exists GidPriv(SoftwareOpenGL)] } {
	puts $fileid "SoftwareOpenGL $GidPriv(SoftwareOpenGL)"
    }
    
    puts $fileid "GUIAppearance $GidPriv(GUIAppearance)"

    foreach i "NormalFont FixedFont BigFont" {
	set family [font conf $i -family]
	set size [font conf $i -size]
	set weight [font conf $i -weight]
	set slant [font conf $i -slant]
	puts $fileid [ list $i $family $size $weight $slant]
    }

    if { $::_DebugDefaultValues} {
	DebugText "in WriteTCLDefaultsInFile: array get GidPriv *Geom"
    }

  catch {
    # write toolbars in order
    set visited [list]
    foreach f {top left right bottom} {
      foreach tb $::GiDTBLayout(.gid.$f) {
	lappend visited [set name $::GiDTBLayout(tb,$tb)]
	puts $fileid "$name $GidPriv($name)"
	# why not unset GidPriv($name)?
      }
    }
    foreach {name val} [array get GidPriv *Geom] {
      if {[lsearch $visited $name] != -1} continue
	if { $::_DebugDefaultValues} {
	    DebugText "  processing: $name $val"
	}
	regsub {Geom$} $name {} geomwinname
	if { $::_DebugDefaultValues} {
	    DebugText "  geomwinname = $geomwinname ([info exists GidPriv($geomwinname)]) ([info exists GiDPrivToolBar($name)])"
	}
	 if { [info exists GidPriv($geomwinname)] } {
	    set w $GidPriv($geomwinname)
	    if { [winfo exists $w] && [winfo class $w] == "Toplevel" } {
		    set InitComm $GidPriv($geomwinname,InitComm)
		    WriteYourGeomToVar $w OPEN $name $InitComm
		    set val $GidPriv($name)
	    } 
	} elseif { [info exists GiDPrivToolBar($name)] } {
	    set w $GiDPrivToolBar($name)
	    if { $::_DebugDefaultValues} {
		DebugText "    $GiDPrivToolBar($name) is a toolbar"
	    }
	      set type [lindex $GidPriv($name) 0]
	      if { [regexp {(INSIDE[a-zA-Z]+)[0-9]*} $type {} type] } {
		set vv $GidPriv($name)
#                   set par [lindex [pack info $w] 1] 
#                   set type $type[lsearch [pack slaves $par] $w]
		array set  grid_info [grid info $w]
		if {$grid_info(-sticky) eq "ns" ||
		    $grid_info(-sticky) eq "sn" } {
		  set type $type$grid_info(-column)
		} else {
		  set type $type$grid_info(-row)
		}
		set GidPriv($name) "$type [lrange $vv 1 end]"
		set val $GidPriv($name)
	      }
	}
	puts $fileid "$name $val"
    }
  } msg 

    if { $::_DebugDefaultValues} {
	DebugText "msg = $msg"
	DebugText "in WriteTCLDefaultsInFile: end of array get GidPriv *Geom"
    }

    if { [info exist RunProcInfo] } {
	#set res [catch { PWQuittingGiD } aa]
	#if { $res } {
	#    #ERROR--$aa
	#}
	if { $RunProcInfo != "" } {
	    puts $fileid "RunProcInfo $RunProcInfo"
	}
    }
    foreach i "remoteserver remotegroup remoteuser remoteauto remotetime remoteport \
	remotetimeout remoteresultsdownload remoteresultsdelete" {
	if { [info exists GidProcWin($i)] } {
	    puts $fileid "$i $GidProcWin($i)"
	}
    }

    if {[info exists GidPriv(ToglWinMode)]} {
      puts $fileid "ToglWinMode $GidPriv(ToglWinMode)"
    }

    puts $fileid "Language [GidUtils::GetCurrentLanguage]"

    if {[info exists GidPriv(RealFormat,Default)]} {
	puts $fileid [list RealFormatDefault $GidPriv(RealFormat,Default)]
    }
   
    close $fileid

    #look for it in tk.tcl
    if { [info exists GidPriv(EventFile)] } {
	close $GidPriv(EventFile)
    }

}

proc MakeAlwaysOnTop { w } {
    set ::GidPriv(Transient,$w) 1
    ToggleTransientWin $w
}

proc ToggleTransientWin { w } {
    global GidPriv
    set top [winfo parent [winfo toplevel $w]]
    if { [info exists GidPriv(Transient,$w)] && $GidPriv(Transient,$w) == 1 && $top != "." } {
	#wm transient $w .gid
	wm transient $w $top
    } else {
	wm transient $w ""
    }
    if { $::tcl_platform(platform) == "windows" } {
	    AddAlwaysOnTopFlag $w
    }
}

proc AddAlwaysOnTopFlag { w } {
    global GidPriv
    catch { destroy $w._topmenu }
    menu $w._topmenu -tearoff 0
    $w conf -menu $w._topmenu
    menu $w._topmenu.system -tearoff 0
    $w._topmenu add cascade -menu $w._topmenu.system
    $w._topmenu.system add checkbutton -label [_ "Always on top"] -var \
	    GidPriv(Transient,$w) -command "ToggleTransientWin $w"
}

proc ConfigureWinIdle { w geomname { InitComm ""} } {
    if { [winfo exists $w] } {
	bind $w <Configure> [ list ConfigureWin $w $geomname [ join $InitComm]]
    }
}

proc ConfigureWin { w geomname {InitComm ""}} {
    global GidPriv tcl_platform

    if { [winfo class $w] == "Toplevel" } {
	bind $w <Configure> ""
	
	WriteYourGeomToVar $w OPEN $geomname $InitComm
	after idle ConfigureWinIdle $w $geomname [ list $InitComm]
    }
}

proc DestroyWin { W w geomname {InitComm ""}} {
    if { $W == $w } {
	WriteYourGeomToVar $w NONE $geomname $InitComm
    }
}

proc WriteYourGeomToVar { w what geomname {InitComm ""}} {
    global GidPriv tcl_platform
    if { [wm transient $w] != "" } {
	set trans 1
    } else {
	set trans 0
    }
    set geom [wm geom $w]
    if { $tcl_platform(platform) == "windows" } {
	regexp {([^x]*)x([^+-]*)([+-]*[0-9]*)([+-]*[0-9]*)} $geom \
		trash width height x y
	if { $width == "" } { set width 1 }
	if { $height == "" } { set height 1 }
	set menuheight [GiD_Info menuheight]
	incr height [expr $menuheight*-1]
	if { $height < 1 } { set height 20 }
	set geom ${width}x${height}${x}${y}
    }
    set GidPriv($geomname) "$what $geom $trans $InitComm"
}

#w: toplevel name
#title: window title (marked to translation)
#geomname: must be <prefix>WindowGeom, with <prefix> unique
#InitComm: the command to be called to reopen the window
proc InitWindow { w title geomname {InitComm ""} { class ""} { OnlyPos 0 } {ontop 1} } {
    global GidPriv tcl_platform
   
#    if { [winfo exists .gid.central.s] && ( [.gid.central.s disable windows] || \
#            [.gid.central.s info variables usemorewindows] == 0 ) } { return close }
#if the user set usemorewindows to 0, then cannot re-open the preferences window, layers, etc!!
    if { [winfo exists .gid.central.s] && [.gid.central.s disable windows] } { return close } 
   
    if { ![info exists GidPriv(TopLevelsList)] } {
	set GidPriv(TopLevelsList) ""
    }
    if { [lsearch $GidPriv(TopLevelsList) $w] == -1 } {
	lappend GidPriv(TopLevelsList) $w
    }

    if { [winfo exists $w] } {
	destroy $w 
    }
    if { $class == "" } {
	toplevel $w
    } else { 
	toplevel $w -class $class 
    }

    if { $tcl_platform(platform) == "windows" || $tcl_platform(platform) == "unix" } {
	if { [info exists GidPriv($geomname)] } {
	    set GidPriv(Transient,$w) [lindex $GidPriv($geomname) 2]
	} else { 
	    set GidPriv(Transient,$w) $ontop
	}
	ToggleTransientWin $w
    }

    wm title $w $title
    if { [info exists GidPriv($geomname)] } {
	if { $OnlyPos == 0 } {
	    set posgeom [lindex $GidPriv($geomname) 1]
	} else {
	    regexp {[+-].*} [lindex $GidPriv($geomname) 1] posgeom
	}
	WmGidGeom $w $posgeom
    } else { 
	WmGidGeom $w ""
    }
    bind $w <Destroy> [ list DestroyWin %W $w $geomname $InitComm]
    bind $w <Configure> [ list ConfigureWin %W $geomname $InitComm]

    regsub {Geom$} $geomname {} geomwinname
    set GidPriv($geomwinname) $w
    set GidPriv($geomwinname,InitComm) $InitComm
    if { ![info exists GidPriv(InitWindowList)] } { 
	set GidPriv(InitWindowList) "" 
    }
    if { [lsearch $GidPriv(InitWindowList) $geomwinname] == "-1" } {
	lappend GidPriv(InitWindowList) $geomwinname
    }
    return $w
}

proc InitWindow_ModifyInitComm { w geomname InitComm } {
    global GidPriv

    bind $w <Destroy> [ list DestroyWin %W $w $geomname $InitComm]
    bind $w <Configure> [ list ConfigureWin %W $geomname $InitComm]

    regsub {Geom$} $geomname {} geomwinname
    set GidPriv($geomwinname) $w
    set GidPriv($geomwinname,InitComm) $InitComm
}

proc QuadraticMenu { w } {
    set quadratic [GiD_Info Project Quadratic]
    switch $quadratic {
	0 { set curr Normal }
	1 { set curr Quadratic }
	2 { set curr Quadratic9 }
	default { set curr Normal }
    }

    set names "Normal Quadratic Quadratic9"
    set namesLabel [ list [_ "Normal"] [_ "Quadratic"] [_ "Quadratic9"]]
    $w del 0 end
    
    foreach i $names j $namesLabel {
	if { $i == $curr } {
	    $w add check -label $j -var $w
	    set ::$w 1
	} else {
	    $w add command -label $j -command ".central.s process escape \
		    escape escape escape Meshing quadratic $i escape"
	}
    }
}


proc ViewModeMenu { w what } {  

    if { $w ne "" } {
	set ::CurrentViewMode [GiD_Info Project ViewMode]
	$w del 0 end
    }
    
    if { $what == "PRE" } {
	set names {GEOMETRYUSE MESHUSE "" "" ""}
	set namesLabel [ list [_ "Geometry"] [_ "Mesh"] [_ "Toggle"] --- [_ "Postprocess"]]
	set namesAcc [list "" "" [list Ctrl-m Control-m] "" ""]
	set namesimages [list "" "" [GetImage "view_geom_mesh16x16.gif"] \
	    "" [GetImage "prepost16x16.gif"]]
	set colors [ list "black" "black" "black" "" "black"]
	set commands { {GiD_Process MEscape Geometry} {MeshView} ToggleGeomMesh \
	    "" {GiD_Process MEscape Postprocess} }  

	if { [winfo exists .gid.central.s] && [lindex [GiD_Info mesh] 0] == 0 } {
	    lset colors 1 $::GidPriv(Color,DisabledForegroundMenu)
	    lset colors 2 $::GidPriv(Color,DisabledForegroundMenu)
	    lset colors 4 $::GidPriv(Color,DisabledForegroundMenu)
	}
	
	if { $w eq "" } {
	    foreach k $commands acc $namesAcc {
		if { $acc ne "" } {
		    bind .gid <[lindex $acc 1]> $k
		}
	    }
	} else {
	    foreach i $names j $namesLabel k $commands acc $namesAcc col $colors \
		image $namesimages {
		if { $j eq "---" } {
		    $w add separator
		    continue
		} elseif { $i eq "" } {
		    $w add command -label $j -command $k -accelerator [lindex $acc 0] \
		        -foreground $col -image $image -compound left
		    if { $acc ne "" } {
		        bind .gid <[lindex $acc 1]> $k
		    }
		} else {
		    $w add checkbutton -label $j -variable ::CurrentViewMode \
		        -onvalue $i -command $k -accelerator [lindex $acc 0] \
		        -foreground $col -image $image -compound left
		    if { $acc ne "" } {
		        bind .gid <[lindex $acc 1]> $k
		    }
		}
	    }
	}
    } else {
	set names {POSTUSE GRAFUSE "" ""}
	set namesLabel [ list [_ "Mesh"] [_ "Graphs"] [_ "Toggle"] --- [_ "Preprocess"]]
	set namesimages [list "" "" [ GetImage "postgraf16x16.gif"] "" [ GetImage "prepost16x16.gif"]]
	set commands { {GiD_Process MEscape Results Graphs OptionsGraph ShowGraphs No} \
	    {GiD_Process MEscape Results Graphs OptionsGraph ShowGraphs Yes} \
		           TogglePostGraph  \
	    "" {GiD_Process MEscape preprocess} }
	set colors [ list "black" "black" "black" "" "black"]
	if { [ llength [ GiD_Info postprocess get all_graphs]] == 0} {
	    lset colors 1 $::GidPriv(Color,DisabledForegroundMenu)
	    lset colors 2 $::GidPriv(Color,DisabledForegroundMenu)
	    lset colors 4 $::GidPriv(Color,DisabledForegroundMenu)
	}
	if { $w eq "" } {
	    # nothing by now
	} else {
	    foreach i $names j $namesLabel k $commands image $namesimages col $colors {
		if { $j eq "---" } {
		    $w add separator 
		    continue
		} elseif { $i eq "" } {
		    $w add command -label $j -command $k \
		            -foreground $col -image $image -compound left
		} else {
		    $w add checkbutton -label $j -variable ::CurrentViewMode \
		        -onvalue $i -command $k  \
		        -foreground $col -image $image -compound left
		}
		##  elseif { $j eq [_ "Preprocess"] } {
		    ##         $w add checkbutton -label $j -command $k -image $image -compound left 
		    ## } else {
		    ##         $w add checkbutton -label $j -variable ::CurrentViewMode \
		        ##             -onvalue $i -command $k -image $image -compound left 
		    ## }
	    }
	}
    }
}

proc DrawSizeMenu { w } {
    $w del 0 end
    set backgroundfilename [GiD_Info Project BackgroundFilename]
    if { $backgroundfilename != "" } {
	$w add command -label [_ "Background mesh"] -command ".central.s process escape \
	    escape escape escape Meshing drawsizes background aselems"
    } else {
	set geomlistNames [ list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] \
		"" [_ "All types"] ] 
	set geomlist "Points Lines Surfaces Volumes Separator All"
	
	foreach i $geomlist j $geomlistNames {
	    if { $i=="Separator" } {
		$w add separator  
	    } else {
		$w add command -label $j -command \
		    "GeometryView ; GiD_Process Mescape meshing drawsizes $i"
	    }
	}
	
	set nfig(0) [GiD_Info Geometry NumPoints]
	set nfig(1) [GiD_Info Geometry NumLines]
	set nfig(2) [GiD_Info Geometry NumSurfaces]
	set nfig(3) [GiD_Info Geometry NumVolumes]
	set nfig(4) 1
	set nfig(5) [expr {$nfig(0)+$nfig(1)+$nfig(2)+$nfig(3)}]
	
	for {set i 0} {$i < 6} {incr i} {
	    if { $i == 4 } continue
	    if { !$nfig($i) } {
		$w entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu) 
	    } else {
		$w entryconfigure $i -foreground "black" 
	    }
	}
    }
}

proc DrawMenuMeshNoMesh { w } {
    $w del 0 end
    set geomlistNames [ list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] "" [_ "All types"] ] 
    set geomlist "Points Lines Surfaces Volumes Separator All"
    
    foreach i $geomlist j $geomlistNames {
	if { $i == "Separator" } {
	    $w add separator 
	} else {
	    $w add command -label $j -command \
		    "GeometryView ; GiD_Process Mescape Meshing DrawMeshNoMesh $i"
	}
    }
    set nfig(0) [GiD_Info Geometry NumPoints]
    set nfig(1) [GiD_Info Geometry NumLines]
    set nfig(2) [GiD_Info Geometry NumSurfaces]
    set nfig(3) [GiD_Info Geometry NumVolumes]
    set nfig(4) 1
    set nfig(5) [expr {$nfig(0)+$nfig(1)+$nfig(2)}]
    
    for {set i 0} {$i < 6} {incr i} {
	if { [lindex $geomlist $i] == "Separator" } continue
	if { !$nfig($i) } {
	    $w entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu) 
	} else {
	    $w entryconfigure $i -foreground "black" 
	}
    }
}

proc DrawForcePointsToMesh { w } {
    $w del 0 end
    
    set n 2
    set geomlistNames [ list [_ "Surface mesh#C#menu"] [_ "Volume mesh#C#menu"]] 
    set geomlist "Surface Volume"
    
    foreach i $geomlist j $geomlistNames {
	$w add command -label $j -command \
	    "GeometryView ; GiD_Process Mescape Meshing DrawForcedPointsToMesh $i"
    }
    
    set nfig(0) [GiD_Info Geometry NumSurfaces]
    set nfig(1) [GiD_Info Geometry NumVolumes]
    
    for {set i 0} {$i < $n} {incr i} {
	if { !$nfig($i) } {
	    $w entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu) 
	} else {
	    $w entryconfigure $i -foreground "black" 
	}
    }
}

proc DrawBoundaryLayerMesh { w } {
    $w del 0 end
    
    if { $::BOUNDARY_LAYER_MESH_3D } {
	set n 2
	set geomlistNames [ list [_ "2 dimensional"] [_ "3 dimensional"]] 
	set geomlist "2d 3d"
    } else {
	set n 1
	set geomlistNames [ list [_ "2 dimensional"]] 
	set geomlist "2d"
    }
    
    foreach i $geomlist j $geomlistNames {
	$w add command -label $j -command \
	    "GeometryView ; GiD_Process Mescape Meshing DrawBoundaryLayerMesh $i"
    }

    set nfig(0) [GiD_Info Geometry NumSurfaces]
    set nfig(1) [GiD_Info Geometry NumVolumes]
    
    for {set i 0} {$i < $n} {incr i} {
	if { !$nfig($i) } {
	    $w entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu) 
	} else {
	    $w entryconfigure $i -foreground "black" 
	}
    }
}

proc DrawMenu { w mode } {
    $w del 0 end
    
    set geomlistNames [ list [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] "" [_ "All types"] ] 
    set geomlist "Lines Surfaces Volumes Separator All"
    
    foreach i $geomlist j $geomlistNames {
	if { $i == "Separator" } {
	    $w add separator 
	} else {
	    if { $mode == "ElementType" } {
		$w add command -label $j -command \
		    "GeometryView ; GiD_Process Mescape Meshing DrawElementType $i"
	    } elseif { $mode == "MeshNoMesh" } {
		 $w add command -label $j -command \
		    "GeometryView ; GiD_Process Mescape Meshing DrawMeshNoMesh $i"
	    } elseif { $mode == "StructuredType" } {
		$w add command -label $j -command \
		    "GeometryView ; GiD_Process Mescape Meshing DrawStructuredType $i"
	    } 
	}
    }
    set nfig(0) [GiD_Info Geometry NumLines]
    set nfig(1) [GiD_Info Geometry NumSurfaces]
    set nfig(2) [GiD_Info Geometry NumVolumes]
    set nfig(3) 1
    set nfig(4) [expr {$nfig(0)+$nfig(1)+$nfig(2)}]
    
    for {set i 0} {$i < 5} {incr i} {
	if { [lindex $geomlist $i] == "Separator" } continue
	if { !$nfig($i) } {
	    $w entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu) 
	} else {
	    $w entryconfigure $i -foreground "black" 
	}
    }
}
    
#  proc ChangeQuadraticFlag { newval } {

#      set quadratic [GiD_Info Project Quadratic]

#      switch $newval {
#          info {
#              set word Using
#          }
#          quadratic {
#              set word "Changing to"
#              .central.s process escape escape escape escape Meshing \
#                      quadratic quadratic escape
#              set quadratic 1
#          }
#          quadratic9 {
#              set word "Changing to"
#              .central.s process escape escape escape escape Meshing \
#                      quadratic quadratic9 escape
#              set quadratic 2
#          }
#          normal {
#              set word "Changing to"
#              .central.s process escape escape escape escape Meshing \
#                      quadratic normal escape
#              set quadratic 0
#          }
#      }
#      if { $quadratic == 0 } {
#          set text "$word linear degree elements"
#      } elseif { $quadratic == 1 }  { set text "$word quadratic degree elements" 
#      } else { set text "$word quadratic9 degree elements" }
#      tk_dialogRAM .gid.tempwinq {information window} $text gidquestionhead 0 OK
    
#  }

proc ChooseElemType { etype { ltype ""} } {
    set MoreWin [GiD_Info variables UseMoreWindows]

    set list1 "Default Linear Triangle Quadrilateral Tetrahedra Hexahedra Prism OnlyPoints Sphere Circle"
    set list2 " -      -      surfaces surfaces      volumes    volumes   volumes volumes  -      -"

    set inum [lsearch -exact $list1 $etype]
    if { $inum == 0 ||  $inum == 1 } { #Default or Linear
	GiD_Process Mescape Meshing ElemType $etype $ltype
	set word $ltype
    } elseif { $inum == 8 } { #Sphere
	GiD_Process Mescape Meshing ElemType $etype $ltype
	set word $ltype
    } elseif { $inum == 9 } { #Circle
	GiD_Process Mescape Meshing ElemType $etype $ltype
	set word $ltype
    } elseif { $inum != -1 } { 
	GiD_Process Mescape Meshing ElemType $etype
	set word [lindex $list2 $inum] 
    } else {
	WarnWinText [_ "Unknown element type %s" $etype]
	return
    }

    if { $MoreWin > 1 } {
	tk_dialogRAMFull .gid.tempwiniw [_ "information window"] \
	    [_ "Select %1\$s to assign elem type: %2\$s" $word $etype]  \
	    "" "" gidquestionhead 0 [_ "OK"]
    }
}

# proc ChangeEntryPlace { {geom NONE}  {ini ""} } {
#     set hhscr [winfo screenheight .]
#     set makeMB 0
#     if { ($ini == "yes" && $geom != "NONE") || ([winfo exists .comm] && \
#             [winfo class .comm] != "Toplevel") } {
#         catch { destroy .comm }
#         set hh [expr [winfo x .]]
#         set vv [expr [winfo y .]+[winfo height .]+25]
#         toplevel .comm 
#         if { $geom != "NONE" } { wm geometry .comm $geom } else {
#             wm geometry .comm [winfo width .]x100+$hh+$vv
#         }
#         wm minsize .comm  1 1
#         set makeMB 1
#         bind .comm <Destroy> { set BottomEntryGeom [wm geom .comm] }
	
#     } else {
#         if { [winfo exists .comm] } {
#             destroy .comm
#             set makeMB 1
#         }
#         frame .comm
#         pack forget .gid.central
#         pack .comm -fill x -side bottom
#         catch { pack .gid.central -fill both -expand yes }
#     }
#     entry .comm.e -relief sunken
#     frame .comm.fwarn
#     scrollbar .comm.fwarn.yscroll -relief sunken  -takefocus 0 \
#             -command ".comm.fwarn.list yview"
#     listbox .comm.fwarn.list -yscroll ".comm.fwarn.yscroll \
#             set"  -relief groove  -borderwidth 4 -takefocus 0
#     if {[winfo class .comm] != "Toplevel" } { 
#         .comm.fwarn.list conf -width 20 -height 2
#     }
    
#     .comm.fwarn.list conf -selectmode single
#     pack .comm.fwarn.yscroll -side right -fill y
#     pack .comm.fwarn.list -expand yes -fill both 
    
#     pack .comm.e -fill x -side bottom -pady 5 -padx 5
#     pack .comm.fwarn -fill both -side top -pady 5 -expand yes
    

#     if { $makeMB == 1 } { MakeBindings }

# }


proc EnterCommand { w command { entry -1 } } {
    #if entry!=-1 => uses entryconfigure
    
    #if { $command == "" } { return }

    switch -- $entry {
	-1 {
	    set comm "$w conf -command "
	}
	-2 {
	    set comm "$w conf -postcommand "
	}
	default {
	    set comm "$w entryconfigure $entry -command "
	}
    }
    if { $command == "" } { 
	lappend comm ""
	eval $comm
	return
    }

    switch -- [lindex $command 0] {
	"-npw-" {
	    set vareval "$comm {[lrange $command 1 end] $w}"
	}
	"-np-" {
	    set vareval "$comm {[lrange $command 1 end]}"
	}
	"-ne-" {
	    set vareval "$comm {.central.s process [lrange $command 1 end]}"
	}
	default {
	    set vareval "$comm {.central.s process escape escape escape escape \
		    $command}"
	}
    }
    set vareval2 ""
    while 1 {
	if { [string first %W $vareval] == -1 } {
	    append vareval2 $vareval
	    break
	}
	append vareval2 [string range $vareval 0 [expr [string first %W $vareval]-1]]
	append vareval2 $w
	set vareval [string range $vareval [expr [string first %W $vareval]+2] end]
    }

    eval $vareval2

}

proc AddAccelerator { key menu index } {

    regsub -all Control $key Ctrl keyname
    $menu entryconf $index -accelerator $keyname

    if { [llength $key] == 1 } {
	bind .gid <$key> "$menu invoke $index ; break"
    } elseif { [llength $key] == 2 } {
	bind .gid <[lindex $key 0]><[lindex $key 1]> "$menu invoke $index ; break"
    }
}

proc AddLittleIcon { icon menu index } {
    set es_imagen_tk 0

    if { ( [ llength $icon] > 1) && ( "[ lindex $icon 0]" == "-np-")} {
	set icon [ eval [ lrange $icon 1 end]]
	# pueden devolvernos la imagen en si, en vez del archivo
	set res [ catch { 
	    set pp [ image type $icon]
	}]
	if { !$res} {
	    set es_imagen_tk 1
	}
    }

    if { !$es_imagen_tk} {
	if { $icon != "" } {
	    $menu entryconf $index -image [GetImage $icon] -compound left
	} else {
	    #add a dummy image to left align all rows
	    $menu entryconf $index -image [GetImage dummy16x16.gif] -compound left
	}
    } else {
	$menu entryconf $index -image $icon -compound left
    }
}

proc CreateTopMenusUnix { w } {
    global MenuNames MenuNamesP        ViewMenuPriv

    if { ![winfo exists $w] } { return }
    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "GEOMETRYUSE"
    }

    if { $infoProj != "POSTUSE" } {
	set MenuNamesInto $MenuNames
	set MenuEntriesName MenuEntries
	set MenuCommandsName MenuCommands
	set MenuAccelerName MenuAcceler
	set MenuLittleIcons MenuIcons
    } else {
	set MenuNamesInto $MenuNamesP
	set MenuEntriesName MenuEntriesP
	set MenuCommandsName MenuCommandsP
	set MenuAccelerName MenuAccelerP
	set MenuLittleIcons MenuIconsP

	if { [ info exist ViewMenuPriv] } {
	    unset ViewMenuPriv
	}
    }

    set NumMenus [llength $MenuNamesInto]

    set UnderLineList ""

    for {set i $NumMenus } { $i< 15 } { incr i } {
	if { [winfo exists $w.button$i] } {
	    destroy $w.button$i
	}
    }


    for {set i 0 } { $i< $NumMenus } { incr i } {
	set text [lindex $MenuNamesInto $i]
	set under [FindUnderChar text UnderLineList]

	if { ![winfo exists $w.button$i] } {
	    # set col_act [CCColorActivo [ $w cget -bg]]
	    menubutton $w.button$i -text $text -underline $under \
		-menu $w.button$i.m.m -relief flat -bd 0 \
		-highlightthickness 0 
	    set col_act [ CCColorActivo \
		              [ option get $w.button$i activeBackground color] \
		              33]
	    $w.button$i conf -activebackground $col_act
	    frame $w.button$i.m -class OverlayWindowN
	    wm colormapwindows .gid $w.button$i.m
	    menu $w.button$i.m.m -tearoff no -bd 2
	} else { 
	    $w.button$i conf -text $text -underline $under
	    if { [$w.button$i.m.m index end] != "none" } {
		ClearMenu $w.button$i.m.m
		#$w.button$i.m.m del 0 end
	    }
	    $w.button$i.m.m configure -postcommand ""
	}

	upvar \#0 $MenuEntriesName ME
	upvar \#0 $MenuCommandsName MC
	upvar \#0 $MenuLittleIcons MI

	if { $ME($i) == "" } {
	    if { [lindex $MC($i) 0] != "-np-" } {

		WarnWin [_ "Error creating menu '%s'. Needed -np-" $text]
	    }
	    set comm [lrange $MC($i) 1 end]
	    regsub -all %W $comm $w.button$i.m.m comm

	    $w.button$i.m.m conf -postcommand $comm
	}

	

	CreateCascadedMenu $w.button$i.m.m $MenuEntriesName $MenuCommandsName \
		$MenuAccelerName $i $MenuLittleIcons
    }

    for {set i 0 } { $i< $NumMenus } { incr i } {
	if { $i < [expr $NumMenus-1] } {
	    pack $w.button$i -side left
	} else { 
	    pack $w.button$i -side right
	}
    }
    GidHelpT $w [_ "Menu with the main commands."]

    if { $infoProj != "POSTUSE" } {
	foreach name { "Files#C#menu" "View#C#menu" "Geometry#C#menu" \
	    "Utilities#C#menu" "Mesh#C#menu" "Calculate#C#menu"} \
	    comm { SetFilesMenuState SetViewMenuState SetGeometryMenuState \
		SetUtilitiesMenuState SetMeshMenuState SetCalculateMenuState } {
	    set i [GiDMenu::_FindIndex $name PRE]
	    if { $i != -1 } {
		if { [winfo exists $w.button$i.m.m] } {
		    $w.button$i.m.m configure -postcommand $comm
		}
	    }
	}
    }
}

proc CreateTopMenusWin { w } {
    global MenuNames MenuNamesP ViewMenuPriv

    if { ![winfo exists $w] } { return }

    set top [winfo toplevel $w]
    if { $top != ".gid" } { 
	set t $top 
	if { [.gid cget -menu] != "" } { 
	    .gid conf -menu "" 
	    destroy .mm1
	}
    } else { 
	set t "" 
    }

    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "GEOMETRYUSE"
    }

    if { $infoProj != "POSTUSE" } {
	set MenuNamesInto $MenuNames
	set MenuEntriesName MenuEntries
	set MenuCommandsName MenuCommands
	set MenuAccelerName MenuAcceler
	set MenuLittleIcons MenuIcons
    } else {
	set MenuNamesInto $MenuNamesP
	set MenuEntriesName MenuEntriesP
	set MenuCommandsName MenuCommandsP
	set MenuAccelerName MenuAccelerP
	set MenuLittleIcons MenuIconsP

	if { [ info exist ViewMenuPriv] } {
	    unset ViewMenuPriv
	}
    }

    set NumMenus [llength $MenuNamesInto]

    set UnderLineList ""

    if { ![winfo exists $t.mm1] } { 
	menu $t.mm1 -type menubar 
#        $top conf -menu $t.mm1
    } else {
	$t.mm1 del 0 end
    }
    $top conf -menu $t.mm1

    foreach i [winfo children $t.mm1] {
	destroy $i
    }   

    for {set i 0 } { $i< $NumMenus } { incr i } {

	set text [lindex $MenuNamesInto $i]
	set under [FindUnderChar text UnderLineList]
	set namemenu $t.mm1.button$i
	if { ![winfo exists $namemenu] } {
	    menu $namemenu
	} else { 
	    ClearMenu $namemenu
	    $namemenu configure -postcommand ""
	    #$namemenu del 0 end
	}
	$t.mm1 add cascade -label $text -underline $under -menu $namemenu

	
	upvar \#0 $MenuEntriesName ME
	upvar \#0 $MenuCommandsName MC
	upvar \#0 $MenuLittleIcons MI
	if { $ME($i) == "" } {
	    if { [lindex $MC($i) 0] != "-np-" } {
		WarnWin [_ "Error creating menu '%s'. Needed -np-" $text]
	    }
	    set comm [lrange $MC($i) 1 end]
	    regsub -all %W $comm $namemenu comm
	    $namemenu conf -postcommand $comm
	}
	
	CreateCascadedMenu $namemenu $MenuEntriesName $MenuCommandsName \
		$MenuAccelerName $i $MenuLittleIcons
    }

    # for debugging
    #InstrumentMenu $t.mm1

    if { $infoProj != "POSTUSE" } {
	set idata [GiDMenu::_FindIndex "Data" PRE]
	if { $idata != -1 } {
	    #force create Data menu (else first user click will be slow)
	    #create FastCacheFromProblemtypeMenu, read customized problemtype images for the 
	    #file browser (specified in .xml file), and create the menus
	    after idle "PostDataMenu $t.mm1.button$idata ; SetProblemTypeMenu $t.mm1.button$idata.0"
	}
    }

    $t.mm1 configure -postcommand SetMenusStates

}

proc CreateTopMenus { { w .gid.menu } } {
    global tcl_platform GidPriv

    #if { [winfo exists .gid.central.s] && [.gid.central.s disable windows] } return
    set GidPriv(POSTDATA) 1
    if { $tcl_platform(platform) == "windows" } {
	set GidPriv(MenuType) "Win"
	CreateTopMenusWin $w
	#set GidPriv(MenuType) "Unix"
	#CreateTopMenusUnix $w        
    } else {
	set GidPriv(MenuType) "Unix"
	CreateTopMenusUnix $w
    }

}

# for debugging
proc InstrumentCommand { label script } {
    set time [time {uplevel \#0 $script}]
    WarnWinText "$label $script $time"
}
proc InstrumentMenu { menu {label "" } } {

    set comm [$menu cget -postcommand]
    if { $comm != "" } {
	set comm [ list InstrumentCommand $label $comm]
	$menu configure -postcommand $comm
    }
    set ni [$menu index end]
    if { $ni == "none" } { set ni -1 }
    for { set i 0 } { $i <= $ni } { incr i } {
	if { [$menu type $i] == "cascade" } {
	    InstrumentMenu [$menu entrycget $i -menu] [$menu entrycget $i -label] 
	}
    }
}

proc CreateGlobalBinding { keys command } {

    regsub -all {Ctrl} $keys {Control} scbind
    regsub { } $scbind {><} scbind
    set scbind <$scbind>
    bind .gid $scbind $command
}

proc ViewGlobalBinding { keys } {
    regsub -all {Ctrl} $keys {Control} scbind
    regsub { } $scbind {><} scbind
    set scbind <$scbind>
    set oldcomm [bind .gid $scbind]
    return $oldcomm
}

proc pepe { msg } {
    set fo [ open /tmp/pp.pp a]
    puts $fo $msg
    close $fo
}

proc ChangeTopMenusAndWindows {} {
    set DisableGraphics [.central.s disable graphics]
    .central.s disable graphics 1
    CreateTopMenus

    if { [GiD_Info Project ViewMode] != "POSTUSE" } {
	OpenOutsideWindows PRE
	Create3rdButtonMenu .gid.phantom.display3dbutton

	# UpdateDataOptions

    } else {
	OpenOutsideWindows POST
	Create3rdButtonMenu .gid.phantom.display3dbutton POST
    }
    update idletasks
    .central.s disable graphics $DisableGraphics 
    GiD_Redraw
}

set DelayedConfigureButsVar 0
set DelayedConfigureSize "0 0"

proc DelayedConfigureButsFunc {} {
    global DelayedConfigureButsVar DelayedConfigureSize GidPriv

    incr DelayedConfigureButsVar -1
    if { $DelayedConfigureButsVar > 0 } {  return }
    if { ![winfo exists .gid.buts] } {
	set GidPriv(PrePostRightButWindowGeom) "NONE {} RightButtons"
    } elseif { [winfo class .gid.buts] == "Toplevel" } {
	set GidPriv(PrePostRightButWindowGeom) "OPEN [wm geom .gid.buts] RightButtons"
    } else {
	set GidPriv(PrePostRightButWindowGeom) "INSIDE {} RightButtons"
    }
    if { [winfo exists .gid.buts] } {
	set w [winfo width .gid.buts]
	set h [winfo height .gid.buts]
	if { $w != [lindex $DelayedConfigureSize 0] || \
		$h != [lindex $DelayedConfigureSize 1] } {
	    SetButtons
	    set DelayedConfigureButsVar 0
	    set DelayedConfigureSize "$w $h"
	}
    }
}

proc SetButtons { { first 0 } } {
    global ActionsMenuVar mxbutwidth PostButStack

    set w .gid.buts.f
#    if { [winfo class .gid.buts] == "Toplevel" } { append w .f }
    
    if { $first == 0 } { set PostButStack "" }
    if { $PostButStack != "" } {
###puts " . .. pack & conf"
	set num [lindex $PostButStack [expr [llength $PostButStack]-1]]
	pack $w.al -side left -padx 0
	$w.al conf -command "set PostButStack \"[lrange $PostButStack \
		0 [expr [llength $PostButStack]-2]]\" ; SetButtons $num"
    } else { pack forget $w.al }
    pack forget $w.ar

###puts " . .. places"

    set xpos 3 ; set ypos 5
    set butpady 2
###puts " . .. [winfo width $w] == 1 -> { update idletasks }"
    if { [winfo width $w] == 1 } { 
	#sobra?# update idletasks
    }
###puts " . .. sets"
    set tothei [expr [winfo height $w]-20]
    set totwid [winfo width $w]
    set overflow 0
    set last [llength $ActionsMenuVar]
    for {set k 0 } { $k<100 } { incr k } {
	if {$k < $first || $k>=$last || $overflow } {
###puts " . .. $k forget"
	    place forget $w.b$k
	} else {
###puts " . .. $k conf & place"
	    set name [lindex $ActionsMenuVar $k]
	    # traduccion de los botones de la derecha
	    # $w.b$k conf -text [_ $name] -command ".central.s processlist [list $name]"
	    $w.b$k conf -text $name -command ".central.s processlist [list $name]"
	    place $w.b$k -x $xpos -y $ypos
	    set reqh [ winfo reqheight $w.b$k]
	    set ypos [ expr $ypos + $reqh + $butpady]
	    #incr ypos 31
	    if { $ypos > $tothei && $k < [expr $last-1] } {
		incr xpos 130
		if { [expr $xpos+80] > $totwid } { 
		    pack $w.ar -side right -padx 0
		    $w.ar conf -command \
		            "lappend PostButStack $first ; SetButtons [expr $k+1]"
		    set overflow 1
		}
		set ypos 5
	    }
	}
    }
###puts " . .. FIN"
}

proc MenuInCommandButton { e x y } {
    global ActionsMenuVar

    set w .gid.menuincommandwindow
    catch { destroy $w }
    menu $w
    set words ""
    foreach i $ActionsMenuVar {
	regsub -all {([a-z])([A-Z])} $i "\\1 \[string tolower \\2]" name
	set name [subst $name]
	set name [string toupper [string index $name 0]][string range $name 1 end]
	if { $i == "Join" || $i == "NoJoin" } { append name " Ctrl-a" }
	lappend words [ list $name "$e insert end \" $i\""]
    }
    rmenuconf $w $words
    
    $w add separator
    $w add command -label [_ "Previous"] -command [bind $e <Up>]
    $w add command -label [_ "Next"] -command [bind $e <Down>]

    GiD_PopupMenu $w $x $y
}


proc AutoCompleteGiDEntry { e } {
    global ActionsMenuVar

    if { [string trim [$e get]] == "" } {
	if { [winfo exists .gid.wSetCoord.xe] } {
	    focus .gid.wSetCoord.xe
	}
	return
    }
    if { [catch {
	set beginwords [lrange [$e get] 0 end-1]
	set lastword [lindex [$e get] end]
    }] } { return }

    if { $beginwords != "" } {
	$e del 0 end
	$e ins end $beginwords
	uplevel #0 [bind $e <Return>]
	$e ins end $lastword
	after idle AutoCompleteGiDEntry $e
	return
    }

    set idx [lsearch -glob [string tolower $ActionsMenuVar] [string tolower ${lastword}*]]
    set word [lindex $ActionsMenuVar $idx]
    set wordoptions $word
    set isfull 1
    if { $idx != -1 } {
	set idx2 $idx
	while 1 {
	    set idx2new [lsearch -glob [string tolower [lrange $ActionsMenuVar [expr $idx2+1] end]] \
		          [string tolower ${lastword}*]]
	    if { $idx2new == -1 } { break }
	    set idx2 [expr $idx2+1+$idx2new]
	    set newword [lindex $ActionsMenuVar $idx2]
	    lappend wordoptions $newword
	    for { set i [string length $lastword] } { $i < [string length $word] } { incr i } {
		set l1 [string tolower [string index $word $i]]
		set l2 [string tolower [string index $newword $i]]
		if {  $l1 != $l2 } { break }
	    }
	    if { $i < [string length $word] } { set isfull 0 }
	    set word [string range $word 0 [expr $i-1]]
	}
	set text [lrange [$e get] 0 end-1]
	$e del 0 end
	if { $isfull } {
	    $e ins end "$word "
	} else {
	    $e ins end "$word"
	    ChangeWarnLineVar [_ "Options: %s" $wordoptions]
	}
	$e icursor end
	$e xview moveto 1
    }
}


proc SaveNewBitmapsGeom { Nombre bw StartFunc { prefix PrePost } } {
    global GidPriv

    regexp {[0-9]+x[0-9]+(.*)} [winfo geom $bw] trash rest
    bind $bw <Configure> ""
    
    set NumBitmaps 0
    while { [winfo exists $bw.$NumBitmaps] } { 
	incr NumBitmaps 
    }

    set w [winfo width $bw]
    set h [winfo height $bw]

    if { ![ info exists toolbarmacros::preferences(toolbar_one_col)]} {
	 toolbarmacros::ReadMacros
    }

    set w_width 0
    set w_height 0
    set sep_height 0
    set sep_width 0
    
    set term Separadores
    if { $w >= $h} {
	foreach icn $GidPriv($Nombre$term) {
	    $bw.$icn conf -image [GetImage sep_vert.gif]
	}
	set sep_width $GidPriv(BitmapsSepVWidth)
	set sep_height $GidPriv(BitmapsSepVHeight)

	# buscamos la maxima altura: $h ha de ser >= max_height
	set max_height 0
	set new_width 0
	for { set icn 0} { $icn < $NumBitmaps} { incr icn} {

	    if { ( "$Nombre" != "StdBar") || ( [ expr $icn + 1] != $NumBitmaps)} {
		set b_width [ expr $GidPriv(BitmapsWidth) + 2]
		set b_height [ expr $GidPriv(BitmapsHeight) + 2]
	    } else {
		# ancho y altura del bitmap info_gid en la barra estandard definidos en GrDataMenu.tcl
		# set GidPriv(InfoGiDBitmapWidth) 77
		# set GidPriv(InfoGiDBitmapHeight) 24
		set b_width [ expr $GidPriv(InfoGiDBitmapsWidth) + 2]
		set b_height [ expr $GidPriv(InfoGiDBitmapsHeight) + 2]
		set GidPriv(InfoGiDButtonName) $bw.$icn
		ShowGiDVersionType
	    }
	    set term Separadores
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		## las lineas de los botoncitos
		set b_width [ expr $sep_width + 2]
		set b_height [ expr $sep_height + 2]
	    }
	    set term MenusList
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		set b_width [ expr $b_width + 1]
		set b_height [ expr $b_height + 1]
	    }
	    if { $b_height > $max_height} {
		set max_height $b_height
	    }
	    set new_width [ expr $new_width + $b_width]
	}
	if { $h < $max_height} {
	    set h $max_height
	    set w $new_width
	}

	set row(0) ""
	set width(0) 0
	set height(0) 0
	set nrow 0
	#buscamos que dibujitos van horizontalmente
	for { set icn 0} { $icn < $NumBitmaps} { incr icn} {
	    ## las lineas de los botoncitos
	    if { ( "$Nombre" != "StdBar") || ( [ expr $icn + 1] != $NumBitmaps)} {
		set b_width [ expr $GidPriv(BitmapsWidth) + 2]
		set b_height [ expr $GidPriv(BitmapsHeight) + 2]
	    } else {
		# ancho y altura del bitmap info_gid en la barra estandard definidos en GrDataMenu.tcl
		# set GidPriv(InfoGiDBitmapWidth) 77
		# set GidPriv(InfoGiDBitmapHeight) 24
		set b_width [ expr $GidPriv(InfoGiDBitmapsWidth) + 2]
		set b_height [ expr $GidPriv(InfoGiDBitmapsHeight) + 2]
	    }
	    set term Separadores
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		## las lineas de los botoncitos
		set b_width [ expr $sep_width + 2]
		set b_height [ expr $sep_height + 2]
	    }
	    set term MenusList
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		set b_width [ expr $b_width + 1]
		set b_height [ expr $b_height + 1]
	    }
	    set width($nrow) [ expr $width($nrow) + $b_width]
	    if { $width($nrow) <= $w} {
		lappend row($nrow) $icn
		if { $b_height > $height($nrow)} {
		    set height($nrow) $b_height
		}
	    } else {
		if { $toolbarmacros::preferences(toolbar_one_col)} {
		    break
		}
		# no lo utilizaremos, pero por si acaso
		set width($nrow) [ expr $width($nrow) - $b_width]
		if { $width($nrow) > $w_width} {
		    set w_width $width($nrow)
		}
		set w_height [ expr $w_height + $height($nrow)]
		incr nrow
		set width($nrow) $b_width
		set row($nrow) $icn
		set height($nrow) $b_height
	    }
	}
	if { $width($nrow) > $w_width} {
	    set w_width $width($nrow)
	}
	set w_height [ expr $w_height + $height($nrow)]
	incr nrow
	
	#enpaquetamos
	eval [concat "grid forget [winfo children $bw]"]

	set ult 1

	for { set ir 0} { $ir < $nrow} { incr ir} {
	    set ic 0
	    foreach icn $row($ir) {
		if { "$Nombre" == "StdBar" } { #special tricks
		    if { $ult == 14 } {
		        grid $bw.$icn  -in $bw.f$ir -row 0 -column $ic -sticky w
		        #truco combobox layers
		        set viewmode [GiD_Info Project ViewMode]
		        if { $viewmode == "GEOMETRYUSE" || $viewmode == "MESHUSE" } {
		            if { ![winfo exists $bw.flayer] } {
		                frame $bw.flayer
		                CreateLayerCombo $bw.flayer
		            }
		            incr ic
		            grid $bw.flayer -in $bw.f$ir -row 0 -column $ic -sticky w
		        }
		        #
		    } elseif { $ult == $NumBitmaps } {
		        grid $bw.$icn -in $bw.f$ir -row 0 -column $ic -sticky e
		        grid columnconfig $bw.f$ir $ic -weight 1
		    } else {
		        grid $bw.$icn -in $bw.f$ir -row 0 -column $ic -sticky w
		    }
		} else {
		    grid $bw.$icn -in $bw.f$ir -row 0 -column $ic -sticky w
		}
		incr ic
		if { [ winfo class $bw.$icn] == "Label"} {
		    grid configure $bw.$icn -pady 2
		}
		incr ult
	    }
	    grid rowconfig $bw.f$ir 0 -weight 1
	    if { ( "$Nombre" != "StdBar") || ( [ expr $ir + 1] != $nrow)} {
		grid $bw.f$ir -sticky nw
	    } else {
		grid $bw.f$ir -sticky new
	    }
	    grid rowconfig $bw $ir -weight 1
	    if { ( "$Nombre" == "StdBar") && ( $ult >= $NumBitmaps)} {
	    }
	}
	grid columnconfig $bw 0 -weight 1


    } else {
	if { "$Nombre" == "StdBar"} {
	    set NumBitmaps [ expr $NumBitmaps - 1]
	}
	# es mas vertical que horizontal ( enpaquetmos de arriba a abajo)
	set term Separadores
	foreach icn $GidPriv($Nombre$term) {
	    $bw.$icn conf -image [GetImage sep_hori.gif]
	}
	set sep_width $GidPriv(BitmapsSepHWidth)
	set sep_height $GidPriv(BitmapsSepHHeight)

	# buscamos la maxima anchura: $w ha de ser >= max_width
	set max_width 0
	set new_height 0
	for { set icn 0} { $icn < $NumBitmaps} { incr icn} {
	    set b_width [ expr $GidPriv(BitmapsWidth) + 2]
	    set b_height [ expr $GidPriv(BitmapsHeight) + 2]
	    set term Separadores
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		## las lineas de los botoncitos
		set b_width [ expr $sep_width + 2]
		set b_height [ expr $sep_height + 2]
	    }
	    set term MenusList
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		set b_width [ expr $b_width + 1]
		set b_height [ expr $b_height + 1]
	    }
	    if { $b_width > $max_width} {
		set max_width $b_width
	    }
	    set new_height [ expr $new_height + $b_height]
	}
	if { $w < $max_width} {
	    set w $max_width
	    set h $new_height
	}
	
	set col(0) ""
	set width(0) 0
	set height(0) 0
	set ncol 0     
	#buscamos que dibujitos van horizontalmente
	for { set icn 0} { $icn < $NumBitmaps} { incr icn} {
	    ## las lineas de los botoncitos
	    set b_width [ expr $GidPriv(BitmapsWidth) + 1]
	    set b_height [ expr $GidPriv(BitmapsHeight) + 1]
	    set term Separadores
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		## las lineas de los botoncitos
		# las barras tiene un pady de 2
		set b_width [ expr $sep_width + 1]
		set b_height [ expr $sep_height + 1]
	    }
	    set term MenusList
	    if { [ lsearch $GidPriv($Nombre$term) $icn] != -1} {
		set b_width [ expr $b_width + 1]
		set b_height [ expr $b_height + 1]
	    }

	    set height($ncol) [ expr $height($ncol) + $b_height]
	    if { $height($ncol) <= $h} {
		lappend col($ncol) $icn
		if { $b_width > $width($ncol)} {
		    set width($ncol) $b_width
		}
	    } else {
		if { $toolbarmacros::preferences(toolbar_one_col)} {
		    break
		}
		# no lo utilizaremos, pero por si acaso
		set height($ncol) [ expr $height($ncol) - $b_height]
		if { $height($ncol) > $w_height} {
		    set w_height $height($ncol)
		}
		set w_width [ expr $w_width + $width($ncol)]
		incr ncol
		set height($ncol) $b_height
		set col($ncol) $icn
		set width($ncol) $b_width
	    }
	}
	if { $height($ncol) > $w_height} {
	    set w_height $height($ncol)
	}
	set w_width [ expr $w_width + $width($ncol)]
	if { $ncol <= $NumBitmaps} {
	    incr ncol
	}
	
	#enpaquetamos
	eval [concat "grid forget [winfo children $bw]"]

	for { set ic 0} { $ic < $ncol} { incr ic} {
	    #WarnWinText "ic = $ic / $ncol"
	    #catch { destroy $bw.f$ir}
	    #frame $bw.f$ir -bd 0
	    set ir 0
	    foreach icn $col($ic) {
		#WarnWinText "Using $bw.$icn -in $bw.f$ic -sticky w"
		grid $bw.$icn -in $bw.f$ic -column 0 -row $ir -sticky nw
		incr ir
		grid rowconfig $bw.f$ic $ir -weight 1
		if { [ winfo class $bw.$icn] == "Label"} {
		    grid configure $bw.$icn -padx 3
		}
	    }
	    grid columnconfig $bw.f$ic 0 -weight 1
	    grid $bw.f$ic -sticky nw -column $ic -row 0
	    grid columnconfig $bw $ic -weight 1
	}

	grid rowconfig $bw 0 -weight 1

    }

    if { [winfo class $bw] == "Toplevel" } {

	if { "$Nombre" == "StdBar"} {
	    incr w_width [winfo width $GidPriv(ComboLayers,entry)]
	}

	wm geom $bw ${w_width}x${w_height}$rest
	# OJO!!!!!  a nadie se le ocurra quitar este 'update', sin el
	# un segfault es generado en Tk_MapWindow mientras se hace un
	# resize de la ventana.
	update
	if { [wm transient $bw] != "" } {
	    set trans 1
	} else {
	    set trans 0
	}
	set term WindowGeom
	set GidPriv($prefix$Nombre$term) \
		"OPEN [wm geom $bw] $trans $StartFunc"
    }
    
    bind $bw <Configure> [ list SaveNewBitmapsGeom $Nombre $bw $StartFunc]
}

#
# Destroy a given widget. If the parent container remains empty it is
# also "grid_removed". Later on the parent container could be
# remembered if needed.
# 
proc DestroySlave { w } {
  catch {array set grid_info [grid info $w]}
  destroy $w
  if {[info exists grid_info(-in)]} {
    # if w is gridded its parent must be a member of GiDTBLayout
    set parent $grid_info(-in)
    set idx [lsearch $::GiDTBLayout($parent) $w]
    
    if {![llength [set ::GiDTBLayout($parent) \
		       [lreplace $::GiDTBLayout($parent) $idx $idx]]]} {
      grid remove $parent
    }
  }
}

# GridInside --
#
# gid a widget w inside a master, making sure master is remembered in
# case it was previouly removed. master must be prevously gridded with
# initial options.
#
proc GridInside { name w master sticky } {
    grid $master
    set slaves $::GiDTBLayout($master)
    if {$sticky eq "ns" || $sticky eq "sn"} {
	set c 0
	foreach s $slaves {
	    if {[winfo exists $s]} {
		grid $s -in $master -row 0 -column $c -sticky $sticky
		incr c
	    }
	}
	grid $w -in $master -row 0 -column $c -sticky $sticky
    } else {
	set r 0
	foreach s $slaves {
	    if {[winfo exists $s]} {
		grid $s -in $master -row $r -column 0 -sticky $sticky
		incr r
	    }
	}
	grid $w -in $master -row $r -column 0 -sticky $sticky
    }
    lappend ::GiDTBLayout($master) $w
    set ::GiDTBLayout(tb,$w) $name
}

proc CreateBitmaps { { what "DEFAULT" } } {

    global BitMapsNames BitMapsCommands GIDDEFAULT BitmapsHelp
    global BitMapsNamesP BitMapsCommandsP GIDDEFAULT BitmapsHelpP
    global GidPriv GiDPrivToolBar

    set w .gid.bitmaps
    set title [_ "Geometry & View bar"]
    AddNewToolbar "Geometry & View bar" PrePostBitmapsWindowGeom CreateBitmaps $title

    set vv $GidPriv(PrePostBitmapsWindowGeom)
    # esta parte intenta corregir algunos gidDefaults corruptos
    if {[lindex $vv end] ne "CreateBitmaps"} {
	set ll [llength $vv]
	set vv [concat $vv  [lrange {INSIDELEFT 1 1 CreateBitmaps} $ll end]]
	set GidPriv(PrePostBitmapsWindowGeom) $vv
    }
    if { $what != "DEFAULT" } {
	set GidPriv(PrePostBitmapsWindowGeom) \
		"$what [lrange $vv 1 end]"
    }
    DestroySlave $w
    if { [lindex $GidPriv(PrePostBitmapsWindowGeom) 0] == "NONE" } {
      return
    }

    set separadores ""
    set num_separadores  0
    set round_bar 0

    if { ![string match INSIDE* [lindex $GidPriv(PrePostBitmapsWindowGeom) 0]] } {
	InitWindow $w [_ "Geom & view operations"] PrePostBitmapsWindowGeom CreateBitmaps
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0

	wm transient $w .gid
	catch { wm attributes $w -toolwindow 1 }
	$w config -bg $GidPriv(Color,ToolbarBg)

	set bitmap_geom "[ winfo width $w]x[ winfo height $w]"
	#puts $bitmap_geom
	if { [ winfo width $w] > [ winfo height $w]} {
	    set separadores V
	    if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
		set round_bar 1
	    }
	} else {
	    set separadores H
	}
    } else {
	set vv $GidPriv(PrePostBitmapsWindowGeom)
	set GidPriv(PrePostBitmapsWindowGeom) $vv
	set GiDPrivToolBar(PrePostBitmapsWindowGeom) $w

	frame $w -bd 0 -bg $GidPriv(Color,ToolbarBg)
	bind $w <Destroy> "+ unset GiDPrivToolBar(PrePostBitmapsWindowGeom)"

#        if { [winfo exists .gid.central.wins] } {
#            pack forget .gid.central.wins
#        } else {
#            pack forget .gid.central.s
#        }


	regexp {([a-zA-Z]*)([0-9]*)} [lindex $GidPriv(PrePostBitmapsWindowGeom) 0] {} \
		where pos
	set widget ""
#        if { $pos != "" } {
#            set widget [lindex [pack slaves .gid.central] $pos]
#        }
	switch $where {
	    INSIDETOP {
#                pack $w -in .gid.central -side top -anchor nw \
#                        -fill both -expand 0 -padx 0 -pady 0
	      GridInside PrePostBitmapsWindowGeom $w .gid.top "we"

#                if { [winfo exists .gid.central.wins] } {
#                    pack .gid.central.wins -expand 1 -fill both -side top
#                } else {
#                    pack .gid.central.s -fill both -expand yes -side top
#                }

		set separadores V
		if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
		    set round_bar 1
		}
	    }
	    INSIDELEFT {
#                 pack $w -in .gid.central -side left -anchor nw \
#                         -fill both -expand 0 -padx 0 -pady 0
#                 if { [winfo exists .gid.central.wins] } {
#                     pack .gid.central.wins -expand 1 -fill both -side left
#                 } else {
#                     pack .gid.central.s -fill both -expand yes -side left
#                 }
	      GridInside PrePostBitmapsWindowGeom .gid.bitmaps .gid.left "ns"

		set separadores H
	    }
	    INSIDEBOTTOM {
#                 pack $w -in .gid.central -side bottom -anchor sw \
#                         -fill both -expand 0 -padx 0 -pady 0
#                 if { [winfo exists .gid.central.wins] } {
#                     pack .gid.central.wins -expand 1 -fill both -side bottom
#                 } else {
#                     pack .gid.central.s -fill both -expand yes -side bottom
#                 }
#                 pack .gid.central.s -fill both -expand yes -side bottom
	      GridInside PrePostBitmapsWindowGeom .gid.bitmaps .gid.bottom "we"
	      set separadores V
		if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
		    set round_bar 1
		}
	    }
	    INSIDERIGHT {
#                 pack $w -in .gid.central -side right -anchor ne \
#                         -fill both -expand 0 -padx 0 -pady 0
#                 if { [winfo exists .gid.central.wins] } {
#                     pack .gid.central.wins -expand 1 -fill both -side right
#                 } else {
#                     pack .gid.central.s -fill both -expand yes -side right
#                 }
	      GridInside PrePostBitmapsWindowGeom .gid.bitmaps .gid.right "ns"
		set separadores H
	    }
	}
#         if { $widget != ""} {
#             pack configure $w -before $widget
#         }
    }

    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "GEOMETRYUSE"
    }
    
    if { $infoProj != "POSTUSE" } {
	upvar #0 BitMapsNames names
	upvar #0 BitMapsCommands commands
	upvar #0 BitmapsHelp thehelp
    } else {
	upvar #0 BitMapsNamesP names
	upvar #0 BitMapsCommandsP commands
	upvar #0 BitmapsHelpP thehelp
    }

    set i 0
    set GidPriv(BitmapMenus) 0
    set GidPriv(BitmapsSeparadores) ""
    set GidPriv(BitmapsMenusList) ""
    set reqmbw 0
    set reqmbh 0
    set reqmbbw 0
    set reqbw 0
    set reqbh 0
    set lst_botones ""
    foreach j $names(0) {
	# para rehacer los bitmaps con separadores
	frame $w.f$i -bd 0 -bg $GidPriv(Color,ToolbarBg)
	if { $round_bar} {
	    $w.f$i conf -relief round_raised
	}

	if { "$j" != "---" } {
	    if { ![info exists names(0,$i)] } {
		button $w.$i -image [GetImage $j] -bd 0 -highlightthickness 0 \
		    -relief flat -bg $GidPriv(Color,ToolbarBg) -width 24 -height 24
	
		global tcl_platform
		if { $tcl_platform(platform) == "windows" } {
		    bind $w.$i <ButtonRelease-1> "catch \"focus .gid.comm.e\" "
		}
		set comm [lindex $commands(0) $i]
		EnterCommand $w.$i $comm
		set reqbw [ winfo reqwidth $w.$i]
		set reqbh [ winfo reqheight $w.$i]
		lappend lst_botones $w.$i
		
	    } else {
		menubutton $w.$i \
		    -image [GetImage $j] \
		    -menu $w.$i.m -relief flat -bd 0 \
		    -highlightthickness 0 -direction right -bg $GidPriv(Color,ToolbarBg) \
		    -width 24 -height 24
		
		menu $w.$i.m -tearoff no -bg $GidPriv(Color,ToolbarBg) \
		    -bd 1 -relief raised -activeborderwidth 0
		
		
		set comm [lindex $commands(0) $i]
		if { $comm != "" } { 
		    EnterCommand $w.$i.m $comm -2
		}
		set ii 0
		foreach jj $names(0,$i) {
		    $w.$i.m add command -image [GetImage $jj] -hidemargin 1
		    EnterCommand $w.$i.m [lindex $commands(0,$i) $ii] $i
		    incr ii
		}
		incr GidPriv(BitmapMenus)
		set reqmbw [winfo reqwidth $w.$i]
		set reqmbh [winfo reqheight $w.$i]
		set reqmbbw [$w.$i cget -borderwidth]
	    }
	    #nos pone un borde uno en los menubuttons que queda feo
	    lappend GidPriv(BitmapsMenusList) $i
	    
	    set help [lindex $thehelp(0) $i]
	    if { $round_bar} {
		bind $w.$i <ButtonPress-1> "$w.$i conf -relief round_sunken"
		bind $w.$i <Enter> "catch { wm title $w \"$help\" } ; \
		    $w.$i conf -relief round_raised"
		bind $w.$i <Leave> "$w.$i conf -relief round_raised"
		bind $w.$i <ButtonRelease-1> "$w.$i conf -relief round_raised"
	    } else {
		bind $w.$i <ButtonPress-1> "$w.$i conf -relief flat"
		#bind $w.$i <Enter> [ list catch { wm title $w $help } ]
		bind $w.$i <Enter> "catch { wm title $w \"$help\" } ; \
		    $w.$i conf -relief raised"
		bind $w.$i <Leave> "$w.$i conf -relief flat"
		bind $w.$i <ButtonRelease-1> "$w.$i conf -relief raised"
	    }

	    ChangeToNewLook $w.$i

	    GidHelpT $w.$i $help

	} else {
	    if { "$separadores" != "" } {
		if { "$separadores" == "H" } {
		    set archivo sep_hori.gif
		} else {
		    set archivo sep_vert.gif
		}
		
		label $w.$i -image [GetImage $archivo] -bd 0 -highlightthickness 0\
		        -relief flat -bg [$w cget -background]
		incr num_separadores
		lappend GidPriv(BitmapsSeparadores) $i
	    } else {
		label $w.$i -height 0 -width 0 -bd 0 -highlightthickness 0\
		        -relief flat -bg [$w cget -background] -padx 0 -pady 0
	    }
	}
	if { $round_bar} {
	    $w.$i conf -relief round_raised
	}
	#sep# bind $w.$i <ButtonPress-1> "$w.$i conf -relief flat"
	#sep# bind $w.$i <Enter> "catch { wm title $w $help } ; \
	#sep#           $w.$i conf -relief raised"
	#sep# bind $w.$i <Leave> "$w.$i conf -relief flat"

	#set help [lindex $thehelp(0) $i]
	#GidHelp $w.$i $help
#         bind $w.$i <ButtonPress-2> [ list DisplayHelp $help %W %x %y]
#         bind $w.$i <ButtonPress-3> [ list DisplayHelp $help %W %x %y]
	incr i
    }

    #WarnWin "B: ${reqbw}x${reqbh} <-> MB: ${reqmbw}x${reqmbh}"
    # quitamos los bordes
    set reqmbw [ expr $reqmbw - 2 * $reqmbbw]
    set reqmbh [ expr $reqmbh - 2 * $reqmbbw]
    foreach bb $lst_botones {
	$bb conf -width $reqmbw -height $reqmbh
    }

    set GidPriv(BitmapsNumSeparadores) $num_separadores
    bind $w <Configure> "SaveNewBitmapsGeom Bitmaps $w CreateBitmaps"
    if { [winfo class $w] == "Toplevel" } {
	bind $w <Leave> [ list wm title $w "Geom & view operations"]
	
	foreach i [bind .gid] {
	    bind $w $i [bind .gid $i]
	}
    }
    
    #GidHelpT $w $title
}



# -np- WarnWin [ winfo reqheight .gid.bitmaps.4]
# -np- WarnWin [ winfo reqheight .gid.bitmapsStdBar.4]
# -np- WarnWin [ winfo reqheight .gid.bitmapsMacrosToolbar.w4]

#Procedure to create an additional toolbar
#Nombre: internal toolbar name "<name> bar" (must finish with 'bar', and a unique name)
#      this name must be the same used by AddNewToolbar and ReleaseToolbar 
#title: a text (marked for translation)
#BitMapsNames: name of a global variable that store a list of filenames with the button images (.gif)
#BitMapsCommands: name of a global variable that store a list of button commands
#BitmapsHelp: name of a global variable that store a list of button commands help strings (translated)
#BitmapsPath: the full path to the images 
#StartFunc: the proc name the create the bar calling CreateOtherBitmaps and AddToolbar
#type: bar location: "DEFAULT ?INSIDELEFT|INSIDERIGHT|INSIDETOP|INSIDEBOTTOM|OUTSIDE?"
#      e.g. "DEFAULT INSIDELEFT"
#prefix: where to show bar: Pre|Post|PrePost

proc CreateOtherBitmaps { Nombre title BitMapsNames BitMapsCommands BitmapsHelp \
		              BitmapsPath StartFunc { what "DEFAULT" } { prefix PrePost} } {
    global GidPriv GiDPrivToolBar GIDDEFAULT

    set term WindowGeom
    set w .gid.bitmaps
    DestroySlave $w$Nombre
    if { [ info exist GidPriv($prefix$Nombre$term)] } {
	set vv $GidPriv($prefix$Nombre$term)
	set _head_ [lrange $vv 0 3]
	set _ll_ [llength $vv]
	if {$_ll_ == 1} {
	    lappend _head_ {} 1
	} elseif {$_ll_ == 2} {
	    lappend _head_ 1
	}
	if { [lindex $what 0] != "DEFAULT" } {
	    set GidPriv($prefix$Nombre$term) "$what [lrange $_head_ 1 2] $StartFunc"
	} else {
	    set GidPriv($prefix$Nombre$term) "[lrange $_head_ 0 2] $StartFunc"
	}
	if { [lindex $GidPriv($prefix$Nombre$term) 0] == "NONE" } {
	    return ""
	}
    } elseif { [lindex $what 0] == "DEFAULT" && [lindex $what 1] != "" } {
	set GidPriv($prefix$Nombre$term) "[lindex $what 1] {} 1 $StartFunc"
    }
    # y en este else que puede pasar ?

    if { [GiD_Info Project ViewMode] != "POSTUSE" } {
	set what PRE
    } else { set what POST }
    
    # busco si hay otra TB previa con este proc asociado, si existe le
    # dejo solo la posicion.
    set _this_ $prefix$Nombre$term
    set _this_cmd_ [lindex $StartFunc 0]
    foreach _tb_ "$::GiDTBLayout(INSIDE,PRE) $::GiDTBLayout(INSIDE,POST)" {
	if {$_tb_ eq $_this_} continue
	set _cmd_ [lindex [lindex $GidPriv($_tb_) 3] 0]
	if {$_cmd_ eq $_this_cmd_} {
	    # esta es una TB que se crea con mi mismo proc, por tanto
	    # solo le dejo la posicion para que no me moleste.
	    set GidPriv($_tb_) [lindex $GidPriv($_tb_) 0]
	}
    }

    if { $prefix != "PrePost" } {
	if { ($what == "POST" && ![string match *Post* $prefix]) || \
		($what == "PRE" && [string match *Post* $prefix]) } {
	    if { [info exists GidPriv($prefix$Nombre$term)] } {
		set aa [lindex $GidPriv($prefix$Nombre$term) 0]
	    }
#             catch {
#                 DestroySlave $w$Nombre
#             }
	    if { [info exists GidPriv($prefix$Nombre$term)] } {
		set GidPriv($prefix$Nombre$term) \
		        "$aa [lrange $GidPriv($prefix$Nombre$term) 1 end]"
	    }
	    return
	}
    }


    set separadores ""
    set num_separadores  0
    set round_bar 0

    if { ![ info exist GidPriv($prefix$Nombre$term)] || \
	    ![string match INSIDE* [lindex $GidPriv($prefix$Nombre$term) 0]] } {
	InitWindow $w$Nombre $title $prefix$Nombre$term $StartFunc
	if { ![winfo exists $w$Nombre] } return ;# windows disabled || usemorewindows == 0

	wm transient $w$Nombre .gid
	catch { wm attributes $w$Nombre -toolwindow 1 }
	$w$Nombre config -bg $GidPriv(Color,ToolbarBg)
	set bitmap_geom "[ winfo width $w$Nombre]x[ winfo height $w$Nombre]"
	#puts $bitmap_geom
	if { [ winfo width $w$Nombre] > [ winfo height $w$Nombre]} {
	    set separadores V
	    if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
		set round_bar 1
	    }
	} else {
	    set separadores H
	}
    } else {
	set vv $GidPriv($prefix$Nombre$term)
	set GidPriv($prefix$Nombre$term) $vv
	set GiDPrivToolBar($prefix$Nombre$term) $w$Nombre

	set GidPriv($prefix${Nombre}Window) $w$Nombre

	frame $w$Nombre -bd 0 -bg $GidPriv(Color,ToolbarBg) 
	# -bd 2 -relief round_raised

	bind $w$Nombre <Destroy> "+ unset GiDPrivToolBar($prefix$Nombre$term)"

#         if { [winfo exists .gid.central.wins] } {
#             pack forget .gid.central.wins
#         } else {
#             pack forget .gid.central.s
#         }

	regexp {([a-zA-Z]*)([0-9]*)} [lindex $GidPriv($prefix$Nombre$term) 0] {} \
		where pos
	set widget ""
#         if { $pos != "" } {
#             set widget [lindex [pack slaves .gid.central] $pos]
#         }
	 switch $where {
	    INSIDETOP {
#               set slaves [pack slaves .gid.central]
#               foreach slave  $slaves {
#                 set pack_info($slave) [pack info $slave]
#               }
#               foreach slave $slaves {
#                 pack forget $slave
#               }
	      
#               pack $w$Nombre -in .gid.central -side top -anchor nw \
#                   -fill both -expand 0 -padx 0 -pady 0
#               foreach slave $slaves {
#                 eval pack $slave $pack_info($slave)
#               }
#               if { [winfo exists .gid.central.wins] } {
#                 pack .gid.central.wins -expand 1 -fill both -side top
#               } else {
#                 pack .gid.central.s -fill both -expand yes -side top
#               }
	      GridInside $prefix$Nombre$term $w$Nombre .gid.top "we" 
	      set separadores V
	      if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
		set round_bar 1
	      }
	    }
	    INSIDELEFT {
#                 pack $w$Nombre -in .gid.central -side left -anchor nw \
#                         -fill both -expand 0 -padx 0 -pady 0
#                 if { [winfo exists .gid.central.wins] } {
#                     pack .gid.central.wins -expand 1 -fill both -side left
#                 } else {
#                     pack .gid.central.s -fill both -expand yes -side left
#                 }
	      GridInside $prefix$Nombre$term $w$Nombre .gid.left "ns"
		set separadores H
	    }
	    INSIDEBOTTOM {
#                 pack $w$Nombre -in .gid.central -side bottom -anchor sw \
#                         -fill both -expand 0 -padx 0 -pady 0
#                 if { [winfo exists .gid.central.wins] } {
#                     pack .gid.central.wins -expand 1 -fill both -side bottom
#                 } else {
#                     pack .gid.central.s -fill both -expand yes -side bottom
#                 }
	      GridInside $prefix$Nombre$term $w$Nombre .gid.bottom "we"
		set separadores V
		if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
		    set round_bar 1
		}
	    }
	    INSIDERIGHT {
#                 pack $w$Nombre -in .gid.central -side right -anchor ne \
#                         -fill both -expand 0 -padx 0 -pady 0
#                 if { [winfo exists .gid.central.wins] } {
#                     pack .gid.central.wins -expand 1 -fill both -side right
#                 } else {
#                     pack .gid.central.s -fill both -expand yes -side right
#                 }
	      GridInside $prefix$Nombre$term $w$Nombre .gid.right "ns"
		set separadores H
	    }

	}
#         if { $widget != ""} {
#             pack configure $w$Nombre -before $widget
#         }
    }

    upvar #0 $BitMapsNames names
    upvar #0 $BitMapsCommands commands
    upvar #0 $BitmapsHelp thehelp

    if { $round_bar} {
	$w$Nombre conf -relief round_raised        
    }


    #already not set problemtype to use: [lindex [GiD_Info project] 0]
    if { $BitmapsPath == "$GIDDEFAULT/bitmaps" } {
	set ptpref ""
    } elseif { [regexp {problemtypes[/|\\](.*)} $BitmapsPath dummy problemtype] } {
	set ptpref "$problemtype/"
    } else {
	set ptpref "$BitmapsPath/"
    }

    set i 0
    #set GidPriv(BitmapMenus) 0
    set term_sep Separadores
    set term_ml  MenusList
    set GidPriv($Nombre$term_sep) ""
    set GidPriv($Nombre$term_ml) ""
    foreach j $names(0) {
	# para rehacer los bitmaps con separadores
	frame $w$Nombre.f$i -bd 0 -bg $GidPriv(Color,ToolbarBg)
	if { $round_bar} {
	    $w$Nombre.f$i conf -relief round_raised
	}

	if { "$j" != "---" } {
	    #warning: not use GetImage because image must be outside /bitmaps
	    set err_img [ catch {
		if { ![info exist GidPriv(image_$ptpref$j)] } {
		    set GidPriv(image_$ptpref$j) [ image create photo -file \
		                                       [file join $BitmapsPath $j]]
		}
	    }]
	    if { ![info exists names(0,$i)] } {
		button $w$Nombre.$i -image  $GidPriv(image_$ptpref$j) \
		    -bd 0 -highlightthickness 0\
		    -relief flat -bg [$w$Nombre cget -background]
		global tcl_platform
		if { $tcl_platform(platform) == "windows" } {
		    bind $w$Nombre.$i <ButtonRelease-1> "catch \"focus .gid.comm.e\" "
		}
		set comm [lindex $commands(0) $i]
		EnterCommand $w$Nombre.$i $comm
		
	    } else {
		menubutton $w$Nombre.$i \
		    -image  $GidPriv(image_$ptpref$j) \
		    -menu $w$Nombre.$i.m -relief flat -bd 0 \
		    -highlightthickness 0 -direction right -bg $GidPriv(Color,ToolbarBg)
		menu $w$Nombre.$i.m -tearoff no -bg $GidPriv(Color,ToolbarBg) \
		    -bd 1 -relief raised -activeborderwidth 0
		set comm [lindex $commands(0) $i]
		if { $comm != "" } { EnterCommand $w$Nombre.$i.m $comm -2}
		set ii 0
		foreach jj $names(0,$i) {
		    set err_img [ catch {
		        if { ![info exist GidPriv(image_$ptpref$jj)] } {
		            set GidPriv(image_$ptpref$jj) [ image create photo -file \
		                                                [file join $BitmapsPath $jj]]
		        }
		    }]
		    $w$Nombre.$i.m add command -image $GidPriv(image_$ptpref$jj) -hidemargin 1
		    EnterCommand $w$Nombre.$i.m [lindex $commands(0,$i) $ii] $ii
		    incr ii
		}
		#incr GidPriv(BitmapMenus)
	    }
	    #nos pone un borde uno en los menubuttons que queda feo
	    lappend GidPriv($Nombre$term_ml) $i
	    
	    set help [lindex $thehelp(0) $i]
	    
	    #bind $w$Nombre.$i <Enter> [ list catch { wm title $w$Nombre $help } ]
	    if { $round_bar} {
		bind $w$Nombre.$i <ButtonPress-1> "$w$Nombre.$i conf -relief round_sunken"
		bind $w$Nombre.$i <Enter> "catch { wm title $w$Nombre \"$help\" }"
		bind $w$Nombre.$i <ButtonRelease-1> "$w$Nombre.$i conf -relief round_raised"
	    } else {
		bind $w$Nombre.$i <ButtonPress-1> "$w$Nombre.$i conf -relief flat"
		bind $w$Nombre.$i <Enter> "catch { wm title $w$Nombre \"$help\" }"
		bind $w$Nombre.$i <ButtonRelease-1> "$w$Nombre.$i conf -relief raised"
	    }
	    
	    ChangeToNewLook $w$Nombre.$i
	    
	    GidHelpT $w$Nombre.$i $help
	    
	} else {
	    if { "$separadores" != "" } {
		if { "$separadores" == "H" } {
		    set archivo sep_hori.gif
		} else {
		    set archivo sep_vert.gif
		}
		
		label $w$Nombre.$i -image [GetImage $archivo] -bd 0 -highlightthickness 0\
		        -relief flat -bg [ $w$Nombre cget -background]
		incr num_separadores
		lappend GidPriv($Nombre$term_sep) $i
	    } else {
		label $w$Nombre.$i -height 0 -width 0 -bd 0 -highlightthickness 0\
		        -relief flat -bg [ $w$Nombre cget -background] -padx 0 -pady 0
	    }
	}
	if { $round_bar} {
	    $w$Nombre.$i conf -relief round_raised
	}
	#sep# bind $w$Nombre.$i <ButtonPress-1> "$w$Nombre.$i conf -relief flat"
	#sep# bind $w$Nombre.$i <Enter> "catch { wm title $w$Nombre $help } ; \
	#sep#           $w$Nombre.$i conf -relief raised"
	#sep# bind $w$Nombre.$i <Leave> "$w$Nombre.$i conf -relief flat"

	#set help [lindex $thehelp(0) $i]
	#GidHelp $w$Nombre.$i $help
#         bind $w$Nombre.$i <ButtonPress-2> [ list DisplayHelp $help %W %x %y]
#         bind $w$Nombre.$i <ButtonPress-3> [ list DisplayHelp $help %W %x %y]
	incr i
    }
  
    #set GidPriv(BitmapsNumSeparadores) $num_separadores
    bind $w$Nombre <Configure> [ list SaveNewBitmapsGeom $Nombre \
	    $w$Nombre $StartFunc $prefix]
    if { [winfo class $w$Nombre] == "Toplevel" } {
	bind $w$Nombre <Leave> [ list wm title $w$Nombre \
		"$title"]
	
	foreach i [bind .gid] {
	    bind $w$Nombre $i [bind .gid $i]
	}
    }
   
    #GidHelpT $w$Nombre $title
    return $w$Nombre
}

proc SetAllLatersTo { onoff } {

    .central.s process layers 
    if { $onoff == "ON" } {
	foreach i [GiD_Info layers -off] {
	    .central.s process on $i
	}
    } else {
	foreach i [GiD_Info layers -on] {
	    .central.s process off $i
	}
    }
    .central.s process escape
    FillInfoLayers
}

proc AddLayerOnOff { onoff w basecommand } {
    global GidPriv
    $w del 0 end
    if { $onoff == "ON" } {
	set words [ list [ list "-All-" "SetAllLatersTo $onoff" ]]
	set LayerNames [GiD_Info layers -off]
	if { [info exists GidPriv(LayersAlphabetic)] && $GidPriv(LayersAlphabetic) } {
	    set LayerNames [lsort -dictionary $LayerNames]
	}

	set ic 0
	foreach i $LayerNames {
	    lappend words [ list $i "$basecommand $i escape ; FillInfoLayers"]
	    incr ic
	    if { $ic > 6 } {
		lappend words [ list "..." ChangeLayers]
		break
	    }
	}
	rmenuconf $w $words
    } elseif { $onoff == "OFF" } {
	set words [ list [ list "-All-" "SetAllLatersTo $onoff" ]]
	set LayerNames [GiD_Info layers -on]
	if { [info exists GidPriv(LayersAlphabetic)] && $GidPriv(LayersAlphabetic) } {
	    set LayerNames [lsort -dictionary $LayerNames]
	}
	set ic 0
	foreach i $LayerNames {
	    lappend words [ list $i "$basecommand $i escape ; FillInfoLayers"]
	    incr ic
	    if { $ic > 6 } {
		lappend words [ list "..." ChangeLayers]
		break
	    }
	}
	rmenuconf $w $words
    } else {
	set layertouse [GiD_Info Project LayerToUse]
	set LayerNames [GiD_Info layers]
	if { [info exists GidPriv(LayersAlphabetic)] && $GidPriv(LayersAlphabetic) } {
	    set LayerNames [lsort -dictionary $LayerNames]
	}
	set words ""
	set ic 0
	foreach i $LayerNames {
	    if { $i != $layertouse } {
		lappend words [ list $i "$basecommand $i escape ; FillInfoLayers"]
	    } else {
		lappend words [ list $i "$basecommand $i escape ; FillInfoLayers" check]
	    }
	    incr ic
	    if { $ic > 6 } {
		lappend words [ list "..." ChangeLayers]
		break
	    }
	}
	rmenuconf $w $words
    }
}

proc LayerEntitiesSub { w entity } {
    global GidPriv

    set LayerNames [GiD_Info layers]
    if { [info exists GidPriv(LayersAlphabetic)] && $GidPriv(LayersAlphabetic) } {
	set LayerNames [lsort -dictionary $LayerNames]
    }
    set words ""
    if { $GidPriv(LayersAlsoLower) == 0 } {
	set ic 0
	foreach i $LayerNames {
	    lappend words [ list $i \
		    ".central.s process layers entities $i $entity"]
	    incr ic
	    if { $ic > 6 } {
		lappend words [ list "..." ChangeLayers]
		break
	    }
	}
    } else {
	set ic 0
	foreach i $LayerNames {
	    lappend words [ list $i \
		    ".central.s process layers entities $i LowerEntities $entity"]
	    incr ic
	    if { $ic > 6 } {
		lappend words [ list "..." ChangeLayers]
		break
	    }
	}
    }
    rmenuconf $w $words
}

proc LayerEntities { w } {
    global GidPriv

    set geomlistNames [ list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] [_ "All types"] ] 
    set geomlist "Points Lines Surfaces Volumes All"
    set meshlistNames [ list [_ "Nodes"] [_ "Elements"] [_ "All types"] ]
    set meshlist "Nodes Elements  All"


    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end

    $w add check -label [_ "Also lower entities"] \
	    -var GidPriv(LayersAlsoLower)
    if { ![info exists GidPriv(LayersAlsoLower)] } {
	set GidPriv(LayersAlsoLower) 0
    }

    switch $whatuse {
	GEOMETRYUSE {
	    foreach i $geomlist j $geomlistNames {
		set menuname [join "$w . [string tolower $i]" ""]
		$w add cascade -label $j -menu $menuname
		if { ![winfo exist $menuname] } {
		    menu $menuname -postcommand "LayerEntitiesSub $menuname $i"
		} else {
		    $menuname conf -postcommand "LayerEntitiesSub $menuname $i"
		}
	    }
	}
	MESHUSE {
	    foreach i $meshlist j $meshlistNames {
		set menuname [join "$w . [string tolower $i]" ""] 
		$w add cascade -label $j -menu $menuname
		if { ![winfo exist $menuname] } {
		    menu $menuname -postcommand "LayerEntitiesSub $menuname $i"
		} else {
		    $menuname conf -postcommand "LayerEntitiesSub $menuname $i"
		}
	    }
	}
    }
}

proc AddChangeToRender { w } {
    $w del 0 end
    set rendertype [GiD_Info Project RenderMode]

    set whatuse [GiD_Info Project ViewMode]
    switch $whatuse {
	GEOMETRYUSE - MESHUSE {
	    #set labels [ list [_ "Normal"] [_ "Flat lighting"] [_ "Smooth lighting"]]
	    set labels [ list [_ "Normal"] [_ "Flat#C#render"] [_ "Smooth#C#render"]]
	    set commands  { Normal FlatLighting SmoothLighting }
	    foreach i $labels j $commands {
		$w add command -label $i -command  [list GiD_Process render $j]
	    }
	    foreach i [GiD_Info customrenders] {
		$w add command -label $i -command  [list GiD_Process render Customized ChangeCustom $i]
	    }
	}
	POSTUSE {
	    foreach i [ list [_ "Normal"] [_ "Flat#C#render"] [_ "Smooth#C#render"]] j \
		    { Normal Flat Smooth } {
		$w add command -label $i -command  [list GiD_Process escape escape escape escape utilities render $j]
	    }
	}
	default {
	    $w add command -label $whatuse -command  ""
	}
    }

    switch $rendertype {
	normal {
	    $w entryconf 0 -state disabled
	}
	polygons {
	    $w entryconf 1 -state disabled
	}
	render {
	    $w entryconf 2 -state disabled
	}
	custom {
	    #$w entryconf 3 -state disabled
	    set customrenders [GiD_Info customrenders]
	    set current [GiD_Info customrenders current]
	    set j 3
	    foreach i $customrenders {
		if { $i == $current } {
		    $w entryconf $j -state disabled
		    break
		}
		incr j
	    }
	}
	postprocess {
	    # miguel 1999-01-21 correctec filled render mode in preprocess ( it uses the postprocess mode)
	    if { ( $whatuse == "GEOMETRYUSE") || ( $whatuse == "MESHUSE") } {
		#$w entryconf 1 -state disabled
	    } else {
		set postrender [GiD_Info postprocess get cur_display_render]
		switch $postrender {
		    Normal { 
		        $w entryconf 0 -state disabled
		    }
		    Flat { 
		        $w entryconf 1 -state disabled
		    }
		    Smooth { 
		        $w entryconf 2 -state disabled
		    }
		}
		
	    }
	}
    }

    $w add separator
    if { ( "$whatuse" == "GEOMETRYUSE") || ( "$whatuse" == "MESHUSE") } {
	$w add command -label [_ "Change light dir"] -command \
		"GiD_Process escape escape escape escape changelightvec"
	$w add command -label [_ "Customize"]... -command "CustomizeRender"
    }
    if { "$whatuse" == "POSTUSE" } {
	$w add command -label [_ "Change light dir"] -command \
		"GiD_Process escape escape escape escape utilities render changelightvec"
    }

    $w add separator

    set reflejo [ GiD_Reflection get state]
    set ::VREPriv(HayReflejoEntorno) [ lindex $reflejo 0]
    set cmd "On"
    if { [ string equal -nocase $::VREPriv(HayReflejoEntorno) "on"]} {
	set cmd "Off"
    }
    $w add checkbutton -label [_ "Reflection#C#render"] \
	-command [ list GiD_Process render ReflectEnvironment $cmd] \
	-variable ::VREPriv(HayReflejoEntorno) \
	-onvalue On -offvalue Off

    #set ind_val 0
    #set cmd "RenderFast"
    #if { [ string equal -nocase [ lindex $reflejo 1] "fast"]} {
    #        set ind_val 1
    #        set cmd "RenderNice"
    #}
    #$w add checkbutton -label [_ "Fast reflection#C#render"] -indicatoron $ind_val \
    #        -command [ list GiD_Process render ReflectEnvironment $cmd]
    #

    $w add command -label [_ "Define reflection...#C#render"] -command VentanaReflejoEntorno
}

proc CreateCascadedMenu { menu varname varcommandname acceleratorname \
	currentindex {littleiconname ""} } {
    upvar \#0 $varname var $varcommandname varcomm
    if { $acceleratorname != ""} {
	upvar \#0 $acceleratorname varaccelerator
    }
    if {$littleiconname != "" } {
	upvar \#0 $littleiconname varicons
    }
    global GidPriv
    set UnderLineList ""
    set j 0

    foreach i $var($currentindex) {
	set comm [lindex $varcomm($currentindex) $j]
	if { [string match ---* $i] } {
	    $menu add separator
	} elseif { [lindex $i 0] == "-CB-" } {
	    set text [lrange $i 1 end]
	    set under [FindUnderChar text UnderLineList]
	    $menu add checkbutton -label $text -underline $under \
		    -variable GidPriv([lindex $i 1])
	    EnterCommand $menu $comm $j
	    if { [info exists varaccelerator($currentindex)] } {
		AddAccelerator [lindex $varaccelerator($currentindex) $j] \
		    $menu $j
	    } else { AddAccelerator "" $menu $j }
	    if { [info exists varicons($currentindex)] } {
		AddLittleIcon [lindex $varicons($currentindex) $j] $menu $j
	    } else { 
		AddLittleIcon "" $menu $j 
	    }
	} elseif { ![info exists var($currentindex,$j)] } {
	    set text $i
	    set under [FindUnderChar text UnderLineList]
	    $menu add command -label $text -underline $under
	    EnterCommand $menu $comm $j
	    if { [info exists varaccelerator($currentindex)] } {
		AddAccelerator [lindex $varaccelerator($currentindex) $j] \
		        $menu $j
	    } else { AddAccelerator "" $menu $j }
	    if { [info exists varicons($currentindex)] } {
		AddLittleIcon [lindex $varicons($currentindex) $j] $menu $j
	    } else { 
		AddLittleIcon "" $menu $j 
	    }
	} else {
	    set text $i
	    set under [FindUnderChar text UnderLineList]
	    $menu add cascade -label $text -underline $under -menu $menu.$j 
	    if { [winfo exists $menu.$j] } { 
		#$menu.$j del 0 end 
		destroy $menu.$j
		menu $menu.$j -tearoff no
	    } else {
		menu $menu.$j -tearoff no
	    }
	    EnterCommand $menu.$j $comm -2
	    CreateCascadedMenu $menu.$j $varname $varcommandname $acceleratorname \
		"$currentindex,$j" $littleiconname
	    
	    if { [info exists varicons($currentindex)] } {
		AddLittleIcon [lindex $varicons($currentindex) $j] $menu $j
	    } else { 
		AddLittleIcon "" $menu $j 
	    }
	}
	incr j
    }

}

proc Create3rdButtonMenu { w { what PRE } } {
    global 3dButtonNames 3dButtonCommands 3dButtonAcceler 3dButtonIcons
    global 3dButtonDisabled 3dButtonDisabledP
    if { [$w index end] == "none" } {
	CreateCascadedMenu $w 3dButtonNames 3dButtonCommands 3dButtonAcceler 0 3dButtonIcons
    }
    set last [$w index end]
    if { $last == "none" } { set last -1}
    if { $what == "PRE" } {
	for { set i 0 } { $i < $last } { incr i} {
	    catch {
		if { [lsearch $3dButtonDisabled [$w entrycget $i -label]] != -1 } {
		    $w entryconf $i -state disabled
		} else { $w entryconf $i -state normal }
	    }
	}
    } else {
	for { set i 0 } { $i < $last } { incr i} {
	    catch {
		if { [lsearch $3dButtonDisabledP [$w entrycget $i -label]] \
		        != -1 } {
		    $w entryconf $i -state disabled
		} else { $w entryconf $i -state normal }
	    }
	}
    }
}

proc TryToFindVector { vec } {

    set num {([+-]?[0-9]*\.?[0-9]*([eE][+-]?[0-9]+)?)}
    set sep [join {[ " " , \n ] +} ""]
    set sepv [join {[ " " , \n 
] *} ""]

    set res [regexp ^$sepv$num$sep$num$sep$num$sepv$ $vec trash v1 trash2 \
	    v2 trash3 v3]

    if { $res && $v1 != "" && $v2 != "" && $v3 != ""} { return $v1,$v2,$v3 }
    return $vec
}


proc MakeBottomEntryBindings {} {
    global uparrow

    focus .gid.comm.e

    bind .gid.comm.e <Shift-Return> {
	set aa [TryToFindVector [.gid.comm.e get]]
	.gid.comm.e delete 0 end 
	.gid.comm.fwarn.list ins end "->$aa"
	set uparrow [.gid.comm.fwarn.list size]
	.gid.comm.fwarn.list yview [expr $uparrow-2]
	.central.s process $aa
	break
    }


    bind .gid.comm.e <Return> {
	set aa [.gid.comm.e get]
	.gid.comm.e delete 0 end 
	.gid.comm.fwarn.list ins end "->$aa"
	set uparrow [.gid.comm.fwarn.list size]
	.gid.comm.fwarn.list yview [expr $uparrow-2]
	
	if { [string match -np-* $aa] } {
	    regsub -- {-np-} $aa {} aa
	    eval $aa
	} else {
	    .central.s process $aa
	}
    }
    bind .gid.comm.e <Escape> {
	.gid.comm.e delete 0 end 
	.central.s process escape
    }
    set uparrow 0
    bind .gid.comm.e <Up> {
	.gid.comm.e delete 0 end
	while 1 {
	    incr uparrow -1
	    if { $uparrow < 0 } {
		set uparrow -1
		break
	    } elseif { [regexp {^->.*[^ ]+.*$} [.gid.comm.fwarn.list get $uparrow]] } {
		.gid.comm.e insert 0 [string range [.gid.comm.fwarn.list get $uparrow] \
		        2 end]
		break
	    }
	}
    }
    bind .gid.comm.e <Down> {
	.gid.comm.e delete 0 end
	while 1 {
	    incr uparrow 1
	    if { $uparrow >= [.gid.comm.fwarn.list size] } { 
		set uparrow [.gid.comm.fwarn.list size]
		break
	    } elseif { [regexp {^->.*[^ ]+.*$} [.gid.comm.fwarn.list get $uparrow]] } {
		.gid.comm.e insert 0 [string range [.gid.comm.fwarn.list get $uparrow] \
		        2 end]
		break
	    }
	}
    }
    bind .gid.comm.e <Alt-KeyPress-3> {
	set x [expr [winfo rootx .gid.comm.e]+%x+2]
	set y [expr [winfo rooty .gid.comm.e]+%y]
	GiD_PopupMenu .gid.phantom.display3dbutton $x $y
    }
    bind .gid.comm.e <Tab> "AutoCompleteGiDEntry .gid.comm.e; break"

    #para dejar el 'pegar' con el boton del medio del raton
    bind .gid.comm.e <ButtonPress-2> ""
    #.gid.comm.e ins insert \"[selection get]\"; break;"
    #bind .gid.comm.e <ButtonRelease-2> ""
}

proc AddEntitiesToMenu { w basecommand args} {
    set geomlist "Points Lines Surfaces Volumes"
    set geomlistnames [ list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] ]
    set meshlist "Nodes Elements"
    set meshlistnames [ list [_ "Nodes"] [_ "Elements"] ]
    set graphlist "Nodes"
    set graphlistnames [ list [_ "Nodes"] ]
    if { [llength $args] == 4 } {
	set geomlistnames [lindex $args 0]
	set geomlist [lindex $args 1]
	set meshlistnames [lindex $args 2]
	set meshlist [lindex $args 3]
    } elseif { [llength $args] == 6 } {
	set geomlistnames [lindex $args 0]
	set geomlist [lindex $args 1]
	set meshlistnames [lindex $args 2]
	set meshlist [lindex $args 3]
	set graphlist [lindex $args 4]
	set graphlistnames [lindex $args 5]
    }
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    foreach name $geomlistnames i $geomlist {
		$w add command -label $name -command "$basecommand $i"
	    }
	}
	MESHUSE {
	    if { [llength $meshlist] == 1 } {
		$w add command -label $meshlistnames -command "$basecommand $meshlist"
	    } else {
		foreach name $meshlistnames i $meshlist {
		    $w add command -label $name -command "$basecommand $i"
		}
	    }
	}
	POSTUSE {
	    if { [llength $meshlist] == 1 } {
		$w add command -label $meshlistnames -command "$basecommand $meshlist"
	    } else {
		foreach name $meshlistnames i $meshlist {
		    $w add command -label $name -command "$basecommand $i"
		}
	    }
	}
	GRAFUSE  {
	    foreach name $graphlistnames i $graphlist {
		$w add command -label $name -command "$basecommand $i"
	    }
	}
    }
}

proc AddEntitiesToMenuCollapse { w } {
    set basecommand ".central.s process escape escape escape escape utilities Collapse"
    set geomlist "Model Points Lines Surfaces Volumes"
    set geomlistnames [ list [_ "Model"] [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] ]
    set meshlist "Mesh Edges Nodes Elements"
    set meshlistnames [ list [_ "Mesh"] [_ "Edges"] [_ "Nodes"] [_ "Elements"] ]
    
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    set colors [list "black" "black" "black" "black" "black"]
	    if { [winfo exists .gid.central.s] } {
		set npoints [GiD_Info Geometry NumPoints]
		set nlines [GiD_Info Geometry NumLines]
		set nsurfs [GiD_Info Geometry NumSurfaces]
		set nvols [GiD_Info Geometry NumVolumes]
		set i [lindex [split $w .] end]
		set parent [winfo parent $w]
		if { !$npoints } {
		    $parent entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 0  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 3  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 4  $::GidPriv(Color,DisabledForegroundMenu)
		} else {
		    $parent entryconfigure $i -foreground "black"
		    lset colors 0  "black"
		    lset colors 1  "black"
		    if { $nvols } {
		        lset colors 2  "black"
		        lset colors 3  "black"
		        lset colors 4  "black"
		    } else {
		        lset colors 4  $::GidPriv(Color,DisabledForegroundMenu)
		        if { $nsurfs } {
		            lset colors 2  "black"
		            lset colors 3  "black"
		        } else {
		            lset colors 3  $::GidPriv(Color,DisabledForegroundMenu)
		            if { $nlines } {
		                lset colors 2  "black"
		            } else {
		                lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		            }
		        }
		    }
		}
	    }
	    foreach name $geomlistnames i $geomlist  col $colors {
		$w add command -label $name -command "$basecommand $i" -foreground $col
	    }
	}
	MESHUSE {
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    $parent entryconfigure $i -foreground "black"
	    foreach name $meshlistnames i $meshlist {
		$w add command -label $name -command "$basecommand $i" -foreground "black"
	    }
	}
    }
}

proc AddEntitiesToMenuSwapNormals { w } {
    set basecommand ".central.s process escape escape escape escape utilities SwapNormals"
    set geomlist [ list "$basecommand Lines Select" "$basecommand Surfaces" ]
    set geomlistnames [ list [_ "Lines"] [_ "Surfaces"] ]
    set meshlist [ list "$basecommand Select" "$basecommand MakeGroupCoherent" "$basecommand SelByNormal"]
    set meshlistnames [ list [_ "Select"] [_ "Make coherent"] [_ "Select by normal"]]
 
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    set colors [list "black" "black" ]
	    if { [winfo exists .gid.central.s] } {
		set nlines [GiD_Info Geometry NumLines]
		set nsurfs [GiD_Info Geometry NumSurfaces]
		set i [lindex [split $w .] end]
		set parent [winfo parent $w]
		if { !$nlines } {
		    $parent entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 0  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		} else {
		    $parent entryconfigure $i -foreground "black"
		    lset colors 0  "black"
		    if { $nsurfs } {
		        lset colors 1  "black"
		    } else {
		        lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		    }                    
		}
	    }
	    foreach name $geomlistnames i $geomlist col $colors {
		if { $name == [_ "Lines"] } {
		    $w add command -label $name -command "$i" -foreground $col
		} elseif { $name == [_ "Surfaces"] } {
		    set id_menu 1
		    $w add cascade -label $name -menu $w.m$id_menu -foreground $col
		    if { [ winfo exist $w.m$id_menu ] } {
		        $w.m$id_menu del 0 end
		    } else {
		        menu $w.m$id_menu
		    }
		    $w.m$id_menu add command -label [_ "Select"] -command  "$i Select" -foreground $col
		    $w.m$id_menu add command -label [_ "Make coherent"] -command  "$i MakeGroupCoherent" -foreground $col
		    $w.m$id_menu add command -label [_ "Select by normal"] -command  "$i SelByNormal" -foreground $col
		}
	    }
	}
	MESHUSE {
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    if { [lindex [GiD_Info Mesh] 0] == 0 } {
		set color $::GidPriv(Color,DisabledForegroundMenu)
	    } else {
		set color "black"
	    }
	    $parent entryconfigure $i -foreground $color
	    foreach name $meshlistnames j $meshlist {
		$w add command -label $name -command "$j" -foreground $color
		
		#set id_menu 0
		#set col "black"
		#$w add cascade -label $name -menu $w.m$id_menu -foreground $col
		#if { [ winfo exist $w.m$id_menu ] } {
		#    $w.m$id_menu del 0 end
		#} else {
		#    menu $w.m$id_menu
		#}
		#$w.m$id_menu add command -label [_ "Select"] -command "$j Select" -foreground $col
		#$w.m$id_menu add command -label [_ "Make coherent"] -command "$j MakeGroupCoherent" -foreground $col
		#$w.m$id_menu add command -label [_ "Select by normal"] -command "$j SelByNormal" -foreground $col
	    }
	}
    }
}

proc AddEntitiesToMenuDrawNormals { w } {
    set basecommand ".central.s process escape escape escape escape utilities Drawnormals"
    #set geomlist [ list "$basecommand Lines" "$basecommand Surfaces" "$basecommand Surfaces Color"]
    #set geomlistnames [ list [_ "Lines"] [_ "Surfaces"] [_ "Color"] ]
    set geomlist [ list "$basecommand Lines" "$basecommand Surfaces" ]
    set geomlistnames [ list [_ "Lines"] [_ "Surfaces"] ]
    set meshlist [ list "$basecommand" "$basecommand Color"]
    set meshlistnames [ list [_ "Elements"] [_ "Colored elements"]]
 
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    set colors [list "black" "black" ]
	    if { [winfo exists .gid.central.s] } {
		set nlines [GiD_Info Geometry NumLines]
		set nsurfs [GiD_Info Geometry NumSurfaces]
		set i [lindex [split $w .] end]
		set parent [winfo parent $w]
		if { !$nlines } {
		    $parent entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 0  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		    #lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		} else {
		    $parent entryconfigure $i -foreground "black"
		    lset colors 0  "black"
		    if { $nsurfs } {
		        lset colors 1  "black"
		    } else {
		        lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		    }
		    #lset colors 2  "black"
		}
	    }
	    foreach name $geomlistnames i $geomlist col $colors {
		if { $name == [_ "Lines"] } {
		    $w add command -label $name -command "$i" -foreground $col
		} elseif { $name == [_ "Surfaces"] } {
		    #$w add command -label $name -command "$i" -foreground $col
		    set id_menu 1
		    $w add cascade -label $name -menu $w.m$id_menu -foreground $col
		    if { [ winfo exist $w.m$id_menu ] } {
		        $w.m$id_menu del 0 end
		    } else {
		        menu $w.m$id_menu
		    }
		    $w.m$id_menu add command -label [_ "Normal"] \
		        -command  "$i" -foreground $col
		    $w.m$id_menu add command -label [_ "Colored#C#Surfaces"] \
		        -command  "$i Color" -foreground $col
		}
	    }           
	}
	MESHUSE {
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    $parent entryconfigure $i -foreground "black"
	    foreach name $meshlistnames i $meshlist {
		$w add command -label $name -command "$i" -foreground "black"
	    }
	}
    }
}

proc AddEntitiesToMenuDrawHigher { w } {
    set basecommand ".central.s process escape escape escape escape utilities DrawHigher"
    set geomlist "Points Lines Surfaces Separator All"
    set geomlistnames [list [_ "Points"] [_ "Lines"] [_ "Surfaces"] "Separator" [_ "All"]]
    set meshlist "Nodes"
    set meshlistnames [list [_ "Nodes"]]
 
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    set colors [list "black" "black" "black" "black" "black" ]
	    if { [winfo exists .gid.central.s] } {
		set npoints [GiD_Info Geometry NumPoints]
		set nlines [GiD_Info Geometry NumLines]
		set nsurfs [GiD_Info Geometry NumSurfaces]
		set i [lindex [split $w .] end]
		set parent [winfo parent $w]
		if { !$npoints } {
		    $parent entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 0  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		    lset colors 4  $::GidPriv(Color,DisabledForegroundMenu)
		} else {
		    $parent entryconfigure $i -foreground "black"
		    lset colors 0  "black"
		    lset colors 4  "black"
		    if { $nsurfs } {
		        lset colors 1  "black"
		        lset colors 2  "black"
		    } else {
		        lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		        if { $nlines } {
		            lset colors 1  "black"
		        } else {
		            lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		        }
		    }
		}
	    }
	    foreach name $geomlistnames i $geomlist col $colors {
		if { $i=="Separator" } {
		    $w add separator
		} else {
		    $w add command -label $name -command "$basecommand $i" -foreground $col
		}
	    }
	}
	MESHUSE {
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    $parent entryconfigure $i -foreground "black"
	    foreach name $meshlistnames i $meshlist {
		$w add command -label $name -command "$basecommand $i" -foreground "black"
	    }
	}
    }
}

proc AddEntitiesToMenuDrawCurvatures { w } {
    global GidPriv
    if { ![info exists GidPriv(DrawMainDirections)] } {
	set GidPriv(DrawMainDirections) 0
    }
    if { !$GidPriv(DrawMainDirections) } {
	set basecommand "GiD_Process Mescape utilities DrawCurvature NoShowMainDir"
    } else {
	set basecommand "GiD_Process Mescape utilities DrawCurvature ShowMainDir"
    }
    set geomlist "CurvatureMax CurvatureMin CurvatureMean CurvatureGauss"
    set geomlistnames [list [_ "Max.#C#curvature"] [_ "Min.#C#curvature"] [_ "Mean#C#curvature"] [_ "Gaussian#C#curvature"]]
 
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    if { $whatuse == "GEOMETRYUSE" && [GiD_Info Geometry NumSurfaces] > 0 } {
	set col "black"
    } else {
	set col $::GidPriv(Color,DisabledForegroundMenu)
    }
    set i [lindex [split $w .] end]
    set parent [winfo parent $w]
    $parent entryconfigure $i -foreground $col
    foreach name $geomlistnames i $geomlist {
	if { $i=="Separator" } {
	    $w add separator
	} else {
	    $w add command -label $name -command "$basecommand $i" -foreground $col
	}
    }
    $w add separator
    if { !$GidPriv(DrawMainDirections) } {
	$w add checkbutton -label [_ "Main directions"] -variable GidPriv(DrawMainDirections) -foreground $col \
		-command "GiD_Process Mescape utilities DrawCurvature ShowMainDir"
    } else {
	$w add checkbutton -label [_ "Main directions"] -variable GidPriv(DrawMainDirections) -foreground $col \
		-command "GiD_Process Mescape utilities DrawCurvature NoShowMainDir"
    }
}

#create a name to show abreviated the full path of a file, for example:  "d:\Documents\...\myfile"
proc ShortFilenameToShowMenu { filename {maxlength 40} } {
    if { [string length $filename] > $maxlength } {
	set splitname [file split $filename] 
	set l [llength $splitname]
	set availablelength $maxlength
	set shortname [lindex $splitname 0]
	incr availablelength [expr -[string length $shortname]-[string length [lindex $splitname [expr $l-1]]-5]]
	for {set i 1 } { $i < [expr $l-1] } {incr i} {
	    incr availablelength [expr -[string length [lindex $splitname $i]]-1]
	    if { $availablelength >=0 } {
		set shortname [file join $shortname [lindex $splitname $i]]
	    } else {
		break
	    }
	}
	set shortname [file join $shortname {...} [lindex $splitname [expr $l-1]]]
	set shortname [file nativename $shortname]
    } else {
	set shortname [file nativename $filename]
    }
    return $shortname
}

proc UltimoTrozoNombreArchivo { archivo} {
    set nombre [ file tail $archivo]
    set dir [ file tail [ file dirname $archivo]]
    if { "$dir" != ""} {
	set nombre [ file join $dir $nombre]
    }
    return $nombre
}

proc AddEntitiesToMenuRecentProjects { w } {
    set names [ GiD_Info RecentProjects]
    $w del 0 end
    foreach i $names {
	$w add command -label [ ShortFilenameToShowMenu $i] -command [ list LoadFileInGid "$i.gid"]
    }
}

proc AddEntitiesToMenuRecentPostFiles { w } {
    set names [ GiD_Info RecentPostFiles]
    $w del 0 end
    foreach i $names {
	if { [catch {
		# es multiple si hay varios .msh y .res o .bin
		set varias_msh [ regexp {^.*\.msh.*\.msh\}?$} $i]
		set varios_res [ regexp {^.*\.res.*\.res\}?$} $i]
		set varios_bin [ regexp {^.*\.bin.*\.bin\}?$} $i]
		if { !$varias_msh && !$varios_res && !$varios_bin} {
		    $w add command -label [ ShortFilenameToShowMenu $i] \
		        -command [ list LoadFileInGidPost "$i"]
		} else {
		    # es un open multiple
		    set a1 [ lindex $i 0]
		    set a2 [ lindex $i end]
		    $w add command -label [ format [_ "Multiple: %s, ..., %s"] \
		            [ UltimoTrozoNombreArchivo $a1] \
		            [ UltimoTrozoNombreArchivo $a2]] \
		        -command [ list ExecuteFileCommand ReadMultiple POST "" "$i"]
		}
	    } err_txt] } {                
	  WarnWin $err_txt
      }
  }
}

proc AddEntitiesToMenuRecentViewFiles { w } {
    set names [ GiD_Info RecentViewFiles]
    $w del 0 end
    foreach i $names {
	$w add command -label [ ShortFilenameToShowMenu $i] -command [ list LoadFileInGidPost "$i"]
    }
}

proc AddRecentViewFiles2IconBar { w } {
    $w del 0 end

    # Vistas predefinidas:
    set img [ list vl_zframe.gif \
		  vl_view_rotate_xy.gif vl_view_rotate_xz.gif vl_view_rotate_yz.gif \
		  vl_view_rotate_-xy.gif vl_view_rotate_-xz.gif vl_view_rotate_-yz.gif \
		  vl_isometric.gif]
    set txt [ list [_ "Zoom frame"] \
		  [_ "View XY plane (original)"] [_ "View XZ plane"] [_ "View YZ plane"] \
		  [_ "View -XY plane"] [_ "View -XZ plane"] [_ "View -YZ plane"] \
		  [_ "Isometric"]]
    set cmd [ list "GiD_Process Zoom Frame" \
		  "GiD_Process Rotate Angle 270 90" \
		  "GiD_Process Rotate Angle 270 0" \
		  "GiD_Process Rotate Angle 0 0" \
		  "GiD_Process Rotate Angle 90 -90" \
		  "GiD_Process Rotate Angle 90 0" \
		  "GiD_Process Rotate Angle 180 0" \
		  "GiD_Process Rotate Angle 45 45"]

    foreach i $img t $txt c $cmd {
	if { "$i" != ""} {
	    $w add command -image [ GetImage $i] -label $t -command $c -compound left
	} else {
	    $w add command -label $t -command $c
	}
    }
    $w add separator

    set names [GiD_Info RecentViewFiles]
    foreach i $names {
	if { ![ regexp {^\{([^\}]*).*\{([^\}]*)\}$} $i tonto a1 a2]} {
	    $w add command -label [ShortFilenameToShowMenu $i] -command [list LoadFileInGidPost "$i"]
	} else {
	    # es un open multiple
	    $w add command -label [ format [_ "Multiple: %s, ..., %s"] [ UltimoTrozoNombreArchivo $a1] [ UltimoTrozoNombreArchivo $a2]]\
		-command [ list ExecuteFileCommand ReadMultiple POST "" "$i"]
	}
    }
}

proc AddEntitiesToMenuAfter { num w basecommand args} {
    set geomlist "Points Lines Surfaces Volumes"
    set geomlistnames [ list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"] ]
    set meshlist "Nodes Elements"
    set meshlistnames [ list [_ "Nodes"] [_ "Elements"] ]
    

    if { [llength $args] == 4 } {
	set geomlistnames [lindex $args 0]
	set geomlist [lindex $args 1]
	set meshlistnames [lindex $args 2]
	set meshlist [lindex $args 3]
    }
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    incr num
    if { [$w index $num] == $num } {
	$w del $num end
    }
    switch $whatuse {
	GEOMETRYUSE {
	    set colors [list "black" "black" "black" "black" ]
	    #SendLayerEntitiesToBackAA Layers->To Back            
	    if { [string trim $basecommand] != "SendLayerEntitiesToBackAA" } {
		if { [winfo exists .gid.central.s] } {
		    set npoints [GiD_Info Geometry NumPoints]
		    set nlines [GiD_Info Geometry NumLines]
		    set nsurfs [GiD_Info Geometry NumSurfaces]
		    set nvols [GiD_Info Geometry NumVolumes]
		    set i [lindex [split $w .] end]
		    set parent [winfo parent $w]
		    if { [string is integer $i] && $i >=0 && $i < [$parent index end] } {
		        $parent entryconfigure $i -foreground "black"
		    }
		    if { !$npoints } {
		        #if { [string is integer $i] && $i >=0 && $i < [$parent index end] } {
		        #    $parent entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu)
		        #}
		        lset colors 0  $::GidPriv(Color,DisabledForegroundMenu)
		        lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		        lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		        lset colors 3  $::GidPriv(Color,DisabledForegroundMenu)
		    } else {
		        if { [string is integer $i] && $i >=0 && $i < [$parent index end] } {
		            $parent entryconfigure $i -foreground "black"
		        }
		        lset colors 0  "black"
		        if { $nvols } {
		            lset colors 3 "black"
		            lset colors 2 "black"
		            lset colors 1 "black"
		        } else {
		            lset colors 3  $::GidPriv(Color,DisabledForegroundMenu)
		            if { $nsurfs } {
		                lset colors 2 "black"
		                lset colors 1 "black"
		            } else {
		                lset colors 2 $::GidPriv(Color,DisabledForegroundMenu)
		                if { $nlines } {
		                    lset colors 1 "black"
		                } else {
		                    lset colors 1 $::GidPriv(Color,DisabledForegroundMenu)
		                }
		            }
		        }
		    }
		}
	    }
	    
	    foreach name $geomlistnames i $geomlist col $colors {
		$w add command -label $name -command "$basecommand $i" -foreground $col \
		        -image [GetImage "dummy16x16.gif"] -compound left
	    }
	}
	MESHUSE {
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    if { [string is integer $i] && $i >=0 && $i < [$parent index end] } {
		$parent entryconfigure $i -foreground "black"
	    }
	    foreach name $meshlistnames i $meshlist {
		$w add command -label $name -command "$basecommand $i" -foreground "black" \
		        -image [GetImage "dummy16x16.gif"] -compound left
	    }
	}
	POSTUSE {
	    foreach name $meshlistnames i $meshlist {
		$w add command -label $name -command "$basecommand $i" \
		        -image [GetImage "dummy16x16.gif"] -compound left
	    }
	}
    }
}

proc AddEntitiesToMenuGeomOrMeshOrGraph { w namesG namesM namesF commandsG commandsM commandsF } {
    set nlist(GEOMETRYUSE) $namesG
    set nlist(MESHUSE) $namesM
    set nlist(GRAFUSE) $namesF 

    set clist(GEOMETRYUSE) $commandsG
    set clist(MESHUSE) $commandsM 
    set clist(GRAFUSE) $commandsF 

    set whatuse ""
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    if { $whatuse == "POSTUSE" } {
      set whatuse MESHUSE
    }

    $w del 0 end

    foreach i $nlist($whatuse) j $clist($whatuse) {
	$w add command -label $i -command $j
    }
}


proc AddCutsToMenu { w basecommand} {
#     global ViewMenuPriv
    
    #if { ![GiD_Info postprocess get changed_geom_list]} { return}

    set cutlist [GiD_Info postprocess get all_cuts]
#    set what cuts2

#    if { [ info exist ViewMenuPriv($what)]} {
#        if { ![ string compare $ViewMenuPriv($what) $cutlist ]} {
#            return
#        }
#    }

#    set ViewMenuPriv($what) $cutlist

    $w del 0 end
    foreach i $cutlist {
	regsub -all {_} $i { } cutlabel
	$w add command -label $cutlabel -command "$basecommand $i"
    }
    if { [ llength $cutlist] > 1} {
	$w add command -label [_ "All cuts"] -command "$basecommand ALL_CUTS"
    }
    if { $cutlist == ""} {
	$w add command -label [_ "There are no Cuts"] -state disabled
    }
}

proc AddGraphsToMenu { w basecommand { todos 0}} {
#    global ViewMenuPriv

    set graphlist [GiD_Info postprocess get all_graphs]
#    set what graphs

#    if { [ info exist ViewMenuPriv($what)]} {
#        if { ![ string compare $ViewMenuPriv($what) $graphlist ]} {
#            return
#        }
#    }

#    set ViewMenuPriv($what) $graphlist

    $w del 0 end
    foreach i $graphlist {
	regsub -all {_} $i { } graphlabel
	$w add command -label $graphlabel -command "$basecommand $i"
    }
    if { $todos && ( [ llength $graphlist] > 1) } {
	$w add command -label [_ "All"] -command "$basecommand all"
    }
    if { $graphlist == ""} {
	$w add command -label [_ "There are no Graphs"] -state disabled
    }
}

proc AddLineGraphsToMenu { w basecommand { todos 0}} {
    set graphlist [GiD_Info postprocess get all_line_graphs]

    $w del 0 end
    foreach i $graphlist {
	regsub -all {_} $i { } graphlabel
	$w add command -label $graphlabel -command "$basecommand $i"
    }
    if { $todos && ( [ llength $graphlist] > 1) } {
	$w add command -label [_ "All"] -command "$basecommand all"
    }
    if { $graphlist == ""} {
	$w add command -label [_ "There are no Graphs"] -state disabled
    }
}

proc AddPostGeomToMenu { w basecommand what color_opt color_labels} {
    global ViewMenuPriv

    set geomlist [GiD_Info postprocess get all_$what]

    if { [ info exist ViewMenuPriv($w,$what)]} {
	if { ![ string compare $ViewMenuPriv($w,$what) $geomlist ]} {
	    return
	}
    }

    $w del 0 end

    if { $geomlist == ""} {
	$w add command -label [_ "There are no %s" $what] -state disabled
	return
    }

    set ViewMenuPriv($w,$what) $geomlist
    
    set id_menu 0
    foreach i $geomlist {
	regsub -all {_} $i { } geomlabel
	if { $color_opt == ""} {
	    $w add command -label $geomlabel -command "$basecommand $i"
	} else {
	    $w add cascade -label $geomlabel -menu $w.m$id_menu
	    if { [ winfo exist $w.m$id_menu ]} {
		$w.m$id_menu del 0 end
	    } else {
		menu $w.m$id_menu
	    }

	    foreach j $color_opt k $color_labels {
		$w.m$id_menu add command -label $k -command "$basecommand $i $j"
	    }
	}
	incr id_menu
    }
}


proc AddListEntitiesToMenuExec { i } {
    global GidPriv

    if { !$GidPriv(ListEntitiesMass) } {
	.central.s process escape escape escape \
		escape utilities list $i
    } else {
	.central.s process escape escape escape \
		escape utilities list MassProp $i
    }
}

proc AddListEntitiesToMenu { w } {
    global GidPriv
    set geomlistNames [ list [_ "Points"] [_ "Lines"] [_ "Surfaces"] [_ "Volumes"]] 
    set geomlist "Points Lines Surfaces Volumes"
    set meshlistNames [ list [_ "Nodes"] [_ "Elements"]]
    set meshlist "Nodes Elements"
    set mcolor  "black"

    set colors [list "black" "black" "black" "black" ]

    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse GEOMETRYUSE
    }
    $w del 0 end
    
    if { [winfo exists .gid.central.s] } {
	if { $whatuse == "GEOMETRYUSE" } {
	    
	    set npoints [GiD_Info Geometry NumPoints]
	    set nlines [GiD_Info Geometry NumLines]
	    set nsurfs [GiD_Info Geometry NumSurfaces]
	    set nvols [GiD_Info Geometry NumVolumes]
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    if { !$npoints } {
		$parent entryconfigure $i -foreground $::GidPriv(Color,DisabledForegroundMenu)
		lset colors 0  $::GidPriv(Color,DisabledForegroundMenu)
		lset colors 1  $::GidPriv(Color,DisabledForegroundMenu)
		lset colors 2  $::GidPriv(Color,DisabledForegroundMenu)
		lset colors 3  $::GidPriv(Color,DisabledForegroundMenu)
		set mcolor  $::GidPriv(Color,DisabledForegroundMenu)
	    } else {
		$parent entryconfigure $i -foreground "black"
		lset colors 0  "black"
		set mcolor  "black"
		if { $nvols } {
		    lset colors 3 "black"
		    lset colors 2 "black"
		    lset colors 1 "black"
		} else {
		    lset colors 3  $::GidPriv(Color,DisabledForegroundMenu)
		    if { $nsurfs } {
		        lset colors 2 "black"
		        lset colors 1 "black"
		    } else {
		        lset colors 2 $::GidPriv(Color,DisabledForegroundMenu)
		        if { $nlines } {
		            lset colors 1 "black"
		        } else {
		            lset colors 1 $::GidPriv(Color,DisabledForegroundMenu)
		        }
		    }
		}
	    }
	} else {
	    set i [lindex [split $w .] end]
	    set parent [winfo parent $w]
	    $parent entryconfigure $i -foreground "black"
	}
    }
    
    $w add checkbutton -label [_ "Mass"] -variable GidPriv(ListEntitiesMass) -foreground $mcolor
    $w add separator

    switch $whatuse {
	GEOMETRYUSE {
	    foreach i $geomlist j $geomlistNames col $colors {
		$w add command -label $j -command "AddListEntitiesToMenuExec $i" -foreground $col
	    }
	}
	MESHUSE {
	    foreach i $meshlist j $meshlistNames {
		$w add command -label $j -command "AddListEntitiesToMenuExec $i"  -foreground "black"
	    }
	}
    }
}

proc HardcopyToPrint { basecommand { prncomm ""} { parent ""}} {
    global tcl_platform GidPriv
    if { $tcl_platform(platform) == "windows" } {
	.central.s process hardcopy print
	return
    } else {
	##WarnWin "hola '$basecommand' '$prncomm' '$parent'"
	if { ! [ info exist GidPriv(PrinterCommand)] } {
	    set GidPriv(PrinterCommand) "lpr"
	}
	if { $prncomm != ""} {
	    set GidPriv(PrinterCommand) $prncomm
	}
	set wname .gid.tempwinw
	if { ( $parent != "" ) && ( $parent != ".") } {
	    set wname $parent.tempwinw
	}
	if { "$prncomm" != ""} {
	    set pepe "ok"
	} else {
	    set pepe [ tk_dialogEntryRAM $wname [_ "Enter value window"] \
		           [_ "Which command do you use to print files?"] gidquestionhead  "any" \
		    $GidPriv(PrinterCommand)]
	}
	# nos devuelve pepe entre comillas si hay mas de una palabra
	if { "$pepe" != "--CANCEL--"} {
	    set tmp_filename [ file join [GiD_Info Project TmpDirectory] \
		    print_[ clock format [clock seconds] -format "%Y-%m-%d-%H:%M:%S"].ps]
	    GiD_Process $basecommand $tmp_filename
	    if { [ file exists $tmp_filename] } {
		#para quitar las comillas de pepe si hay mas de una palabra,
		# pues exec toma pepe como nombre de programa ( con espacios y todo)
		set res [ eval [ join "exec $GidPriv(PrinterCommand) $tmp_filename"]]
		if { $res != ""} {
		    WarnWin "$res"
		} else {
		    WarnWin [_ "Job succesfully printed or queued."]
		}
		file delete -force $tmp_filename
		#set GidPriv(PrinterCommand) $pepe
	    }
	}
	return 
    }
}


proc AddBitmapsToMenu { w basecommand args } {
    global GIDDEFAULT
    global GidPriv

    set geomlist "point.gif line.gif surface.gif volume.gif"
    set meshlist "node.gif element.gif"
    set postlist "node.gif element.gif resultonview.gif"
    set graflist "node.gif resultonview.gif"

    if { [llength $args] == 2 } {
	set geomlist [lindex $args 0]
	set meshlist [lindex $args 1]
    }
    set whatuse [GiD_Info Project ViewMode]
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    foreach i $geomlist {
		set command [file rootname $i]
		$w add command -image [GetImage $i] -command "$basecommand $command" -hidemargin 1
	    }
	}
	MESHUSE {
	    foreach i $meshlist {
		set command [file rootname $i]
		$w add command -image [GetImage $i] -command "$basecommand $command" -hidemargin 1
	    }
	}
	POSTUSE {
	    if { $args == "list"} {
		set postlist $meshlist
	    }
	    foreach i $postlist {
		set command [file rootname $i]
		$w add command -image [GetImage $i] -command "$basecommand $command" -hidemargin 1
	    }
	}
	GRAFUSE {            
	    foreach i $graflist {
		set command [file rootname $i]
		$w add command -image [GetImage $i] -command "$basecommand $command" -hidemargin 1
	    }
	}
    }
}

proc AddBitmapsDeleteMenu { w } {
    set whatuse [GiD_Info Project ViewMode]
    $w del 0 end
    switch $whatuse {
	GEOMETRYUSE {
	    set images {point.gif line.gif coonsurf.gif volume.gif alltypes.gif}
	    set commands { {geometry delete point} {geometry delete line} \
		    {geometry delete surface}  {geometry delete volume} {geometry delete all} }
	}
	MESHUSE {
	    set images {element.gif node.gif}
	    set commands { {Meshing EditMesh DeleteElems} {Meshing EditMesh DeleteNodes} }
	}
	POSTUSE -
	GRAFUSE {
	    set images {}
	    set commands {}
	}
	default {
	    set images {}
	    set commands {}
	}
    }
    foreach image $images command $commands {
	$w add command -image [GetImage $image] -command "GiD_Process Mescape $command" -hidemargin 1
    }
}

proc AddBitmapsToCamera { w basecommand args} {
    $w del 0 end
    set hardcopylist "eps.gif ps.gif bmp.gif gif.gif jpeg.gif png.gif tga.gif tiff.gif vrml.gif"
    foreach i $hardcopylist {
	set command [file rootname $i]
	$w add command -image [GetImage $i] -command "$basecommand $command" -hidemargin 1
    }
}

proc CambiaImagen { w toca command} {
    $w conf -image [GetImage $toca]
    eval $command
}

proc AddBitmapsToDisplayStyle { w basecommand args} {
    global GidPriv
    set idx [ string last "." $w]
    if { $idx != -1 } {
	set GidPriv(PostDisplayStyleBitmap) [ string range $w 0 [ expr $idx - 1]] 
    } else {
	set GidPriv(PostDisplayStyleBitmap) ""
    }

    set displaystylelist "boundaries hiddenbound alllines hiddenlines body bodybound bodylines points pointsbound"
    
    set var [ string tolower [GiD_Info postprocess get cur_display_style]]
    regsub {_} $var {} var
    
    $w del 0 end
    foreach i $displaystylelist {
	if { $i != $var} {
	    $w add command -image [GetImage $i.gif] -command \
		[list CambiaImagen $GidPriv(PostDisplayStyleBitmap) $i.gif "$basecommand $i"] -hidemargin 1
	} else {
	    $w add command -image [GetImage $i.gif] -state disabled -hidemargin 1
	}
    }
}

proc AddBitmapsToCullStyle { w basecommand args} {
    global GidPriv
    set idx [ string last "." $w]
    if { $idx != -1 } {
	set GidPriv(PostDisplayCullBitmap) [ string range $w 0 [ expr $idx - 1]] 
    } else {
	set GidPriv(PostDisplayCullBitmap) ""
    }
    set cullstylelist "none frontfaces backfaces"
    set var [string tolower [GiD_Info postprocess get cur_display_culling]]

    $w del 0 end
    foreach i $cullstylelist {
	if { $i != $var} {
	    $w add command -image [GetImage cull$i.gif] \
		-command [list CambiaImagen $GidPriv(PostDisplayCullBitmap) cull$i.gif "$basecommand $i"] \
		        -hidemargin 1
	} else {
	    $w add command -image [GetImage cull$i.gif] -state disabled -hidemargin 1
	}
    }
}

proc AddBitmapsToCutDivide { w basecommand que args} {
    set displaystylelist "cut2p.gif cut3p.gif"
    $w del 0 end
    $w add command -image [GetImage cut2p.gif] -command "$basecommand $que" -hidemargin 1
    $w add command -image [GetImage cut3p.gif] -command "$basecommand $que ThreePoints" -hidemargin 1
    if { $que == "CutPlane"} {
	$w add command -image [GetImage cut_succesion.gif] -command "$basecommand Succesion" -hidemargin 1
    }
}

proc AddBitmapsToMeshSetCut { w basecommand que args} {
    global GIDDEFAULT GidPriv

    $w del 0 end
    set all_que [GiD_Info postprocess get all_$que]
    if { $all_que == ""} {
	$w add command -label [_ "There are no %s" $que] -state disabled
	return
    }
    if { [ llength $all_que] < 17 } {
	set all_colors [GiD_Info postprocess get all_$que\_colors]
	set que_on [GiD_Info postprocess get cur_$que]
	set jc 0
	foreach j $all_que {
	    regsub {_} $j { } lab
	    set col_set [ lindex $all_colors $jc]
	    # es una lista { ambient diffuse specular shininess}
	    set col_diffuse [ lindex $col_set 1]
	    if { [lsearch $que_on $j] != -1 } {
		$w add command -bitmap @$GIDDEFAULT/bitmaps/$que\_on -background $col_diffuse \
		        -command "$basecommand $que $j escape" -label $lab \
		        -hidemargin 1
	    } else {
		$w add command -bitmap @$GIDDEFAULT/bitmaps/$que -background $col_diffuse \
		        -command "$basecommand $que $j escape" -label $lab \
		        -hidemargin 1
	    }
	    set jc [ expr $jc + 1]
	}
	if { [ llength $all_que] > 1} {
	    $w add command -label [_ "All On"] -command "$basecommand $que all_$que escape" \
		        -hidemargin 1
	    $w add command -label [_ "All Off"] -command "$basecommand $que no_$que escape" \
		        -hidemargin 1
	}
    } else {
	$w add command -label [_ "Style Window"] -command " \
	    set GidPriv(WinList_volumes) 0; \
	    set GidPriv(WinList_surfaces) 0; \
	    set GidPriv(WinList_cuts) 0; \
	    set GidPriv(WinList_$que) 1; \
	    PostDisplay" \
		        -hidemargin 1
    }
}

proc AddMeshSetCutMenu { w basecommand que args} {
    global GIDDEFAULT GidPriv

    $w del 0 end
    set all_que [GiD_Info postprocess get all_$que]
    if { $all_que == ""} {
	$w add command -label [_ "There are no %s" $que] -state disabled
	return
    }
   
    set all_colors [GiD_Info postprocess get all_$que\_colors]
    set jc 0
    foreach j $all_que {
	regsub {_} $j { } lab
	set col_set [ lindex $all_colors $jc]
	# es una lista { ambient diffuse specular shininess}
	set col_diffuse [ lindex $col_set 1]
	$w add command -background $col_diffuse \
		-command "$basecommand $que $j" -label $lab
	set jc [ expr $jc + 1]
    }
    if { [ llength $all_que] > 1} {
	$w add command -label [_ "All"] -command "$basecommand $que all_$que"
    }
}

proc ToDoWhenChoosingStep { comm } {
    global ViewMenuPriv

    eval [ concat $comm]
    set ViewMenuPriv(DirtyMenus) 1
}

#proc ListCurrentResults { } {
#
#    set cur_anal [GiD_Info postprocess get cur_analisis]
#
#    set cur_step [GiD_Info postprocess get cur_step $cur_anal] 
#
#    set cur_res [GiD_Info postprocess get cur_result]
#
#    if { "$cur_res" == ""} {
#         set info_res [_ "First select any result representation, please"]
#         WarnWin $info_res
#    } else {
#         set info_res [GiD_Info list_entities results $cur_anal $cur_step $cur_res 1:1 196:197 ]
#         #WarnWinText $info_res
#         foreach res_val $info_res {
#             WarnWinText "[ llength $res_val] -> $res_val"
#         }
#    }
#}

proc AddAnalysisStepsToMenu { w args} {
    global ViewMenuPriv GidPriv RMenuRPriv

    rmenuinit
    set all_anal [lsort -dictionary [GiD_Info postprocess get all_analisis]]
    # set all_anal [ GiD_Info postprocess get all_analisis]

    if { ![GiD_Info postprocess get changed_analysis_step] && \
	    ![GiD_Info postprocess get changed_vista_resultados] && \
	    [ info exists ViewMenuPriv($w,all_anal)] && \
	    ( $ViewMenuPriv($w,all_anal) == $all_anal) } {
	set igual 1
	foreach i $all_anal {
	    if { [ info exists ViewMenuPriv($w,$i)]} {
		set all_step [GiD_Info postprocess get all_steps $i]        
		if { $ViewMenuPriv($w,$i) != $all_step} {
		    set igual 0
		    break
		}
	    } else {
		set igual 0
		break
	    }
	}
	if { $igual} {
	    return
	}
    }
    set ViewMenuPriv($w,all_anal) $all_anal

    set cur_anal [GiD_Info postprocess get cur_analisis]
    
    $w del 0 end

    set cur_step [GiD_Info postprocess get cur_step $cur_anal] 

    set cur_res [GiD_Info postprocess get cur_result]
    set cur_res_view  [GiD_Info postprocess get cur_results_view]
    set cur_comp [GiD_Info postprocess get cur_component]

    regsub -all {_} $cur_res_view {} pp
    set restcommand "escape escape escape escape results $pp $cur_res $cur_comp"
    set no_res_command "Geometry NoResults Geometry Original escape escape escape escape View Label Off"
    set iso_hecho 0
    switch $cur_res_view {
	"Iso_Surfaces" {
	    set values [GiD_Info postprocess get iso_cur_result_values]
	    set restcommand "$pp exact $cur_res $cur_comp [lindex $values 0] [lindex $values 1]"
	    set iso_hecho 1
	}
	"Stream_Lines" {
	    set restcommand ""
	}
	"No_Result" {
	    set restcommand ""
	}
	"Display_Vectors" {
	    set value [GiD_Info postprocess get cur_vector_factor $cur_res_view \
		    $cur_res $cur_anal $cur_step ]
	    set restcommand "$restcommand $value escape"
	}
	default {
	}
    }
    
    set iso_cmd ""
    set iso_res ""
    if { !$iso_hecho} {
	set iso_values [ lindex [.central.s info postprocess get iso_cur_result_values] 0]
	if { [ llength $iso_values] != 0} {
	    set iso_res [ lindex $GidPriv(PostAnimateIsoSurfaces) 2]
	    set iso_cmd "escape escape escape escape results IsoSurfaces Exact [ lindex $iso_res 0] [ lindex $iso_res 1]" 
	    set iso_cmd "$iso_cmd [ lindex $iso_values 0] [ lindex $iso_values 1]"
	}
    }

    set basecommand ".central.s process escape escape escape escape results analysissel"

    if { $all_anal == ""} {
	$w add command -label [_ "Please load a MODEL first"] -state disabled
	return
    }

    foreach i $all_anal {
	set menu_txt [= $i]
	regsub -all {_} $menu_txt { } menu_txt
	regsub -all {\.} $i {_} menu_name

	set img $GidPriv(VerifNoImage)
	if { $cur_anal == $i } {
	    set img $GidPriv(VerifImage)
	}
	$w add cascade -label $menu_txt -menu $w.m$menu_name  -image $img -compound left

	set all_step [GiD_Info postprocess get all_steps $i]
	set ViewMenuPriv($w,$i) $all_step
	set cur_step [GiD_Info postprocess get cur_step $i] 
	if { [ winfo exist $w.m$menu_name ]} {
	    $w.m$menu_name del 0 end
	} else {
	    menu $w.m$menu_name
	}
	# foreach j $all_step {
	#     $w.m$menu_name add command -label $j -command "$basecommand $i $j"
	# }
	set words ""

	set entrada_ventana 0
	if { ( [ llength $all_step] > $RMenuRPriv(MaxAbsolute)) && ( $RMenuRPriv(MaxAbsolute) > 2)} {
	    set all_step [ lrange $all_step 0 [ expr $RMenuRPriv(MaxAbsolute) - 2]]
	    set entrada_ventana 1
	}

	foreach j $all_step {
	    set all_res [ GiD_Info postprocess get results_list $cur_res_view $i $j]
	    set cmds $restcommand
	    if { [ lsearch $all_res $cur_res] == -1} {
		set cmds $no_res_command
	    }
	    if { ( $iso_res != "") && ( [ lsearch $all_res [ lindex $iso_res 0]] != -1)} {
		set cmds "$iso_cmd; .central.s process $cmds"
	    }
	    if { ( $cur_anal != $i) || ( $cur_step != $j) } {
		lappend words [ list $j "$basecommand $i $j $cmds"]
	    } else {
		lappend words [ list $j "$basecommand $i $j $cmds" check]
	    }
	}
	rmenuconf $w.m$menu_name $words

	if { $entrada_ventana} {
	    $w.m$menu_name add separator
	    $w.m$menu_name add command -label ...[_ "More"]... -command PostGeom::Create
	}
    }
}

proc ResultsNoStep { } {
    WarnWin [_ "Please select a Step first"]
}

proc GetResultMenuFromPath { w name img} {
    global GidPriv

    if { ![ regexp {^/?/?(.*)//(.*?)$} $name dum path res_name]} {
	#WarnWinText "nothing to do"
	return $w
    }
    regsub -all { } $path {_} nname
    #WarnWinText "doing '$name' $path + $nname"    
    set wm $w
    #set nname [ split $nname /]
    regsub -all {//} $nname { } nname
    #set menu_name ""

    #WarnWinText "Exists $w ? [ winfo exist $w]"
    # WarnWinText "[ llength $nname] - $nname"
    # en windows parece haber un nivel de anidamiento máximo de 10 para el top menu bar
    # 8 del path del resultado + 1 del resultado + 1 del componente del resultado
    if { [ llength $nname] >= 7} {
	return ""
    }
    foreach subm $nname {
	if { "$subm" == ""} { continue}
	#append menu_name /$subm
	set txt [= $subm]
	regsub -all {_} $txt { } txt
	regsub -all {\.} $subm {_} menu_name
	if { [ catch { $wm index "$txt"}] || ( [ $wm index "$txt"] == "none")} {
	    #WarnWinText "$wm Adding cascade m$menu_name LABEL: $txt"
	    if { "$img" != ""} {
		$wm add cascade -label $txt -menu $wm.m$menu_name \
		        -image $img -compound left
	    } else {
		$wm add cascade -label $txt -menu $wm.m$menu_name
	    }
	    if { [ winfo exist $wm.m$menu_name ]} {
		$wm.m$menu_name del 0 end
	    } else {
		menu $wm.m$menu_name
	    }
	} else {
	    #WarnWinText "[winfo class $wm] $wm index '$txt' [ $wm index [ list $txt]]" 
	}
	#WarnWinText "set wm $w.m$menu_name"
	#WarnWinText "set wm $wm.m$menu_name"
	set wm $wm.m$menu_name
    }
    return $wm
}

proc AutomaticConvertCutToMesh { w idx} {
    # WarnWin "$w - [ winfo parent $w]"
    set nuevo 1
    if { [ GiD_Info variables PostCutToMesh]} {
	set nuevo 0
    }
    GiD_Process escape escape escape escape utilities variables PostCutToMesh $nuevo
    
    $w entryconf $idx -image [ IconoAutomaticConvertCutToMesh] -compound left
}

proc IconoAutomaticConvertCutToMesh {} {
    global GidPriv
    set img $GidPriv(VerifNoImage)
    if { [ GiD_Info variables PostCutToMesh]} {
	set img $GidPriv(VerifImage)    
    }
    return $img
}

proc AddResultsToMenu { w basecommand que_res { sub 0}} {
    global ViewMenuPriv GidPriv RMenuRPriv

    rmenuinit
    set all_anal [lsort -dictionary [GiD_Info postprocess get all_analisis]]
    # set all_anal [ GiD_Info postprocess get all_analisis]
    if { $all_anal == ""} {
	set msg [_ "Please load a MODEL first"]
    } else {
	set msg [_ "Please select a STEP first"]
    }
    set cur_anal [GiD_Info postprocess get cur_analisis]
    if { $cur_anal == ""} { 
	$w del 0 end  
	$w add command -label $msg -state disabled
	return
    }

    set cur_step [GiD_Info postprocess get cur_step $cur_anal] 
    if { $cur_step == ""} { 
	$w add command -label [_ "Please select a STEP first"] -state disabled
	$w del 0 end
	return
    }

    if { $que_res != "Deformation" } {
	set all_res [ lsort -dictionary [ GiD_Info postprocess get cur_results_list $que_res]]
    } else {
	set all_res [ lsort -dictionary [ GiD_Info postprocess get show_geom_all_deform $cur_anal \
		                              $cur_step]]
    }
    if { $all_res == ""} { 
	$w del 0 end
	$w add command -label [_ "No useful results found"] -state disabled
	return
    }

    if { ![ info exist ViewMenuPriv($w,OldResView)]} {
	set ViewMenuPriv($w,OldResView) ""
    }

    set cur_res_view [GiD_Info postprocess get cur_results_view]
    if { $que_res == "Deformation" } {
	# mas que nada para que cree los menus
	if { "[GiD_Info postprocess get main_geom_state]" == "Deformed"} {
	    set cur_res_view $que_res
	}
    }

    ## miramos si el menu que ya tenemos aun es valido
    # para el conyazo de windows, que es mas lento que una tortuga con reuma
    if { ( $cur_anal != "") && ( $cur_step != "") && ( $all_res != "") } {
	if { [ info exist ViewMenuPriv($w,$que_res/$sub/analstep)] && \
		 [ info exist ViewMenuPriv($w,$que_res/$sub/allres)] } {
	    if { ![ string compare $ViewMenuPriv($w,$que_res/$sub/analstep) "$cur_anal $cur_step" ] && \
		     ![ string compare $ViewMenuPriv($w,$que_res/$sub/allres) $all_res] } { 
		# esto para actualizar el checkbox del resultado elegido
		if { ( $cur_res_view != $que_res) && ( $ViewMenuPriv($w,OldResView) != $que_res)} {
		    # ahora a mirar los componentes
		    set igual 1
		    foreach i $all_res {
		        set all_comp [ GiD_Info postprocess get components_list $que_res $i $cur_anal $cur_step ]
		        regsub -all { } $i {_} idx
		        if { [ info exists ViewMenuPriv($w,$que_res/$sub/$idx)] } {
		            if { [ string compare $ViewMenuPriv($w,$que_res/$sub/$idx) $all_comp]} {
		                # no son iguales
		                set igual 0
		                break
		            }
		        } else {
		            # no son iguales
		            set igual 0
		            break
		        }
		    }
		    return
		}
	    }
	}
    }

    # para ahorrarnos construir cosas
    set ViewMenuPriv($w,$que_res/$sub/analstep) "$cur_anal $cur_step" 
    set ViewMenuPriv($w,$que_res/$sub/allres) $all_res
    foreach i $all_res {
	set all_comp [ GiD_Info postprocess get components_list $que_res $i $cur_anal $cur_step ]
	regsub -all { } $i {_} idx
	set ViewMenuPriv($w,$que_res/$sub/$idx) $all_comp
    }

    $w del 0 end

    if { $que_res != "Deformation" } {
	set cur_res [GiD_Info postprocess get cur_result]
	set cur_comp [GiD_Info postprocess get cur_component]
    } else {
	set cur_res [GiD_Info postprocess get main_geom_cur_deform]
	set cur_comp ""
    }
    
    set entrada_ventana 0
    if { [ llength $all_res] > $RMenuRPriv(MaxNumCol)} {
	set all_res [ lrange $all_res 0 [ expr $RMenuRPriv(MaxNumCol) - 2]]
	set entrada_ventana 1
    }
#WarnWinText "$cur_res <------------------------------------------"
    if { [ info exists entrara_ya_anyadida]} {
	unset entrara_ya_anyadida
    }
    foreach i $all_res {
	
	#WarnWinText "components_list '$que_res' '$i' '$cur_anal' '$cur_step'"
	#WarnWinText "+++[encoding system]\n([string length $i]:[string bytelength $i]) '$i'\n---"

	set img ""
	if { ( $cur_res != $i) || ( $cur_res_view != $que_res) } {
	    set img $GidPriv(VerifNoImage)
	} else {
	    set img $GidPriv(VerifImage)
	}

	#WarnWinText "GetResultMenuFromPath $w $i $img"
	set wm [ GetResultMenuFromPath $w $i $img]
	if { $wm == ""} {
	    # set entrada_ventana 1
	    # break
	    if { [ regexp {^/?/?([^/]*)} $i dum path_tmp ] } {
		regsub -all {_} $path_tmp { } path_parcial
		if { ![ info exists entrara_ya_anyadida($path_tmp)]} {
		    set trad [_ "More"]
		    $w add command -label "${path_parcial}/... ${trad}..." \
		        -command "PostGeom::Create; PostResultsSetResultsView $que_res"
		    set entrara_ya_anyadida($path_tmp) 1
		}
		continue
	    } else {
		set entrada_ventana 1
		break
	    }
	}
	#WarnWinText "$w getting path '$wm' from '$i'"

	set txt [= $i]
	regsub -all {_} $txt { } txt
	if { "$wm" != "$w"} {    
	    if { [ regexp {^/?/?(.*)//(.*?)$} $i dum path res_txt]} {
		regsub -all {_} $res_txt { } txt
	    }
	}

	set all_comp [GiD_Info postprocess get components_list $que_res $i $cur_anal $cur_step ]

	if { ( [ llength $all_comp] > 1) && \
		( $que_res != "Stream_Lines") && ( $que_res != "Deformation") } {
	    
	    # anyadimos el menu de las componentes
	    #WarnWinText "###[encoding system]\n([string length $i]:[string bytelength $i]) '$i'\n***"
	    regsub -all {\.} $i {_} menu_name
	    #set txt $i

	    #WarnWinText "adding cascade"
	    $wm add cascade -label $txt -menu $wm.m$menu_name \
		    -image $img -compound left

	    #WarnWinText "$wm added cascade '$txt' menu '$wm.m$menu_name'"
	    
	    if { [ winfo exist $wm.m$menu_name ]} {
		$wm.m$menu_name del 0 end
	    } else {
		menu $wm.m$menu_name
	    }
	    
	    foreach j $all_comp {
		set txt_comp [= $j]
		regsub -all {_} $txt_comp { } txt_comp
		set img ""
		if { ( $cur_res != $i) || ( $cur_res_view != $que_res) || \
		        ( $cur_comp != $j ) } {
		    set img $GidPriv(VerifNoImage)
		} else {
		    set img $GidPriv(VerifImage)
		    set ViewMenuPriv($w,OldResView) $que_res
		}
		$wm.m$menu_name add command -label $txt_comp -command "$basecommand $i $j" \
		        -image $img -compound left
	    }
	} else {
	    set postcmd ""
	    #if { $que_res == "Deformation"} {
	    #         set postcmd "escape"
	    #}
	    # anyadimos el comando
	    set img ""
	    if { ( $cur_res != $i) || ( $cur_res_view != $que_res) } {
		set img $GidPriv(VerifNoImage)
	    } else {
		set img $GidPriv(VerifImage)
		set ViewMenuPriv($w,OldResView) $que_res
	    }
	    #WarnWinText "adding command"
	    $wm add command -label $txt -command "$basecommand $i $postcmd" \
		    -image $img -compound left

	    #WarnWinText "$wm added command '$txt'"
	}
    }
    if { [ info exists entrara_ya_anyadida]} {
	unset entrara_ya_anyadida
    }

    if { $entrada_ventana} {
	$w add separator
	$w add command -label ...[_ "More"]... \
	    -command "PostGeom::Create; PostResultsSetResultsView $que_res"
    }

}

proc AddGraphsResultsToMenu { w basecommand que_res args} {
    global ViewMenuPriv

    set all_anal [lsort -dictionary [GiD_Info postprocess get all_analisis]]
    # set all_anal [ GiD_Info postprocess get all_analisis]
    if { $all_anal == ""} {
	set msg [_ "Please load a MODEL first"]
    } else {
	set msg [_ "Please select a STEP first"]
    }
    set cur_anal [GiD_Info postprocess get cur_analisis]
    if { $cur_anal == "" } {
	$w del 0 end
	$w add command -label $msg -state disabled
	return
    }

    set cur_step [GiD_Info postprocess get cur_step $cur_anal] 
    if { $cur_step == ""} { 
	$w del 0 end
	$w add command -label [_ "Please select a STEP first"] -state disabled
	return
    }

    set all_res [ lsort -dictionary [ GiD_Info postprocess get cur_results_list $que_res]]
    if { $all_res == ""} { 
	$w del 0 end
	return
    }

    ## miramos si el menu que ya tenemos aun es valido
    # para el conyazo de windows, que es mas lento que una tortuga con reuma
    if { ( $cur_anal != "") && ( $cur_step != "") && ( $all_res != "") } {
	if { [ info exist ViewMenuPriv($w,$que_res//analstep)] && \
		[ info exist ViewMenuPriv($w,$que_res//allres)] } {
	    if { ![ string compare $ViewMenuPriv($w,$que_res//analstep) "$cur_anal $cur_step" ] || \
		    ![ string compare $ViewMenuPriv($w,$que_res//allres) $all_res] } { 
		# ahora a mirar los componentes
		set igual 1
		foreach i $all_res {
		    set all_comp [ GiD_Info postprocess get cur_components_list $i ]
		    regsub -all { } $i {_} idx
		    if { [ info exists ViewMenuPriv($w,$que_res//$idx)] } {
		        if { [ string compare $ViewMenuPriv($w,$que_res//$idx) $all_comp]} {
		            # no son iguales
		            set igual 0
		            break
		        }
		    } else {
		        # no son iguales
		        set igual 0
		        break
		    }
		}
		return
	    }
	}
    }

    # para ahorrarnos construir cosas
    set ViewMenuPriv($w,$que_res//analstep) "$cur_anal $cur_step" 
    set ViewMenuPriv($w,$que_res//allres) $all_res
    foreach i $all_res {
	set all_comp [ GiD_Info postprocess get cur_components_list $i ]
	regsub -all { } $i {_} idx
	set ViewMenuPriv($w,$que_res//$idx) $all_comp
    }

    $w del 0 end

    set entrada_ventana 0
    if { $all_res != ""} {
	foreach i $all_res {
	    set txt_res [= $i]
	    regsub -all {_} $txt_res { } txt_res
	    set all_comp [GiD_Info postprocess get cur_components_list $i ]
	    
	    set wm [ GetResultMenuFromPath $w $i ""]
	    if { $wm == ""} {
		set entrada_ventana 1
		break
	    }

	    if { $wm != $w} {    
		if { [ regexp {^/?/?(.*)//(.*?)$} $i dum path res_txt]} {
		    set txt_res [= $res_txt]
		    regsub -all {_} $txt_res { } txt_res
		}
	    }

	    if { [ llength $all_comp] > 1 } {
		# anyadimos el menu de las componentes
		regsub -all {\.} $i {_} menu_name
		$wm add cascade -label $txt_res -menu $wm.m$menu_name
		if { [ winfo exist $wm.m$menu_name ]} {
		    $wm.m$menu_name del 0 end
		} else {
		    menu $wm.m$menu_name
		}
		foreach j $all_comp {
		    set txt_comp [= $j]
		    regsub -all {_} $txt_comp { } txt_comp
		    $wm.m$menu_name add command -label $txt_comp -command "$basecommand $i $j"
		}
	    } else {
		# anyadimos el comando
		#if { ( $cur_res != $i) || ( $cur_res_view != $que_res) } {
		$wm add command -label $txt_res -command "$basecommand $i"
		#} else {
		#$w add checkbutton -label $i -onvalue 1 -variable pp -command "$basecommand $i" 
		#set pp 1
		#}
	    }
	}
    }
    
    if { $entrada_ventana} {
	$w add separator
	$w add command -label ...[_ "More"]... -command PostGraphs::Create
    }
}

proc PointEvolutionSetAxisTo { que_eje res res_comp} {
    global ViewMenuPriv

    if { [ info exist ViewMenuPriv(point_evolution_ax$que_eje)]} {
	set old_res [ lindex $ViewMenuPriv(point_evolution_ax$que_eje) 0]
	set old_res_comp [ lindex $ViewMenuPriv(point_evolution_ax$que_eje) 1]
	set ViewMenuPriv(point_evolution_ax$que_eje//$old_res//$old_res_comp) 0
    }
    set ViewMenuPriv(point_evolution_ax$que_eje) "$res $res_comp"
    set ViewMenuPriv(point_evolution_ax$que_eje//$res//$res_comp) 1
}

proc PointEvolutionDoIt { basecommand} {
    global ViewMenuPriv
    
    if { ![ info exist ViewMenuPriv(point_evolution_axX)]} {
	WarnWin [_ "Please set X axis first"]
	return
    }
    if { ![ info exist ViewMenuPriv(point_evolution_axY)]} { 
	WarnWin [_ "Please set Y axis first"]
	return
    }

    
    set comm "$basecommand $ViewMenuPriv(point_evolution_axX) $ViewMenuPriv(point_evolution_axY)"
    if { $::ViewMenuPriv(pointgraphallsteps) } {
	lappend comm AllSteps
    }

    eval [ concat $comm]
}

proc BorderGraphSetAxisTo { que_eje res { res_comp ""}} {
    global ViewMenuPriv

    if { [ info exist ViewMenuPriv(border_graph_ax$que_eje)]} {
	set old_res [ lindex $ViewMenuPriv(border_graph_ax$que_eje) 0]
	set old_res_comp [ lindex $ViewMenuPriv(border_graph_ax$que_eje) 1]
	set ViewMenuPriv(border_graph_ax$que_eje//$old_res//$old_res_comp) 0
    }
    if { $que_eje == "Y"} {
	set ViewMenuPriv(border_graph_ax$que_eje) "$res $res_comp"
	set ViewMenuPriv(border_graph_ax$que_eje//$res//$res_comp) 1
    }                         
    if { $que_eje == "X"} {         
	set ViewMenuPriv(border_graph_ax$que_eje) "$res"
	foreach i "X_Variation Y_Variation Z_Variation LineVariation" {
	    set ViewMenuPriv(border_graph_ax$que_eje//$i) 0
	}
	set ViewMenuPriv(border_graph_ax$que_eje//$res) 1
    } 
}

proc BorderGraphDoIt { basecommand} {
    global ViewMenuPriv
    
    if { ![ info exist ViewMenuPriv(border_graph_axX)]} { 
	WarnWin [_ "Please set X axis first"]
	return
    }
    if { ![ info exist ViewMenuPriv(border_graph_axY)]} { 
	WarnWin [_ "Please set Y axis first"]
	return
    }

    set comm "$basecommand $ViewMenuPriv(border_graph_axX) $ViewMenuPriv(border_graph_axY)"

    eval [ concat $comm]
}

proc AddGraphsXYResultsToMenu { w basecommand que_res args} {
    global ViewMenuPriv

    if { ![info exists ::ViewMenuPriv(pointgraphallsteps)] } {
	set ::ViewMenuPriv(pointgraphallsteps) 1
    }
    set all_anal [lsort -dictionary [GiD_Info postprocess get all_analisis]]
    # set all_anal [ GiD_Info postprocess get all_analisis]
    if { $all_anal == ""} {
	set msg [_ "Please load a MODEL first"]
    } else {
	set msg [_ "Please select a STEP first"]
    }
    set cur_anal [GiD_Info postprocess get cur_analisis]
    if { $cur_anal == ""} { 
	$w del 0 end
	foreach eje "X Y" {
	    $w add cascade -label [_ "Set %s axis" $eje] -menu $w.m$eje
	    if { [ winfo exist $w.m$eje ]} {
		$w.m$eje del 0 end
		$w.m$eje add command -label $msg -state disabled
	    } else {
		menu $w.m$eje
		$w.m$eje add command -label $msg -state disabled
	    }    
	} 
	if { $que_res == "Point_Graph"} {
	    $w add command -label [_ "Select Points"] -command ResultsNoStep
	    $w add checkbutton -label [_ "All steps"] -variable ::ViewMenuPriv(pointgraphallsteps) -indicatoron 1
	}
	if { $que_res == "Border_Graph"} {
	    $w add command -label [_ "Select Border"] -command ResultsNoStep 
	}
	return
    }

    set cur_step [GiD_Info postprocess get cur_step $cur_anal] 
    if { $cur_step == ""} { 
	$w del 0 end
	foreach eje "X Y" {
	    $w add cascade -label [_ "Set %s axis" $eje] -menu $w.m$eje
	    if { [ winfo exist $w.m$eje ]} {
		$w.m$eje del 0 end
		$w.m$eje add command -label [_ "Please select a STEP first"] -state disabled
	    } else {
		menu $w.m$eje
		$w.m$eje add command -label [_ "Please select a STEP first"] -state disabled
	    }    
	}
	if { $que_res == "Point_Graph"} {
	    $w add command -label [_ "Select Points"]  -command ResultsNoStep
	    $w add checkbutton -label [_ "All steps"] -variable ::ViewMenuPriv(pointgraphallsteps) -indicatoron 1
	}
	if { $que_res == "Border_Graph"} {
	    $w add command -label [_ "Select Border"]  -command ResultsNoStep
	}
	return
    }

    set all_res [ lsort -dictionary [ GiD_Info postprocess get cur_results_list $que_res]]
    if { $all_res == ""} { 
	$w del 0 end
	foreach eje "X Y" {
	    $w add cascade -label [_ "Set %s axis" $eje] -menu $w.m$eje
	    if { [ winfo exist $w.m$eje ]} {
		$w.m$eje del 0 end
	    } else {
		menu $w.m$eje
	    }    
	}
	if { $que_res == "Point_Graph"} {
	    $w add command -label [_ "Select Points"]  -command ResultsNoStep
	    $w add checkbutton -label [_ "All steps"] -variable ::ViewMenuPriv(pointgraphallsteps) -indicatoron 1
	}
	if { $que_res == "Border_Graph"} {
	    $w add command -label [_ "Select Border"]  -command ResultsNoStep
	}
	return
    }

    ## miramos si el menu que ya tenemos aun es valido
    # para el conyazo de windows, que es mas lento que una tortuga con reuma
    if { ( $cur_anal != "") && ( $cur_step != "") && ( $all_res != "") } {
	if { [ info exist ViewMenuPriv($w,$que_res//analstep)] && \
		[ info exist ViewMenuPriv($w,$que_res//allres)] } {
	    if { ![ string compare $ViewMenuPriv($w,$que_res//analstep) "$cur_anal $cur_step" ] || \
		    ![ string compare $ViewMenuPriv($w,$que_res//allres) $all_res] } {  
		# ahora a mirar los componentes
		set igual 1
		foreach i $all_res {
		    set all_comp [ GiD_Info postprocess get cur_components_list $i ]
		    regsub -all { } $i {_} idx
		    if { [ info exists ViewMenuPriv($w,$que_res//$idx)] } {
		        if { [ string compare $ViewMenuPriv($w,$que_res//$idx) $all_comp]} {
		            # no son iguales
		            set igual 0
		            break
		        }
		    } else {
		        # no son iguales
		        set igual 0
		        break
		    }
		}
		return
	    }
	}
    }

    # para ahorrarnos construir cosas
    set ViewMenuPriv($w,$que_res//analstep) "$cur_anal $cur_step" 
    set ViewMenuPriv($w,$que_res//allres) $all_res
    foreach i $all_res {
	set all_comp [ GiD_Info postprocess get cur_components_list $i ]
	regsub -all { } $i {_} idx
	set ViewMenuPriv($w,$que_res//$idx) $all_comp
    }

    if { ( [ $w type 0] != "cascade") || ( [ $w type 1] != "cascade") } {
	$w del 0 end
	foreach eje "X Y" {
	    $w add cascade -label [_ "Set %s axis" $eje] -menu $w.m$eje
	    if { [ winfo exist $w.m$eje ]} {
		$w.m$eje del 0 end
	    } else {
		menu $w.m$eje
	    }    
	}
	if { $que_res == "Point_Graph"} {
	    $w add command -label [_ "Select Points"]  -command ResultsNoStep
	    $w add checkbutton -label [_ "All steps"] -variable ::ViewMenuPriv(pointgraphallsteps) -indicatoron 1
	}
	if { $que_res == "Border_Graph"} {
	    $w add command -label [_ "Select Border"]  -command ResultsNoStep
	}
    }

    set lista_ejes "X Y"
    if { $que_res == "Border_Graph"} {
	set lista_ejes "Y"
	set eje X
	# construyamos el menu del eje X
	if { [ winfo exist $w.m$eje ]} {
	    $w.m$eje del 0 end
	}
	set cmds {X_Variation Y_Variation Z_Variation LineVariation}
	set labels [list [_ "X Variation"] [_ "Y Variation"] [_ "Z Variation"] [_ "Line Variation"]]
	foreach i $cmds l $labels {
	    $w.m$eje add checkbutton -label $l -command "BorderGraphSetAxisTo $eje $i" \
		    -onvalue 1 -variable ViewMenuPriv($w,border_graph_ax$eje//$i)
	    set ViewMenuPriv($w,border_graph_ax$eje//$i) 0
	}
    }

    set entrada_ventana 0
    foreach eje $lista_ejes {
	if { [ winfo exist $w.m$eje ]} {
	    $w.m$eje del 0 end
	}

	foreach i $all_res {
	    set txt_res [= $i]
	    regsub -all {_} $txt_res { } txt_res
	    set all_comp [GiD_Info postprocess get cur_components_list $i ]

	    set wm [ GetResultMenuFromPath $w.m$eje $i ""]
	    if { $wm == ""} {
		set entrada_ventana 1
		break
	    }

	    if { "$wm" != "$w.m$eje"} {    
		if { [ regexp {^/?/?(.*)//(.*?)$} $i dum path res_txt]} {
		    set txt_res [= $res_txt]
		    regsub -all {_} $txt_res { } txt_res
		}
	    }

	    if { [ llength $all_comp] > 1 } {
		# anyadimos el menu de las componentes
		regsub -all {\.} $i {_} menu_name
		$wm add cascade -label $txt_res -menu $wm.m$menu_name
		if { [ winfo exist $wm.m$menu_name ]} {
		    $wm.m$menu_name del 0 end
		} else {
		    menu $wm.m$menu_name
		}
		foreach j $all_comp {
		    set txt_comp [= $j]
		    regsub -all {_} $txt_comp { } txt_comp
		    if { $que_res == "Point_Graph"} {
		        $wm.m$menu_name add checkbutton -label $txt_comp \
		                -command "PointEvolutionSetAxisTo $eje $i $j" \
		                -onvalue 1 -variable ViewMenuPriv($w,point_evolution_ax$eje//$i//$j)
		        set ViewMenuPriv($w,point_evolution_ax$eje//$i//$j) 0
		    }
		    if { $que_res == "Border_Graph"} {
		        $wm.m$menu_name add checkbutton -label $txt_comp \
		                -command "BorderGraphSetAxisTo $eje $i $j" \
		                -onvalue 1 -variable ViewMenuPriv($w,border_graph_ax$eje//$i//$j)
		        set ViewMenuPriv($w,border_graph_ax$eje//$i//$j) 0
		    }
		}
	    } else {
		if { $que_res == "Point_Graph"} {
		    $wm add checkbutton -label $txt_res \
		            -command "PointEvolutionSetAxisTo $eje $i $i" \
		            -onvalue 1 -variable ViewMenuPriv($w,point_evolution_ax$eje//$i//$i)
		    set ViewMenuPriv($w,point_evolution_ax$eje//$i//$i) 0
		}
		if { $que_res == "Border_Graph"} {
		    $wm add checkbutton -label $txt_res \
		            -command "BorderGraphSetAxisTo $eje $i $i" \
		            -onvalue 1 -variable ViewMenuPriv($w,border_graph_ax$eje//$i//$i)
		    set ViewMenuPriv($w,border_graph_ax$eje//$i//$i) 0
		}
	    }
	}
    }
    
    if { $que_res == "Point_Graph"} {
	$w entryconfigure 2 -command "PointEvolutionDoIt [ list $basecommand]"
    }
    if { $que_res == "Border_Graph"} {
	$w entryconfigure 2 -command "BorderGraphDoIt [ list $basecommand]"
    }
    
    if { $entrada_ventana} {
	$w add separator
	$w add command -label ...[_ "More"]... -command PostGraphs::Create
    }
}


# Position a dialog box at a reasonable place on the screen.

proc dpos { w } {
    set hh [expr [winfo screenheight .gid]*.55]                
    set hh [string range $hh 0 [expr [string first . $hh]-1]]
    wm geometry $w -0+400        
}

proc GiveDpos {} {
    return "-0+400"
}

proc AddMoreWelcome {} {
    global GIDDEFAULTTCL tcl_platform GidPriv
    set base .gid.pepe

    wm geometry .gid.pepe 630x650
    text $base.text \
	  -height 7 -width 1 -yscroll "$base.scrollbar set"

    scrollbar $base.scrollbar \
	-orient v -command "$base.text yview"

    $base.text ins end [_ "%s internal version: %s" $GidPriv(ProgName) [GiD_Info GiDVersion]]\n
    
    if { 0 } {
	set compilationtime ""
	if { ![ catch { set fcomp [ open [ file native [ file join $GIDDEFAULTTCL Compiled]]]}]} {
	    while { ![ eof $fcomp] } {
		set compilationtime [read $fcomp]
	    }
	    close $fcomp
	    $base.text ins end [_ "Last compilation time: %s" $compilationtime]\n
	}
    }

    $base.text ins end [_ "Operating system"]:
    foreach i "machine platform os osVersion" {
	 $base.text ins end " '$tcl_platform($i)'"
    }
    $base.text ins end \n

    catch { $base.text ins end [GiD_Info internalstate] }

    if { ![ catch { set fcomp [ open [ file join $GIDDEFAULTTCL Compiled-more]]}] } {
	while { ![ eof $fcomp] } {
	     $base.text ins end [ gets $fcomp]
	     $base.text ins end \n
	}
	close $fcomp
    }


    grid $base.text -in $base        -row 3 -column 1  \
	    -columnspan 3 \
	    -sticky nesw
    grid $base.scrollbar -in $base        -row 3 -column 3  \
	    -sticky nse
    
    grid rowconfigure $base 3 -weight 1 
    
    $base.button_more conf -text [_ "Less"] -command { 
	destroy .gid.pepe.text
	destroy .gid.pepe.scrollbar
	wm geometry .gid.pepe 630x500

	.gid.pepe.button_more conf -text [_ "More"] -command "AddMoreWelcome" } 
}



proc CreateWelcomeWin { { overw .gid } } {
    global GIDDEFAULT GidPriv
    global tcl_platform

    toplevel .gid.pepe
    wm title .gid.pepe $GidPriv(ProgName)
    set base .gid.pepe
    wm minsize $base 630 500
    wm maxsize $base 630 10000
	
    wm geometry .gid.pepe +[expr [winfo x $overw]+[winfo width $overw]/2-315]+[ \
	    expr [winfo y $overw]+[winfo height $overw]/2-315]
    
    set imatge_about [image create photo -file $GidPriv(AboutWindow)]  
    set imatge_tapa  [image create photo -file [file join $GIDDEFAULT bitmaps about_grey.gif]] 
    
    if { $GidPriv(TextOfVersion) != "" } {
	set GidVersion $GidPriv(TextOfVersion)
    } else {
	set GidVersion [GiD_Info GiDVersion]
    }
    
   
    label $base.label_img -image $imatge_about 
    label $base.label_tapa -image $imatge_tapa -bd 0 -bg #595959 -anchor s
    
    place $base.label_tapa -x 448 -y 144 -in $base
    
    button $base.button_ok \
	    -text [_ "OK"] -width 6 -padx 2  -command "destroy .gid.pepe"
    
    button $base.button_more \
	    -text [_ "More"] -padx 2  -command "AddMoreWelcome"
    
    global credits_about
    set credits_about 0
    
    
    button $base.button_credits \
	    -text [_ "Credits"] -padx 2  -command "start_credits $base 0"                    
    
    
    SetWidgetsWidthFromText $base.button_ok $base.button_more $base.button_credits
    
    # Geometry management
    
    grid $base.label_img -in $base        -row 1 -column 1  \
	    -columnspan 3
    grid $base.button_ok -in $base        -row 2 -column 1  \
	    -sticky e -padx 8 -pady 8 
    
    grid $base.button_credits -in $base        -row 2 -column 2  \
	    -padx 8 -pady 8 
    
    grid $base.button_more -in $base        -row 2 -column 3  \
	    -sticky w -padx 8 -pady 8 
    
    
    # ponemos el numero de version encima de la imagen
    canvas $base.c -borderwidth 0 -highlightthickness 0 -width 400 -height 35 -bg black
    
    place $base.c -x 8 -y 422 -in $base
	
    
    # frame $base.fv -bd 0
    # label $base.fv.label_version -text $GidVersion -font BigFont -fg #cccccc -bg black
    # grid $base.fv.label_version -sticky we
    
    #set rw [ winfo reqwidth $base.fv.label_version]
    #place $base.label_version -x [ expr [ winfo reqwidth $base.label_img] - $rw - 16] -y 420 -in $base
    
    
    # Resize behavior management
    
    grid rowconfigure $base 1 -weight 0 -minsize 30
    grid rowconfigure $base 2 -weight 0 -minsize 30
    grid rowconfigure $base 3 -weight 1 
    
    
    grid columnconfigure $base 1 -weight 1 -minsize 30
    grid columnconfigure $base 2 -weight 1 -minsize 30
    grid columnconfigure $base 3 -weight 1 -minsize 30
    
    bind $base <Return> "destroy .gid.pepe"
    bind $base <Escape> "destroy .gid.pepe"
    

    tkwait visibility .gid.pepe
    grab  .gid.pepe 


    # a mover las letras
    update idletasks
    if { ![ winfo exist $base.c ]} {
	break
    }
    set txt [_ "Version %s" $GidVersion]
    set len_txt [ string length $txt]
    set dest_x [ expr [ $base.c cget -width] - 36]
    for { set is 0} { $is < $len_txt} { incr is} {
	frame $base.fv$is -bd 0
	label $base.fv$is.l -text [ string index $txt [ expr $len_txt - $is - 1]] \
		-font "Helvetica 14 bold italic" -fg #aaaaaa -bg black -bd 0 -highlightthickness 0
	grid $base.fv$is.l -sticky we -padx 0 -pady 0 
	
	set rw [ winfo reqwidth $base.fv$is.l]
	set dest_x [ expr $dest_x - $rw]
	set ini_x 0
	set c_id [ $base.c create window $ini_x 0 -anchor nw -window $base.fv$is]
	set dif_x [ expr $dest_x - $ini_x]
	update
	update idletasks
	for { set ix $ini_x} { $ix < $dest_x} { set ix [ expr $ix + 5]} {
	    # por si han cerrado la ventana antes de acabar con las letritas
	    if { ![ winfo exist $base.c ]} {
		break
	    }
	    $base.c coords $c_id $ix 0 
	    update
	    update idletasks
	}
	# por si han cerrado la ventana antes de acabar con las letritas
	if { ![ winfo exist $base.c ]} {
	    break
	}
	$base.c coords $c_id $dest_x 0 
	
    }


    # por si han cerrado la ventana antes de acabar con las letritas
    if { [ winfo exist .gid.pepe]} {
	tkwait window .gid.pepe
    }
}


proc start_credits {base i} {
    global credits_about
    if {$credits_about==0} {  show_credits $base $i }
}

proc show_credits {w i} {
    #credits
    global credits_about
    
    set credits_about 1
    
    if {[winfo exist $w.label_tapa]} {
	$w.label_tapa configure -height [expr {143-$i}]
	place $w.label_tapa -x 448 -y [expr 144+$i] -in $w
    } else {set i 9998}
    incr i
    if {$i<143} {after 60 "show_credits $w $i"}   
}




proc ChangeWindowTitle { { projectname "" } } {
    global GidPriv
    #if { [winfo exists .gid.central.s] && [.gid.central.s disable windows] } return

    if { $projectname == "" } {
	set projectname [GiD_Info Project ModelName]
    }
    set que64 ""
    if { [ lsearch {"amd64" "x86_64"} "$::tcl_platform(machine)"] != -1 } {
	set que64 x64
    }
    wm title .gid [join [ list "$GidPriv(ProgName) $que64                       " \
		  [_ "Project: %s" [file tail $projectname] ] ]]
    wm iconname .gid [file tail $projectname]
}

proc MakeNoise { name } {
    global GIDDEFAULT
#     catch { exec playaiff $GIDDEFAULT/audio/$name & }
}


proc PrintLayerName {} {
    set layer [GiD_Info Project LayerToUse]
    if { $layer != "*****" } { 
	ChangeWarnLineVar [_ "Layer in use is '%s'" $layer]
    } else {
	ChangeWarnLineVar [_ "There is no layer in use"]
    }
}

proc CreateIcon {} {
    global GIDDEFAULT tcl_platform

    # is it necessary for windows ?
    if { $tcl_platform(platform) == "windows" } {
	wm iconbitmap .gid -default $GIDDEFAULT/bitmaps/gid.ico
	wm iconbitmap .gid $GIDDEFAULT/bitmaps/gid.ico
    } else {
	wm iconbitmap .gid @$GIDDEFAULT/bitmaps/GiDicono64.xbm
    }
#    wm iconbitmap . @$GIDDEFAULT/bitmaps/gid_icon.xbm
##    wm iconbitmap . @$GIDDEFAULT/bitmaps/GiDicono.xbm 
#-background yellow -foreground black
#     catch {
#         set imm [image create bitmap -background black -foreground \
#                 AntiqueWhite3 -file $GIDDEFAULT/bitmaps/gid_icon.xbm]
#         catch { destroy .iconw }
#         toplevel .iconw
#         label .iconw.l -image $imm
#         pack .iconw.l
#         wm iconwindow . .iconw
#     }
#     if { $tcl_platform(platform) != "windows" } {
#          catch {
#        set imm [ image create bitmap -file $GIDDEFAULT/bitmaps/GiDicono.xbm \
#                 -background yellow -foreground black ]
#              set imm [ image create photo -file $GIDDEFAULT/bitmaps/GiDicono64.gif ]
#              toplevel .__t__icw
#              label .__t__icw.l -image $imm -bd 0
#              pack .__t__icw.l
#              wm iconwindow . .__t__icw
#          }
#     }
}

proc OpenTemporaryWindow { text commands {names ""} { help "" } } {

    if { $names == "" } { regsub -all escape $commands Cancel names}

    if {$help == "" } {
	set var [ list tk_dialogRAMFull .gid.tempwinw [_ "Dialog window"] $text \
		"" "" gidquestionhead 0]
    } else {
	set HelpCommand [ list HelpWindow "" $help .gid.tempwinw.helpwin]
	
	set var [ list tk_dialogRAMHelp .gid.tempwinw [_ "Dialog window"] $text \
		"" "" gidquestionhead $HelpCommand 0]
    }
    append var " $names"
    set retval [eval $var]

    after idle .central.s process [lindex $commands $retval]
}

proc OpenTemporaryEntryWindow { text {default ""} {type any} {btnstxt ""} } {

    if { $btnstxt == "" } {
	set retval [tk_dialogEntryRAM .gid.tempwine [_ "Enter value window"] $text gidquestionhead \
		$type $default]
    } else {
	set retval [tk_dialogEntryRAM .gid.tempwine [_ "Enter value window"] $text gidquestionhead \
		$type $default $btnstxt]
    }

    if { $retval == "--CANCEL--" } { 
	GiD_Process escape 
    } else { 
	GiD_Process $retval
    }
}

proc OpenTemporaryComboWindow { text AcceptNew args} {
    if { [llength $args] } {
	set retval [tk_dialogComboRAM .gid.tempwinec [_ "Enter value window"] \
		$text gidquestionhead $AcceptNew $args]
    } else {
	set retval [tk_dialogComboRAM .gid.tempwinec [_ "Enter value window"] $text \
		gidquestionhead $AcceptNew ]
    }
    if { $retval == "--CANCEL--" } { 
	GiD_Process escape 
    } else { 
	GiD_Process $retval
    }
}

proc AddProblemtypeNamesToVar { prefix dir varname recurse } {
    global GiDVisitedDir
    if { [info exists GiDVisitedDir($dir)] } {
	return ;#to avoid infinite loops with links
    }
    set GiDVisitedDir($dir) 1
    upvar #0 $varname var
    set dirs [glob -nocomplain -tails -directory $dir -types d *]
    foreach link [glob -nocomplain -tails -directory $dir -types f *.lnk] {
	if { [GidUtils::IsLink [file join $dir $link]] == 2 } {
	    lappend dirs $link
	}
    }
    foreach i [lsort -dictionary $dirs] {
	set ifull [file join $dir $i]
	if { [file extension $i] == ".lnk" && [GidUtils::IsLink $ifull] == 2 } {
	    set i [file root $i] ;#to remove .lnk
	    set ifull [GidUtils::ReadLink $ifull] 
	}
	if { [file extension $i] == ".gid" } {
	    if { $prefix == "" } {
		set name [file root $i]
	    } else { 
		set name [file join $prefix [file root $i]] 
	    }
	    set var($name) ""
	} elseif { $recurse } {
	    if { $prefix == "" } {
		set name $i
	    } else { 
		set name [file join $prefix $i]
	    }
	    AddProblemtypeNamesToVar $name $ifull $varname 1
	}
    }
}

proc OpenTemporaryTransformProblemTypeWindow { text } {
    
    global GIDDEFAULT GidPriv TransformPrtypeVar
    global GiDVisitedDir
    
    set current [GiD_Info Project ProblemType]
    
    unset -nocomplain TransformPrtypeVar
    unset -nocomplain GiDVisitedDir
    #AddProblemtypeNamesToVar "" $GIDDEFAULT TransformPrtypeVar 0
    AddProblemtypeNamesToVar "" [file join $GIDDEFAULT problemtypes] TransformPrtypeVar 1

    set args [lsort -dictionary [array names TransformPrtypeVar]]
    catch { unset TransformPrtypeVar }
    
    if { [llength $args] } {
	set retval [tk_dialogComboRAM .gid.tempwinetp [_ "Enter name"] \
		$text gidquestionhead 1 $args]
    } else {
	set retval [tk_dialogComboRAM .gid.tempwinetp [_ "Enter name"] $text \
		gidquestionhead 1]
    }
    if { $retval == "--CANCEL--" } { 
	.central.s process escape 
    } else { 
	.central.s process $retval
    }
}

proc OpenPasswordEntryWindow { text {default ""} {type any} } {
  global GidPriv
  
  if { $default != "MODULO"} {
    set default ""
    set fin 0
    while { !$fin} {
      set retval [tk_dialogEntryPASS .gid.tempwineps \
		      [_ "Enter password window"] $text gidquestionhead \
		      $type $default]
      if { $retval == "--CANCEL--" } { 
	GiD_Info setpassword "--CANCEL--abcdef" 
	set fin 1
      } elseif { $retval == "--ACADEMIC--" } { 
	GiD_Info setpassword "--ACADEMIC--" 
	set fin 1
      } else { 
	# longitud password v7 = 24 chars, v6 y anteriores = 16
	# accept also a password-server valid IP
	if { [ string length $retval] > 20 ||
	     [regexp {([0-9]+)[.]([0-9]+)[.]([0-9]+)[.]([0-9]+):?([0-9]+)?} $retval c1 c2 c3 c4 c5] } {
	  GiD_Info setpassword $retval
	  set fin 1
	} else {
	  set msg [_ "This password is not valid for GiD 7.x."]
	  append msg "\n"
	  append msg [_ "Passwords and licences from version 6.x or older are incompatible with this version."]
	  append msg "\n"
	  append msg [_ "Get a temporal password at %s" $GidPriv(PassWeb)]
	  append msg "\n"
	  append msg [_ "Please, contact your dealer to get a new licence."]
	  WarnWin $msg
	}
      }
    }
    
  } else {
    set default ""
    set retval [tk_dialogEntryPASS .gid.tempwineps [_ "Register Problem type"] $text gidquestionhead \
		    $type $default 1]
    set GidPriv(pass_info) $retval
    if { $retval != ""} {
	if {$retval != "--CANCEL--" && ( $GidPriv(RegisterModuletext) == "" || 
	$GidPriv(RegisterModuletext) == "..." ) } {
	    WarnWin [_ "Before entering the password, it is necessary to select one or more Problem types"]
	} 
	GiD_Info setpassword "$retval"
    }
  }
}


proc SavePassword { pass_file machine password date module } {
    set exists 0
    if { [llength [split $password .]] == 4 } {
	#for an IP save only a word: the IP, avoiding the machine
	set buf "$password # $date Password for module '$module'"
    } else {
	set buf "$machine $password # $date Password for module '$module'"
    }
    if { [file exists $pass_file] } {
	set fp [open $pass_file r]
	if { $fp != "" } {
	    while { [gets $fp line] >= 0} {
		if { [string trim $line] == $buf } {
		    set exists 1
		    break
		}
	    }
	    close $fp
	}
    }
    if { !$exists } {
	set fp [open $pass_file a]
	if { $fp != "" } {
	    puts $fp $buf
	    close $fp
	}
    }
}

#return fail==1 if some password module fail, -1 if can't be verified or 0 else
proc GiDValidateModulePassword {def_pass_file id_name academic usb_pass_file password modules disablewindows } {
  global GidPriv GIDDEFAULT

  #set pass_info $GidPriv(pass_info)
  #if {[llength $pass_info] != 2} {
  #  return 0
  #}
  #set pass [string trim [lindex $pass_info 0]]
  #set pt_list [lindex $pass_info 1]
  set pass [string trim $password]
  if {$pass eq ""} {
    # empty password is invalid
    return 0
  }
  set pt_list $modules
  if {$pass eq "--ACADEMIC--"} {
    set pass $academic
  }
  set date [clock format [clock second] -format "%Y %m %d"]
  set invalid_list [list]
  set valid_list [list]
  set unverified_list [list]
  set tdom_bad [catch {package require tdom}]
  proc trunc_fname {name} {
    set items [file split $name]
    if {[llength $items]>3} {
      return [eval file join ... [lrange $items end-2 end]]
    }
    return $name
  }
  foreach dir $pt_list {
    if { [file exists $dir.lnk] && [GidUtils::IsLink $dir.lnk] == 2 } {
	set dir [GidUtils::ReadLink $dir.lnk] 
    }
    set pass_file $def_pass_file
    # load validation routine
    set xmlfile [file join $dir [file root [file tail $dir]]].xml
    set possible_msg ""
    if {!$tdom_bad && [file exists $xmlfile]} {
      set xml [tDOM::xmlReadFile $xmlfile]
      if {[info exists xmldoc]} {
	$xmldoc delete
      } 
      set xmldoc [dom parse $xml]
      set root [$xmldoc documentElement]  
      set vp_node [$root select Program/ValidatePassword]
      if {$vp_node ne ""} {
	set vp_script [$vp_node text]
	if {$vp_script eq ""} {
	  lappend unverified_list [list $dir [_ "Node ValidatePassword doesn't have script"]]
	  continue
	}
	# now try to validate the password
	namespace eval GiD_TmpNS set dir [list $dir]
	namespace eval GiD_TmpNS set key [list $pass]
	namespace eval GiD_TmpNS set computername $id_name
	set tcl_code [catch {
	  namespace eval GiD_TmpNS $vp_script
	} catch_res]
	namespace delete GiD_TmpNS
	if {$tcl_code == 1} {
	  lappend unverified_list \
	      [list $dir [_ "tcl error: %s" $catch_res]]
	  continue
	}
	if {[llength $catch_res] == 2} {
	  set it0 [lindex $catch_res 0]
	  if {[string is integer $it0]} {
	    if {!$it0} {
	      lappend invalid_list \
		  [list $dir [lindex $catch_res 1]]
	      continue
	    }
	    if {$it0 == 2 && [lindex $catch_res 1] ne ""} {
	      lappend valid_list \
		  [list $dir [lindex $catch_res 1]]
	      continue
	    }
	    set possible_msg [lindex $catch_res 1]
	  }
	} elseif {$catch_res eq "0"} {
	  lappend invalid_list [list $dir "password rejected"]
	} elseif {$catch_res eq "2"} {
	  # 2 means password has being written.
	  continue
	}
      } else {
	  #lappend unverified_list [list $dir [_ "Node ValidatePassword not exists"]]
      }
      set passdir_node [$root select Program/PasswordPath]
      if {$passdir_node ne ""} {
	set passdir_txt [$passdir_node text]
	if {$passdir_txt eq ""} {
	  lappend invalid_list \
	      [list $dir [_ "Node PasswordPath is empty"]]
	  continue
	}
	if {[file pathtype $passdir_txt] eq "absolute"} {
	  set dir $passdir_txt
	} else {
	  set dir [file join $dir $passdir_txt]
	}
      }
    } else {
	lappend unverified_list [list $dir [_ "The password has not been verified now. It will be verified when running the analysis code"]]
    }

    # write the password in the default way
    # now write the password info.
    set pass_file [file join $dir $pass_file]
    set exists_file [file exists $pass_file]
    if {!$exists_file} {
      if {![file writable [file dir $pass_file]]} {
	lappend invalid_list \
	    [list $dir \
		 [_ "couldn't write into %s" [trunc_fname $pass_file]]]
	continue
      }          
    } elseif {![file writable $pass_file]} {
      lappend invalid_list \
	  [list $dir \
	       [_ "file %s isn't writable" [trunc_fname $pass_file]]]
      continue          
    }
    
    set dir [file normalize $dir]
    if { $usb_pass_file != {} } { 
	set lst_mod [ file split $dir]
	set lst_gid [ file split $GIDDEFAULT]
	lappend lst_gid problemtypes
	set idx 0
	set fin [ llength $lst_gid]
	set fin2 [ llength $lst_mod]
	if { $fin2 < $fin} { set fin $fin2}
	for { set idx 0 } { ( $idx < $fin) && \
		( [ string tolower [ lindex $lst_mod $idx]] == [ string tolower [ lindex $lst_gid $idx]])} { incr idx} {
	}
	set module_name [ eval file join [ lrange $lst_mod $idx end]]
	regsub -all { } $module_name {_} tmp_mod
	regsub -all {\\} $tmp_mod {/} tmp_mod2
	set module [ string tolower $tmp_mod2]

	SavePassword $usb_pass_file $id_name $pass $date $dir
	# y en el archivo normal tambien:
	SavePassword $pass_file $id_name $pass $date $dir
    } else {
	SavePassword $pass_file $id_name $pass $date $dir
    }
    if {$possible_msg ne ""} {
      lappend valid_list [list $dir $possible_msg]
    }
  }

  if { !$disablewindows } {
      set title [_ "Password Validation"]
      set rows [list]

      set GidPriv(RegisterModuletext) ""
      set GidPriv(RegisterModule) ""
      foreach i $invalid_list {
	  set pt_dir [lindex $i 0]
	  set pt_name [file root [file tail $pt_dir]]
	  lappend GidPriv(RegisterModuletext) $pt_name
	  lappend GidPriv(RegisterModule) $pt_dir
	  lappend rows [list $pt_name [_ "fail"] [lindex $i 1]]
      }
      foreach i $unverified_list {
	  set pt_dir [lindex $i 0]
	  set pt_name [file root [file tail $pt_dir]]
	  lappend rows [list $pt_name [_ "unverified"] [lindex $i 1]]
      }
      foreach i $valid_list {
	  set pt_dir [lindex $i 0]
	  set pt_name [file root [file tail $pt_dir]]
	  lappend rows [list $pt_name [_ "ok"] [lindex $i 1]]
      }
    
      if {[llength $rows]} {
	  if {[winfo exists .gid.tempwineps]} {
	      set ivpw .gid.tempwineps.ivpw
	  } else {
	      set ivpw .gid.ivpw
	  }
	  if {[llength $rows]==1} {
	      set i [lindex $rows 0]
	      set pt_name [lindex $i 0]
	      MessageDlg $ivpw -title $title -type ok \
		      -message [lindex $i 2]
	  } else {
	      dialogwin_snit $ivpw -title $title -cancelname -
	      set f [$ivpw giveframe]
	      tablelist::tablelist $f.tb -width 60 \
		      -columns [list 10 [_ "Problem type"] 10 [_ "Status"] 40 [_ "Reason"]] 
	      grid $f.tb -sticky snew
	      grid rowconfigure $f 0 -weight 1
	      grid columnconfigure $f 0 -weight 1
	      eval $f.tb insert end $rows
	      $ivpw createwindow 
	      destroy $ivpw
	  }
      }
  }

  set has_invalid [llength $invalid_list]
  set has_valid [llength $valid_list]
  set has_unverified [llength $unverified_list]
  if { $has_invalid > 0 } {
      set fail 1
  } else {
      if { $has_unverified > 0 } {
	  set fail -1
      } else {
	  set fail 0
      }
  }
  return $fail
}

proc RegisterFromFileW { } {
    global GidPriv
    if { ![info exists GidPriv(filename,RegisterFromFileW)] } {
	set GidPriv(filename,RegisterFromFileW) ""
    }
    set w .gid
    set types [list [list [_ "Passwords file"] ".pwd"] [list [_ "All files"] ".*"]]
    set filename [Browser-ramR file read $w [_ "read passwords file"] \
	    $GidPriv(filename,RegisterFromFileW) $types .pwd 0]
    if { $filename != "" } {
	set GidPriv(filename,RegisterFromFileW) [file tail $filename]
	set msg [_ "Register status"]\n
	foreach item [RegisterFromFile $filename] {
	    if { [lindex $item 0] == 0 } {
		set res [_ "ok"]
	    } elseif { [lindex $item 0] == 1 } {
		set res [_ "fail"]
	    } elseif { [lindex $item 0] == -1 } {
		#unverified
		set res "??"
	    } else {
		#unexpected
		set res "!!"
	    }
	    set type [lindex $item 1]
	    if { $type == "GID" } {
		set msgi "$GidPriv(ProgName) for [lindex $item 2] ([lindex $item 3])"
	    } elseif { $type == "GID_PASSERVER" } {
		set msgi "$GidPriv(ProgName) from server [lindex $item 2]"
	    } elseif { $type == "MODULE" } {
		set msgi "[lindex $item 4] for [lindex $item 2] ([lindex $item 3])"
	    } else { ;#UNKNOWN
		set msgi "unknown $line"
	    }
	    append msg "$res ----- $msgi\n"
	}
	WarnWinText $msg
    }
}

proc RegisterFromFile { filename } {
    # sample of file contents:
    # mediador 33137efc1jqvyvy0 # 2006 01 10 Password for Problem type 'C:/gid project/problemtypes/tdyn4.10/tdyn3d.gid'
    # mediador a6b0c6561jpp4v70 # 2005 05 03 Password for module 'C:/gid project/problemtypes/Tdyn5/tdyn2d.gid'
    # mediador bq9akuu1p851cub9vnmmv2kc1g # V8 Registration data: Mar 16 2006; Expiration data: Jan  1 2035.
    # usb_17 badakuu1p851cub9vnmmv2kc1g # V7 Registration data: Mar 16 2006; Expiration data: Jan  1 2035.
    # 142.183.133.108:1625
    set res ""
    set fp [open $filename r]
    if { $fp != "" } {
	while { [gets $fp line] >= 0} {
	    set line [string trim $line]
	    if { [string index $line 0] == "#" } continue
	    #si el comentario contiene llaves abiertas o cosas asi no seria una lista, y llength daria error!!!
	    if { [llength $line] == 1 } {
		set type GID_PASSERVER
		set ip [lindex $line 0]                
		set fail [gidregister GiD $ip]
	    } elseif { [llength $line] > 1 } {
		set machine [lindex $line 0]
		set password [lindex $line 1]
		set comments [lrange $line 2 end]
		if { [regexp -nocase {(?:Password for module|Password for Problem type)\s+'(.*)'} $comments dummy pt] } {  
		    set type MODULE   
		    set ptsplit [file split $pt]
		    set pos [lsearch $ptsplit problemtypes]
		    if { $pos != -1 } {
		        set ptrelative [eval file join [lrange $ptsplit [expr $pos+1] end]]
		    } else {
		        set ptrelative $pt
		    }
		    set ptfull [file join $::GIDDEFAULT problemtypes $ptrelative]
		    set existspt 0
		    if { [file exists $ptfull] && [file isdirectory $ptfull] } {
		        set existspt 1                      
		    } elseif { [file exists $ptfull.lnk] && [GidUtils::IsLink $ptfull.lnk] == 2 } {
		        set existspt 1
		        set ptfull [GidUtils::ReadLink $ptfull.lnk] 
		    } elseif { [file exists [file dirname $ptfull].lnk] && \
		            [GidUtils::IsLink [file dirname $ptfull].lnk] == 2 } {
		        set existspt 1
		        set ptname [file tail $ptfull]
		        set ptfull [file join [GidUtils::ReadLink [file dirname $ptfull].lnk] $ptname]
		    }
		    if { $existspt } {
		        set fail [gidregister module $machine $password [list $ptfull]]
		    } else {
		        #molule not found
		        set fail 1
		    }
		} else {
		    #try for GiD
		    set type GID
		    set fail [gidregister GiD $machine $password]
		}
	    } else {
		set type UNKNOWN
		set fail 1
	    }
	   
	    if { $type == "GID" } {
		lappend res [list $fail $type $machine $password]
	    } elseif { $type == "GID_PASSERVER" } {
		lappend res [list $fail $type $ip]
	    } elseif { $type == "MODULE" } {
		lappend res [list $fail $type $machine $password $ptrelative]
	    } else { ;#UNKNOWN
		lappend res [list $fail $type $line]
	    }
	   
	}
	close $fp
    }
    return $res
}

proc PrintTimeTable { tt } {
    upvar #0 $tt aa
    puts "\n--------------TIME TABLE--------\n"
    foreach {name val} [array get aa] {
	puts "Time for [format %-20s:%lf $name \
		[expr 1e-6*[lindex $val 0]]] seconds"
    }
}

#ChooseType == "project" or "file"
#OpenMode == "read" or "write"
#w == parent
#title == "title of window"
#DefaultFile == "Name of file"
#filetypes == list of names and extensions, e.g: {{"Jpeg Image" ".jpg .jpeg"} {"All files" ".*"} }
#defaultextension == extension to automatically append to name, e.g: ".jpg
#Multiple == 1 can return multiple filenames , Multiple == 0 only one
#NOTE: if ChooseType == "project" then filetypes and defaultextension are set automatically for GiD
proc Browser-ram { ChooseType OpenMode { w .gid }  { title "Browser" } 
		   { DefaultFile {} }  { filetypes {} } { defaultextension {} }
		   { Multiple 0 } 
		   { MoreOpts {} } } {
    
    # Traduccion argumentos por defecto
    if { $title == "Browser" } {
	set title [_ "Browser"]
    }
    set ret [Browser-ramR $ChooseType $OpenMode $w $title $DefaultFile $filetypes $defaultextension $Multiple [lrange $MoreOpts 0 1]]


    update
    #update idletasks

    if { $ret == "" } { 
	GiD_Process escape 
    } else { 
	
	if { $ChooseType == "project" && $OpenMode == "write" } {
	    set SelectedSaveVersion \
		$::gidtk::dialog::file::__tk_filedialog(typeCombo,textvar)
	    if { $SelectedSaveVersion == [concat [_ "GiD project"] (*.gid)] } {
		GiD_Process "geoversion:current"
	    } elseif { $SelectedSaveVersion == [concat [_ "Old GiD project 7"] (*.gid)] } {
		GiD_Process "geoversion:2.24"
	    } else {
		GiD_Process "geoversion:current"
	    }
	}
	if {[llength $MoreOpts]>2} {
	    # MoreOptions[2] is responsible of inserting in .process
	    # the extra arguments.
	    eval [lindex $MoreOpts 2]
	    if {[string index $ret 0] eq "-"} {
		GiD_Process "--"
	    }
	}
	GiD_Process $ret
    }
}

array set ::VXLOptions {
    ab,dir            "open"
    isov            0
    method,names    {"Marching cubes" "Uniform Hexahedra" 
		     "Orthogonal Hexahedra"}
    method,codes    {"mc" "hu" "fd"}
    method,current  0
}

proc VxlOptsChangeMethod { w } {
    set ::VXLOptions(method,current) [$w getvalue]
}

#
# CreateReadVxlOpts --
#
# Este proc es invocado por el dialogo de importar volumen de voxels
# (.vtk). Recibe como argumentos un frame donde se deben crear las
# opciones extras necesarias para importar los datos. El resultado del
# proc debe ser una variable donde dejar el estado del frame
# contenedor de opciones extras: open o close.
#
proc CreateReadVxlOpts { f } {
    set CCBG [CCColorActivo [$f cget -background]]
    label $f.liso -text [_ "Iso-surface:"] -bg $CCBG
    entry $f.eiso -textvariable ::VXLOptions(isov)
    label $f.lmet -text [_ "Method:"] -bg $CCBG
    ComboBox $f.cbmet -values $::VXLOptions(method,names) -editable false \
	-modifycmd "VxlOptsChangeMethod $f.cbmet" 
    #$f.cbmet configure -selectbackground [$f.cbmet cget -background]
    $f.cbmet setvalue "@$::VXLOptions(method,current)"
    grid $f.liso $f.eiso $f.lmet $f.cbmet
    return "VXLOptions(ab,dir)"
}

#
# ProcessReadVxlOpts --
#
# Este proc es invocado al aceptar el dialogo de importar de importar
# volumen de voxels (.vtk). Debe encargarse de enviar al GiD_Process
# los argumentos extras necesarios para importar los datos.
#
proc ProcessReadVxlOpts { } {
    set meth [lindex $::VXLOptions(method,codes) $::VXLOptions(method,current)]
    GiD_Process "-isovalue:$::VXLOptions(isov)" "-method:$meth" "--"
}

array set ::JPEGOptions {
    dir "close"
    widget ""
}

proc CreateExportJPEGOpts { f } {
    set CCBG [CCColorActivo [$f cget -background]]
    label $f.lquality -text [_ "Quality:"] -bg $CCBG
    set ::JPEGOptions(widget) \
	[scale $f.quality \
	     -bg $CCBG -from 0 -to 100 \
	     -orient "horizontal"]
    $f.quality set [.central.s info variables "JPEGQuality"]
    grid $f.lquality $f.quality -sticky "s"
    return "JPEGOptions(dir)"
}

proc ProcessExportJPEGOpts { } {
    GiD_Process "-quality:[$::JPEGOptions(widget) get]" "--"
}

array set ::ImportGeomOptions {
    dir "close"
    var,collapse 0
    var,ignore 0
    var,tolerance 0
    var,autotol 0
    widget,f ""
}

proc OnCheckIfCollapse { } {
    set f $::ImportGeomOptions(widget,f)

    if {$::ImportGeomOptions(var,collapse)} {
	$f.lf.iglayer configure -state "normal"
	$f.lf.autotol configure -state "normal"
	OnCheckAutoTol
    } else {
	$f.lf.iglayer configure -state "disabled"
	$f.lf.autotol configure -state "disabled"
	$f.lf.etol    configure -state "disabled"
    }
}

proc OnCheckAutoTol { } {
    set f $::ImportGeomOptions(widget,f)

    set state [expr {$::ImportGeomOptions(var,autotol)?"disabled":"normal"}]
    $f.lf.etol    configure -state $state
}

proc CreateImportGeomOpts { f } {
    set ::ImportGeomOptions(widget,f) $f
    set CCBG [CCColorActivo [$f cget -background]]
    checkbutton $f.chkCollapse -text [_ "Automatic collapse after import"] \
	-variable ::ImportGeomOptions(var,collapse) -bg $CCBG \
	-command OnCheckIfCollapse
    labelframe $f.lf -labelwidget $f.chkCollapse -bg $CCBG
    checkbutton $f.lf.iglayer -text [_ "Collapse ignoring layers"] \
	-variable ::ImportGeomOptions(var,ignore) -bg $CCBG
    checkbutton $f.lf.autotol -text [_ "Use automatic collapse tolerance"] \
	-variable ::ImportGeomOptions(var,autotol) -bg $CCBG \
	-command OnCheckAutoTol
    label $f.lf.ltol -text [_ "Tolerance:"] -bg $CCBG
    entry $f.lf.etol -textvariable ::ImportGeomOptions(var,tolerance)
    set ::ImportGeomOptions(var,collapse) [GiD_Set AutoCollapseAfterImport]
    set ::ImportGeomOptions(var,ignore) [GiD_Set CollapseIgnoringLayers]
    set ::ImportGeomOptions(var,tolerance) [GiD_Set ImportTolerance]
    set ::ImportGeomOptions(var,autotol) [GiD_Set AutoImportTolerance]
    OnCheckIfCollapse
    grid columnconfigure $f 1 -weight 1
    grid $f.lf.autotol -sticky w -columnspan 2
    grid $f.lf.ltol $f.lf.etol
    grid $f.lf.iglayer -sticky w -columnspan 2
    grid $f.lf -sticky "snew" -row 0 -column 0 -columnspan 3
    return "ImportGeomOptions(dir)"
}

proc ProcessImportGeomOpts { } {
    GiD_Process \
	"-collapse:$::ImportGeomOptions(var,collapse)" \
	"-ignorelayer:$::ImportGeomOptions(var,ignore)" \
	"-tolerance:$::ImportGeomOptions(var,tolerance)" \
	"-autotol:$::ImportGeomOptions(var,autotol)" \
	"--"
}

proc GetPrjThumbnail { gidprj }  {
    set name [file root [file tail $gidprj]]
    return [file join $gidprj $name].png
}

proc GetImagePath { img_path }  {
    set img_path
}

proc GetPostThumbnail { post_file } {
    return [file root $post_file].png
}

proc GetThumbnailCmd { filetypes } {
    if {[lindex [lindex $filetypes 0] 0] eq "GiD postprocess"} {
	return GetPostThumbnail
    }
    foreach {ft ext} $filetypes {
	if {[lsearch {GIF JPEG PNG TGA TIFF Image} [lindex $ft 0]]!=-1} {
	    return GetImagePath
	}
    }
    return ""
}

proc Browser-ramR { ChooseType OpenMode {w .gid}  {title ""} 
		    {DefaultFile {}} {filetypes {}} {defaultextension {}} 
		    {Multiple 0} {MoreOpts {}} } {
    global DefaultSearchDirectory DefaultSearchDirectoryP

    # Se traduce el parametro por defecto
    if { $title == "" } {
	set title [_ "Browser"]
    }

    if { $ChooseType == "project" } {
	if { ![info exists DefaultSearchDirectoryP] || \
		![file isdirectory $DefaultSearchDirectoryP]} {
	    set DefaultSearchDirectoryP [pwd]
	}
	set deft $DefaultSearchDirectoryP
    } else {
	if { ![info exists DefaultSearchDirectory] || \
		![file isdirectory $DefaultSearchDirectory]} {
	    set DefaultSearchDirectory [pwd]
	}
	set deft $DefaultSearchDirectory
    }

    if { $OpenMode == "read" } {
	set comm [ list ::gidtk::dialog::file:: open]
    } else {
	set comm [ list ::gidtk::dialog::file:: save]
    }
    lappend comm -title $title -initialdir $deft -parent $w

    if { $Multiple } {
	lappend comm -multiple 1
    }

    if { $ChooseType == "project" } {
	lappend comm -previewcmd GetPrjThumbnail
	if { $OpenMode == "read" } {
	    set types [list \
		    [list [_ "GiD project"] {.gid}]]
	    lappend comm \
		-filetypes $types -dirprojectext .gid \
		-defaultextension .gid 
	} else {
	    set types [list \
		    [list [_ "GiD project"] {.gid}]\
		    [list [_ "Old GiD project 7"] {.gid}]]
	    lappend comm \
		    -filetypes $types -dirprojectext .gid -defaultextension .gid
	}
    } else {
	#filename, not GiD project
	if { $filetypes != "" } {
	    lappend comm -filetypes $filetypes \
		-previewcmd [GetThumbnailCmd $filetypes]
	}
	if { $defaultextension != "" } {
	    lappend comm -defaultextension $defaultextension
	}
    }

    if { $DefaultFile != "" } {
	lappend comm -initialfile $DefaultFile
    }

    if {$MoreOpts ne ""} {
	lappend comm -moreopt $MoreOpts
    }
    set fileName [eval $comm]
    if { $fileName != "" } {
	if { $ChooseType == "project" } {
	    set DefaultSearchDirectoryP [file dirname $fileName]
	} else { 
	    set file1 $fileName
	    if { $Multiple} {
		set file1 [ lindex $fileName 0]
	    }
	    set DefaultSearchDirectory [file dirname $file1] 
	}
    }
    update 
    update idletasks
    return $fileName
}

proc TopMenuFrame { {action ""} { w .gid.menu } } {
    global GidPriv tcl_platform

    if { $action == "" } {
	if { [winfo exist $w] } { return }
	if { ![info exists GidPriv(PrePostTopMenuWindowGeom)] } {
	    set action INSIDE
	} else {
	    set action [lindex $GidPriv(PrePostTopMenuWindowGeom) 0]
	}
    }

    if { $action == "INSIDE" } {
	 if { [winfo exists $w] } {
	     if { [winfo class $w] != "frame" } {
	       catch {DestroySlave $w}
	     }  else { return }
	 }

	if { ( $tcl_platform(platform) != "windows") && [ string eq $GidPriv(Color,Linux) "brown"] } {
	    frame $w -relief raised -borderwidth 1
	} else { frame $w -borderwidth 0 }

	#pack $w -side top -fill x -before .gid.central
	set GidPriv(PrePostTopMenuWindowGeom) "INSIDE {} 1 TopMenuFrame"
	catch {GridInside PrePostTopMenuWindowGeom $w .gid.topm "we"}
    } else {
	if { ![info exists GidPriv(PrePostTopMenuWindowGeom)] || \
		[lindex $GidPriv(PrePostTopMenuWindowGeom) 1] == "" } {
	    set hh [winfo x .gid]
	    set vv [expr [winfo y .gid]+25]
	    set GidPriv(PrePostTopMenuWindowGeom) \
		    "OPEN [winfo width .gid]x25+$hh+$vv 1 TopMenuFrame"
	}
      catch {DestroySlave $w}
	InitWindow $w $GidPriv(ProgName) PrePostTopMenuWindowGeom TopMenuFrame
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
	#sobra?# update idletasks
	wm transient $w .gid
	wm protocol $w WM_DELETE_WINDOW "wm withdraw $w"

	if { $action == "NONE" } {
	    set GidPriv(PrePostTopMenuWindowGeom) "NONE {} 1 TopMenuFrame"
	    wm withdraw $w
	    .gid conf -menu ""
	}
	#sobra?# update idletasks
    }
    CreateTopMenus
    if { [winfo class $w] == "Toplevel" } {
	foreach i [bind .gid] {
	    bind $w $i [bind .gid $i]
	}
    }

}

proc DefineCoordFrame { { crear 1}} {
    global GidPriv

    set w $GidPriv(CoordFrame) 
    GidHelp "$w label $w.lx label $w.ly label $w.lz" \
	[_ "Display the current coordinates.\nTo hide, see Utilities->Preferences->General"]
    if { $crear} {
	label $w.lx -bg $GidPriv(Color,ToolbarBg) -anchor w -width 10
	label $w.ly -bg $GidPriv(Color,ToolbarBg) -anchor w -width 10
	label $w.lz -bg $GidPriv(Color,ToolbarBg) -anchor w -width 10
	pack $w -side right -fill y
    }
    grid $w.lx -pady 0 -padx 2 -sticky ewn
    grid $w.ly -pady 0 -padx 2 -sticky ew
    grid $w.lz -pady 0 -padx 2 -sticky ew
    grid columnconf $w 0 -weight 1
    grid columnconf $w 0 -minsize 65
}

proc OlvidaCoordFrame {} {
    global GidPriv

    set w $GidPriv(CoordFrame) 
    grid forget $w.lx
    grid forget $w.ly
    grid forget $w.lz
}

proc BottomEntryFrame { {action ""} { w .gid.comm } } {
    global GidPriv tcl_platform GIDDEFAULT

#     if { $action == "" && [winfo exists $w] } { return }

    if { [winfo exists $w.fwarn.list] } {
	set ListboxContents [$w.fwarn.list get 0 end]
    }

    if { $action == "" } {
	if { [winfo exist $w] } { return }
	if { ![info exists GidPriv(PrePostEntryWindowGeom)] } {
	    set action INSIDE
	} else {
	    set action [lindex $GidPriv(PrePostEntryWindowGeom) 0]
	}
    }

    if { $action == "INSIDE" } {
	if { [winfo exists $w] } {
	    if { [winfo class $w] != "frame" } {
		destroy $w
	    } else { return }
	}
	frame $w
#         pack forget .gid.central
#         pack $w -fill x -side bottom
#         catch { pack .gid.central -fill both -expand yes }
	set GidPriv(PrePostEntryWindowGeom) "INSIDE {} 1 BottomEntryFrame"
      GridInside PrePostEntryWindowGeom $w .gid.bottomc "we"
    } else {
	if { ![info exists GidPriv(PrePostEntryWindowGeom)] || \
		[lindex $GidPriv(PrePostEntryWindowGeom) 1] == "" } {
	    set hh [winfo x .gid]
	    set vv [expr [winfo y .gid]+[winfo height .gid]+25]
	    set GidPriv(PrePostEntryWindowGeom) \
		    "OPEN [winfo width .gid]x100+$hh+$vv 1 BottomEntryFrame"
	}
	InitWindow $w [_ "Enter commands"] PrePostEntryWindowGeom BottomEntryFrame
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
	#sobra?# update idletasks
	wm protocol $w WM_DELETE_WINDOW "wm withdraw $w"

	foreach i [bind .gid] {
	    bind $w $i [bind .gid $i]
	}
	if { $tcl_platform(platform) != "windows" } {
	    bind .gid <FocusIn> "catch \"focus $w.e\" "
	    bind $w <Destroy> "+ bind .gid <FocusIn> {}"
	}

	if { $action == "NONE" } {
	    set GidPriv(PrePostEntryWindowGeom) "NONE {} 1 BottomEntryFrame"
	    wm withdraw $w
	}
    }

    $w conf -background $GidPriv(Color,ToolbarBg)
    frame $w.fle -bg $GidPriv(Color,ToolbarBg) 
    label $w.l -text [_ "Command"]: -bg $GidPriv(Color,ToolbarBg) 
    bind $w.l <1> "MenuInCommandButton $w.e %X %Y"
    entry $w.e -relief sunken -background $GidPriv(Color,CommandLineBg)
    checkbutton $w.grid -image [GetImage grid.gif] -variable ::GidPriv(Grid,Show) -indicatoron 0 \
	    -offrelief flat -overrelief raised -bd 0 \
	    -background $GidPriv(Color,ToolbarBg) -pady 0 -padx 0


    trace variable ::GidPriv(Grid,Show) w "UpdateGridShow ;#"
    proc UpdateGridShow { } {
	GiD_Set Grid(Show) $::GidPriv(Grid,Show)
	#from the grid button set both variables Show and Active to same value 
	GiD_Set Grid(Active) $::GidPriv(Grid,Show)
	set ::ChangeVarInfo(Grid,Show) $::GidPriv(Grid,Show)
	set ::ChangeVarInfo(Grid,Active) $::GidPriv(Grid,Show)
	GiD_Redraw
    }
    trace variable ::GidPriv(Grid,Active) w "UpdateGridActive ;#"
    proc UpdateGridActive { } { 
	GiD_Set Grid(Active) $::GidPriv(Grid,Active)
	set ::ChangeVarInfo(Grid,Active) $::GidPriv(Grid,Active)
    }

    #label $w.lcoord -bg $GidPriv(Color,ToolbarBg) -textvariable GidPriv(Coordinates)


    frame $w.fwarn -relief groove -bd 2
    scrollbar $w.fwarn.yscroll  -takefocus 0 \
	    -command "$w.fwarn.list yview" -background $GidPriv(Color,ToolbarBg) \
	    -bd 1 -relief sunken
    listbox $w.fwarn.list -yscroll "$w.fwarn.yscroll \
	    set" -takefocus 0 -background $GidPriv(Color,ToolbarBg) -relief groove -borderwidth 4

    if { [info exists GidPriv(havecontextualbutton)] && $GidPriv(havecontextualbutton)} {
	frame $w.fwarn.buts
	menubutton $w.fwarn.buts.cont -bitmap @$GIDDEFAULT/bitmaps/contextual \
		-bg white -fg black -bd 1 -highlightthickness 0 \
		-menu $w.fwarn.buts.cont.m -direction above
	menu $w.fwarn.buts.cont.m 
	
	bind $w.fwarn.buts.cont <1> {
	    ContextualSubMenu %W.m
	    if {$tkPriv(inMenubutton) != ""} { 
		tkMbPost $tkPriv(inMenubutton) %X %Y
	    }
	    break
	}
    
	GidHelpT $w.fwarn.buts.cont [_ "This menu contains all contextual options\
		that current command has."]
	buttonT $w.fwarn.buts.escape -bitmap @$GIDDEFAULT/bitmaps/escape \
		-bg white -fg black -bd 1 -highlightthickness 0 -relief raised \
		-command ".central.s process escape"
	GidHelpT $w.fwarn.buts.escape [_ "This buttons is the 'Escape' button. Which\
		finishes most commands.\nAccelerator: ESC"]
	pack $w.fwarn.buts.cont -side top -padx 1 -pady 1
	pack $w.fwarn.buts.escape -side top -padx 1 -pady 1
	pack $w.fwarn.buts -side right
    }

    if { [info exists ListboxContents] } {
	foreach i $ListboxContents {
	    $w.fwarn.list ins end $i
	}
	$w.fwarn.list see end
    }

    if {[winfo class $w] != "Toplevel" } { 
	$w.fwarn.list conf -width 20 -height 2
    }
    
    #coordenadas en posicion vertical
    frame $w.fcoord -bg $GidPriv(Color,ToolbarBg) -relief groove -bd 2
    set GidPriv(CoordFrame) $w.fcoord
    set GidPriv(CoordFrame,uso) coordenadas
    DefineCoordFrame 

    proc UpdateLabelCoordinates { w  } { 
	global GidPriv GIDDEFAULT
	if { "$GidPriv(CoordFrame,uso)" != "coordenadas" || ! [winfo exist $w.fcoord] } { return}
	if { [scan $GidPriv(Coordinates) "x=%g y=%g z=%g" x y z] == 3 } { 
	    $w.fcoord.lx configure -text x=[GiD_FormatReal $GidPriv(RealFormat,Default) $x] -state normal
	    $w.fcoord.ly configure -text y=[GiD_FormatReal $GidPriv(RealFormat,Default) $y] -state normal
	    $w.fcoord.lz configure -text z=[GiD_FormatReal $GidPriv(RealFormat,Default) $z] -state normal
	    $w.fcoord.lz configure -image ""
	} else {
	    if { $GidPriv(Coordinates) == "" } {
		$w.fcoord.lx configure -state disabled
		$w.fcoord.ly configure -state disabled
		$w.fcoord.lz configure -state disabled
	    } else {
		$w.fcoord.lx configure -text [_ "Orthogonal"] -state disabled
		$w.fcoord.ly configure -text [_ "view to Z"] -state disabled
		$w.fcoord.lz configure -image [GetImage no_create_point.gif] -state disabled
	    }
	}
	set maxlength 0
	foreach item "$w.fcoord.lx $w.fcoord.ly $w.fcoord.lz" {
	    set length [string length [$item cget -text]]
	    if { $maxlength < $length } {
		set maxlength $length
	    }
	}
	if { [$w.fcoord.lx cget -width] < $maxlength } {
	    foreach item "$w.fcoord.lx $w.fcoord.ly $w.fcoord.lz" {
		$item configure -width $maxlength
	    }
	}
    }

    trace var GidPriv(Coordinates) w "UpdateLabelCoordinates $w ;#"
    
    $w.fwarn.list conf -selectmode single
    pack $w.fwarn.yscroll -side right -fill y
    pack $w.fwarn.list -expand yes -fill both -side left
    
    pack $w.fle -fill x -side bottom -expand yes -pady 2 -padx 2
    pack $w.fwarn -fill both -side top -expand yes

    grid $w.l $w.e $w.grid -in $w.fle -pady 0 -padx 5 -sticky ew
    grid conf $w.l -sticky w
    grid configure $w.grid -sticky e -padx 0

    proc ShowMenuGrid { x y preffocus } {
	set w .gid.__gridmenu
	if { [winfo exists $w] } { destroy $w }
	menu $w

	set text [_ "Auxiliar grid to aid in points definition"]
	$w add command -label [_ "help"] -command "DisplayGidHelp $w [list $text] menu $x $y"

	if { [GiD_Info variables Grid(Show)] == "0" } {
	    $w add command -label [_ "Show grid"] -command {set ::GidPriv(Grid,Show) 1}
	} else {
	    $w add command -label [_ "Hide grid"] -command {set ::GidPriv(Grid,Show) 0}
	}
	if { [GiD_Info variables Grid(Active)] == "0" } {
	    $w add command -label [_ "Enable snap"] -command {set ::GidPriv(Grid,Active) 1}
	} else {
	    $w add command -label [_ "Disable snap"] -command {set ::GidPriv(Grid,Active) 0}
	}
	if { [GiD_Info variables Grid(DrawAxis)] == "0" } {
	    $w add command -label [_ "Show axis"] -command {GiD_Set Grid(DrawAxis) 1 ; GiD_Redraw}
	} else {
	    $w add command -label [_ "Hide axis"] -command {GiD_Set Grid(DrawAxis) 0 ; GiD_Redraw}
	}
	$w add separator
	if { $preffocus == "grid" } {
	    $w add command -label [_ "Settings"]... -command \
		    {ChangeVariables .gid.wvar Grid ; focus .gid.wvar.f.nb.fGrid.frame1.cb1}
	} else {
	    $w add command -label [_ "Settings"]... -command \
		    {ChangeVariables .gid.wvar Grid ; focus .gid.wvar.f.nb.fGrid.frame1.cb2}
	}
	GiD_PopupMenu $w $x $y
    }
    
    bind $w.grid <3> [list ShowMenuGrid %X %Y grid]
    
    grid columnconf $w.fle 1 -weight 1
    
    #set same minsize in ChangeVariables.tcl if GidPriv(Displaycoordinates) is modified
    #after read preferences, if not displaycoordinates hide coord. label
    if { !$GidPriv(Displaycoordinates) } {
	pack forget .gid.comm.fcoord
    }

    
    set helptxt [_ "Here are output the process messages. At the same time, it is\
	    possible to enter written commands or point coordinates or entities\
	    labels. Use <Tab> to expand commands."]

    GidHelpT "$w.fwarn $w.e" $helptxt

    proc WarnMessagesContextualMenu { helptxt listbox w idx0 x y } {

	$listbox selection clear 0 end
	$listbox selection set $idx0
	set idx [$listbox curselection]
	if { [llength $idx] == 1 } {
	    set txt [$listbox get $idx]
	    if { [regexp {x=(\S+)\s*,\s*y=(\S+)\s*,\s*z=(\S+)} $txt {} xc yc zc] } {
		set txtpoint "$xc $yc $zc"
	    }
	}
	catch { destroy .gid.__helpmenu }
	menu .gid.__helpmenu
	if { [info exists txt] } {
	    .gid.__helpmenu add command -label [_ "Copy"] -command \
		"clipboard clear ; [list clipboard append $txt]"
	}
	if { [info exists txtpoint] } {
	    .gid.__helpmenu add command -label [_ "Copy point coords"] -command \
		"clipboard clear ; [list clipboard append $txtpoint]"
	}
	.gid.__helpmenu add separator
	.gid.__helpmenu add command -label [_ "help"] -command \
	    "DisplayGidHelp $w [ list $helptxt] menu $x $y"
	.gid.__helpmenu add command -label [_ "Configure toolbars"] -command \
	    TOBegin
	GiD_PopupMenu .gid.__helpmenu $x $y
    }

    foreach i [concat $w.fwarn [winfo children $w.fwarn]] {
	bind $i <ButtonPress-3> [list WarnMessagesContextualMenu $helptxt $w.fwarn.list \
		                     %W @%x,%y %X %Y]
    }

    GidHelpT $w.l [_ "Press mouse button 1 over it to get a commands menu."]

    MakeBottomEntryBindings

    # para dejar que el boton del medio mueva la barra de scroll
    bind $w.fwarn.yscroll <ButtonPress-2> ""
    bind $w.fcoord <1> "SetCoord"
    bind $w.fcoord.lx <1> "SetCoord"
    bind $w.fcoord.ly <1> "SetCoord"
    bind $w.fcoord.lz <1> "SetCoord"

# check new tab options in entry
#     if { [winfo exists .wSetCoord] } {
#         bind $w.e <Tab> "focus .wSetCoord.xe ; break"
#     }
}


proc RightButtons { {action ""} { w .gid.buts } } {
    global GidPriv GIDDEFAULT ActionsMenuVar
###puts " . -> '$action' '$w'"
    if { $action == "" && [winfo exists $w] } { return }
    if { $action == "INSIDE" || ($action == "" && \
	    (![info exists GidPriv(PrePostRightButWindowGeom)] || \
	    [lindex $GidPriv(PrePostRightButWindowGeom) 0] == "INSIDE")) } {
	if { [winfo exists $w] } {
	    if { [winfo class $w] != "frame" } {
		destroy $w
	    } else { return }
	}
	frame $w -relief flat  -borderwidth 0
#         pack forget .gid.central.s
#         pack $w -side right -fill y -in .gid.central
#         catch { pack .gid.central.s -fill both -expand yes -side left}
	set GidPriv(PrePostRightButWindowGeom) "INSIDE {} 1 RightButtons"
      GridInside PrePostRightButWindowGeom $w .gid.rightb "ns"
    } elseif { $action == "OPEN" || \
	    ($action == "" && [lindex $GidPriv(PrePostRightButWindowGeom) 0] \
	    == "OPEN") } {
	if { ![info exists GidPriv(PrePostRightButWindowGeom)] || \
		[lindex $GidPriv(PrePostRightButWindowGeom) 1] == "" } {
	    set GidPriv(PrePostRightButWindowGeom) \
		    "OPEN 130x300+[expr [winfo x .gid]+[winfo width .gid] \
		    +10]+[expr [winfo y .gid]+25] RightButtons"
	}
###puts " . -> InitWindow"
	catch {DestroySlave $w}
	InitWindow $w [_ "Command buttons"] PrePostRightButWindowGeom RightButtons
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
	wm transient $w .gid
    } else {
	catch { DestroySlave $w }
	set GidPriv(PrePostRightButWindowGeom) "NONE {} 1 RightButtons"
	return
    }
###puts " . -> frames"
    frame .gid.buts.f -width 130 -height 100 -relief flat -bd 2 -background $GidPriv(Color,ToolbarBg)
    pack propagate .gid.buts.f no
    pack .gid.buts.f -side top -fill both -expand yes -padx 0 -pady 0
    bind .gid.buts <Configure> { 
	incr DelayedConfigureButsVar
	after idle DelayedConfigureButsFunc
    }
    for {set i 0 } { $i<100 } { incr i } {
	button $w.f.b$i -anchor w -width 15 -takefocus 0 -bd 1 -background $GidPriv(Color,ToolbarBg) \
		-padx 2 -font SmallFont -highlightthickness 0 -relief flat

	ChangeToNewLook $w.f.b$i
    }
    set reqwidth [ winfo reqwidth $w.f.b0]
    .gid.buts.f conf -width [ expr $reqwidth + 8]
    button $w.f.ar -bitmap @$GIDDEFAULT/bitmaps/arrow-r-16 -takefocus 0 -bd 1 -bg $GidPriv(Color,ToolbarBg) -fg blue
    button $w.f.al -bitmap @$GIDDEFAULT/bitmaps/arrow-l-16 -takefocus 0 -bd 1 -bg $GidPriv(Color,ToolbarBg) -fg blue

    ChangeToNewLook $w.f.ar
    ChangeToNewLook $w.f.al

    foreach i [bind .gid] {
	bind $w $i [bind .gid $i]
    }
###puts " . -> GidHelpT"
    GidHelpT $w [_ "These buttons are a direct access to the GiD commands"]
    bind $w <2> { .gid.central.s button 2 %x %y }
    
    bind $w <Right> "if \[winfo ismapped $w.f.ar] \"$w.f.ar invoke\""
    bind $w <Left> "if \[winfo ismapped $w.f.al] \"$w.f.al invoke\""
    
###puts " . -> [info exists ActionsMenuVar] -> { SetButtons }"
    update idletasks
    if { [info exists ActionsMenuVar] } { 
      #WarnWinText $::ActionsMenuVar
      SetButtons 
    }
###puts " . -> FIN"
}

proc UnMapAllWindows { w } {
    global GidPriv
    if { $w != ".gid" || ![info exists GidPriv(TopLevelsList)] } { return }
    foreach i $GidPriv(TopLevelsList) {
	if { [winfo exists $i] && [winfo ismapped $i] } {
	    wm withdraw $i
	}
    }
}

proc MapAllWindows { w } {
    global GidPriv
    if { $w != ".gid"  || ![info exists GidPriv(TopLevelsList)] } { return }
    foreach i $GidPriv(TopLevelsList) {
	if { [winfo exists $i] && ![winfo ismapped $i] } {
	    catch { wm deiconify $i }
	}
    }
}

proc OpenTextFromFile { title FileNamebase offset { offset2 -1} { quit no } } {
    global tcl_platform
    if { [winfo exists .gid] } {
	if { [winfo exists .textwin] } { 
	    destroy .textwin 
	}
	set w .gid.textwin
    } else { 
	set w .textwin 
    }

    if { $offset2 != -1 } {
	set FileName [lindex $FileNamebase 0]
	set FileName2 [lindex $FileNamebase 1]
    } else {
	set FileName $FileNamebase
    }

    set FileNameGood 1
    if { ![file readable $FileName] } { 
	set FileNameGood  0 
    } elseif { $offset >= [file size $FileName] } { 
	set FileNameGood  0 
    }
    if { $FileNameGood == 1 } {
	set f [open $FileName r]
	seek $f $offset
	if { [eof $f] } { 
	    close $f 
	    set FileNameGood  0 
	}
    }
    
    if { $offset2 == -1 && $FileNameGood == 0 } { return }

    set FileNameGood2 0
    if { $offset2 != -1 } { 
	set FileNameGood2 1
	if { ![file readable $FileName2] } { 
	    set FileNameGood2  0 
	} elseif { $FileNameGood2 == 1 && $offset2 >= [file size $FileName2] } { 
	    set FileNameGood2 0
	}
	if { $FileNameGood2 == 1 } {
	    set f2 [open $FileName2 r]
	    seek $f2 $offset2
	    if { [eof $f2] } { 
		close $f2 
		set FileNameGood2 0 
	    }
	}
	if { $FileNameGood == 0 && $FileNameGood2 == 0 } { 
	    return 
	}
    }

    set CreatingWindow 0
    if { ![winfo exists $w] || $title != [wm title $w] } {

	set CreatingWindow 1

	InitWindow $w $title PrePostTextWindowGeom
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
	
	frame $w.f
	text $w.f.t -width 50  -yscroll "$w.f.yscroll set" -bd 1
	scrollbar $w.f.yscroll -relief sunken -command "$w.f.t yview" -bd 1 \
	    -elementborderwidth 1
	pack $w.f.yscroll -side right -fill y -pady 5
	pack $w.f.t  -expand yes -fill both -anchor nw -padx 5 -pady 5
	button $w.b -text [_ "Close"] -command "destroy $w" -bd 1
	SetWidgetsWidthFromText $w.b
	focus $w.b
	pack $w.b -side bottom -pady 2
	pack $w.f -side top -fill both -expand yes -side top
	update
	
	scan [ winfo geometry $w] %dx%d ww hh
	if { ( $ww < 200) || ( $hh < 200)} {
	    wm geometry $w 200x200
	}
    }

    $w.f.t conf -state normal

    if { $FileNameGood == 1 } {
	while { ![eof $f] } {
	    $w.f.t ins end [gets $f]\n
	}
	close $f 
    }
    if { $FileNameGood2 == 1 } {
	while { ![eof $f2] } {
	    $w.f.t ins end [gets $f2]\n
	}
	close $f2
    }
    $w.f.t see [$w.f.t index "end-2l"]

    regsub {\..*} [$w.f.t index end] {} NumLines
    incr NumLines -1
    if { $NumLines > 20 } { set NumLines 20 }

    if { $tcl_platform(platform) != "windows" } {
	$w.f.t conf -state disabled
    }
    
    if { $CreatingWindow == 1 } {
	set hei $NumLines
	if { $hei < 6 } { set hei 6 }
	$w.f.t conf -height $hei
    } 

    if { $quit != "no" } {
	$w.b conf -text [_ "Quit"]
	tkwait visibility $w
	raise $w
	grab $w
	tkwait window $w
    }
}

package require progressbar 1.2

proc ProgressInMeshing { NumSurfs NumVols isurf ivol SurfNum VolNum SurfFactor VolFactor \
	Bar3Title Bar3Factor NumNodes NumElems } {
    global GidPriv
    
    set w .gid.wprogress
    
    if { ![winfo exists $w] } {
	
	if { $isurf == $NumSurfs && $ivol == $NumVols } { return }
	
	toplevel $w -class OverlayWindow
	
	#bind $w <Destroy> "catch { unset GidPriv(StopMeshing) }"
	
	if { [winfo cells $w] < 6 } {
	    set metercolor white
	} else { set metercolor blue }
	
	
	wm title $w [_ "Progress in meshing"]
	set pare [winfo parent $w]
	set x [expr [winfo x $pare]+[winfo width $pare ]/2- [winfo reqwidth $w]/2]
	set y [expr [winfo y $pare]+[winfo height $pare ]/2- [winfo reqheight $w]/2]
	
	WmGidGeom $w +$x+$y
	
	$w conf -cursor watch
	
	foreach i "surface volume bar3" {
#             if { $i == "volume" && $NumVols == 0 } { break }
	    if { $i == "volume" && $NumVols == 0 } { continue }
	    if { $i == "bar3" && $Bar3Title == "" } { break }
	    frame $w.$i -relief ridge -bd 2
	    frame $w.$i.fr
	    if { $i == "surface" } {
		label $w.$i.fr.l1 -text [_ "Meshed#C#surface"] -width 7
		label $w.$i.fr.l2 -text 0 -width 5
		label $w.$i.fr.l3 -text [_ "surfaces of  %s" $NumSurfs]
		label $w.$i.l4 -text [_ "Meshing surface number:"]
		label $w.$i.l5 -text 0 -relief sunken -width 5
	    } elseif { $i == "volume" } {
		label $w.$i.fr.l1 -text [_ "Meshed#C#volume"] -width 7
		label $w.$i.fr.l2 -text 0  -width 5
		label $w.$i.fr.l3 -text [_ "volumes of  %s" $NumVols]
		label $w.$i.l4 -text [_ "Meshing volume number:"]
		label $w.$i.l5 -text 0 -relief sunken -width 5
	    } else {
		label $w.$i.fr.l1 -text [_ "Percentage done:"] -width 7
		label $w.$i.fr.l2 -text 0  -width 5
#                 label $w.$i.fr.l3 -text [_ "volumes of  %s" $NumVols]
		label $w.$i.l4 -text $Bar3Title
#                 label $w.$i.l5 -text 0 -relief sunken -width 5
	    } 
	    set wp [::progressbar::progressbar $w.$i.f -variable ::progressbar::var$i \
		    -width 200 -background [ $w.$i cget -background]]
	    set ::progressbar::var$i 0
	    
	    if { $i == "bar3" } {
		grid $w.$i.l4 -row 0 -column 1 -sticky w -pady 3
#                 grid $w.$i.l5 -row 0 -column 2 -sticky w -pady 3
		grid $wp -columnspan 4
		#grid $w.$i.l1 $w.$i.l2 $w.$i.l3 -pady 3
		grid $w.$i.fr.l1 $w.$i.fr.l2 -pady 3
		grid $w.$i.fr -sticky ew -columnspan 4
		grid $w.$i -sticky ewns
	    } else {
		#grid $w.$i.l4 $w.$i.l5 -sticky w -pady 3
		grid $w.$i.l4 -row 0 -column 1 -sticky w -pady 3
		grid $w.$i.l5 -row 0 -column 2 -sticky w -pady 3
		grid $wp -columnspan 4
		#grid $w.$i.l1 $w.$i.l2 $w.$i.l3 -pady 3
		grid $w.$i.fr.l1 $w.$i.fr.l2 $w.$i.fr.l3 -pady 3
		grid $w.$i.fr -sticky ew -columnspan 4
		grid $w.$i -sticky ewns
	    }
	}
	frame $w.n
	label $w.n.lnnodes -text [_ "Number of nodes:"]
	label $w.n.nnodes -text $NumNodes -relief sunken -width 8 -bd 1 -justify right
	label $w.n.lnelems -text [_ "Number of elements:"]
	label $w.n.nelems -text $NumElems -relief sunken -width 8 -bd 1 -justify right
	grid $w.n.lnnodes $w.n.nnodes -sticky w -pady 3
	grid $w.n.lnelems $w.n.nelems -sticky w -pady 3
	grid $w.n -sticky ew
	
	grid columnconf $w 0 -weight 1
	grid rowconf $w 0 -weight 1
	if { $Bar3Title != "" } {
	    grid rowconf $w 2 -weight 1
	} else {
	    if {$NumVols != 0 } { grid rowconf $w 1 -weight 1 }
	}

	#lower buttons
	set def_back [$w cget -background]
	frame $w.frmButtons -bg [CCColorActivo $def_back]
	button $w.frmButtons.btnclose -text [_ "Stop"] \
	    -command {set GidPriv(StopMeshing) 1} \
	    -underline 0 -width 6
	SetWidgetsWidthFromText $w.frmButtons.btnclose
	
	grid $w.frmButtons -sticky ews -columnspan 7
	if { $::tcl_version >= 8.5 } { grid anchor $w.frmButtons center }
	grid $w.frmButtons.btnclose -padx 5 -pady 6
	grid columnconf $w 0 -weight 1
	grid rowconfigure $w 0 -weight 1
	
	focus $w.frmButtons.btnclose
	bind $w <Alt-c> "tkButtonInvoke $w.frmButtons.btnclose"
	bind $w <Escape> "tkButtonInvoke $w.frmButtons.btnclose"
	
	bind $w <Destroy> [list +DestroyProgressInMeshing %W $w] ;# + to add to previous script
	grab $w
    }
    
    if { $NumSurfs != 0 && $SurfNum != [$w.surface.l5 cget -text] } {
	$w.surface.l5 conf -text $SurfNum
    }
    if { $NumSurfs != 0 && ($isurf != [$w.surface.fr.l2 cget -text] || $SurfFactor>0.05) } {
	$w.surface.fr.l2 conf -text $isurf
	set factor [expr (double($isurf)+$SurfFactor)/$NumSurfs]
	set ::progressbar::varsurface [expr int($factor*100)]
    }
    if { $NumVols != 0 && $VolNum != [$w.volume.l5 cget -text] } {
	$w.volume.l5 conf -text $VolNum
    }
    if { $NumVols != 0 && ($ivol != [$w.volume.fr.l2 cget -text] || $VolFactor>0.05) } {
	$w.volume.fr.l2 conf -text $ivol
	set factor [expr (double($ivol)+$VolFactor)/$NumVols]
	set ::progressbar::varvolume [expr int($factor*100)]
    }
    
    if { $Bar3Title != "" && $Bar3Factor != [$w.bar3.fr.l2 cget -text] } {
	$w.bar3.fr.l2 conf -text $Bar3Factor
#         set factor [expr (double($ivol)+$VolFactor)/$NumVols]
	set ::progressbar::varbar3 [expr int($Bar3Factor*100)]
    }
    

    $w.n.nnodes configure -text $NumNodes
    $w.n.nelems configure -text $NumElems
    
    update 
    
    if { $isurf == $NumSurfs && $ivol == $NumVols } {
	after idle EndProgressInMeshing
    }
    
}

proc DestroyProgressInMeshing { W w } {
    unset -nocomplain ::GidPriv(StopMeshing)
}

proc EndProgressInMeshing { } {
    set w .gid.wprogress
    if { [winfo exists $w] } {
	destroy $w
    }
}

proc WmGidGeom { w geom } {
    global tcl_platform

    if { [info exists ::GidPriv(OffScreen)] && $::GidPriv(OffScreen)} { 
	wm withdraw $w
	return
    }

    set x 5
    set y 20
    set yeswh 1

    set iszoomed 0
    if { [lindex $geom 0] == "ZOOMED" } {
	set geom [lindex $geom 1]
	set iszoomed 1
    }

    set screenwidth [winfo screenwidth $w]
    set screenheight [winfo screenheight $w]

    if { $geom ==  "" } {
      set x [expr $screenwidth/3]
      set y [expr $screenheight/3]
      set geom +${x}+$y
      set ::GidPriv($w,has,geom) 0
    } else {
      set ::GidPriv($w,has,geom) 1
    }

    if { ![regexp {^([0-9]*)x([0-9]*)([-+]*[0-9]*)([-+]*[0-9]*)$} $geom trash \
	    width height x y] } {
	if { ![regexp {([-+]*[0-9]*)([-+]+[0-9]*)$} $geom trash x y] } {
	    wm geom $w +32+32
	    return
	    error "couldn't compute geometry: $geom"
	}
	set width [winfo reqwidth $w]
	set height [winfo reqheight $w]
	set yeswh 0
    }

    if { ![regsub {[+]-} $x {-} x] } {
	if { $x<0 } { set x [expr $screenwidth+$x-$width] }
    }
    if { ![regsub {[+]-} $y {-} y] } {
	if { $y<0 } { set y [expr $screenheight+$y-$height] }
    }

    if { $x < -15 } { set x 5 }
    if { $y < -15 } { set y 20 }

    if { [expr $x+$width] > [expr $screenwidth+15] } {
	set x [expr $screenwidth-$width]
    }
    if { [expr $y+$height] > [expr $screenheight+15] } {
	set y [expr $screenheight-$height]
    }

    
    if { $x < -15 } { set width [expr $width+$x+5] }
    if { $y < -15 } { set height [expr $height+$y+5] }

    if { ![regexp {[+]} $x] } { set x +$x}
    if { ![regexp {[+]} $y] } { set y +$y}

    if { $yeswh } {
	wm geom $w ${width}x$height$x$y
    } else {
	wm geom $w $x$y
    }

    if { $iszoomed && $tcl_platform(platform) == "windows" } {
	wm state $w zoomed
    }

}

# proc IsWord
#
# w: parent window
# type -        Can be:
#                        word:    word without spaces
#                        any:     anything
#                        real:    real
#                        real+:   positive real
#                        int:     int
#                        int+:    positive int

proc IsWord { type word { w .gid } { PrintError no } } {

    switch $type {
	word {
	    if { ![regexp {^[  ]*[^ \t]+[  ]*$} $word] } {
		set errorRes [_ "A single word must be entered."]
	    }
	}
	real {
	    if { ![regexp \
   {^[  ]*([+-]?([0-9]+\.?[0-9]*|[0-9]*\.?[0-9]+)([eE][+-]?[0-9]+)?)[  ]*$} \
		    $word] } {
		set errorRes [_ "One number must be entered"].
	    }           
	}
	real+ {
	    if { ![regexp \
    {^[  ]*([+]?([0-9]+\.?[0-9]*|[0-9]*\.?[0-9]+)([eE][+-]?[0-9]+)?)[  ]*$} \
		    $word] } {
		set errorRes [_ "One positive number must be entered."]
	    }          
	}
	int {
	    if { ![regexp \
		    {^[  ]*[+-]?[0-9]+[  ]*$} \
		    $word] } {
		set errorRes [_ "One integer number must be entered."]
	    }           
	}
	int+ {
	    if { ![regexp \
		    {^[  ]*[+]?[0-9]*[1-9]+[0-9]*[  ]*$} \
		    $word] } {
		set errorRes [_ "One positive integer number must be entered."]
	    }           
	}
    }
    if { [info exists errorRes] } {
	if { $PrintError != "no" } {
	    tk_dialogRAM $w.tempwinerr [_ "Error window"] "$errorRes $PrintError" \
		error 0 [_ "OK"]
	}
	return 0
    } else {
	return 1
    }
}

proc SetProblemTypeChoose { projecttype } {

    package require dialogwin

    foreach {oldproblemtype probname changes} [GiD_Info Project] break

    if { $changes || $probname != "UNNAMED" } {
	set putyes yes
    } else {
	set putyes ""
    }

    if { $oldproblemtype != "UNKNOWN" && ($changes || $probname != "UNNAMED") } {
	set f [DialogWin::Init .gid [_ "Select new problemtype"] separator]
	set w [winfo toplevel $f]        
	label $f.l1 -text [_ "You are going to load problemtype '%s'.\nWhat do you want to do?" \
		              $projecttype] \
	    -justify left
	radiobutton $f.r1 -text [_ "Update to new problemtype"] -variable todo -value update
	label $f.l2 -text [_ "All data information (materials, conditions, data) will be lost"] \
	    -justify left -wraplength 180
	radiobutton $f.r2 -text [_ "Transform to new problemtype"] -variable todo -value transform
	label $f.l3 -text [_ "Data will be converted to new problemtype if possible. Some data may be lost"]\
	    -justify left -wraplength 180

	grid $f.l1 -column 0
	grid $f.r1 -column 0 -sticky w -padx 10
	grid $f.l2 -column 0 -sticky w -padx 50
	grid $f.r2 -column 0 -sticky w -padx 10
	grid $f.l3 -column 0 -sticky w -padx 50

	if { $oldproblemtype == "UNKNOWN" } {
	    $f.r2 conf -state disabled
	    $f.l3 conf -fg grey
	    set ::todo update
	} elseif { ![info exists ::todo] || [lsearch [list update transform] $::todo] == -1 } {
	    set ::todo update
	}
	#supergrid::go $f

	#change Windows shell link by the pointed name
	#if { [GidUtils::IsLink [file join $::GIDDEFAULT problemtypes $projecttype.gid.lnk]] == 2 } {
	 #   set projecttype [file rootname [GidUtils::ReadLink [file join $::GIDDEFAULT problemtypes $projecttype.gid.lnk]]]
	#}

	set action [DialogWin::CreateWindow]
	DialogWin::DestroyWindow
	switch $action {
	    0 { return }
	    1 {
		switch $::todo {
		    update {
		        if { $putyes != "" } {
		            GiD_Process escape escape escape escape data defaults ProblemType \
		                    $putyes $projecttype escape
		        } else {
		            GiD_Process escape escape escape escape data defaults ProblemType \
		                    $projecttype escape
		        }
		    }
		    transform {
#                         if { $changes || $probname == "UNNAMED" } {
#                             WarnWin [_ "Please, save the problem before transforming it."]
#                             return
#                         }
		        GiD_Process escape escape escape escape data defaults TransfProblem \
		                $projecttype escape
		    }
		}
	    }
	}
    } else {
	if { $putyes != "" } {
	    GiD_Process escape escape escape escape data defaults ProblemType $putyes $projecttype escape
	} else {
	    GiD_Process escape escape escape escape data defaults ProblemType $projecttype escape
	}
    }
}

proc ReadProblemtypeConfigFile { path name } {
    global GidPriv
    if { [file extension $path] == ".lnk" && [GidUtils::IsLink [file join $::GIDDEFAULT problemtypes $path]] == 2 } {
	set fullrealpath [GidUtils::ReadLink [file join $::GIDDEFAULT problemtypes $path]]
	set path [file rootname $path] ;#to remove .lnk
    } else  {
	set fullrealpath [file join $::GIDDEFAULT problemtypes $path]
    }

    if { [file extension $name] == ".lnk" && [GidUtils::IsLink [file join $fullrealpath $name]] == 2 } {
	set realdir [GidUtils::ReadLink [file join $fullrealpath $name]]
	set name [file rootname $name] ;#to remove .lnk
    } else {
	set realdir [file join $fullrealpath $name]
    }
    
    set filename [file join $realdir [file rootname $name]].xml
    set tdom_bad [catch {package require tdom}]
    if { $tdom_bad || ![file exists $filename] } return
    set xml [tDOM::xmlReadFile $filename]
    set doc [dom parse $xml]
    set root [$doc documentElement]
    set pt [file join $path [file rootname $name]]
    foreach item {ImageFileBrowser ImageFileBrowserHidden} {
	if { [catch { set relfilename [[$root selectNodes Program/$item/text()] data] }] } {
	    continue
	}
	set imagefilename [file join $realdir $relfilename]
	if { [file exists $imagefilename] } {
	    set ::GidPriv(problemtype,$item,$pt) [image create photo -file $imagefilename]
	}
    }
    $doc delete
}

proc SetProblemTypeMenuForDir { dir menu current path } {
    global GidPriv    
    set inum 0

    if { ![info exists GidPriv(dirfiles,$dir)] } {
	#set GidPriv(dirfiles,$dir) and subdirectories
	FastCacheFromProblemtypeMenuDir $dir
	#FastCacheFromProblemtypeMenuDir also call ReadProblemtypeConfigFile
    }

    foreach i $GidPriv(dirfiles,$dir) {
	set ifull [file join $dir $i]
	if { [file extension $i] == ".lnk" && [GidUtils::IsLink $ifull] == 2 } {
	    set i [file root $i] ;#to remove .lnk
	}
	if { [file extension $i] == ".gid" } {
	    set pt [file root $i]
	    set fullrelativept [file join $path $pt]
	    if { $pt != $current } {
		if { [info exists ::GidPriv(problemtype,ImageFileBrowser,$fullrelativept)] } {
		    $menu add command -label $pt -command [ list SetProblemTypeChoose $fullrelativept] \
		            -image $::GidPriv(problemtype,ImageFileBrowser,$fullrelativept) -compound left
		} else {
		    $menu add command -label $pt -command [ list SetProblemTypeChoose $fullrelativept]
		}
	    } else {
		if { [info exists ::GidPriv(problemtype,ImageFileBrowser,$fullrelativept)] } {
		    $menu add check -label $pt -command [ list SetProblemTypeChoose $fullrelativept] \
		            -variable GidPriv(PTMFD) -onvalue 1 \
		            -image $::GidPriv(problemtype,ImageFileBrowser,$fullrelativept) -compound left
		} else {
		    $menu add check -label $pt -command [ list SetProblemTypeChoose $fullrelativept] \
		            -variable GidPriv(PTMFD) -onvalue 1
		}
		set GidPriv(PTMFD) 1
	    }
	    incr inum            
	} else {
	    set newmenu $menu.m$inum
	    $menu add cascade -label $i -menu $newmenu
	    menu $newmenu
	    if { $i == [lindex [file split $current] 0] } {
		set inumin [SetProblemTypeMenuForDir $ifull $newmenu [lrange [file split $current] 1 end] \
		        [file join $path $i]]
	    } else {
		set inumin [SetProblemTypeMenuForDir $ifull $newmenu "" [file join $path $i]]
	    }
	    if { $inumin == 0 } {
		$menu delete end
		destroy $newmenu
	    } else {
		incr inum $inumin
	    }
	}
    }
    return $inum
}

proc SetProblemTypeMenu { menu } {
    global GIDDEFAULT GidPriv

    if { ![winfo exists $menu] } return
    set maindir [file join $GIDDEFAULT problemtypes]
    if { [$menu index end] != "none" && [info exists GidPriv(dirfiles,$maindir)] } { return }
    $menu del 0 end
    foreach i [winfo children $menu] { destroy $i }
    
    set current [GiD_Info Project ProblemType]

    set inum 0

    #set dir $GIDDEFAULT
    #if { ![info exists GidPriv(dirfiles,$dir)] } {
    #        set dirs [glob -nocomplain -tails -directory $dir -types d *.gid]
    #        foreach link [glob -nocomplain -tails -directory $dir -types f *.lnk] {
    #            if { [GidUtils::IsLink [file join $dir $link]] == 2 } {
    #                lappend dirs $link
    #            }
    #        }
    #        set GidPriv(dirfiles,$dir) [lsort -dictionary $dirs]
    #}

    #foreach i $GidPriv(dirfiles,$dir) {
    #        if { [file extension $i] == ".gid" } {
    #            set pt [file root $i]
    #        } elseif { [file extension $i] == ".lnk" && [GidUtils::IsLink [file join $dir $i]] == 2 } {
    #            set pt [file root [file root $i]]
    #        }
    #        if { $pt != $current } {
    #            $menu add command -label $pt -command [list SetProblemTypeChoose $pt]
    #            incr inum
    #        } else {
    #            $menu add check -label $pt -command [list SetProblemTypeChoose $pt] \
    #                    -variable GidPriv(PTMFD) -onvalue 1
    #            set GidPriv(PTMFD) 1
    #        }
    #}
    if { $inum } { $menu add separator }
    SetProblemTypeMenuForDir [file join $GIDDEFAULT problemtypes] $menu $current ""

    $menu add separator
    
    if { [lindex [GiD_Info project] 0] == "UNKNOWN" } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
    } else {
	set color black
    }
   
    $menu add command -label [_ "Transform"]... -command \
	"GiD_Process escape escape escape escape data defaults TransfProblem"
    $menu add command -label [_ "Internet Retrieve"]... -command \
	RetrievePT::OpenGUI

    $menu add command -label [_ "Load"]... -command \
	    "GiD_Process escape escape escape escape data defaults ProblemType"

    $menu add command -label [_ "Unload"] -command UnloadProblemtype
 
#-foreground $color

    if { ![info exists GidPriv(RamdebuggerMenuState)] } {
	set GidPriv(RamdebuggerMenuState) 1
    }
    if { $GidPriv(RamdebuggerMenuState) != 0} {
	if { $GidPriv(RamdebuggerMenuState) == -1} {
	    set color $::GidPriv(Color,DisabledForegroundMenu)
	} else {
	    set color black
	}
	$menu add command -label [_ "Debugger"]... -command "package require RamDebugger" -foreground $color
    }
}

#state can be: normal, hidden or disable (to show in gray)
proc SetRamdebuggerMenuState { state } {
    global GidPriv
    if { $state == "normal" } {
	set GidPriv(RamdebuggerMenuState) 1
    } elseif { $state == "hidden" } {
	set GidPriv(RamdebuggerMenuState) 0
    } elseif { $state == "disable" } {
	set GidPriv(RamdebuggerMenuState) -1
    } else {
	set GidPriv(RamdebuggerMenuState) 1
    }
    set menu_name "Data#C#menu"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i != -1 } {
	set menu [GetMenuName $idata].0
	$menu del 0 end
	SetProblemTypeMenu $menu
    }
} 

proc UnloadProblemtype { } {
    if { [GiD_Info Project AreChanges] || [GiD_Info Project ModelName] != "UNNAMED" } {
	set msg [_ "All data information (materials,conditions,data) will be lost."]
	set Ret [tk_dialogRAMFull .gid.askyn [_ "Unload"] \
		$msg "" "" gidquestionhead 0 [_ "Ok"] [_ "Cancel"]]
	if { $Ret == 0 } {
	    GiD_Process escape escape escape escape data defaults ProblemType yes UNKNOWN
	}
    } else {
	GiD_Process escape escape escape escape data defaults ProblemType UNKNOWN
    }
}


proc ClearCacheFormProblemtypeMenuDir {} {
    global GidPriv
    foreach i [array names GidPriv dirfiles,*] {
	unset GidPriv($i)
    }
}

proc FastCacheFromProblemtypeMenuDir { dir } { 
    global GidPriv
    #recreate this information also if exists GidPriv(dirfiles,$dir)
    #can be created or deleted files or directrories
    set startrelpath [string length [file join $::GIDDEFAULT problemtypes]]
    incr startrelpath

#WarnWinText "FastCacheFromProblemtypeMenuDir $dir"
    if { [file extension $dir] == ".lnk" && [GidUtils::IsLink $dir] == 2 } {
	set realdir [GidUtils::ReadLink $dir] 
    } else {
	set realdir $dir
    }
    set dirs [glob -nocomplain -tails -directory $realdir -types d *]
    foreach link [glob -nocomplain -tails -directory $realdir -types f *.lnk] {
	if { [GidUtils::IsLink [file join $realdir $link]] == 2 } {
	    lappend dirs $link
	}
    }
    set GidPriv(dirfiles,$dir) [lsort -dictionary $dirs]
    foreach i $GidPriv(dirfiles,$dir) {
	if { [file extension $i] == ".gid" } {
	    set path [string range $dir $startrelpath end]
	    ReadProblemtypeConfigFile $path $i
	    continue
	} elseif { [file extension $i] == ".lnk" && [GidUtils::IsLink [file join $realdir $i]] == 2 && \
	    [file extension [file rootname $i]] == ".gid" } {
	    set path [string range $dir $startrelpath end]
	    ReadProblemtypeConfigFile $path $i
	    continue
	}
	FastCacheFromProblemtypeMenuDir [file join $dir $i]
    }
}

proc FastCacheFromProblemtypeMenu {} {
    global GIDDEFAULT GidPriv
    #recreate this information also if exists GidPriv(dirfiles,$dir)
    #can be created or deletes files or directrories
    
    set dir $GIDDEFAULT
    #set dir [ file join $::GIDDEFAULT problemtypes]
    set dirs [glob -nocomplain -tails -directory $dir -types d *.gid]
    foreach link [glob -nocomplain -tails -directory $dir -types f *.lnk] {
	if { [GidUtils::IsLink [file join $dir $link]] == 2 } {
	    lappend dirs $link
	}
    }
    set GidPriv(dirfiles,$GIDDEFAULT) [lsort -dictionary $dirs]
    WarnWinText "FastCacheFromProblemtypeMenu - [lsort -dictionary $dirs]"
    FastCacheFromProblemtypeMenuDir [file join $dir problemtypes]
}


proc GetProblemTypes { } {
    global GIDDEFAULT GidPriv
    
    set current [GiD_Info Project ProblemType]

    set probt ""
    foreach i [glob -nocomplain [file join $GIDDEFAULT *.gid]] {
	lappend probt [file tail [file root $i]]
	#WarnWinText [file tail [file root $i]]
    }
    set probt [lsort -dictionary $probt]

    return $probt
}


proc AdvanceBar { numerator divisor  inum { title "" } { maintext ""} { width 0} \
	{ dentro 0} {parent .gid} } {
    global GidPriv tcl_platform

    if { $parent == ".gid" } {
	set w .gid._wadvancebar 
    } else { set w $parent._wadvancebar }

    if { !$dentro && ![winfo exists $w] } {
	if { $numerator != 0 } { return }

	if 1 {
	    
	    toplevel $w -class OverlayWindow
	    
	    if { [winfo cells $w] < 6 } {
		set metercolor white
	    } else { set metercolor blue }
	    
	    
	    wm title $w $title
	    set pare [winfo parent $w]
	    set x [expr [winfo x $pare]+[winfo width $pare ]/2- \
		    [winfo reqwidth $w]/2]
	    set y [expr [winfo y $pare]+[winfo height $pare ]/2- \
		    [winfo reqheight $w]]
	    if { $x < 0 } { set x 0 }
	    if { $y < 0 } { set y 0 }
	    WmGidGeom $w +$x+$y
	    if { $tcl_platform(platform) == "windows" } {
		wm transient $w .gid
	    }
	    $w conf -cursor watch
	} else {
	    frame $w -bd 2 -relief groove
	    place $w -in .gid.comm -x 0 -y 0 -width 100
	    raise $w
	    set metercolor blue
	}

	frame $w.pp -relief ridge -bd 2

	if { [winfo class $w] != "Frame" } {
	    if { $width} {
		label $w.pp.l3 -width $width -text $maintext
	    } else {
		label $w.pp.l3 -text $maintext
	    }
	    label $w.pp.l4 -text 0 -relief sunken -width 8 -bd 1
	}
	set wp [::progressbar::progressbar $w.pp.f -variable ::progressbar::var \
		-width 200 -background [ $w.pp cget -background]]
	set ::progressbar::var 0
#          frame $w.pp.f -relief sunken -bd 2 -height 20
#          frame $w.pp.f.f -bg $metercolor
	
	if { [winfo class $w] != "Frame" } {
	    grid $w.pp.l3 $w.pp.l4 -sticky w -padx 3 -pady 3
	}
#          grid $w.pp.f - -sticky ew -padx 3
#          place $w.pp.f.f -in $w.pp.f -x 0 -y 0 -relheight 1 -relwidth 0
	grid $wp - -padx 3
	    
	grid $w.pp -sticky ewns -padx 5 -pady 5
	grid columnconf $w 0 -weight 1
	grid rowconf $w 0 -weight 1

	button $w.stop -text [_ "Stop"] -command { 
	    set GidPriv(StopLoading) 1
	} -bd 1
	##if { [winfo class $w] != "Frame" } {
	    grid $w.stop -row 1 -column 0 -pady 4
	##} else {
	##    grid $w.stop -row 0 -column 1
	##}
	#tkwait visibility $w
	#grab $w
	focus $w.stop
	return
    } else {
	# dentro
	if { $dentro && ( "$GidPriv(CoordFrame,uso)" != "progreso")} {
	    set w $GidPriv(CoordFrame)
	    OlvidaCoordFrame
	    set GidPriv(CoordFrame,uso) progreso
	    frame $w.pp -relief ridge -bd 0 -background [ $w cget -background]
	    label $w.pp.l3 -text $maintext -background [ $w cget -background]
	    set ancho 100
	    if { $width} {
		set na [ expr $width * 6]
		if { $na > $ancho} {
		    set ancho $na
		}
	    }
	    set wp [::progressbar::progressbar $w.pp.f -variable ::progressbar::var \
		    -width $ancho -background [ $w cget -background]]
	    set ::progressbar::var 0
	    grid $w.pp.l3 -sticky w
	    grid $wp -padx 2 -ipady 0 -pady 0
	    
	    button $w.stop -text [_ "Stop"] -command { 
		set GidPriv(StopLoading) 1
	    } -bd 1 -pady 0
	    
	    grid $w.pp -sticky ewns -padx 0 -pady 0 -row 0 -column 0
	    grid $w.stop -padx 0 -pady 0 -row 1 -column 0

	    #grid columnconf $w 0 -weight 1
	    #grid rowconf $w 0 -weight 1
	    focus $w.stop
	    return
	}
    }

    set proportion [expr double($numerator)/$divisor]

    if { $dentro} {
	set w $GidPriv(CoordFrame)
    }

    if { [winfo exists $w.pp.l4] } {
	$w.pp.l4 conf -text $inum
    }

    if { [winfo exists $w.pp.l3] && ( $maintext != "") } {
	if { $width} {
	    $w.pp.l3 conf -text $maintext -width $width
	} else {
	    $w.pp.l3 conf -text $maintext
	}
    }

#      place $w.pp.f.f -x 0 -y 0 -relheight 1 \
#              -relwidth $proportion

    set ::progressbar::var [expr int($proportion*100)]

    update

    if { $numerator == $divisor} {
	catch { focus .gid.comm.e }
	if { !$dentro} {
	    destroy $w
	} else {
	    set w $GidPriv(CoordFrame)
	    catch {
		grid forget        $w.pp.f
		grid forget        $w.pp.l3
		grid forget        $w.pp
		grid forget        $w.stop
		destroy $w.pp.f
		destroy $w.pp.l3
		destroy $w.pp
		destroy $w.stop
	    }
	    set GidPriv(CoordFrame,uso) coordenadas
	    DefineCoordFrame 0
	}
    }


}

proc SaveWinConfiguration { } {
    set file [Browser-ramR file write .gid \
		  [_ "Write window configuration"]]
    if { $file == "" } { return }
    file delete $file
    WriteTCLDefaultsInFile $file
}

proc SetOptions { } {
    global tcl_platform GidPriv
    #aun no se ha pasado por ReadDefaultValues
    #SetGUIAppearanceColor
    
    # option add *Checkbutton*HighlightThickness 0
    # option add *Radiobutton*HighlightThickness 0
    # option add *Checkbutton*TakeFocus 0
    # option add *Radiobutton*TakeFocus 0
    option add *Entry*BorderWidth 1
    if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
	option add *Button*relief "round_raised"
	option add *Menubutton*relief "round_raised"
	#option add *Radiobutton*relief "round_raised"
	#option add *Radiobutton*Offrelief "round_raised"
	#option add *Radiobutton*Overrelief "round_raised"
	option add *Menu*activeBorderWidth 0
    } else {
	option add *Button*relief "raised"
	option add *Menubutton*relief "raised"
    }

    if { $tcl_platform(platform) != "windows" } {
	option add *Scrollbar*Width 10
	#option add *Scrollbar*BorderWidth 2
	option add *Scrollbar*BorderWidth 1
	option add *Button*BorderWidth 1

	option add *Menu*BorderWidth 1
	option add *Menu*activeBorderWidth 0
	option add *Menu*Relief flat

	#option add *Menu*HighlightThickness 1
	
	#option add *Menu*activeBackground [ CCColorActivo [ .gid cget -bg]]
	#option add *Menu*activeForeground black
	
	menubutton .gid.tonto_menu_button________
	set col_act [ option get .gid.tonto_menu_button________ activeBackground color]
	destroy .gid.tonto_menu_button________

	option add *Menu*activeBackground [ CCColorActivo $col_act 33]
    }
    option add *Menu*TearOff 0
  
    if { ( $tcl_platform(platform) != "windows") && [ string eq $GidPriv(Color,Linux) "brown"] } {
	    # si se cambia esta linea, hay que cambiarla mas abajo: proc SetGUIAppearanceColor ...
	    option add *background AntiqueWhite3
	    option add *foreground black
	    option add *Button*background bisque3
	    option add *Button*foreground black
	    option add *Menu*background bisque3
	    option add *Menu*foreground black
	    option add *DisabledForeground grey60
	    option add *HighlightForeground black
	    option add *HighlightBackground AntiqueWhite3
	    option add *activeBackground [CCColorActivo AntiqueWhite3]
	    option add *Listbox*Background thistle
	    option add *Listbox*selectForeground black
	    option add *Entry*Background thistle
	    option add *Entry*DisabledBackground thistle
	    option add *Entry*DisabledForeground black
	    option add *Entry*Foreground black
    
	    set GidPriv(Color,BotonActivo) [ CCColorActivo AntiqueWhite3]
	    set GidPriv(Color,BotonMarcoActivo) [ CCColorSombra AntiqueWhite3 65]
    
    } else {
	# no se quiere el color marron
	# #from tcl8.3.3 to tcl8.4.1 change the background of the listbox to white
	# #                          and the DisabledBackgruond ot entry to gray
	# option add *Entry*DisabledBackground $GidPriv(Color,DisabledBackgroundEntry)
	# option add *Entry*DisabledForeground $GidPriv(Color,DisabledForegroundEntry)
	# option add *Listbox*Background $GidPriv(Color,BackgroundListbox)

	# de todas maneras necesito definir un color para el active background
	# option add *activeBackground [ CCColorActivo AntiqueWhite3]
	option add *activeBackground [ CCColorActivo orange 129]
	option add *activeForeground Black

	# #option add *Canvas*Background $GidPriv(Color,BackgroundCanvas)
	# 
	# if { [ info exists ::tk_hasRoundRelief] && $::tk_hasRoundRelief} {
	    set GidPriv(Color,BotonActivo) [ CCColorActivo orange 129]
	    set GidPriv(Color,BotonMarcoActivo) [ CCColorActivo orange 193]
	# } else {
	#     set GidPriv(Color,BotonActivo) [ CCColorActivo \#dddddd]
	#     set GidPriv(Color,BotonMarcoActivo) [ CCColorSombra \#dddddd 65]
	# }
	#set GidPriv(Color,BotonActivo) [ option get .gid activeBackground color]
	#set GidPriv(Color,BotonMarcoActivo) [ CCColorSombra $GidPriv(Color,BotonActivo) 65]
    }

    SetFontsDefault
}

proc SetFontsDefault {} {
    global tcl_platform
    set disp_height [winfo screenheight .gid]
#cambio por las pantallas panoramicas
# 768 --> 800
    if { $disp_height > 1024} {
	set fontsize 10
	set bigfontsize 14
    } elseif { $disp_height > 800 && $disp_height <= 1024 } {
	set fontsize 10
	set bigfontsize 14
    } elseif { $disp_height > 600 && $disp_height <= 800 } {
	# for linux (before it was 8)
	set fontsize 10
	set bigfontsize 14
    } elseif { $disp_height > 480 && $disp_height <= 600 } {
	set fontsize 8
	set bigfontsize 11
    } else {
	set fontsize 6
	set bigfontsize 10
    }

    if { $tcl_platform(os) == "IRIX" } {
	font create NormalFont -family helvetica -size 10 -slant roman
	option add *font -*-helvetica-medium-r-*-*-17-*-*-*-*-*-*-*
	font create BoldFont -weight bold -size $fontsize -family helvetica
    } elseif { $tcl_platform(platform) != "windows" } {
	# porque hay un menos en el tamanno?
	#font create NormalFont -family Helvetica -size [expr -$fontsize]
	font create NormalFont -family Helvetica -size [expr $fontsize]
	font create BoldFont -weight bold -size $fontsize -family helvetica
	option add *font NormalFont
    } else {
	set fontsize 8
	catch {
	  #kike, fuente demasiado grande
	  #set size [expr [GiD_Info menuheight]-4]
	  set size [expr [GiD_Info menuheight]-6]
	  font create NormalFont -family "MS Sans Serif" -size -$size
	  set fontsize [font actual NormalFont -size]
	  font delete NormalFont
	}
	set bigfontsize [ expr $fontsize + 4]
	font create NormalFont -family "MS Sans Serif" -size $fontsize
	font create BoldFont -size $fontsize -family "MS Sans Serif" -weight bold
	option add *font NormalFont
    }

    set family [font conf NormalFont -family]
    font create FixedFont -size $fontsize -family Courier
    font create BigFont -family $family -size $bigfontsize
    # [expr int($fontsize*1.5)]
    font create SmallFont -family $family -size [expr int($fontsize-2)]
}

# 
# proc GiveBestGidDefaults {} {
#     global GIDDEFAULTTCL
#     set realw [winfo screenwidth .gid]
#     foreach i [glob [file join $GIDDEFAULTTCL gidDefaults]*] {
#         if { [regexp -nocase {gidDefaults([0-9]*)x([0-9]*)} $i trash w h] } {
#             if { $w == $realw } { return $i }
#             if { ![info exists bestf] || [expr abs($realw-$w)]< $bestdist } {
#                 set bestf $i
#                 set bestdist [expr abs($realw-$w)]
#             }
#         }
#     }
#     if { [info exists bestf] } { return $bestf }
#     return $GIDDEFAULTTCL/gidDefaults
# }

#cambio por las pantallas panoramicas
# pues la ventana de gid se sale de pantalla
proc GiveBestGidDefaults {} {
    global GIDDEFAULTTCL
    set realh [winfo screenheight .gid]
    foreach i [glob [file join $GIDDEFAULTTCL gidDefaults]*] {
	if { [regexp -nocase {gidDefaults([0-9]*)x([0-9]*)} $i trash w h] } {
	    if { $h == $realh } { return $i }
	    if { ![info exists bestf] || [expr abs($realh-$h)]< $bestdist } {
		set bestf $i
		set bestdist [expr abs($realh-$h)]
	    }
	}
    }
    if { [info exists bestf] } { return $bestf }
    return $GIDDEFAULTTCL/gidDefaults
}


proc ContextualSubMenu { w } {
    global ActionsMenuVar ActionsMenuLabels ActionsMenuVarType
    $w del 0 end
    if { $ActionsMenuVarType == "TMP" } {
	set words ""
	foreach i $ActionsMenuVar j $ActionsMenuLabels {
	    #regsub -all {([a-z])([A-Z])} $i "\\1 \[string tolower \\2]" name
	    #set name [subst $name]
	    #set name [string toupper [string index $name 0]][string range $name 1 end]
	    set name $j
	    if { $i == "Join" || $i == "NoJoin" } { append name " Ctrl-a" }
	    lappend words [ list $name ".central.s process $i"]
	}
	rmenuconf $w $words
    } else {
	rmenuconf $w  { { -NONE- "" } }
	$w entryconf end -state disabled
    }
}


proc SelectJoinOrNoJoin {} {
    global ActionsMenuVar

    if { [lsearch -exact $ActionsMenuVar Join] != -1 } {
	.central.s process Join
    } elseif { [lsearch -exact $ActionsMenuVar NoJoin] != -1 } {
	.central.s process NoJoin
    } 
}

proc SelectionOnlyVisibleOrAll {} {
    if { [GiD_Set SelectionOnlyVisible] == 0 } {
	GiD_Set SelectionOnlyVisible 1
	::GidUtils::SetWarnLine "Now only select visible entities"
    } else {
	GiD_Set SelectionOnlyVisible 0
	::GidUtils::SetWarnLine "Now select visible and hidden entities"
    } 
}

proc UpdateGraphFramesBegin { w HorV x y} {
    global GidPriv
    bind $w <ButtonRelease-1> "UpdateGraphFramesEnd %W $x $y %x %y $HorV"
}

proc UpdateGraphFramesEnd { w xold yold x y HorV } {

    set master [lindex [grid info $w] 1]

    regexp -- {-row[ ]*([0-9]+)} [grid info $w] {} irow
    regexp -- {-column[ ]*([0-9]+)} [grid info $w] {} icol
    
    if { $HorV == "v" || $HorV == "hv"} {
	set newwidth1 [expr [lindex [grid bbox $master 0 0] 2]+$x]
	set newwidth2 [expr [lindex [grid bbox $master 2 0] 2]-$x]
	if { $newwidth1 < 0 } { set newwidth1 0 }
	if { $newwidth2 < 0 } { set newwidth2 0 }
	grid columnconf $master 0 -weight $newwidth1
	grid columnconf $master 2 -weight $newwidth2
    } 
    if { $HorV == "h" || $HorV == "hv"} {
	set newheight1 [expr [lindex [grid bbox $master 0 0] 3]+$y]
	set newheight2 [expr [lindex [grid bbox $master 0 2] 3]-$y]
	if { $newheight1 < 0 } { set newheight1 0 }
	if { $newheight2 < 0 } { set newheight2 0 }
	grid rowconf $master 0 -weight $newheight1
	grid rowconf $master 2 -weight $newheight2
    }
}

proc PackToglSeparator { w row col type } {
    switch $type {
	h {
	    $w conf -cursor sb_v_double_arrow
	    bind $w <1> "UpdateGraphFramesBegin %W h %x %y"
	    grid $w -in .gid.central.wins -row $row -column $col \
		    -sticky nsew
	}
	v {
	    $w conf -cursor sb_h_double_arrow
	    bind $w <1> "UpdateGraphFramesBegin %W v %x %y"
	    grid $w -in .gid.central.wins -row $row -column $col \
		    -sticky nsew
	}
	hv {
	    $w conf -cursor cross
	    bind $w <1> "UpdateGraphFramesBegin %W hv %x %y"
	    grid $w -in .gid.central.wins -row $row -column $col \
		    -sticky nsew
	}
    }
}

proc PackToglWindows { type } {
    
    if { ![winfo exists .gid.central.wins] } {
	PrepareToglForSeveralWindows
    }
    
    set totalnum [string index $type 0]
    set i 2
    while { [winfo exists .gid.central.wins.graphwin$i] } {
	incr i
    }
    incr i -1
    if { $i < $totalnum } { return }

    foreach i [grid slaves .gid.central.wins] { grid forget $i }
    for { set i 1 } { $i <= 5 } { incr i } {
	if { ![winfo exists .gid.central.wins.separator$i] } { 
	    frame .gid.central.wins.separator$i
	}
    }
    
    if { $type == "1" } {
	grid .gid.central.wins.graphwin1 -in .gid.central.wins \
		-row 0 -column 0 \
		-sticky nsew -columnspan 3  -rowspan 3 
	.gid.central.s SetPosition 1
#sobra?# puesto idletasks
	update idletasks
	.gid.central.wins.graphwin1 configure -bd 0
    } elseif { $type == "2LR" || $type == "3L" } {
	grid .gid.central.wins.graphwin1 -rowspan 3 -row 0 -column 0 \
		-sticky nsew
	.gid.central.s SetPosition L
	PackToglSeparator .gid.central.wins.separator1 0 1 v
	if { $type == "2LR" } {
	    PackToglSeparator .gid.central.wins.separator2 1 1 v
	} else {
	    PackToglSeparator .gid.central.wins.separator2 1 1 hv
	}
	PackToglSeparator .gid.central.wins.separator3 2 1 v
    } elseif { $type == "2UD" || $type == "3U" } {
	grid .gid.central.wins.graphwin1 -columnspan 3 -row 0 -column 0 \
		-sticky nsew
	.gid.central.s SetPosition B
	PackToglSeparator .gid.central.wins.separator4 1 0 h
	if { $type == "2UD" } {
	    PackToglSeparator .gid.central.wins.separator2 1 1 v
	} else {
	    PackToglSeparator .gid.central.wins.separator2 1 1 hv
	}
	PackToglSeparator .gid.central.wins.separator5 1 2 h
    } else {
	grid .gid.central.wins.graphwin1 -row 0 -column 0 \
		-sticky nsew
	.gid.central.s SetPosition BL
	PackToglSeparator .gid.central.wins.separator1 0 1 v
	PackToglSeparator .gid.central.wins.separator2 1 1 hv
	PackToglSeparator .gid.central.wins.separator4 1 0 h
    }

    set icurr 2
    if { $type == "2UD" || $type == "3D" } {
	grid .gid.central.wins.graphwin$icurr -in .gid.central.wins \
		-columnspan 3 -row 2 -column 0 \
		-sticky nsew
	.gid.central.wins.graphwin${icurr}.s SetPosition T
	incr icurr
    } elseif { $type == "3R" || $type == "3U" || $type == "4" } {
	grid .gid.central.wins.graphwin$icurr -in .gid.central.wins -row 2 -column 0 \
		-sticky nsew
	.gid.central.wins.graphwin${icurr}.s SetPosition TL
	incr icurr
	PackToglSeparator .gid.central.wins.separator3 2 1 v
    }
    if { $type == "2LR" || $type == "3R" } {
	  grid .gid.central.wins.graphwin$icurr -in .gid.central.wins -rowspan 3 -row 0 -column 2 \
		  -sticky nsew
	.gid.central.wins.graphwin${icurr}.s SetPosition R
	  incr icurr
    } elseif { $type == "3L" || $type == "3D" || $type == "4" } {
	grid .gid.central.wins.graphwin$icurr -in .gid.central.wins \
		-rowspan 1 -row 0 -column 2 \
		-sticky nsew
	.gid.central.wins.graphwin${icurr}.s SetPosition BR
	incr icurr
	PackToglSeparator .gid.central.wins.separator5 1 2 h

    }

    if { $type == "3L" || $type == "3U" || $type == "4" } {
	grid .gid.central.wins.graphwin$icurr -in .gid.central.wins -row 2 -column 2 \
		-sticky nsew
	.gid.central.wins.graphwin${icurr}.s SetPosition TR
	incr icurr
    }

    grid columnconf .gid.central.wins 0 -weight 1
    grid rowconf .gid.central.wins 0 -weight 1

    grid columnconf .gid.central.wins 1 -weight 0 -minsize 1
    grid columnconf .gid.central.wins 2 -weight 1
    
    grid rowconf .gid.central.wins 1 -weight 0 -minsize 1
    grid rowconf .gid.central.wins 2 -weight 1


#      if { $type == 1 } {
#          set i 5
#          while { [winfo exists .gid.central.wins.graphwin$i] } { incr i }
#          incr i -5
#          if { $i == 0 } {
#              .gid.central.wins.wins.graphwin1 conf -bd 0
#          }
#      }

}

proc GiveNumToglInteriorWins {} {
    set i 2
    while { [winfo exists .gid.central.wins.graphwin$i] } { incr i }
    incr i -1
    return $i
}

proc DestroyRemainingToglWindows { type } {

    if { "[string tolower $type]" == "ext" } { return }
    
    set totalnum [string index $type 0]
    set i [expr $totalnum+1]
    while { [winfo exists .gid.central.wins.graphwin$i] } {
	  destroy .gid.central.wins.graphwin$i
	incr i
    }
}

proc PrepareToglForSeveralWindows {} {

    frame .gid.central.wins
    #pack .gid.central.wins -expand 1 -fill both
    

    switch [winfo manager .gid.central.s] {
	grid {
	    eval grid .gid.central.wins [grid info .gid.central.s]
	}
	pack {
	    pack .gid.central.wins -in .gid.central -before .gid.central.s \
		-anchor center -expand 1 -fill both -side left
	}
	panedwindow {
	    set ws "."
	    set found 0
	    for { set i 0 } { $i < [llength $ws] } { incr i } {
		foreach j [winfo children $ws] {
		    if { [winfo class $j] eq "Panedwindow" && [lsearch [$j panes] .gid.central.s] != -1 } {
		        set found 1
		        break
		    }
		    lappend ws $j
		}
	    }
	    if { $found } {
		set ipos [lsearch [$j panes] .gid.central.s]
		$j add .gid.central.wins -before .gid.central.s
		$j forget .gid.central.s
	    }
	}
    }
    
    
    frame .gid.central.wins.graphwin1 -bd 2 -relief groove
    pack .gid.central.s -in .gid.central.wins.graphwin1 -expand 1 \
	    -fill both
    .gid.central.s configure -width 0 -height 0
    
    raise .gid.central.s

    grid .gid.central.wins.graphwin1 -in .gid.central.wins -row 0 \
	    -column 0 -sticky nsew
    grid rowconf .gid.central.wins 0 -weight 1
    grid columnconf .gid.central.wins 0 -weight 1
}

proc GiDActualizaVentanasTogl { } {
    .central.s render
    if { [winfo exists .gid.central.wins] } {
	# 1, 2, 3, 4, 5 --> interiores
	# 6 --> exterior
	for { set i 0} { $i < 7} { incr i} {
	    set new_win .gid.central.wins.graphwin$i
	    if { [ winfo exists $new_win] && [ winfo exists ${new_win}.s] } {
		${new_win}.s render
	    }
	}
    }
}

proc CreateNewToglWindow { type } {
    global tcl_platform GidPriv

    if { ![winfo exists .gid.central.wins] } {
	PrepareToglForSeveralWindows
    }

    set type [string toupper $type]

    if { $type == "EXT" } {
	set i 6
    } else {
	set i 2
    }
    while { [winfo exists .gid.central.wins.graphwin$i] } { incr i }
    set w .gid.central.wins.graphwin$i
    
    if { $type == "EXT" } {
	# "" 0 0 <- para permitir que se ponga debajo ( y no la haga transient)
	set geomname NewWindow${i}WindowGeom
	InitWindow $w [_ "View %d" $i] $geomname \
		"after idle GiD_Process NewWin EXT" "" 0 0
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
	if { ![info exists GidPriv($geomname)] } {
	    wm geometry $w "320x240+100+100"
	}
	bind $w <Escape> ".central.s process escape"
	if { $GidPriv(GiDRuedas)} {
	    bind $w <MouseWheel> { MouseWheelEventToButtonEvent %D }
	}
	bind $w <Control-a> SelectJoinOrNoJoin
	#bind $w <Control-h> SelectionOnlyVisibleOrAll
	

	set w $w.f
	frame $w -bd 2 -relief groove
	pack $w -expand 1 -fill both
    } else {
	frame $w -bd 2 -relief groove
	#PackToglWindows $type
    }

    set extra_flags ""
    
    if { $tcl_platform(platform) == "windows" } {
	if { $GidPriv(SoftwareOpenGL)} {
	    set extra_flags "-softrender true"
	}
    }

    if { [catch { \
	     eval [ subst "togl $w.s -rgba true -double true   \
	     -depth true -ident GID -privatecmap false -overlay 1 \
	     -width 100  -height 100 $extra_flags"] }]} {
	 eval [ subst "togl $w.s -rgba true -double true  \
		 -depth true -ident GID -privatecmap false -overlay 0 \
		 -width 100   -height 100 $extra_flags"]
    }
    
    $w.s conf -width 0 -height 0
    pack $w.s -fill both -expand yes

    bind $w.s <1> { 
	set GidPriv(PressingButton1) 1
	%W button 1 %x %y
	after 300 [ list catch {  unset GidPriv(PressingButton1) } ]
    }
    bind $w.s <ButtonRelease-1> {
	if { [info exists GidPriv(PressingButton1)] } {
	    %W button -r -f 1 %x %y 
	    break
	}
	%W button -r 1 %x %y 
    }
    bind $w.s <2> { %W button 2 %x %y }
    bind $w.s <Motion> { %W motion %x %y }
    bind $w.s <Leave> { %W leave %x %y }
    bind $w.s <Enter> { %W enter %x %y }

    bind $w <Control-a> SelectJoinOrNoJoin
    #bind $w <Control-h> SelectionOnlyVisibleOrAll
    
    bind $w.s <3> {
	#%W gidmakecurrent
	set x [expr [winfo rootx %W]+%x+2]
	set y [expr [winfo rooty %W]+%y]
	catch {
	    GiD_PopupMenu .gid.phantom.display3dbutton $x $y
	}
    }
    foreach i [bind .gid] {
	bind $w $i [bind .gid $i]
    }
#     bind $w <FocusIn> "catch { focus .gid.comm.e }"

    SetButtonMeaning $w.s

    return $w
}



# to initialize
set GidHelp(HelpSystemActive) "" 
set GidHelp(helpcanceller) ""
set GidPriv(JustHelpActived) 0
set GidHelp(HelpLeaveBefore) ""
set GidHelp(JustDeleted) ""
set GidPriv(HelpSystem) .gid.helpsystem
set GidPriv(HelpSystemmenu) 0

proc DisplayGidHelp { w text but x y } {
    global GidPriv GidHelp

    if { $but != "menu" && ($GidHelp(JustDeleted) == $w || $GidHelp(JustDeleted) == "all") } { return }

    if { ![winfo exists $w] } { return }

    if { [winfo exists $GidPriv(HelpSystem)] } {
	if { $w == $GidHelp(HelpSystemActive) } { return }
	destroy $GidPriv(HelpSystem)
    }
    set tag ""
    if { [llength $w] == 2 } {
	set tag [lindex $w 1]
	set w [lindex $w 0]
    }

    set x [winfo pointerx $w]
    set y [winfo pointery $w]

    if { $but != "menu" } {
	if { $x < [winfo rootx $w] || $y < [winfo rooty $w] || \
		$x > [expr [winfo rootx $w]+[winfo width $w]] || \
		$y > [expr [winfo rooty $w]+[winfo height $w]] } {
	    return
	}
    }

    if { $y < [expr [winfo screenheight $w]/2.0] } {
	set y +[expr $y+3]
    } else {
	set y -[expr [winfo screenheight $w]-$y+3]
    }
    if { $x < [expr [winfo screenwidth $w]/2.0] } {
	set x +[expr $x+3]
    } else {
	set x -[expr [winfo screenwidth $w]-$x+3]
    }

    toplevel $GidPriv(HelpSystem)

    set GidHelp(HelpSystemActive) $w
    wm overrideredirect $GidPriv(HelpSystem) 1
    wm geom $GidPriv(HelpSystem) $x$y
    frame $GidPriv(HelpSystem).f -relief ridge -bg white -bd 1
    label $GidPriv(HelpSystem).f.l -bg white -fg black \
	    -justify left -text $text -wraplength 10c
    set link [ToolbarHelp $text]
    if { $link != "" } {
	button $GidPriv(HelpSystem).f.b -image [GetImage points3.gif] \
	    -command "destroy $GidPriv(HelpSystem); directHelp $link" -bg white -bd 1
	bind $GidPriv(HelpSystem).f.b <1> {
	    uplevel #0 [ list %W invoke]
	    destroy $GidPriv(HelpSystem)
	    break
	}
    }
    pack $GidPriv(HelpSystem).f.l -fill both -expand 1
    if { [winfo exists $GidPriv(HelpSystem).f.b] } {
	pack $GidPriv(HelpSystem).f.b -anchor e -padx 1 -pady 1
    }
    pack $GidPriv(HelpSystem).f -fill both -expand 1

    if { $but == "enter" } {
	bind $GidPriv(HelpSystem) <1> "destroy $GidPriv(HelpSystem)"
	#bind $w <Leave> "catch { destroy $GidPriv(HelpSystem) } ; \
		#    bind $w <Leave> {}"
    } elseif { $but == "menu" } {
	if { [catch { tkwait visibility $GidPriv(HelpSystem) }] } { return }
	set oldfocus [focus]
	focus $GidPriv(HelpSystem)
	grab -global $GidPriv(HelpSystem)
	set GidPriv(HelpSystemmenu) 1
	bind $GidPriv(HelpSystem) <Destroy> "set GidPriv(HelpSystemmenu) 0"
 
	bind $GidPriv(HelpSystem) <1> "destroy $GidPriv(HelpSystem) ; focus $oldfocus"
	bind $GidPriv(HelpSystem) <2> "destroy $GidPriv(HelpSystem) ; focus $oldfocus"
	bind $GidPriv(HelpSystem) <3> "destroy $GidPriv(HelpSystem) ; focus $oldfocus"
	bind $GidPriv(HelpSystem) <Escape> "destroy $GidPriv(HelpSystem) ; focus $oldfocus"
    } else {
	bind $GidPriv(HelpSystem) <1> "destroy $GidPriv(HelpSystem)"
	if { $tag == "" } {
	    bind $w <ButtonRelease-$but> " \
		    catch { destroy $GidPriv(HelpSystem) } ; \
		    bind $w <ButtonRelease-$but> {}"
	} else {
	    $w tag bind $tag <ButtonRelease-$but> " \
		    catch { destroy $GidPriv(HelpSystem) } ; \
		    $w tag bind $tag <ButtonRelease-$but> {}"
	}   
    }
}

proc DisplayGidHelpMenu { w text type xbase ybase } {
    global GidHelp

    after cancel $GidHelp(helpcanceller)
    set GidHelp(helpcanceller) ""

    if { $GidHelp(HelpLeaveBefore) != "" } {
	bind $w <Leave> $GidHelp(HelpLeaveBefore)
    }
    set GidHelp(HelpLeaveBefore) ""

    set x $xbase ; set y $ybase

    catch { destroy .gid.__helpmenu }
    menu .gid.__helpmenu
    .gid.__helpmenu add command -label [_ "help"] -command \
	    "DisplayGidHelp $w [ list $text] menu $xbase $ybase"
    if { $type == "helpandtools" } {
	.gid.__helpmenu add command -label [_ "Configure toolbars"] -command \
		TOBegin
    }
    GiD_PopupMenu .gid.__helpmenu $x $y
}

proc DisplayGidHelpLeave { w x y { force no } } {
    global GidPriv GidHelp

    if ![winfo exists $w] {
	return
    }
    if { $force == "no" && $x >= [winfo rootx $w] && $y >= [winfo rooty $w] && \
	    $x <= [expr [winfo rootx $w]+[winfo width $w]] && \
	    $y <= [expr [winfo rooty $w]+[winfo height $w]] } {
	return
    }

    if { [winfo exists $GidPriv(HelpSystem)] } {
	if { $force == "no" && $x >= [winfo rootx $GidPriv(HelpSystem)] && $y >= \
		[winfo rooty $GidPriv(HelpSystem)] && \
		$x <= [expr [winfo rootx $GidPriv(HelpSystem)]+[winfo width $GidPriv(HelpSystem)]] && \
		$y <= [expr [winfo rooty $GidPriv(HelpSystem)]+[winfo height $GidPriv(HelpSystem)]] } {
	    return
	}
    }
    

    bind all <Motion> ""

    after cancel $GidHelp(helpcanceller)
    set GidHelp(helpcanceller) ""

    if { [winfo exist $GidPriv(HelpSystem)] } {
	set GidPriv(JustHelpActived) 1  
	after 1000 set GidPriv(JustHelpActived) 0  
    }

    if { !$GidPriv(HelpSystemmenu) } {
	catch { destroy $GidPriv(HelpSystem)}  
    }
    if { $GidHelp(HelpLeaveBefore) != "" } {
	bind $w <Leave> $GidHelp(HelpLeaveBefore)
    }
    set GidHelp(HelpLeaveBefore) ""
    after cancel set GidHelp(JustDeleted) ""

    if { $force == "no" } {
	set GidHelp(JustDeleted) $w
    } else {
	set GidHelp(JustDeleted) all
    }
    after 3000 "set GidHelp(JustDeleted) {}"
}

proc DisplayGidHelpLauncher { w x y text } {
    global GidPriv GidHelp

    #if { $GidHelp(JustDeleted) == $w || $GidHelp(JustDeleted) == "all" } { return }

    if { $GidPriv(JustHelpActived) } {
	DisplayGidHelp $w $text enter $x $y
    }

    if { [winfo exist $GidPriv(HelpSystem)] } {
	DisplayGidHelp $w "$text" enter $x $y
    } else {
	after cancel $GidHelp(helpcanceller)
	set GidHelp(helpcanceller) [after 1500 DisplayGidHelp $w [ list $text] enter $x $y]
    }
    #set GidHelp(HelpLeaveBefore) [bind $w <Leave>]

    bind all <Motion> "DisplayGidHelpLeave $w %X %Y"
    #bind $w <Leave> "$GidHelp(HelpLeaveBefore) ; DisplayGidHelpLeave $w"
    bind $w <ButtonPress-1> "+DisplayGidHelpLeave $w %X %Y yes"
}

proc GidHelpL { wlist type args } {
    set text [join $args]

    # use bwidget
    #DynamicHelp::configure -bg white

    foreach w $wlist {
	set tag ""
	if { [llength $w] == 2 } {
	    set tag [lindex $w 1]
	    set w [lindex $w 0]
	}
	if { ![winfo exists $w] } { continue }

	set haveimage 0
	catch {
	    if { [$w cget -image] != "" } { set haveimage 1 }
	}
	catch {
	    if { [$w cget -bitmap] != "" } { set haveimage 1 }
	}

	if { $haveimage } {
	    if { $tag == "" } {
		# using BWidgets
		#DynamicHelp::register $w balloon $text
		#return
		bind $w <Enter> "+DisplayGidHelpLauncher $w %X %Y [ list $text]"
		foreach i [winfo children $w] {
		    GidHelpL $i $type $text
		}
	    } else {
		$w tag bind $tag <Enter> "+DisplayGidHelpLauncher $w %X %Y [ list $text]"
	    }
	}

	if { $tag == "" } {
	    bind $w <ButtonPress-2> \
		          "[ list DisplayGidHelpMenu $w $text $type %X %Y]; break"
	    bind $w <ButtonPress-3> \
		    "[ list DisplayGidHelpMenu $w $text $type %X %Y]; break"
	    foreach i [winfo children $w] {
		GidHelpL $i $type $text
	    }
	} else {
	    $w tag bind $tag <ButtonPress-2> \
		          "[ list DisplayGidHelpMenu $w $text $type %X %Y]; break"
	    $w tag bind $tag <ButtonPress-3> \
		    "[ list DisplayGidHelpMenu $w $text $type %X %Y]; break"
	}
    }
}

proc GidHelp { wlist args } {
    GidHelpL $wlist help [join $args]
}

proc GidHelpT { wlist args } {
    GidHelpL $wlist helpandtools [join $args]
}

proc GidHelpRecursive { w args } {
    GidHelpL $w help [join $args]
    foreach i [winfo children $w] {
	eval GidHelpRecursive $i $args
    }
}

proc NURBSSurfaceMove {} {
    global GidPriv
    set w .gid.nurbsurfmove

    InitWindow $w [_ "NURBS surface movement"] NURBSSurfaceMoveWindowGeom NURBSSurfaceMove
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0


    set GidPriv(NURBSsurfoldmove) 0
    scale $w.s -from -10 -to 10 -resolution 1 -showvalue 0 \
	    -command NURBSSurfaceMoving -orient horizontal

    proc NURBSSurfaceMoving { val } {
	global GidPriv
	set incre [expr $val-$GidPriv(NURBSsurfoldmove)]
	.central.s process normal=$incre
	set GidPriv(NURBSsurfoldmove) $val
    }
	

    button $w.b -text [_ "Reset slide"] -command "\
	    set GidPriv(NURBSsurfoldmove) 0 ; $w.s set 0"

    button $w.close -text [_ "Close"] -command "destroy $w"
    SetWidgetsWidthFromText $w.b $w.close

    grid $w.s -row 1 -column 1 -sticky ew
    grid $w.b -row 2 -column 1
    grid $w.close -row 2 -column 2
    grid columnconf $w 1 -weight 1


}
#NURBSSurfaceMove

proc PostGraphResultsFill { combo_anal combo_step combo_result frame_components type } {
    variable ViewMenuPriv

    if { ![winfo exists $combo_result] } { return }

    if { $type eq "anal" } {
	set all [lsort -dictionary [.central.s info postprocess get all_analisis]]
	# set all [ .central.s info postprocess get all_analisis]
	set curr [ .central.s info postprocess get cur_analisis]
	
	set all_analisis {}
	set max_len 0
	set l 0
	foreach i $all {
	    regsub -all {_} $i { } var
	    lappend all_analisis $var
	    set l [ string length $var]
	    if { $l > $max_len} {
		set max_len $l
	    }
	}
	$combo_anal configure -values $all_analisis
	if { $l > 20} {
	    grid configure $combo_anal -sticky ew
	}
	if { $curr != "" } {
	    regsub -all {_} $curr { } curr
	} else { regsub -all {_} [lindex $all 0] { } curr }
	set var [$combo_anal cget -textvariable]
	uplevel #0 [list set $var $curr]
	return
    } else {
	set var [$combo_anal cget -textvariable]
	set analisis [uplevel #0 [list set $var]]
	regsub -all { } $analisis {_} analisis
	if { [lsearch -exact [.central.s info postprocess get all_analisis] \
	    $analisis] == -1 } {
	    set analisis [.central.s info postprocess get cur_analisis]
	}
    }

    if { $type eq "anal" || $type eq "step" } {
	if { $analisis ne ""} {
	    set all [.central.s info postprocess get all_steps $analisis]
	    set current [.central.s info postprocess get cur_step $analisis]
	    if { ( $current == "") || ( [ lsearch -exact $all $current] == -1) } { 
		regsub -all {_} [lindex $all 0] { } current
	    }
	    $combo_step configure -values $all
	    set var [$combo_step cget -textvariable]
	    uplevel #0 [list set $var $current]
	    $combo_step configure -state normal
	    set step $current
	} else {
	    $combo_step configure -values ""
	    set var [$combo_step cget -textvariable]
	    uplevel #0 [list set $var ""]
	    $combo_step configure -state disabled
	    set step ""
	}
	if { $type eq "step" } { return }
    } else {
	set var [$combo_step cget -textvariable]
	set step [uplevel #0 [list set $var]]
	if { [lsearch -exact [.central.s info postprocess get all_steps $analisis] \
	    $step] == -1 } {
	    set step [.central.s info postprocess get cur_step $analisis]
	}
    }
    if { $analisis ne "" && $step ne "" } {
	set res [ lsort -dictionary [GiD_Info postprocess get results_list Contour_Fill $analisis $step]]
	set var [$combo_result cget -textvariable]
	if { $res eq "" } {
	    uplevel #0 [list set $var ""]
	    $combo_result configure -values ""
	    $combo_result configure -state disabled
	    set result ""
	} else {
	    set result [uplevel #0 [list set $var]]
	    if { [lsearch -exact $res $result] == -1 } {
		set result [lindex $res 0]
		uplevel #0 [list set $var $result]
	    }
	    $combo_result configure -values $res
	    $combo_result configure -state normal
	}
    } else {
	set result ""
	uplevel #0 [list set $var ""]
	$combo_result configure -values ""
	$combo_result configure -state disabled
    }
    foreach i [winfo children $frame_components] {
	catch { destroy $i }
    }
    if { $result eq "" } { return }
    set irow 1
    set componentes_txt [GiD_Info postprocess get components_list Contour_Fill \
	    $result $analisis $step]
    foreach i $componentes_txt {
	set ww [ string tolower $i]$irow
	regsub -all {\.} $ww · w
	regsub -all {_} $i { } comp
	radiobutton $frame_components.r$w -text $comp -value $i -variable \
	    ViewMenuPriv(border_comp)
	grid $frame_components.r$w -sticky nw -row $irow -column 1
	grid rowconf $frame_components $irow -weight 0
	incr irow
    }
    set ViewMenuPriv(border_comp) [lindex $componentes_txt 0]
    grid rowconf $frame_components $irow -weight 1
    grid columnconf $frame_components 1 -weight 1
    if { [ llength $componentes_txt] != 4} {
	set ViewMenuPriv(border_comp) [ lindex $componentes_txt 0]
    } else {
	set ViewMenuPriv(border_comp) [ lindex $componentes_txt 3]
    }
}

proc PostGraphResultsWindow { {w .gid.graph} } {
    global ViewMenuPriv

    InitWindow $w [_ "View graphs"] PostGraphResultsWindowGeom PostGraphResultsWindow
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
#PostGraphResultsWindow 
#<- para abrir una ventana sin resultados?

    set cur_anal [GiD_Info postprocess get cur_analisis]

    labelframe $w.m1 -text [_ "graph type"]
    set m1 [frame $w.m1.m]
    pack $m1 -fill both -expand 1
    
    #TitleFrame $w.m1 -text [_ "graph type"]
    #set m1 [$w.m1 getframe]

    radiobutton $m1.r1 -text [_ "Line variation"] -variable ViewMenuPriv(border_graph_axX) \
	-value LineVariation
    radiobutton $m1.r2 -text [_ "X variation"] -variable ViewMenuPriv(border_graph_axX) \
	-value X_Variation
    radiobutton $m1.r3 -text [_ "Y variation"] -variable ViewMenuPriv(border_graph_axX) \
	-value Y_Variation
    radiobutton $m1.r4 -text [_ "Z variation"] -variable ViewMenuPriv(border_graph_axX) \
	-value Z_Variation

    set ViewMenuPriv(border_graph_axX) LineVariation

    grid $m1.r1 $m1.r2 -sticky w
    grid $m1.r3 $m1.r4 -sticky w

    labelframe $w.m2 -text [_ "result"]
    set m2 [frame $w.m2.m]
    pack $m2 -fill both -expand 1

#     TitleFrame $w.m2 -text [_ "Result"]
#     set m2 [$w.m2 getframe]

    label $m2.l1 -text [_ "Analysis:"]
    ComboBox $m2.am1 -textvariable ViewMenuPriv(BG_PostResultsAnal) -editable 0 \
	-width 20 -bd 1

    label $m2.lstep -text [_ "Step:"]
    ComboBox $m2.cbstep -textvariable ViewMenuPriv(BG_PostResultsStep) \
	-editable 0 -width 8 -bd 1

    label $m2.prr -text [_ "Result:"]
    ComboBox $m2.m2 -textvariable ViewMenuPriv(BG_PostResultsRes) -width 20 \
	-editable 0

    grid $m2.l1 $m2.am1 $m2.lstep -sticky ew -pady 2
    grid $m2.prr $m2.m2 $m2.cbstep -sticky ew -pady 2
    grid columnconfigure $m2 1 -weight 1

    labelframe $w.fr -text [_ "components"]
    set fr [frame  $w.fr.m]
    pack $fr -fill both -expand 1

#     TitleFrame $w.fr -text [_ "components"]
#     set fr [$w.fr getframe]
    $fr conf -height 100

#    frame $w.fr -bd 2 -relief ridge -height 100

    set cmd [list PostGraphResultsFill $m2.am1 $m2.cbstep $m2.m2 $fr]
    foreach "n t" [list BG_PostResultsAnal step BG_PostResultsStep result \
	    BG_PostResultsRes components] {
	trace add variable ViewMenuPriv($n) write "$cmd $t;#"
	bind $fr <Destroy> [list trace remove variable \
		ViewMenuPriv($n) write "$cmd $t;#"]
    }
    eval $cmd anal

#     if { ( "$res" != "-No results-") && ( "$res" != "") } {
#         set cmd "AddComponentsToMenu $m2 $fr ;#"
#         trace add variable ViewMenuPriv(border_graph_axY) write $cmd
#         set ViewMenuPriv(border_graph_axY) [lindex $res 0]
#         bind $fr <Destroy> [list trace remove variable \
#                 ViewMenuPriv(border_graph_axY) write $cmd]
#     }
    frame $w.buts -bg [CCColorActivo [$w cget -bg]]


    button $w.buts.b0 -text [_ "Select 2 points"] -width 10 \
	    -command "ViewGraphs:BorderGraphTwoPointsInternal $w.buts"
    button $w.buts.b -text [_ "Select border"] -width 10 -command "ViewGraphs:SelectBorder $w.buts"
    focus $w.buts.b

    if { ![info exists ViewMenuPriv(image,arrow_down_vsmall.gif)] } {
	set ViewMenuPriv(image,arrow_down_vsmall.gif) [image create photo -file \
		[file join $::GIDDEFAULT bitmaps arrow_down_vsmall.gif]]
    }
    menubutton $w.buts.mopt -text [_ "Actions"] -width 10 -menu $w.buts.mopt.m \
	-image $ViewMenuPriv(image,arrow_down_vsmall.gif) \
	-compound right
    menu $w.buts.mopt.m
    $w.buts.mopt.m add command -label [_ "Draw graphs"] -command {
	.central.s process escape escape escape escape Results Graphs \
		        OptionsGraph ShowGraphs yes
    }
    $w.buts.mopt.m add command -label [_ "No draw graphs"] -command {
	.central.s process escape escape escape escape Results Graphs \
		        OptionsGraph ShowGraphs No
    }
    $w.buts.mopt.m add command -label [_ "Clear graphs"] -command {
	.central.s process escape escape escape escape Results Graphs \
		        OptionsGraph ClearGraphs
    }

    button $w.buts.close -text [_ "Close"] -width 10 -command "destroy $w"

    SetWidgetsWidthFromText $w.buts.b0 $w.buts.b $w.buts.mopt $w.buts.close

    # necessary as it is a compound
    $w.buts.mopt configure -width 50

    grid $w.buts.b0 $w.buts.b $w.buts.mopt $w.buts.close -padx 3 -pady 3
    

    grid $w.m1 -sticky nwe -row 1 -column 1
    grid $w.m2 -sticky nwe -row 2 -column 1
    grid $w.fr -sticky nsew -row 3 -column 1
    grid $w.buts -sticky nwe -row 5 -column 1
    if { $::tcl_version >= 8.5 } { grid anchor $w.buts center }
    grid rowconf $w 3 -weight 1
    grid columnconf $w 1 -weight 1
}

proc ViewGraphs:SelectBorder { frame } {
    global ViewMenuPriv
    
    if {[info exist ViewMenuPriv(border_comp)]} {
	regsub -all { } $ViewMenuPriv(BG_PostResultsAnal) {_} analysis
	GiD_Process escape escape escape escape results AnalysisSel \
	    $analysis $ViewMenuPriv(BG_PostResultsStep)

	GiD_Process escape escape escape escape results graphs \
	    bordergraph $ViewMenuPriv(border_graph_axX) \
	    $ViewMenuPriv(BG_PostResultsRes) $ViewMenuPriv(border_comp)
	set SmallWinSelecting [GiD_Info variables SmallWinSelecting]
	FinishButton [winfo toplevel $frame] $frame \
	    [_ "Press 'Finish' to end selection"] \
	    "" disableall $SmallWinSelecting
    }
}

proc ViewGraphs:BorderGraphTwoPointsInternal { frame } {
    global ViewMenuPriv
    if {[info exist ViewMenuPriv(border_comp)]} {
	GiD_Process escape escape escape escape results graphs linegraph \
		$ViewMenuPriv(BG_PostResultsRes) $ViewMenuPriv(border_comp)
	
	set SmallWinSelecting [GiD_Info variables SmallWinSelecting]
	FinishButton [winfo toplevel $frame] $frame \
		[_ "Press 'Finish' to end selection"] \
		"" disableall $SmallWinSelecting
    }
}

#obsolete proc using auxiliary tcl commands, best use BorderGraphTwoPointsInternal
proc ViewGraphs:BorderGraphTwoPoints {} {
    variable ViewMenuPriv

    if {![info exist ViewMenuPriv(border_comp)]} { return }

    if { [GiD_Info project ViewMode] == "GRAFUSE" } {
	GiD_Process escape escape Results Graphs OptionsGraph ShowGraphs No
    }
    set p1 [GidUtils::GetCoordinates [_ "Enter first point (ESC to leave)"]]
    if { $p1=="" } return
    set p2 [GidUtils::GetCoordinates [_ "Enter second point (ESC to leave)"]]
    if { $p2=="" } return

    if { [MathUtils::VectorDistance $p1 $p2] < $MathUtils::Epsilon } {
	WarnWin [= "Points cannot be the same"]
	return
    }
    set v1 [MathUtils::VectorDiff $p1 $p2]
    if { [lindex $v1 0] != 0.0 || [lindex $v1 1] != 0.0 } {
	set p3 [MathUtils::VectorSum $p1 [list [expr {-1*[lindex $v1 1]}] \
		    [lindex $v1 0] 0.0]]
    } else {
	set p3 [list [lindex $v1 2] 0 0]
    }
    set n [MathUtils::VectorVectorialProd [MathUtils::VectorDiff $p2 $p1] \
	    [MathUtils::VectorDiff $p3 $p1]]
    set p4 [MathUtils::VectorSum $p1 $n]

    set old_cuts [GiD_Info postprocess get all_cutsets]
    set old_surfacesets [GiD_Info postprocess get all_surfacesets]
    set old_volumesets [GiD_Info postprocess get all_volumesets]

    GidUtils::WaitState .gid
    GidUtils::DisableGraphics

    if { $old_volumesets != "" } {
	################################################################################
	#    First cut
	################################################################################

	GiD_Process escape escape escape escape DoCut CutPlane ThreePoints \
	    $p1 $p2 $p3
	GiD_Process escape escape escape escape DoCut TurnIntoMeshes

	set newsets ""
	foreach set [GiD_Info postprocess get all_surfacesets] {
	    if { [lsearch -exact $old_surfacesets $set] == -1 } {
		lappend newsets $set
	    }
	}
	
	################################################################################
	#    Delete 2nd cuts
	################################################################################
	
	set cmd [list GiD_Process escape escape escape escape \
		Utilities Delete CutSets]
	foreach cut [GiD_Info postprocess get all_cutsets] {
	    if { [lsearch -exact $old_cuts $cut] == -1 } {
		lappend cmd $cut yes
	    }
	}
	lappend cmd escape
	eval $cmd
	
	################################################################################
	#    Sets to off and on
	################################################################################
	
	set active_vols [GiD_Info postprocess get cur_volumesets]
	set active_surfs [GiD_Info postprocess get cur_surfacesets]
	
	set cmd [list GiD_Process escape escape escape escape select \
		VolumeSets]
	eval lappend cmd $active_vols
	lappend cmd escape
	eval $cmd
	
	set cmd [list GiD_Process escape escape escape escape select \
		SurfaceSets]
	foreach surf $active_surfs {
	    if { [lsearch -exact $newsets $surf] == -1 } {
		lappend cmd $surf
	    }
	}
	foreach surf $newsets {
	    if { [lsearch -exact $active_surfs $surf] == -1 } {
		lappend cmd $surf
	    }
	}
	lappend cmd escape
	eval $cmd        
    
	################################################################################
	#    Second cut
	################################################################################
	
	GiD_Process escape escape escape escape DoCut CutPlane ThreePoints \
	    $p1 $p2 $p4
	GiD_Process escape escape escape escape DoCut TurnIntoMeshes
	
	################################################################################
	#    Delete 2nd cuts
	################################################################################
	
	set cmd [list GiD_Process escape escape escape escape \
		Utilities Delete CutSets]
	foreach cut [GiD_Info postprocess get all_cutsets] {
	    if { [lsearch -exact $old_cuts $cut] == -1 } {
		lappend cmd $cut yes
	    }
	}
	lappend cmd escape
	eval $cmd
	
	################################################################################
	#    Delete surface sets
	################################################################################
	
	set cmd [list GiD_Process escape escape escape escape \
		Utilities Delete SurfaceSets]
	foreach set $newsets {
	    lappend cmd $set yes
	}
	lappend cmd escape
	eval $cmd
	
	################################################################################
	#    Activate created surface sets and rename
	################################################################################
	
	set allsets [GiD_Info postprocess get all_surfacesets]
	set newsets ""
	foreach set [GiD_Info postprocess get all_surfacesets] {
	    if { [lsearch -exact $old_surfacesets $set] == -1 } {
		set idx ""
		while { [lsearch -exact $allsets S_border_graph$idx] != -1 } {
		    if { $idx eq "" } { set idx 2 } else { incr idx }
		}
		set newname border_graph$idx
		GiD_Process escape escape escape escape Utilities ChangeName \
		    SurfaceSets $set $newname
		lappend newsets S_$newname
		lappend active_surfs S_$newname
	    }
	}
	
	################################################################################
	#    Leave activated sets as they were
	################################################################################
	
	set active_vols_new [GiD_Info postprocess get cur_volumesets]
	set active_surfs_new [GiD_Info postprocess get cur_surfacesets]
	
	set cmd [list GiD_Process escape escape escape escape select \
		VolumeSets]
	foreach vol [GiD_Info postprocess get all_volumesets] {
	    if { [lsearch -exact $active_vols_new $vol] == -1 && \
		[lsearch -exact $active_vols $vol] != -1} {
		lappend cmd $vol
	    }
	}
	eval $cmd
	
	set cmd [list GiD_Process escape escape escape escape select \
		SurfaceSets]
	foreach surf [GiD_Info postprocess get all_surfacesets] {
	    if { [lsearch -exact $active_surfs_new $surf] == -1 && \
		[lsearch -exact $active_surfs $surf] != -1} {
		lappend cmd $surf
	    }
	}
	lappend cmd escape
	eval $cmd
    } elseif { $old_surfacesets != "" } {
	# create cut
	GiD_Process escape escape escape escape DoCut CutPlane $p1 $p2
	# convert to new set 
	GiD_Process escape escape escape escape DoCut TurnIntoMeshes
	set newsets ""
	foreach set [GiD_Info postprocess get all_surfacesets] {
	    if { [lsearch -exact $old_surfacesets $set] == -1 } {
		lappend newsets $set
	    }
	}
	# delete cut 
	set cmd [list GiD_Process escape escape escape escape \
		Utilities Delete CutSets]
	foreach cut [GiD_Info postprocess get all_cutsets] {
	    if { [lsearch -exact $old_cuts $cut] == -1 } {
		lappend cmd $cut yes
	    }
	}
	lappend cmd escape
	eval $cmd     
	#set set to on (by default, when converting cut, are off!!)
	eval GiD_Process escape escape escape escape select SurfaceSets $newsets escape      
    } else {
	set newsets ""
    }
    ################################################################################
    #    bye bye
    ################################################################################
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    GiD_Redraw
    if { $newsets != "" } {
	regsub -all { } $ViewMenuPriv(BG_PostResultsAnal) {_} analysis
	GiD_Process escape escape escape escape results AnalysisSel \
	    $analysis $ViewMenuPriv(BG_PostResultsStep)
	eval [list GiD_Process escape escape escape escape results graphs \
		bordergraph $ViewMenuPriv(border_graph_axX) \
		$ViewMenuPriv(BG_PostResultsRes) $ViewMenuPriv(border_comp)] \
	    $newsets escape

	#delete new sets
	eval GiD_Process escape escape escape escape Utilities Delete SurfaceSets $newsets yes escape
    } else {
	GidUtils::SetWarnLine [_ "It not exists any volume or surface set to create the graph"]
    }
}

# proc AddComponentsToMenu { m2 frame } {
#     global ViewMenuPriv
#     
#     set res [GiD_Info postprocess get cur_results_list Border_Graph]   
#     if { "$res" != "" } {
#         set reslabels $res
#     } else {
#         set res {"-No results-"}
#         set reslabels [_ "-No results-"]
#     }
#     if { [lsearch -exact $res $ViewMenuPriv(border_graph_axY)] == -1 } {
#         ComboBoxRAM $m2.m2 ViewMenuPriv(border_graph_axY) $reslabels $res
#         set ViewMenuPriv(border_graph_axY) [lindex $res 0]
#     }
#     if { $ViewMenuPriv(border_graph_axY) == "-No results-" || \
#         $ViewMenuPriv(border_graph_axY) == "" } {
#         return
#     }
#     foreach i [winfo children $frame] {
#         catch { destroy $i }
#     }
#     set irow 1
#     set componentes_txt [GiD_Info postprocess get cur_components_list \
#             $ViewMenuPriv(border_graph_axY)]
#     foreach i $componentes_txt {
#         set ww [ string tolower $i]$irow
#         regsub -all {\.} $ww · w
#         radiobutton $frame.r$w -text $i -value $i -variable \
#             ViewMenuPriv(border_comp)
#         grid $frame.r$w -sticky nw -row $irow -column 1
#         incr irow
#     }
#     grid rowconf $frame $irow -weight 1
#     grid columnconf $frame 1 -weight 1
#     if { [ llength $componentes_txt] != 4} {
#         set ViewMenuPriv(border_comp) [ lindex $componentes_txt 0]
#     } else {
#         set ViewMenuPriv(border_comp) [ lindex $componentes_txt 3]
#     }
# }

# rename wm wm_old
# proc wm { args } {
#     set ret [eval [concat wm_old $args]]
#     if { [lindex $args 1] == ".wlay" } {
#         if { ![winfo exists .tt.text] } {
#             toplevel .tt
#             pack [text .tt.text]
#         }
#         .tt.text ins end "wm args=//$args// return value=$ret\n"
#     }
#     return $ret
# }

proc PostContLimitsIncr { } {
    global GidPriv
    
    set GidPriv(PostContLimitsUserMin) [ format "%g" [ expr $GidPriv(PostContLimitsUserMin) - $GidPriv(PostContLimitsUserDiv)]]
    set GidPriv(PostContLimitsUserMax) [ format "%g" [ expr $GidPriv(PostContLimitsUserMax) + $GidPriv(PostContLimitsUserDiv)]]
}

proc PostContLimitsDecr { } {
    global GidPriv
    
    if { [ expr $GidPriv(PostContLimitsUserMax) - $GidPriv(PostContLimitsUserMin)] > \
	    [ expr 2.0 * $GidPriv(PostContLimitsUserDiv)] } {
	set GidPriv(PostContLimitsUserMin) [ format "%g" [ expr $GidPriv(PostContLimitsUserMin) + $GidPriv(PostContLimitsUserDiv)]]
	set GidPriv(PostContLimitsUserMax) [ format "%g" [ expr $GidPriv(PostContLimitsUserMax) - $GidPriv(PostContLimitsUserDiv)]]
    }
}

proc PostContLimitsReset { } {
    global GidPriv

    set GidPriv(PostContLimitsUserMin) [ format "%g" [ lindex $GidPriv(PostContLimitsOrgLimits) 0]]
    set GidPriv(PostContLimitsUserMax) [ format "%g" [ lindex $GidPriv(PostContLimitsOrgLimits) 1]]
}

proc PostContLimitsInitVariables {} {
    global GidPriv

    if { ![ info exist GidPriv(PostContLimitsApplying)]} {
      set GidPriv(PostContLimitsApplying) 0
    }

    if { !$GidPriv(PostContLimitsApplying)} {
	set GidPriv(PostContLimitsOrgLimits) [GiD_Info postprocess get cur_contour_limits]
	set GidPriv(PostContLimitsUserMin) [ format "%g" [ lindex $GidPriv(PostContLimitsOrgLimits) 0]]
	set GidPriv(PostContLimitsUserMax) [ format "%g" [ lindex $GidPriv(PostContLimitsOrgLimits) 1]]
	set GidPriv(PostContLimitsUserDiv) [ expr ( $GidPriv(PostContLimitsUserMax) - \
		$GidPriv(PostContLimitsUserMin)) * 0.05]

	set pp [ lindex [GiD_Info postprocess get contour_limits] 0]
	set GidPriv(PostContLimitsUseMax) 0
	set GidPriv(PostContLimitsUseMin) 0
	if { "[ lindex $pp 0]" == "USER" } {
	    set GidPriv(PostContLimitsUseMin) 1
	    if { [ info exist GidPriv(PostContLimitsWindow)] && \
		    [ winfo exist $GidPriv(PostContLimitsWindow).f.emin] } {
		$GidPriv(PostContLimitsWindow).f.emin conf -state normal -foreground black
	    }
	}
	if { "[ lindex $pp 2]" == "USER" } {
	    set GidPriv(PostContLimitsUseMax) 1
	    if { [ info exist GidPriv(PostContLimitsWindow)] && \
		    [ winfo exist $GidPriv(PostContLimitsWindow).f.emax] } {
		$GidPriv(PostContLimitsWindow).f.emax conf -state normal -foreground black
	    }
	}
    }
}

proc PostContLimitsApply {} { 
    global GidPriv

    set cur_res [GiD_Info postprocess get cur_result]
    set cur_res_view  [GiD_Info postprocess get cur_results_view]
    set cur_comp [GiD_Info postprocess get cur_component]
    regsub -all {_} $cur_res_view {} tipo_contour
    
    if { ( "$cur_res_view" == "Contour_Fill") || ( "$cur_res_view" == "Contour_Lines") || \
	( "$cur_res_view" == "Smooth_Contour_Fill") } {
	set GidPriv(PostContLimitsApplying) 1
	.central.s process escape escape escape escape results \
		contoptions
	if { $GidPriv(PostContLimitsUseMin) } { 
	    .central.s process setminoptions setvalue $GidPriv(PostContLimitsUserMin)
	} else { 
	    .central.s process setminoptions resetvalue escape
	}

	if { $GidPriv(PostContLimitsUseMax) } { 
	    .central.s process setmaxoptions setvalue $GidPriv(PostContLimitsUserMax)
	} else {
	    .central.s process setmaxoptions resetvalue escape
	}
	.central.s process escape $tipo_contour $cur_res $cur_comp
	#.central.s process escape escape escape escape results \
	#         contoptions setminoptions resetvalue escape \
	#         setmaxoptions resetvalue escape
	set GidPriv(PostContLimitsApplying) 0

	#.central.s process escape escape escape escape Utilities Redisplay
    }
}

proc PostContLimitsCloseWin { w } {

    set GidPriv(PostContLimitsHistoryMin) [ [ $w.f.emin cget list] get 0 end]
    set GidPriv(PostContLimitsHistoryMax) [ [ $w.f.emax cget list] get 0 end]
    destroy $w
}
    
proc PostContLimitsWin { { w .gid.wPostContLim} { solo 1}} {
    global GidPriv

    if { $solo} {
	InitWindow $w [_ "Contour Limits"] PostContLimitsWindowGeom PostContLimitsWin
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
    }

    if { $solo} {
	set GidPriv(PostContLimitsWindow) $w
    }

    #if { [ winfo exists .postContLim] } {
    #         raise .postContLim
    #         return
    #}
    #set donde [ wm geometry $w]
    #set nodecor 1
    #if { ![ regexp {([0-9]*)x([0-9]*)[\+,-]([0-9]*)[\+,-]([0-9]*)} $donde dummy w h x y] } {
    #         set nodecor 0
    #}

    #toplevel $w 
    #wm geometry $w +[expr $x + 50]+[expr $y + 50]

    frame $w.f -relief ridge -bd 2
    #label $w.f.lmax -text Max:
    checkbutton $w.f.lmax -text [_ "Max:"] -variable GidPriv(PostContLimitsUseMax) -command {
	if { $GidPriv(PostContLimitsUseMax)} {
	    $GidPriv(PostContLimitsWindow).f.emax conf -state normal -foreground black
	} else {
	    $GidPriv(PostContLimitsWindow).f.emax conf -state disabled -foreground grey60
	}
    }
    #entry $w.f.emax -width 11 -bd 1 -textvariable GidPriv(PostContLimitsUserMax)
    if { ![info exist GidPriv(PostContLimitsHistoryMax)] } {
	set GidPriv(PostContLimitsHistoryMax) ""
    }
    combobox $w.f.emax -width 11 -bd 1 -textvariable GidPriv(PostContLimitsUserMax) \
	    -editable 1 -keephistory 1 -history "$GidPriv(PostContLimitsHistoryMax)"
    #label $w.f.lmin -text [_ "Min:"]
    checkbutton $w.f.lmin -text [_ "Min:"] -variable GidPriv(PostContLimitsUseMin) -command {
	if { $GidPriv(PostContLimitsUseMin)} {
	    $GidPriv(PostContLimitsWindow).f.emin conf -state normal -foreground black
	} else {
	    $GidPriv(PostContLimitsWindow).f.emin conf -state disabled -foreground grey60
	}
    }
    #entry $w.f.emin -width 11 -bd 1 -textvariable GidPriv(PostContLimitsUserMin)
    if { ![info exist GidPriv(PostContLimitsHistoryMin)] } {
	set GidPriv(PostContLimitsHistoryMin) ""
    }
    combobox $w.f.emin -width 11 -bd 1 -textvariable GidPriv(PostContLimitsUserMin) \
	    -editable 1 -keephistory 1 -history "$GidPriv(PostContLimitsHistoryMin)"
    
    frame $w.f.fi
    button $w.f.fi.bmas -text + -bd 1 -padx 2 -pady 0 -highlightthickness 0 \
	    -command PostContLimitsIncr
    button $w.f.fi.breset -text o -bd 1 -padx 2 -pady 0 -highlightthickness 0 \
	    -command PostContLimitsReset 
    button $w.f.fi.bmenos -text - -bd 1 -padx 2 -pady 0 -highlightthickness 0 \
	    -command PostContLimitsDecr
    grid $w.f.fi.bmas -row 0 -column 0 -padx 3 -sticky news
    grid $w.f.fi.breset -row 1 -column 0 -padx 3 -sticky news
    grid $w.f.fi.bmenos -row 2 -column 0 -padx 3 -sticky news

    if { $solo} {
	button $w.f.bapply -text [_ "Apply"] -bd 1 -command PostContLimitsApply
	button $w.f.bclose -text [_ "Close"] -bd 1 -command "PostContLimitsCloseWin $w"
	SetWidgetsWidthFromText $w.f.bapply $w.f.bclose
    }

    grid $w.f.lmax $w.f.emax 
    grid $w.f.lmin $w.f.emin
    grid $w.f.fi -row 0 -column 2 -rowspan 2

    if { $solo} {
	grid $w.f.bapply -row 0 -column 3 -padx 4 -pady 2
	grid $w.f.bclose -row 1 -column 3 -padx 4 -pady 2
    }

    grid $w.f

    set GidPriv(PostContLimitsApplying) 0
    PostContLimitsInitVariables 
    $w.f.emax conf -state disabled -foreground grey60
    $w.f.emin conf -state disabled -foreground grey60

    if { $solo} {
	bind $w <Return> PostContLimitsApply
    }

}

proc PostContLimitsApplyMinEntry { w } {
    global GidPriv
    set err 0 
    
    if { "$GidPriv(PostContLimitsUserMin)" != "" } {
	set err [ catch { format "%g" $GidPriv(PostContLimitsUserMin) } ]
	if { !$err} {
	    set cur_res [GiD_Info postprocess get cur_result]
	    set cur_res_view  [GiD_Info postprocess get cur_results_view]
	    set cur_comp [GiD_Info postprocess get cur_component]
	    regsub -all {_} $cur_res_view {} tipo_contour
	    
	    if { ( "$cur_res_view" == "Contour_Fill") || ( "$cur_res_view" == "Contour_Lines") || \
		    ( "$cur_res_view" == "Smooth_Contour_Fill") } {
		#set GidPriv(PostContLimitsApplying) 1
		.central.s process escape escape escape escape results \
		        contoptions setminoptions setvalue $GidPriv(PostContLimitsUserMin) 
		#        escape $tipo_contour $cur_res $cur_comp
		#.central.s process escape escape escape escape results \
		#         contoptions setminoptions resetvalue escape 
		.central.s process escape escape escape escape Utilities Redisplay \
		        escape escape
		#PostContLimitsInitVariables
		#set GidPriv(PostContLimitsApplying) 0
	    }
	} else {
	    WarnWin [_ "A Numerical value must be entered."]
	}
    }
    
    #WarnWin "$GidPriv(PostContLimitsUserMin) - $err"        
    if { !$err} {
	set GidPriv(PostContLimitsHistoryMin) [ [ $w.f.emin cget list] get 0 end]
	destroy $w
	#focus .gid
    }
}

proc PostContLimitsMinEntry { w} {
    global GidPriv tcl_platform

    #set donde [ wm geometry $w]
    #set donde [ winfo geometry $w]
    #set nodecor 1
    #if { ![ regexp {([0-9]*)x([0-9]*)[\+,-]([0-9]*)[\+,-]([0-9]*)} $donde dummy w h x y] } {
    #          set nodecor 0
    #}
    set x [ winfo x $w]
    set y [ winfo y $w]
    set wi [ winfo width $w]
    set w [ winfo parent $w]
    while { "$w" != "" } {
	set xp [ winfo x $w]
	set yp [ winfo y $w]
	set x [ expr $x + $xp]
	set y [ expr $y + $yp + 16]
	if { "[ winfo class $w]" == "Toplevel"} {
	    break
	}
	set w [ winfo parent $w]
    }
    set x [expr $x + $wi]
    set y [expr $y - 16]
    #WarnWin "$w = +$x+$y"

    toplevel .wPostLimMin
    wm geometry .wPostLimMin +$x+$y
    wm title .wPostLimMin [_ "Minimum"]
    wm minsize .wPostLimMin 1 1
    wm overrideredirect .wPostLimMin 1

    frame .wPostLimMin.f -bd 2 -relief ridge
    if { ![info exist GidPriv(PostContLimitsHistoryMin)] } {
	set GidPriv(PostContLimitsHistoryMin) ""
    }
    combobox .wPostLimMin.f.emin -width 11 -bd 1 -textvariable GidPriv(PostContLimitsUserMin) \
	    -editable 1 -keephistory 1 -history "$GidPriv(PostContLimitsHistoryMin)"

    set limites_org [GiD_Info postprocess get cur_contour_limits]
    set GidPriv(PostContLimitsUserMin) [ format "%g" [ lindex $limites_org 0]]
    .wPostLimMin.f.emin sel to end

    bind .wPostLimMin <Return> "PostContLimitsApplyMinEntry .wPostLimMin"
    bind .wPostLimMin <Escape> "destroy .wPostLimMin"
    bind .wPostLimMin <1> "+PostContLimitsBut1 %W %x %y"

    if { $tcl_platform(platform) == "windows" } {
	button .wPostLimMin.f.b -text X -command "destroy .wPostLimMin" -padx 2 -bd 1 \
		-pady 0 -highlightthickness 0
    } else {
	button .wPostLimMin.f.b -text X -command "destroy .wPostLimMin" -padx 4 -bd 1 -pady 2
    }

    #grid .wPostLimMin.f.emin .wPostLimMin.f.b
    grid .wPostLimMin.f.emin -padx 2 -pady 1
    grid .wPostLimMin.f.b -column 1 -row 0 -padx 1 -pady 1

    grid .wPostLimMin.f
    focus -force .wPostLimMin.f.emin
    #raise .wPostLimMin .gid
    update
    grab .wPostLimMin

}

proc PostContLimitsApplyMaxEntry { w }  {
    global GidPriv
    set err 0

    if { "$GidPriv(PostContLimitsUserMax)" != "" } {
	set err [ catch { format "%g" $GidPriv(PostContLimitsUserMax) } ]
	if { !$err} {
	    set cur_res [GiD_Info postprocess get cur_result]
	    set cur_res_view  [GiD_Info postprocess get cur_results_view]
	    set cur_comp [GiD_Info postprocess get cur_component]
	    regsub -all {_} $cur_res_view {} tipo_contour
	    
	    if { ( "$cur_res_view" == "Contour_Fill") || ( "$cur_res_view" == "Contour_Lines") || \
		    ( "$cur_res_view" == "Smooth_Contour_Fill") } {
		#set GidPriv(PostContLimitsApplying) 1
		#WarnWin "Set value"
		.central.s process escape escape escape escape results \
		        contoptions setmaxoptions setvalue $GidPriv(PostContLimitsUserMax)
		#        escape $tipo_contour $cur_res $cur_comp
		#.central.s process escape escape escape escape results \
		#         contoptions setmaxoptions resetvalue escape 
		#WarnWin "Redisplay"
		.central.s process escape escape escape escape Utilities Redisplay \
		        escape escape
		#WarnWin "init vars"
		# PostAnimateInitVariables
		# PostContLimitsInitVariables
		# FillPostLegendInfo
		#set GidPriv(PostContLimitsApplying) 0
	    }
	} else {
	    WarnWin [_ "A Numerical value must be entered."]
	}
    }

    #WarnWin "$GidPriv(PostContLimitsUserMax) - $err"
    if { !$err} {
	set GidPriv(PostContLimitsHistoryMax) [ [ $w.f.emax cget list] get 0 end]
	destroy $w
    }
    #focus .gid
}


proc PostContLimitsMaxEntry { w } {
    global GidPriv tcl_platform

    #set donde [ wm geometry $w]
    #set donde [ winfo geometry $w]
    #set nodecor 1
    #if { ![ regexp {([0-9]*)x([0-9]*)[\+,-]([0-9]*)[\+,-]([0-9]*)} $donde dummy w h x y] } {
    #          set nodecor 0
    #}
    set x [ winfo x $w]
    set y [ winfo y $w]
    set wi [ winfo width $w]
    set w [ winfo parent $w]
    while { "$w" != "" } {
	set xp [ winfo x $w]
	set yp [ winfo y $w]
	set x [ expr $x + $xp]
	set y [ expr $y + $yp + 16]
	if { "[ winfo class $w]" == "Toplevel"} {
	    break
	}
	set w [ winfo parent $w]
    }
    set x [expr $x + $wi]
    set y [expr $y - 16]
    #WarnWin "$w = +$x+$y"

    toplevel .wPostLimMax
    wm geometry .wPostLimMax +$x+$y
    wm title .wPostLimMax [_ "Maximum"]
    wm minsize .wPostLimMax 1 1
    wm overrideredirect .wPostLimMax 1

    frame .wPostLimMax.f -bd 2 -relief ridge
    if { ![info exist GidPriv(PostContLimitsHistoryMax)] } {
	set GidPriv(PostContLimitsHistoryMax) ""
    }   
    combobox .wPostLimMax.f.emax -width 11 -bd 1 -textvariable GidPriv(PostContLimitsUserMax) \
	    -editable 1 -keephistory 1 -history "$GidPriv(PostContLimitsHistoryMax)"

    set limites_org [GiD_Info postprocess get cur_contour_limits]
    set GidPriv(PostContLimitsUserMax) [ format "%g" [ lindex $limites_org 1]]
    .wPostLimMax.f.emax sel to end

    bind .wPostLimMax <Return> "PostContLimitsApplyMaxEntry .wPostLimMax"
    bind .wPostLimMax <Escape> "destroy .wPostLimMax"
    bind .wPostLimMax <1> "+PostContLimitsBut1 %W %x %y"

    if { $tcl_platform(platform) == "windows" } {
	button .wPostLimMax.f.b -text X -command "destroy .wPostLimMax" -padx 2 -bd 1 \
		-pady 0 -highlightthickness 0
    } else {
	button .wPostLimMax.f.b -text X -command "destroy .wPostLimMax" -padx 4 -bd 1 -pady 2
    }

    grid .wPostLimMax.f.emax -padx 2 -pady 1
    grid .wPostLimMax.f.b -column 1 -row 0 -padx 1 -pady 1
    grid .wPostLimMax.f
    focus -force .wPostLimMax.f.emax
    #raise .wPostLimMax .gid
    update
    grab .wPostLimMax

}

proc PostContLimitsBut1 { w x y } {
    if { $x < 0 || $y < 0 || $x > [winfo width $w] || $y > [winfo height $w] } {
	destroy $w
    }
}

proc PostOutMinMaxColourMenu { w que } {
    global GidPriv

    set actual_fuera [GiD_Info postprocess get results_options]
    set que [ string toupper [ string index $que 0]][ string tolower [ string range $que 1 end]]
    if { "$que" == "Min"} {
	set curr [ lindex $actual_fuera 3]
	set names "Black White MinColor Transparent Material"
	set namesLabel [ list [_ "Black"] [_ "White"] [_ "Min Color"] [_ "Transparent"] [_ "Material"]]
    } else {
	set curr [ lindex $actual_fuera 4]
	set names "Black White MaxColor Transparent Material"
	set namesLabel [ list [_ "Black"] [_ "White"] [_ "Max Color"] [_ "Transparent"] [_ "Material"]]
    }    
    

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results ContOptions Set${que}Options Out${que}Color $i \
	#             escape escape escape escape Utilities Redisplay escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results ContOptions Set${que}Options Out${que}Color $i \
		escape escape escape escape Utilities Redisplay escape"
    }
}

proc PostSmoothContourTypeMenu { w } {
    global GidPriv

    set opciones [GiD_Info postprocess get results_options]
    set curr [ lindex $opciones 8]
    
    set names "MinimumSmooth MaximumSmooth MeanSmooth"
    set namesLabel [ list [_ "Minimum value"] [_ "Maximum value"] [_ "Mean value"]]

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results ContOptions SmoothingType $i escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results ContOptions SmoothingType $i escape"
    }
}

proc PostDiagramOptionsMenu { w } {
    global GidPriv

    set opciones [GiD_Info postprocess get diagram_options]
    set curr [ lindex $opciones 0]
    
    set names "None NodesOnly WholeLine FilledLine ContFillLine"
    set namesLabel [ list [_ "None"] [_ "Nodes only"] [_ "Whole line"] \
	    [_ "Filled line"] [_ "Contour filled"] ]

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results LineDiagram Options ShowElevations $i escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results LineDiagram Options ShowElevations $i escape"
    }
}

proc PostResultSurfaceOptionsMenu { w } {
    global GidPriv

    set opciones [GiD_Info postprocess get result_surface_options]
    # nos devuelve: {ExtrudedEdges [ContourFill]} {None ExtrudedNodes ExtrudedEdges SwitchContFill}
    set actual [ lindex $opciones 0]
    set curr [ lindex $actual 0]
    set hay_cfill 0
    if { [ llength $actual] != 1} {
	set hay_cfill 1
    }

    set names "None ExtrudedNodes ExtrudedEdges SwitchContFill"
    set namesLabel [ list [_ "None"] [_ "Extruded nodes"] [_ "Extruded edges"] [_ "Contour Fill"]]

    $w del 0 end
    foreach i $names j $namesLabel {
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	if { $hay_cfill && ( "$i" == "SwitchContFill") } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results Options ResSurfShowElev $i escape"
    }
}

proc PostStreamLinesColorOptionsMenu { w } {
    global GidPriv

    set opciones [GiD_Info postprocess get result_stream_lines_options]
    # nos devuelve: {4SidedPrism} {Points Ribbons 4SidedPrism 8SidedPrism} {Monochrome} {Monochrome StreamContourFilled ResultContourFilled} 

    # primero tipo:
    set curr [ lindex $opciones 0]
    
    $w del 0 end
    
    if { 0} {
	#set names "Points Ribbons 4SidedPrism 8SidedPrism"
	#set namesLabel [ list [_ "Points"] [_ "Ribbons"] [_ "4SidedPrism"] [_ "8SidedPrism"]]
	set names "Points 4SidedPrism"
	set namesLabel [ list [_ "Points"] [_ "4SidedPrism"]]
	
	foreach i $names j $namesLabel {
	    set img $GidPriv(VerifNoImage)
	    if { $i == $curr } {
		set img $GidPriv(VerifImage)
	    }
	    $w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results Options StreamColor $i escape"
	}
	$w add separator
    }

    set curr [ lindex $opciones 2]

    set names "Monochrome StreamContourFilled ResultContourFilled"
    set namesLabel [ list [_ "Monochrome"] [_ "Stream contour filled"] [_ "Result contour filled"]]

    foreach i $names j $namesLabel {
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results Options StreamColor $i escape"
    }
}

proc PostVectorsDetailMenu { w } {
    global GidPriv

    set opciones [GiD_Info postprocess get vector_detail]
    set curr [ lindex $opciones 0]
    
    set names "NoArrow Points Lines 2Triangles 4Triangles 8Triangles"
    set namesLabel [ list [_ "No arrow head"] [_ "Point"] [_ "Lines"] \
	    [_ "2 Triangles"] [_ "4 Triangles"] [_ "8 Triangles"] ]

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results Options VectorDetail $i escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results Options VectorDetail $i escape"
    }
}

proc PostVectorsSizeMenu { w } {
    global GidPriv

    # ramsan: problems mixing versions
    set opciones [list "" ""]
    catch { set opciones [GiD_Info postprocess get vector_size] }
    set curr [ lindex $opciones 0]
    
    set names [lindex $opciones 1]
#"FixedSize ModulusDependent"
    set namesLabel [ list [_ "Fixed size"] [_ "Modulus dependent"]]

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results Options VectorDetail $i escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results Options VectorSize $i escape"
    }
}

proc IconoIsoAlwaysDrawn {} {
    global GidPriv
    set opciones [ GiD_Info postprocess get isosurface_options]
    set lista [ lindex $opciones 8]
    set nombre [ lindex $lista 0]
    set valor [ lindex $lista 1]

    set img $GidPriv(VerifNoImage)
    if { $valor} {
	set img $GidPriv(VerifImage)    
    }
    return $img
}

proc PostIsoSurfaceAlwaysDrawn { w idx} {
    set opciones [ GiD_Info postprocess get isosurface_options]
    set lista [ lindex $opciones 8]
    set nombre [ lindex $lista 0]
    set valor [ lindex $lista 1]

    set nuevo_valor Yes
    if { $valor} {
	set nuevo_valor No
    }
    GiD_Process escape escape escape escape Results isosurfaces AlwaysDrawn $nuevo_valor escape
    
    $w entryconf $idx -image [ IconoIsoAlwaysDrawn] -compound left
}

proc IconoIsoDrawIsolineas {} {
    global GidPriv
    set opciones [ GiD_Info postprocess get isosurface_options]
    set lista [ lindex $opciones 9]
    set nombre [ lindex $lista 0]
    set valor [ lindex $lista 1]

    set img $GidPriv(VerifNoImage)
    if { $valor} {
	set img $GidPriv(VerifImage)    
    }
    return $img
}

proc PostIsoSurfaceDrawIsolines { w idx} {
    set opciones [ GiD_Info postprocess get isosurface_options]
    set lista [ lindex $opciones 9]
    set nombre [ lindex $lista 0]
    set valor [ lindex $lista 1]

    set nuevo_valor Yes
    if { $valor} {
	set nuevo_valor No
    }
    GiD_Process escape escape escape escape Results isosurfaces $nombre $nuevo_valor escape
    
    $w entryconf $idx -image [ IconoIsoDrawIsolineas] -compound left
}

proc PostIsoSurfaceOptionsMenu { w que} {
    global GidPriv

    set opciones [ GiD_Info postprocess get isosurface_options]
    set curr ""
    set names ""
    set namesLabel ""
    switch "[ string tolower $que]" {
	"displaystyle" {
	    set curr [ lindex $opciones 0]
	    set names [ lindex $opciones 1]
	    set namesLabel [ list [_ "All Lines"] [_ "Hidden Lines"] [_ "Body"] [_ "Body Lines"] ]
	}
	"render" {
	    set curr [ lindex $opciones 2]
	    set names [ lindex $opciones 3]
	    set namesLabel [ list [_ "Normal"] [_ "Flat"] ]
	}
	"transparency" {
	    set curr [ lindex $opciones 4]
	    set names [ lindex $opciones 5]
	    set namesLabel [ list [_ "Opaque"] [_ "Transparent"] ]
	}
	"color" {
	    set curr [ lindex $opciones 6]
	    set names [ lindex $opciones 7]
	    set namesLabel [ list [_ "Monochrome"] [_ "Result color"] [_ "Contour fill color"]]
	}
    }

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results isosurfaces DisplayStyle $i escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results isosurfaces DisplayStyle $i escape"
    }
}

proc PostIsosurfaceMenuPrincipal { w } {
    set MenuEntriesP(5,8) [list [_ "Display Style#C#menu"] [_ "Transparency#C#menu"] \
			       [_ "Color Mode#C#menu"] [_ "Change color#C#menu"] \
			       [_ "Convert to cuts#C#menu"] [_ "Draw always#C#menu"] \
			       [_ "Show isolines#C#menu"]]
    set MenuCommandsP(5,8) [ list  \
				 [ list PostIsoSurfaceOptionsMenu %W DisplayStyle] \
				 [ list PostIsoSurfaceOptionsMenu %W Transparency] \
				 [ list PostIsoSurfaceOptionsMenu %W Color] \
				 [ list GiD_Process escape escape escape escape Results isosurfaces DisplayStyle ChangeMonoColour] \
				 [ list GiD_Process escape escape escape escape Results isosurfaces TurnIntoCuts] \
				 [ list PostIsoSurfaceAlwaysDrawn %W 5] \
				 [ list PostIsoSurfaceDrawIsolines %W 6]]

    set MenuEntriesP(5,8,0) ""
    set MenuCommandsP(5,8,0) ""
    
    set MenuEntriesP(5,8,1) ""
    set MenuCommandsP(5,8,1) ""
    
    set MenuEntriesP(5,8,2) ""
    set MenuCommandsP(5,8,2) ""

    set lst_img [ list $::GidPriv(VerifNoImage) $::GidPriv(VerifNoImage) \
		      $::GidPriv(VerifNoImage) $::GidPriv(VerifNoImage) \
		      $::GidPriv(VerifNoImage) \
		      [ IconoIsoAlwaysDrawn] [ IconoIsoDrawIsolineas]]    

    $w del 0 end
    set idx 0
    foreach entrada $MenuEntriesP(5,8) comando $MenuCommandsP(5,8) img $lst_img {
	if { [ info exists MenuEntriesP(5,8,$idx)]} {
	    set m $w.mm$idx
	    if { ![ winfo exist $w.mm$idx]} { 
		menu $m
	    }
	    $w add cascade -label $entrada -image $img -compound left \
		-menu $m
	    regsub -all {%W} $comando $m cmd
	    $m conf -postcommand $cmd
	} else {
	    regsub -all {%W} $comando $w cmd
	    $w add command -label $entrada -image $img -compound left \
		-command $cmd
	}
	incr idx
    }
}

proc PostContourScaleOptionsMenu { w que } {
    global GidPriv

    set opciones [GiD_Info postprocess get cur_contour_color_config]
    set curr ""
    set names ""
    set namesLabel ""
    switch "[ string tolower $que]" {
	"scaleramp" {
	    set curr [ lindex $opciones 5]
	    set names "Tangent Linear ArcTang"
	    set namesLabel [ list [_ "Tangent"] [_ "Linear"] [_ "ArcTangent"] ]
	}
	"scaletype" {
	    set curr [ lindex $opciones 4]
	    set names "RGB HSV"
	    set namesLabel [ list [_ "RGB"] [_ "HSV"] ]
	}
    }

    $w del 0 end
    foreach i $names j $namesLabel {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command ".central.s process \
	#             escape escape escape escape \
	#             Results ContOptions ColorRamp $i \
	#             escape escape escape escape Utilities Redisplay escape"
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command ".central.s process \
		escape escape escape escape \
		Results ContOptions ColorRamp $i \
		escape escape escape escape Utilities Redisplay escape"
    }
}

proc PostContourWhichScaleMenu { w} {
    global GidPriv

    set opciones [GiD_Info postprocess get cur_contour_color_config]
    set curr [ lindex $opciones 1]
    # para que no aparezca un mensaje nuevo en la bd a traducir
    set bw [_ "White"]
    append bw " " [_ "Black"]
    set names "Rainbow {Inv. rainbow} Terrain {Black & White} { White & Black} {User defined}"
    set namesLabel  [list [_ "Standard"] [_ "Inverse Standard"] [_ "Terrain Map"] \
		         [_ "Black White"] $bw \
	    [_ "User defined"]... ]
    set cmds_lst { \
	    { .central.s process \
	    escape escape escape escape Results ContOptions SetMinOptions MinColor Standard \
	    escape escape escape escape Results ContOptions SetMaxOptions MaxColor Standard\
	    escape escape escape escape Results ContOptions ColorRamp Tangent \
	    escape escape escape escape Utilities Redisplay escape } \
	    { .central.s process \
	    escape escape escape escape Results ContOptions SetMinOptions MinColor InverseStd \
	    escape escape escape escape Results ContOptions SetMaxOptions MaxColor InverseStd \
	    escape escape escape escape Results ContOptions ColorRamp Tangent\
	    escape escape escape escape Utilities Redisplay escape } \
	    { .central.s process \
	    escape escape escape escape Results ContOptions ColorRamp TerrainMap \
	    escape escape escape escape Utilities Redisplay escape } \
	    { .central.s process \
	    escape escape escape escape Results ContOptions SetMinOptions MinColor Black \
	    escape escape escape escape Results ContOptions SetMaxOptions MaxColor White \
	    escape escape escape escape Results ContOptions ColorRamp Tangent \
	    escape escape escape escape Utilities Redisplay escape } \
	    { .central.s process \
	    escape escape escape escape Results ContOptions SetMinOptions MinColor White \
	    escape escape escape escape Results ContOptions SetMaxOptions MaxColor Black \
	    escape escape escape escape Results ContOptions ColorRamp Tangent \
	    escape escape escape escape Utilities Redisplay escape } \
		       PostContourColors  }
    
    $w del 0 end
    foreach i $names j $namesLabel  cmd $cmds_lst {
	#if { $i == $curr } {
	#    $w add check -label $j -var $w
	#    set ::$w 1
	#} else {
	#    $w add command -label $j -command $cmd
	#}
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		-command $cmd
    }

    set names "ScaleRamp ScaleType"
    set namesLabel [ list  [_ "Scale Ramp"] [_ "Scale Type"]]

    foreach submenu $names label $namesLabel {
	set menuname [ join "$w . [ string tolower $submenu]" ""]
	# esto es para mantener el alineamiento
	set img $GidPriv(VerifNoImage)
	$w add cascade -label $label -image $img -compound left \
		-menu $menuname
	if { ![winfo exist $menuname] } {
	    menu $menuname -postcommand "PostContourScaleOptionsMenu $menuname $submenu"
	} else {
	    $menuname conf -postcommand "PostContourScaleOptionsMenu $menuname $submenu"
	}
    }
}

proc PostContourBrightColorMenu { w} {
    global GidPriv

    set opciones [GiD_Info postprocess get cur_contour_color_config]
    set curr [ lindex [ lindex $opciones 8] 0]
    # es del tipo {%s {No Metallic Plastic Rubber}}
    set names "No Metallic Plastic Rubber"
    set namesLabel  [list [_ "No"] [_ "Metallic"] [_ "Plastic"] [_ "Rubber"]]
    $w del 0 end
    foreach i $names j $namesLabel {
	set img $GidPriv(VerifNoImage)
	if { $i == $curr } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
	    -command "GiD_Process escape escape escape escape Results ContOptions BrightColors $i"
    }
}

proc PostLegendOptionsMenu { w} {
    global GidPriv

    set opciones [GiD_Info postprocess get legends_state]
    
    set namesLabel [list [_ "Show"] [_ "Opaque"] [_ "Show Title"] [_ "Draw User Limits"] [_ "Border"] \
	    [_ "Change Title"] [_ "Outside"] [_ "Automatic Comments"] ]
    set values_lst ""
    set cmds_lst ""

    # Show Legends ?
    set curr [ lindex $opciones 0]
    if { "$curr" == "Yes"} { 
	set val 1
	set cmd ".central.s process Hardcopy Options ShowLegends No"
    } else {
	set val 0
	set cmd ".central.s process Hardcopy Options ShowLegends Yes"
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    # Opaque Legends ?
    set curr [ lindex $opciones 1]
    if { "$curr" == "Opaque"} { 
	set val 1
	set cmd ".central.s process Hardcopy Options OpaqueLegends No"
    } else {
	set val 0
	set cmd ".central.s process Hardcopy Options OpaqueLegends Yes"
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    # Title showed?
    set curr [ lindex $opciones 2]
    if { "$curr" == "Title"} { 
	set val 1
	set cmd ".central.s process escape escape escape escape Results ContOptions ShowTitleLeg No escape"
    } else {
	set val 0
	set cmd ".central.s process escape escape escape escape Results ContOptions ShowTitleLeg Yes escape"
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    # User limits drawn as boxes??
    set curr [ lindex $opciones 3]
    if { "$curr" == "DrawUserLimits"} { 
	set val 1
	set cmd ".central.s process escape escape escape escape Results ContOptions DrawUserLimitsLeg No escape"
    } else {
	set val 0
	set cmd ".central.s process escape escape escape escape Results ContOptions DrawUserLimitsLeg Yes escape"
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    # draw border??
    set curr [ lindex $opciones 4]
    if { "$curr" == "DrawBorder"} { 
	set val 1
	set cmd ".central.s process Hardcopy Options DrawBorderLegends No"
    } else {
	set val 0
	set cmd ".central.s process Hardcopy Options DrawBorderLegends Yes"
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    # change title 
    set val 0
    set cmd ".central.s process escape escape escape escape Results ContOptions ChangeTitleLeg"
    lappend values_lst $val
    lappend cmds_lst $cmd

    # Outside?
    set val [ ExistPostLegend]
    if { $val} {
	set cmd PostLegendInside
    } else {
	set cmd PostLegend
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    # Automatic Comments
    set opciones [GiD_Info postprocess get comments]
    set curr [ lindex $opciones 1]
    if { "$curr" == "yes"} {
	set val 1
	set cmd ".central.s process escape escape escape escape Utilities Comments Automatic No escape"
    } else {
	set val 0
	set cmd ".central.s process escape escape escape escape Utilities Comments Automatic Yes escape"
    }
    lappend values_lst $val
    lappend cmds_lst $cmd

    $w del 0 end
    foreach j $namesLabel val $values_lst cmd $cmds_lst {
	#if { $val} {
	#    $w add check -label $j -var $w -command $cmd
	#    set ::$w $val
	#} else {
	#    $w add command -label $j -command $cmd
	#}
	set img $GidPriv(VerifNoImage)
	if { $val } {
	    set img $GidPriv(VerifImage)
	}
	$w add command -label $j -image $img -compound left \
		 -command $cmd
    }
}



proc ViewSeveralWindows {} {
    set w .gid.viewseveralwindows

    InitWindow $w [_ "Multiple windows"] PrePostViewMWindowsWindowGeom ViewSeveralWindows "" 1
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
    
    label $w.l -text [_ "Select the desired window configuration"] -font BigFont
    label $w.i -image [GetImage mwindows.gif] -relief ridge -bd 2
    $w.l configure -wraplength [image width [$w.i cget -image]]
    
    frame $w.but -bg [CCColorActivo [$w cget -background]]
    button $w.but.close -text [_ "Close"] -command "destroy $w" -und 0
    SetWidgetsWidthFromText $w.but.close

    grid $w.but.close -padx 5 -pady 5

    grid $w.l -sticky new
    grid $w.i -padx 3 -sticky nsew
    grid $w.but -sticky sew
    if { $::tcl_version >= 8.5 } { grid anchor $w.but center }
    grid rowconfigure $w "1" -weight 1
    grid columnconfigure $w "0" -weight 1

    focus $w.but.close
    bind $w.i <1> "ViewSeveralWindowsSelect $w %x %y"
    bind $w <Alt-c> "tkButtonInvoke $w.but.close"
    bind $w <Escape> "tkButtonInvoke $w.but.close"
}


proc ViewSeveralWindowsSelect { w x y } {
    set type ""
    if { $x > 16 && $x < 77 } {
	if { $y > 21 && $y < 77 } {
	    set type 1
	} elseif { $y > 102 && $y < 157 } {
	    set type 3L
	} elseif { $y > 185 && $y < 236 } {
	    set type 3D
	}
    } elseif { $x > 109 && $x < 169 } {
	if { $y > 21 && $y < 77 } {
	    set type 2LR
	} elseif { $y > 102 && $y < 157 } {
	    set type 3U
	} elseif { $y > 185 && $y < 236 } {
	    set type 4
	}
    } elseif { $x > 200 && $x < 262 } {
	if { $y > 21 && $y < 77 } {
	    set type 2UD
	} elseif { $y > 102 && $y < 157 } {
	    set type 3R
	} elseif { $y > 185 && $y < 236 } {
	    set type EXT
	}
    }
    if { $type != "" } {
	.central.s process NewWin $type
    }
}

proc CreatePointNewOld {} {
    global GidPriv
    if { ![info exists GidPriv(CreatePointNewOld)] } {
	set GidPriv(CreatePointNewOld) 1
    }
    set w .gid.createpointnewold
    InitWindow $w [_ "Create point procedure"] CreatePointNewOldWindowGeom CreatePointNewOld "" 1
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
    label $w.l -text [_ "Click the image to get the existing point or create a new one"] 
    frame $w.centre -relief ridge -bd 2
    label $w.centre.txt_NJ -text [_ "No join"] -font BigFont
    label $w.centre.txt_NJ2 -text [_ "New point is created"]
    label $w.centre.txt_J -text [_ "Join"] -font BigFont
    label $w.centre.txt_J2 -text [_ "Existing point is selected"]
    button $w.centre.nojoin -image [GetImage no_join.gif] -cursor hand2 -relief groove -overrelief raised \
	    -command "set GidPriv(CreatePointNewOld) 0 ; CreatePointNewOldExec $w 1"      
    button $w.centre.join -image [GetImage join.gif] -cursor hand2 -relief groove -overrelief raised \
	    -command "set GidPriv(CreatePointNewOld) 1 ; CreatePointNewOldExec $w 1"
    label $w.note_text -text [_ "NOTE: Ctrl-a toggles between Join and No join modes."]   
    set manual_link "gid_5.html#SEC9"
    button $w.help -text [_ "Help"] -command "directHelp $manual_link"
    grid $w.centre.txt_NJ $w.centre.txt_J
    grid $w.centre.nojoin $w.centre.join -padx 3
    grid $w.centre.txt_NJ2 $w.centre.txt_J2 
    grid $w.l -sticky new -columnspan 2 -padx 3 -pady 3
    grid $w.centre -sticky nsew -columnspan 2 -padx 3
    grid columnconfigure $w.centre "0 1" -weight 1
    grid $w.note_text -sticky sew -columnspan 2 -padx 3
    grid $w.help -sticky se -columnspan 2 -padx 3 -pady 3
    grid columnconfigure $w "0" -weight 1
    grid rowconfigure $w "1" -weight 1
    wm protocol $w WM_DELETE_WINDOW "CreatePointNewOldExec $w 0"
    bind $w <Escape> "CreatePointNewOldExec $w 0"
    focus $w 
    bind $w <Return> "set GidPriv(CreatePointNewOld) 1 ; CreatePointNewOldExec $w 1"
    update idletasks
}

proc CreatePointNewOldExec { w doit } {
    global GidPriv
    if { !$doit } {
	GiD_Process escape
    } elseif { $GidPriv(CreatePointNewOld) } {
	GiD_Process old
    } else {
	GiD_Process new
    }
    destroy $w
}


proc GeometryView { } {
    if { [catch { set whatuse [GiD_Info Project ViewMode] }] } {
	set whatuse "GEOMETRYUSE"
    }
    if { $whatuse != "GEOMETRYUSE" } {
	GiD_Process MEscape Geometry
    }
}

proc MeshView { { showmessage 1 } } {
    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "GEOMETRYUSE"
    }
    if { $infoProj == "GEOMETRYUSE" } {
	if { $showmessage && [lindex [GiD_Info mesh] 0] == "0" } {
	    WarnWin [_ "Sorry. No mesh"]
	} else {
	    GiD_Process Mescape Meshing MeshView
	}
    }
}

proc ToggleGeomMesh {} {
    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "GEOMETRYUSE"
    }
    if { $infoProj == "GEOMETRYUSE" } {
	if { [lindex [GiD_Info mesh] 0] != "0" } {
	    GiD_Process Mescape Meshing MeshView
	} else {
	    #WarnWin [_ "There is no mesh. Generate it?"]
	    GiD_Process Mescape Meshing Generate
	}
    } else {
	GiD_Process Mescape Geometry escape
    }
    set ::CurrentViewMode [GiD_Info Project ViewMode]
}

proc TogglePrePost {} {
    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "GEOMETRYUSE"
    }

    if { ( $infoProj == "GEOMETRYUSE") || ( $infoProj == "MESHUSE") } {
	GiD_Process Mescape Postprocess
    } else {
	GiD_Process Mescape  Preprocess
    }
}

proc TogglePostGraph {} {
    if { [catch { set infoProj [GiD_Info Project ViewMode] }] } {
	set infoProj "POSTUSE"
    }
    
    if { ( $infoProj == "POSTUSE") } {
	GiD_Process Mescape Results Graphs OptionsGraph ShowGraphs Yes
    } else {
	GiD_Process Mescape Results Graphs OptionsGraph ShowGraphs No 
    }
}

proc imagefrom { name } {
    return [GetImage $name]
}

proc bitmapfrom { name } {
    global GIDDEFAULT
    return @[file join $GIDDEFAULT bitmaps $name]
}

proc ibitmapfrom { name color} {
    global GidPriv GIDDEFAULT
    if { ![info exist GidPriv($name)] } {
	set GidPriv($name) [image create bitmap -file \
		[file join $GIDDEFAULT bitmaps $name] -foreground $color]
    }
    return $GidPriv($name)
}


proc GGCompress {} {
    set w .wcomp

    toplevel $w
    frame $w.f -relief groove -bd 2
    label $w.f.l1 -text [_ "Flags:"]
    entry $w.f.e1
    label $w.f.l2 -text [_ "File In:"]
    entry $w.f.e2 
    label $w.f.l3 -text [_ "File Out:"]
    entry $w.f.e3 
    grid $w.f.l1 $w.f.e1
    grid $w.f.l2 $w.f.e2
    grid $w.f.l3 $w.f.e3

    proc GGCDoIt { w} {
	set fl [ $w.f.e1 get]
	set fi [ $w.f.e2 get]
	set fo [ $w.f.e3 get]
	if { "$fl" != ""} {
	    gidcompress $fl $fi $fo
	} else {
	    gidcompress $fl $fi $fo
	}
    }
    button $w.f.b1 -text [_ "Do"] -command "GGCDoIt $w"
    button $w.f.b2 -text [_ "Close"] -command "destroy $w"
    SetWidgetsWidthFromText $w.f.b1 $w.f.b2

    grid $w.f.b1 $w.f.b2

    grid $w.f

}

proc framelabel { w text { rel groove}} {
    frame $w -bd 2 -relief $rel
    set i 0
    set parent [winfo parent $w]
    if { $parent == "." } { set parent "" }
    set label $parent.__label$i
    while { [winfo exists $label] } {
	incr i
	set label $parent.__label$i
    }
    label $label -text $text -pady 0
    place $label -in $w -x 10 -y -1 -anchor w -bordermode outside
    after idle raise $label
    return $w
}

proc framelabelsep {} {
    set FontHeight [font metric NormalFont -linespace]
    return [expr $FontHeight/2+1]
}

proc AskForNormal { w message } {
    global tkPriv tcl_platform

    toplevel $w
    wm title $w [_ "Enter normal"]

    if { $tcl_platform(platform) == "windows" } {
	wm transient $w [winfo toplevel [winfo parent $w]]
    }

    framelabel $w.f [_ "Enter Normal"]
    frame $w.f.f 

    label $w.f.f.l -text $message
    radiobutton $w.f.f.r1 -text [_ "Positive Z"] -var tkPriv(afw_check) -value z
    radiobutton $w.f.f.r2 -text [_ "Positive Y"] -var tkPriv(afw_check) -value y
    radiobutton $w.f.f.r3 -text [_ "Positive X"] -var tkPriv(afw_check) -value x
    radiobutton $w.f.f.r4 -text [_ "Coords:"] -var tkPriv(afw_check) -value coords

    entry $w.f.f.e1 -width 5
    entry $w.f.f.e2 -width 5
    entry $w.f.f.e3 -width 5

    $w.f.f.e1 ins end 0.0
    $w.f.f.e2 ins end 0.0
    $w.f.f.e3 ins end 1.0

    proc ChangeCoords { w } {
	global tkPriv
	if { $tkPriv(afw_check) == "coords" } {
	    $w.f.f.e1 conf -state normal -fg [$w.f.f.r1 cget -fg]
	    $w.f.f.e2 conf -state normal -fg [$w.f.f.r1 cget -fg]
	    $w.f.f.e3 conf -state normal -fg [$w.f.f.r1 cget -fg]
	} else { 
	    $w.f.f.e1 conf -state disabled -fg grey60
	    $w.f.f.e2 conf -state disabled -fg grey60
	    $w.f.f.e3 conf -state disabled -fg grey60
	}
    }

    trace var tkPriv(afw_check) w " ChangeCoords $w ;#"

    set tkPriv(afw_check) z

    grid $w.f.f.l -row 1 -column 1 -columnspan 4 -sticky w -padx 2
    grid  $w.f.f.r1 -row 2 -column 1 -columnspan 4 -sticky w -padx 2
    grid  $w.f.f.r2 -row 3 -column 1 -columnspan 4 -sticky w
    grid  $w.f.f.r3 -row 4 -column 1 -columnspan 4 -sticky w
    grid  $w.f.f.r4 -row 5 -column 1
    grid $w.f.f.e1 -row 5 -column 2 -sticky ew
    grid $w.f.f.e2 -row 5 -column 3 -sticky ew
    grid $w.f.f.e3 -row 5 -column 4 -sticky ew
    grid columnconf $w.f.f 2 -weight 1
    grid columnconf $w.f.f 3 -weight 1
    grid columnconf $w.f.f 4 -weight 1
    
    grid $w.f.f -row 1 -column 1 -sticky ewns -pady 6
    grid columnconf $w.f 1 -weight 1
    grid rowconf $w.f 1 -weight 1

    button $w.ok -text [_ "OK"] -width 8 -command "set tkPriv(afw_button) ok" -und 0
    button $w.scr -text [_ "In screen"] -width 8 -command "set tkPriv(afw_button) scr" -und 0
    button $w.cancel -text [_ "Cancel"] -width 8 -command "set tkPriv(afw_button) cancel" -und 0

    wm protocol $w WM_DELETE_WINDOW "tkButtonInvoke $w.cancel"
    bind $w.ok <Return> "tkButtonInvoke $w.ok"
    bind $w.f.f.e1 <Return> "tkButtonInvoke $w.ok"
    bind $w.f.f.e2 <Return> "tkButtonInvoke $w.ok"
    bind $w.f.f.e3 <Return> "tkButtonInvoke $w.ok"
    bind $w <Escape> "tkButtonInvoke $w.cancel"
    bind $w <Alt-o> "tkButtonInvoke $w.ok"
    bind $w <Alt-c> "tkButtonInvoke $w.cancel"
    bind $w <Alt-i> "tkButtonInvoke $w.scr"

    grid $w.f -row 1 -column 1 -columnspan 3 -pady 6 -sticky ewns
    grid $w.ok -row 2 -column 1 -padx 2
    grid $w.scr -row 2 -column 2 -padx 2
    grid $w.cancel -row 2 -column 3 -padx 2
    grid columnconf $w 1 -weight 1
    grid rowconf $w 1 -weight 1
    grid columnconf $w 2 -weight 1
    grid columnconf $w 3 -weight 1

    wm withdraw $w
    update idletasks

    set pare [winfo toplevel [winfo parent $w]]
    set x [expr [winfo x $pare]+[winfo width $pare ]/2- [winfo reqwidth $w]/2]
    set y [expr [winfo y $pare]+[winfo height $pare ]/2- [winfo reqheight $w]]
    if { $x < 0 } { set x 0 }
    if { $y < 0 } { set y 0 }
    WmGidGeom $w +$x+$y
    #wm geom $w +$x+$y
    update
    wm deiconify $w

    set oldFocus [focus]
    tkwait visibility $w
    grab $w
    focus $w.ok
    after 100 catch [ list "focus -force $w.ok"]

    while 1 {
	tkwait variable tkPriv(afw_button)

	if { $tkPriv(afw_button) != "ok" } { break }
	set value ""
	set isdiffzero 0
	if { $tkPriv(afw_check) == "coords" } {
	    set isbad 0
	    for { set i 1 } { $i <= 3 } { incr i } {
		if { [catch { set val [expr double([$w.f.f.e$i get])] }] } {
		    WarnWin [_ "Coordinate must be a real number"]
		    tkTabToWindow $w.f.f.e$i
		    set isbad 1
		    break
		}
		if { $val != 0.0 } { set isdiffzero 1 }
		if { $value == "" } { 
		    append value $val
		} else { append value ",$val" }
	    }
	    if { $isbad } { 
		focus -force $w
		tkTabToWindow $w.f.f.e$i
		continue 
	    }
	    if { !$isdiffzero } {
		WarnWin [_ "Normal must be different from zero"]
		focus -force $w
		tkTabToWindow $w.f.f.e1
		continue
	    }
	} elseif { $tkPriv(afw_check) == "z" } {
	    set value "0.0,0.0,1.0"
	} elseif { $tkPriv(afw_check) == "y" } {
	    set value "0.0,1.0,0.0"
	} else {
	    set value "1.0,0.0,0.0"
	}
	break
    }
    set oldGrab [grab current $w]
    if {$oldGrab != ""} {
	set grabStatus [grab status $oldGrab]
    }

    if { $oldFocus == "" || [catch {focus $oldFocus}] } {
	if { [winfo exists .gid] } {
	    focus [focus -lastfor .gid]
	}
    }

    destroy $w
    if {$oldGrab != ""} {
	if {$grabStatus == "global"} {
	    catch { grab -global $oldGrab }
	} else {
	    catch { grab $oldGrab }
	}
    }

    switch $tkPriv(afw_button) {
	ok {
	    .central.s process $value
	}
	cancel {
	    .central.s process escape
	}
    }
}

proc ExecuteFileCommand { comm PREPOST lastaction file } {

    set what [GiD_Info Project ViewMode]
    if { $what == "GRAFUSE" } { set what POSTUSE }
    if { $PREPOST == "PRE" && $what == "POSTUSE" } {
	if { $lastaction == "goingpre" } { return }
	.central.s process escape escape escape escape Preprocess
	delayedop changefunc [ list LoadFileInGid $file goingpre]
	return
    }
    if { $PREPOST == "POST" && $what != "POSTUSE" } {
	.central.s process escape escape escape escape Postprocess
    }
    .central.s process escape escape escape escape Files $comm \"$file\"
}

# last action can be: saving; goingpre;


proc LoadFileInGid { file {lastaction ""} } {
    if { ![file exists $file] } {
	WarnWin [_ "File '%s' not exists" $file]
	return
    }

    if {$::tcl_platform(platform) eq "windows"} {
      set file [file native [file attributes $file -longname]]
    } else {
      set file [file native $file]
    }
    if { [file isdir $file] && [string tolower [file extension $file]] == ".gid" } {
	set what [GiD_Info Project ViewMode]
	if { $what == "GRAFUSE" } { set what POSTUSE }
	if { $what == "POSTUSE" } {
	    if { $lastaction == "goingpre" } { return }
	    .central.s process escape escape escape escape Preprocess
	    delayedop changefunc [ list LoadFileInGid $file goingpre]
	    return
	}
	if { [GiD_Info Project AreChanges] } {
	    if { $lastaction == "saving" } { return }
	    .central.s process escape escape escape escape Files New
	    delayedop changefunc [ list LoadFileInGid $file saving]
	} else {
	    ExecuteFileCommand Read PRE $lastaction $file
	}
    } elseif { [string match *.post.bin $file] || [string match *.post.bin $file] || \
		   [string match *.post.msh $file] || [string match *.post.res $file] || \
		   [string match *.flavia.msh $file] || [string match *.flavia.res $file] } {
	ExecuteFileCommand Read POST $lastaction $file
    } else {
	switch [string tolower [file extension $file]] {
	    ".dxf" { ExecuteFileCommand DXFRead PRE $lastaction $file}
	    ".msh" - ".mesh" { ExecuteFileCommand MeshRead PRE $lastaction $file}
	    ".igs" - ".iges" { ExecuteFileCommand IgesRead PRE $lastaction $file}
	    ".nas" { ExecuteFileCommand NASTRANRead ANY $lastaction $file}
	    ".bch" { ExecuteFileCommand BatchFile ANY $lastaction $file}
	    ".x_t" { ExecuteFileCommand ParasolidRead PRE $lastaction $file}
	    ".x_b" { ExecuteFileCommand ParasolidRead PRE $lastaction $file}
	    ".sat" { ExecuteFileCommand AcisRead PRE $lastaction $file}
	    ".vda" { ExecuteFileCommand VDARead PRE $lastaction $file}
	    ".stl" { ExecuteFileCommand STLRead PRE $lastaction $file}
	    ".shp" { ExecuteFileCommand ShapefileRead PRE $lastaction $file}
	    ".3dm" { ExecuteFileCommand RhinoRead PRE $lastaction $file}
	    ".wrl" - ".vrml" { ExecuteFileCommand VrmlRead PRE $lastaction $file}
	    ".3ds" { ExecuteFileCommand 3DStudioRead ANY $lastaction $file}
	    ".cgns" { ExecuteFileCommand CGNSRead PRE $lastaction $file}
	    ".neu" - ".fno" { ExecuteFileCommand ReadFEMAP POST $lastaction $file}
	    ".plt" { ExecuteFileCommand ReadTECPLOT POST $lastaction $file}
	    ".bin" { ExecuteFileCommand Read POST $lastaction $file}
	    ".vv" {  ExecuteFileCommand ReadView ANY $lastaction $file}
	    default {
		if { [info proc LoadFileInGidUnknowExtension] != "" } {
		    LoadFileInGidUnknowExtension $file
		} else {
		    WarnWin [_ "Unknown file type for file '%s'" $file]
		    return
		}
	    }
	}
    }
}

proc LoadFileInGidPost { file {lastaction ""} } {
    if { ![file exists $file] } {
	WarnWin [_ "File '%s' not exists" $file]
	return
    }

    if {$::tcl_platform(platform) eq "windows"} {
      set file [file native [file attributes $file -longname]]
    } else {
      set file [file native $file]
    }
    if { [file isdir $file] && [string tolower [file extension $file]] == ".gid" } {
	set what [GiD_Info Project ViewMode]
	if { $what == "GRAFUSE" } { set what POSTUSE }
	if { $what == "POSTUSE" } {
	    if { $lastaction == "goingpre" } { return }
	    .central.s process escape escape escape escape Preprocess
	    delayedop changefunc [ list LoadFileInGid $file goingpre]
	    return
	}
	if { [GiD_Info Project AreChanges] } {
	    if { $lastaction == "saving" } { return }
	    .central.s process escape escape escape escape Files New
	    delayedop changefunc [ list LoadFileInGid $file saving]
	} else {
	    ExecuteFileCommand Read PRE $lastaction $file
	}
    } elseif { [string match *.post.bin $file] || [string match *.post.bin $file] || \
		   [string match *.post.msh $file] || [string match *.post.res $file] || \
		   [string match *.flavia.msh $file] || [string match *.flavia.res $file] } {
	ExecuteFileCommand Read POST $lastaction $file
    } else {
	switch [string tolower [file extension $file]] {
	    ".nas" { ExecuteFileCommand NASTRANRead POST $lastaction $file}
	    ".bch" { ExecuteFileCommand BatchFile POST $lastaction $file}
	    ".3ds" { ExecuteFileCommand 3DStudioRead POST $lastaction $file}
	    ".cgns" { ExecuteFileCommand CGNSRead PRE $lastaction $file}
	    ".neu" - ".fno" { ExecuteFileCommand ReadFEMAP POST $lastaction $file}
	    ".plt" { ExecuteFileCommand ReadTECPLOT POST $lastaction $file}
	    ".bin" { ExecuteFileCommand Read POST $lastaction $file}
	    ".vv" {  ExecuteFileCommand ReadView ANY $lastaction $file}
	    ".res" { ExecuteFileCommand Read POST $lastaction $file}
	    ".msh" { ExecuteFileCommand Read POST $lastaction $file}
	    ".bon" { ExecuteFileCommand Read POST $lastaction $file}
	    ".dat" { ExecuteFileCommand Read POST $lastaction $file}
	    default {
		if { [info proc LoadFileInGidUnknowExtension] != "" } {
		    LoadFileInGidUnknowExtension $file
		} else {
		    WarnWin [_ "Unknown file type for file '%s'" $file]
		    return
		}
	    }
	}
    }
}

proc MsgcatGetLocales {} {
    global GIDDEFAULT
    # Directorio en el que se encuentran los ficheros msgs
    set dir [file join $GIDDEFAULT scripts msgs]
    
    # Obtenemos los nombres de fichero = locale
    #force to show also en in the list
    set locales {en}
    foreach file [glob -nocomplain -directory $dir *.msg] {
	set name [lindex [file split $file] end]
	set name [string range $name 0 end-4]
	set locales [linsert $locales end $name]
    }
    set locales [lsort -unique $locales]

    return $locales
}

proc MsgcatSetLocale {locale} {
    global GIDDEFAULT
    # Directorio en el que se encuentran los ficheros msgs
    set dir [file join $GIDDEFAULT scripts msgs]
    
    ::msgcat::mclocale $locale
    ::msgcat::mcload $dir
    # load the problemtype's message catalog
    if {![catch {GiD_Info problemtypepath} dir]} {
	::GIDi18nProblemType::load $dir
    }
}

proc MsgcatMcInitialize {} {

    global env

    # Obtener el fichero de configuracion local
    set config [file nativename [GiveGidDefaultsFile]]

    
    # Consultar las preferencias del usuario
    set found -1
    set res [file exists $config]
    if {$res} {
	 
	# Leer el fichero
	set fileid [open "$config" r]
	fconfigure $fileid -encoding "utf-8"
	set data [read $fileid]
	close $fileid
	
	foreach line [split $data "\n"] {
	    
	    set res [string match "Language*" $line]
	    
	    if { $res} {
		
		set found 1
	    
		set language [string range $line 9 end]
	    }
	}
    }
    
    # Consultamos las preferencias del sistema
    if { $found == -1} {
	set preferences [::msgcat::mcpreferences]
	set preferences [linsert $preferences 0 en] ;#until the message catalog is fully translated
	# Obtenemos los locales disponibles
	set locales [MsgcatGetLocales]

	# Buscamos el primer lenguaje disponible para
	# las preferencias
	set found -1
	foreach preference $preferences {
	    set found [lsearch -exact $locales $preference]
	    if { $found > -1 } {
		set language $preference
		break
	    }
	}
	
	if { $found == -1 } {
	    
	    # Si no se ha podido determinar el idioma, 
	    # se define el inglés  por defecto
	    set language en
	}
    }

    ## Establecemos el idioma
    #::msgcat::mclocale $language
    ## Cargamos las cadenas
    #set dir [file join $env(GIDDEFAULTTCL) "msgs"]
    #::msgcat::mcload $dir
    MsgcatSetLocale $language
}


proc SetGUIAppearanceColor { } {
    global tcl_platform GidPriv
    if { ![info exists GidPriv(GUIAppearance)] } {
	set GidPriv(GUIAppearance) 1
    }
    if { ( $tcl_platform(platform) == "windows") || ![ string eq $GidPriv(Color,Linux) "brown"] } {
	if { $GidPriv(GUIAppearance) == "0" } {
	    set GidPriv(Color,ToolbarBg) [.gid cget -background]
	} elseif { $GidPriv(GUIAppearance) == "1" } {
	    #clarified color 
	    set GidPriv(Color,ToolbarBg) [CCColorActivo [.gid cget -background] 25] 
	} else {
	    #set to a fixed value 
	    set GidPriv(Color,ToolbarBg) #DEDFDE
	}
	set GidPriv(Color,CommandLineBg) #FFFFFF
    } else {
	    if { $GidPriv(GUIAppearance) == "0" } {
		#set GidPriv(Color,ToolbarBg) [.gid cget -background]
		# este es el otro sitio que hay que cambiar:
		set GidPriv(Color,ToolbarBg) AntiqueWhite3
	    } elseif { $GidPriv(GUIAppearance) == "1" } {
		#clarified color 
		#set GidPriv(Color,ToolbarBg) [CCColorActivo [.gid cget -background] 25] 
		set GidPriv(Color,ToolbarBg) #E9E2D0
	    } else {
		#set to a fixed value 
		#set GidPriv(Color,ToolbarBg) #DEDFDE
		set GidPriv(Color,ToolbarBg) #E9E2D0
	    }
	    set GidPriv(Color,CommandLineBg) [CCColorSombra $GidPriv(Color,ToolbarBg) 25]
	    #set GidPriv(Color,CommandLineBg) #FFFFFF
    }

    
    set GidPriv(ToolbarBg) $GidPriv(Color,ToolbarBg) ;#defined only for back compatibility (GiD-CEM3.0,etc)

    # tk_setPalette AntiqueWhite3
    # set bg \#116611
    # option add *Background $bg
    # set GidPriv(Color,BotonActivo) [ CCColorActivo $bg 129]
    # set GidPriv(Color,BotonMarcoActivo) [ CCColorSombra $bg 65]

    # .gid conf -background $bg

    set GidPriv(Original,Color,Background) [ .gid cget -bg]
    #set GidPriv(Original,Color,Foreground) [ .gid cget -fg]
    foreach col_idx {BotonActivo BotonMarcoActivo DisabledBackgroundEntry\
		         DisabledForegroundEntry BackgroundListbox\
		         DisabledForegroundMenu CommandLineBg ToolbarBg BackgroundFileDialog} {
	set GidPriv(Original,Color,$col_idx) $GidPriv(Color,$col_idx)
    }

    set GidPriv(AllowBGChange) 0

}

proc ChangeBgWindows { w} {
    global GidPriv

    set org_col $GidPriv(Original,Color,Background)
    set retcol [ GIDChooseColor $w.selectcolor -title [_ "Select color"] -color $org_col]
    if { ![ string eq $retcol ""]} {
	.gid conf -background $retcol

	option add *Background $retcol
	option add *NoteBook*background $retcol

	set bg_entries [ CCColorActivo $retcol 65]
	option add *activeBackground $bg_entries
	option add *ScrollBar*foreground $bg_entries
	option add *ScrollBar*background $retcol
	option add *ScrollBar*troughColor $bg_entries
	option add *Scale*Background $retcol
	option add *Scale*troughColor $bg_entries

	set GidPriv(Color,BackgroundListbox) $bg_entries
	set GidPriv(Color,DisabledBackgroundEntry) $bg_entries
	set GidPriv(Color,DisabledForegroundEntry) {black}
	set GidPriv(Color,DisabledForegroundMenu) {grey60} 
	set GidPriv(Color,BackgroundFileDialog) $bg_entries

	option add *Entry*DisabledBackground $GidPriv(Color,DisabledBackgroundEntry)
	option add *Entry*Background $bg_entries
	option add *ComboBox*Entry*background $bg_entries
	option add *Entry*DisabledForeground $GidPriv(Color,DisabledForegroundEntry)
	option add *Listbox*Background $GidPriv(Color,BackgroundListbox)

	set GidPriv(Color,BotonActivo) [ CCColorActivo $retcol 129]
	set GidPriv(Color,BotonMarcoActivo) [ CCColorSombra $retcol 65]

	if { $GidPriv(GUIAppearance) == "0" } {
	    set GidPriv(Color,ToolbarBg) [ .gid cget -background]
	} elseif { $GidPriv(GUIAppearance) == "1" } {
	    set GidPriv(Color,ToolbarBg) [ CCColorActivo [.gid cget -background] 25] 
	} else {
	    set GidPriv(Color,ToolbarBg) \#DEDFDE
	}
	    set GidPriv(Color,CommandLineBg) [ CCColorSombra $GidPriv(Color,ToolbarBg) 25]
	# nos servimos de esto para actualizar widgets
	UpdateWidgetsLanguage
	# los botones de la derecha
	RightButtons [lindex $::GidPriv(PrePostRightButWindowGeom) 0]
    }
}

proc SetTemplatesMenu { menu } {
    global GIDDEFAULT GidPriv

    set dir [file join $GIDDEFAULT templates]
    
    $menu del 0 end
    foreach i [winfo children $menu] { destroy $i }
    set inum 0
    
    set files [glob -nocomplain -tails -directory $dir -types f *.bas]
    foreach i [lsort -dictionary $files] {
	$menu add command -label [file root $i] -command [list processbasfile [file join $dir $i]]
	incr inum
    }
    proc processbasfile { basfile } {
       .central.s process escape escape escape escape Files WriteForBAS \"$basfile\"
    }

    if { $inum } { $menu add separator }
    $menu add command -label [_ "Others"]... -command \
	    ".central.s process escape escape escape escape \
	    Files WriteForBAS"
}

namespace eval WaitStater {
    variable cursors
}

proc WaitStater::SetWatch { w } {
    SetWatchRecursive $w $w
    update
}

proc WaitStater::UnsetWatch { w } {
    variable cursors

    foreach i [array names cursors $w,*] {
	regexp {[^,]+,(.*)} $i {} win
	$win configure -cursor $cursors($i)
	unset cursors($i)
    }
    update
}

proc WaitStater::SetWatchRecursive { w win } {
    variable cursors


    catch {
	set cursors($w,$win) [$win cget -cursor]
	$win configure -cursor watch
    }
    foreach i [winfo children $win] {
	SetWatchRecursive $w $i
    }
}

proc ShowGiDVersionType { { que ""}} {
    global GidPriv

    # que = NONE | ACADEMIC | PROFESSIONAL
    # el CreateOtherBitmaps hace esto:
    # boton ... -image [GetImage $filename]

    if { "$que" == ""} {
	set que none
	if { [ info exist GidPriv(GiDVersionType)] } {
	    set que $GidPriv(GiDVersionType)
	}
    }

    if { [ regexp -nocase {^(.*)\.gif$} $GidPriv(InfoGiDBitmapName) dummy base_name] } {
	set img ""
	if { [ string match -nocase "none" $que]} {
	    set img [GetImage $GidPriv(InfoGiDBitmapName)]
	    set GidPriv(GiDVersionType) $que
	} elseif { [ string match -nocase "academic" $que]} {
	    set img [GetImage ${base_name}_a.gif]
	    set GidPriv(GiDVersionType) $que
	} elseif { [ string match -nocase "professional" $que]} {
	    set img [GetImage ${base_name}_p.gif]
	    set GidPriv(GiDVersionType) $que
	} elseif { [ string match -nocase "temporallyprofessional" $que]} {
	    set img [GetImage ${base_name}_t.gif]
	    set GidPriv(GiDVersionType) $que
	}
	
	if { [ info exist GidPriv(InfoGiDButtonName)] && ( $GidPriv(InfoGiDButtonName) != "")} {
	    if { [ winfo exists $GidPriv(InfoGiDButtonName)] && ( $img != "")} {
		$GidPriv(InfoGiDButtonName) conf -image $img
	    }
	}
    }
}

proc GetImage { filename } {
    global GidPrivImages
    if { ![info exist GidPrivImages($filename)] } {
	set GidPrivImages($filename) [image create photo -file [file join $::GIDDEFAULT bitmaps $filename]]
    }
    return $GidPrivImages($filename)
}

proc ReloadLastModel { {previousproblemtype ""} } {
    set Reload [GiD_Info variables ReloadLastModel]
    if { $Reload != 0 } {
	if { [GiD_Info variables LastModelSaved] == 0 } {
	    set Reload 0
	}
    }
    if { $Reload != 0 && [GiD_Info Project ModelName] != "UNNAMED" } {
	#already read a backup file after crash or load a model from the command line
	set Reload 0
    }
    if { $Reload != 0 } {
	set names [GiD_Info recentprojects]
	if { [llength $names] > 0 } {
	    if { $previousproblemtype != "" } {
		#only load last model if his probletype is also pt
		set geoinfo [GiD_GetInfoGeoFile [file join [lindex $names 0].gid [file tail [lindex $names 0]].geo]]
		#file join to convert \ to /
		if { [file join $previousproblemtype] != [file join [lindex $geoinfo 2]] } {
		    set Reload 0
		}
	    }
	    if { $Reload != 0 } {
		LoadFileInGid [lindex $names 0].gid
	    }
	}
    }
}

#args are the command line arguments argv
proc BeforeMainLoop { args } {

    #check if used -p problemtype <pt> or -e data defaults problemtype <pt>
    #to not reload the last model if it has a different problemtype
    set previousproblemtype ""
    if { [llength $args] > 0 } {
	#check if try to load a problemtype
	set pos [lsearch $args "-p"]
	if { $pos != -1 } {
	    set previousproblemtype [lindex $args [expr $pos+1]]
	}
	set pos [lsearch $args "-e"]
	if { $pos != -1 } {
	    set pos2 [lsearch [string tolower [lrange $args [expr $pos+1] end]] "problemtype"]
	    if { $pos2 != -1 } {
		set previousproblemtype [lindex $args [expr $pos+$pos2+2]]
	    }
	}
    }
    
    ReloadLastModel $previousproblemtype
    
    if { [info command configure_program_proc] ne "" } {
	uplevel #0 configure_program_proc
    }
}



proc bgerror { err } {
    #avoid disabled graphics
    if { $::_DebugDefaultValues} {
	DebugText "bgerror: $err"
	DebugText "$::errorInfo"
    }
    if { [info command .gid.central.s] ne "" } {
	foreach i [list windows graphinput graphics warnline] {
	    .gid.central.s disable $i 0
	}
    }
    ::tk::dialog::error::bgerror $err
}

#unknown is a copy of init.tcl but including translation (maybe can be alse be enhanced)

proc __unknown args {
    global auto_noexec auto_noload env unknown_pending tcl_interactive
    global errorCode errorInfo

    # If the command word has the form "namespace inscope ns cmd"
    # then concatenate its arguments onto the end and evaluate it.

    set cmd [lindex $args 0]
    if {[regexp "^:*namespace\[ \t\n\]+inscope" $cmd] && [llength $cmd] == 4} {
	set arglist [lrange $args 1 end]
	set ret [catch {uplevel 1 ::$cmd $arglist} result]
	if {$ret == 0} {
	    return $result
	} else {
	    return -code $ret -errorcode $errorCode $result
	}
    }

    # Save the values of errorCode and errorInfo variables, since they
    # may get modified if caught errors occur below.  The variables will
    # be restored just before re-executing the missing command.

    # Safety check in case something unsets the variables 
    # ::errorInfo or ::errorCode.  [Bug 1063707]
    if {![info exists errorCode]} {
	set errorCode ""
    }
    if {![info exists errorInfo]} {
	set errorInfo ""
    }
    set savedErrorCode $errorCode
    set savedErrorInfo $errorInfo
    set name [lindex $args 0]
    if {![info exists auto_noload]} {
	#
	# Make sure we're not trying to load the same proc twice.
	#
	if {[info exists unknown_pending($name)]} {
	    return -code error [_ "self-referential recursion in 'unknown' for command '%s'" $name];
	}
	set unknown_pending($name) pending;
	set ret [catch {auto_load $name [uplevel 1 {::namespace current}]} msg]
	unset unknown_pending($name);
	if {$ret != 0} {
	    append errorInfo [concat "\n    (" [_ "autoloading"] "'$name')"]
	    return -code $ret -errorcode $errorCode -errorinfo $errorInfo $msg
	}
	if {![array size unknown_pending]} {
	    unset unknown_pending
	}
	if {$msg} {
	    set errorCode $savedErrorCode
	    set errorInfo $savedErrorInfo
	    set code [catch {uplevel 1 $args} msg]
	    if {$code ==  1} {
		#
		# Compute stack trace contribution from the [uplevel].
		# Note the dependence on how Tcl_AddErrorInfo, etc. 
		# construct the stack trace.
		#
		set cinfo $args
		set ellipsis ""
		while {[string bytelength $cinfo] > 150} {
		    set cinfo [string range $cinfo 0 end-1]
		    set ellipsis "..."
		}
		append cinfo $ellipsis [concat "\"\n    (" [_ "'uplevel' body line 1"] ")"]
		append cinfo [concat "\n    " [_ "invoked from within"]]
		append cinfo [concat "\n\"" [_ "uplevel"] " 1 \$args\""]
		#
		# Try each possible form of the stack trace
		# and trim the extra contribution from the matching case
		#
		set expect [concat "$msg\n    " [_ "while executing"] "\n\"$cinfo"]
		if {$errorInfo eq $expect} {
		    #
		    # The stack has only the eval from the expanded command
		    # Do not generate any stack trace here.
		    #
		    return -code error -errorcode $errorCode $msg
		}
		#
		# Stack trace is nested, trim off just the contribution
		# from the extra "eval" of $args due to the "catch" above.
		#
		set expect [concat "\n    " [_ "invoked from within"] "\n\"$cinfo"]
		set exlen [string length $expect]
		set eilen [string length $errorInfo]
		set i [expr {$eilen - $exlen - 1}]
		set einfo [string range $errorInfo 0 $i]
		#
		# For now verify that $errorInfo consists of what we are about
		# to return plus what we expected to trim off.
		#
		if {$errorInfo ne "$einfo$expect"} {
		    error [_ "Tcl bug: unexpected stack trace in 'unknown'"] {} \
		        [list CORE UNKNOWN BADTRACE $expect $errorInfo]
		}
		return -code error -errorcode $errorCode \
		        -errorinfo $einfo $msg
	    } else {
		return -code $code $msg
	    }
	}
    }

    if {([info level] == 1) && [string equal [info script] ""] \
	    && [info exists tcl_interactive] && $tcl_interactive} {
	if {![info exists auto_noexec]} {
	    set new [auto_execok $name]
	    if {$new != ""} {
		set errorCode $savedErrorCode
		set errorInfo $savedErrorInfo
		set redir ""
		if {[string equal [info commands console] ""]} {
		    set redir ">&@stdout <@stdin"
		}
		return [uplevel 1 exec $redir $new [lrange $args 1 end]]
	    }
	}
	set errorCode $savedErrorCode
	set errorInfo $savedErrorInfo
	if {[string equal $name "!!"]} {
	    set newcmd [history event]
	} elseif {[regexp {^!(.+)$} $name dummy event]} {
	    set newcmd [history event $event]
	} elseif {[regexp {^\^([^^]*)\^([^^]*)\^?$} $name dummy old new]} {
	    set newcmd [history event -1]
	    catch {regsub -all -- $old $newcmd $new newcmd}
	}
	if {[info exists newcmd]} {
	    tclLog $newcmd
	    history change $newcmd 0
	    return [uplevel 1 $newcmd]
	}

	set ret [catch {set candidates [info commands $name*]} msg]
	if {[string equal $name "::"]} {
	    set name ""
	}
	if {$ret != 0} {
	    return -code $ret -errorcode $errorCode \
		[concat [_ "error in unknown while checking if '%s' is a unique command abbreviation" $name] ":\n$msg"]
	}
	# Filter out bogus matches when $name contained
	# a glob-special char [Bug 946952]
	set cmds [list]
	foreach x $candidates {
	    if {[string range $x 0 [expr [string length $name]-1]] eq $name} {
		lappend cmds $x
	    }
	}
	if {[llength $cmds] == 1} {
	    return [uplevel 1 [lreplace $args 0 0 $cmds]]
	}
	if {[llength $cmds]} {
	    if {[string equal $name ""]} {
		return -code error [concat [_ "empty command name"] " ''"]
	    } else {
		return -code error \
		        [concat [_ "ambiguous command name"] " '$name': [lsort $cmds]"]
	    }
	}
    }
    return -code error [concat [_ "invalid command name"] " '$name'"]
}


proc SetMenusStates { } {
    #return
    global tcl_platform
    
    if { $tcl_platform(platform) == "windows" } {
	set what [GiD_Info Project ViewMode]
	if { $what == "GRAFUSE" } { set what POSTUSE }
	if { $what == "POSTUSE" } {
	    #MENUS DE POST
	} else {
	    SetFilesMenuState
	    SetViewMenuState
	    SetGeometryMenuState
	    SetUtilitiesMenuState
	    SetDataMenuState
	    SetMeshMenuState
	    SetCalculateMenuState
	}
    } else {
	#LINUX
    }
	
}

proc SetFilesMenuState { } {   
    set menu_name "Files#C#menu"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return
    
    set menuname [GetMenuName $i]
    
    if { [winfo exists .gid.central.s] && [lindex [GiD_Info mesh] 0] == 0 } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
    } else {
	set color "black"
    }
    
    set option_name [list "Postprocess#C#menu"]
    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
    if { $position != -1 } {
	ChangeMenuForegroundColor $menuname $position $color
    }
}

proc SetViewMenuState { } {
    set menu_name "View#C#menu,Verb"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return

    set menuname [GetMenuName $i]
    
    set npoints [lindex [.central.s info Geometry NumPoints] 0]
    #set nlines [lindex [.central.s info Geometry NumLines] 0]
    #set nsurfs [lindex [.central.s info Geometry NumSurfaces] 0]
    #set nvols [lindex [.central.s info Geometry NumVolumes] 0]
    
    set option_name [list "Label#C#menu"]
    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
    if { $position != -1 } {
	if { !$npoints } { 
	    ChangeMenuForegroundColor $menuname $position $::GidPriv(Color,DisabledForegroundMenu)
	} else {
	    ChangeMenuForegroundColor $menuname $position "black"
	}
    }
}

proc SetGeometryMenuState { } {
    set menu_name "Geometry#C#menu"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return

    set menuname [GetMenuName $i]
    
    set npoints [lindex [.central.s info Geometry NumPoints] 0]
    set nlines [lindex [.central.s info Geometry NumLines] 0]
    set nsurfs [lindex [.central.s info Geometry NumSurfaces] 0]
    set nvols [lindex [.central.s info Geometry NumVolumes] 0]
    
    set option_name [list "Create#C#menu"]
    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
    if { $position != -1 } {
	SetCreateGeometrySubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
    }

    if { !$npoints } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
	
	set optionstochange [list [list "View geometry#C#menu"] [list "Delete#C#menu"] [list "Edit#C#menu"]]
	foreach option_name $optionstochange {
	    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	    if { $position != -1 } {
		ChangeMenuForegroundColor $menuname $position $color
	    }
	}
    } else {
	set option_name [list "View geometry#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position "black"
	}
	
	set option_name [list "Delete#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetDeleteGeometrySubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
	}
	
	set option_name [list "Edit#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetEditGeometrySubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
	}
    }
}

proc SetCreateGeometrySubMenusState { menuname pos npoints nlines nsurfs nvols } {
    $menuname entryconfigure $pos -foreground "black"

    set optionstochange [list [_ "Polyline#C#menu"] [_ "NURBS surface#C#menu"] [_ "Volume#C#menu"]\
	    [_ "Contact#C#menu"] [_ "Arc#C#menu"] [_ "Contact surface#C#menu"]]

    set submenuname $menuname.$pos
    if { ![winfo exists $submenuname] } { 
	return 
    }
    if { [$submenuname index end] == "none" } { return }
    
    if { $nvols } {
	set colors(0) "black"
	set colors(1) "black"
	set colors(2) "black"
	set colors(3) "black"
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(0) "black"
	    set colors(1) "black"
	    set colors(2) "black"
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(0) "black"
		set colors(1) "black"
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
		if { $npoints } {
		    set colors(0) "black"
		} else {
		    set colors(0) $::GidPriv(Color,DisabledForegroundMenu)
		}
	    }
	}
    }

    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	switch $option {
	    0 { ChangeMenuForegroundColor $submenuname $i $colors(0) } 
	    1 { 
		if { !$npoints } {
		    ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu) 
		} else {
		    $submenuname entryconfigure $i -foreground "black"
		    set subsubmenuname "$submenuname.$i"
		    if { ![winfo exists $subsubmenuname] } {
		        continue
		    }
		    set optionstochange2 [list [_ "By contour#C#menu"] [_ "Automatic#C#menu"] [_ "Trimmed#C#menu"] \
		            [_ "Parallel Lines#C#menu"] [_ "By Points#C#menu"] [_ "By LinePoints#C#menu"] [_ "Search#C#menu"]]
		    set nj [$subsubmenuname index end]
		    if { $nj == "none" } { set nj -1 }
		    for {set j 0} {$j <= $nj} {incr j} {
		        if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		        set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		        if { $option == -1 } { continue }
		        if { $option == 0 || $option == 1 || $option == 3 || $option == 5 || $option == 6 } { 
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(1) 
		        } elseif { $option == 2 } {
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(2) 
		        } elseif { $option == 4 } {
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(0) 
		        }
		    } 
		}
	    } 
	    2 {  
		if { !$nsurfs } {
		    ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu) 
		} else {
		    $submenuname entryconfigure $i -foreground "black"
		    set subsubmenuname "$submenuname.$i"
		    if { ![winfo exists $subsubmenuname] } {
		        continue
		    }
		    set optionstochange2 [list [_ "By contour#C#menu"] [_ "Automatic 6-sided volumes#C#menu"] \
		            [_ "Search#C#menu"] [_ "Hole#C#menu"]]
		    set nj [$subsubmenuname index end]
		    if { $nj == "none" } { set nj -1 }
		    for {set j 0} {$j <= $nj} {incr j} {
		        if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		        set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		        if { $option == -1 } { continue }
		        if { $option == 0 || $option == 1 || $option == 2 } { 
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(2) 
		        } elseif { $option == 3 } {
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(3) 
		        } 
		    } 
		}
	    } 
	    3 { 
		if { !$nsurfs } {
		    ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu) 
		} else {
		    $submenuname entryconfigure $i -foreground "black"
		    set subsubmenuname "$submenuname.$i"
		    if { ![winfo exists $subsubmenuname] } {
		        continue
		    }
		    set optionstochange2 [list [_ "Volume#C#menu"] [_ "Separated volume#C#menu"]]
		    set nj [$subsubmenuname index end]
		    if { $nj == "none" } { set nj -1 }
		    for {set j 0} {$j <= $nj} {incr j} {
		        if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		        set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		        if { $option == -1 } { continue }
		        ChangeMenuForegroundColor $subsubmenuname $j $colors(2) 
		    } 
		}
	    } 
	    4 {
		$submenuname entryconfigure $i -foreground "black"
		set subsubmenuname "$submenuname.$i"
		if { ![winfo exists $subsubmenuname] } {
		    continue
		}
		set optionstochange2 [list [_ "Fillet curves#C#menu"] [_ "By 3 points#C#menu"] ]
		set nj [$subsubmenuname index end]
		if { $nj == "none" } { set nj -1 }
		for {set j 0} {$j <= $nj} {incr j} {
		    if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		    set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		    if { $option == -1 } { continue }
		    if { $option == 0 } { 
		        ChangeMenuForegroundColor $subsubmenuname $j $colors(1) 
		    } 
		} 
	    }
	    5 { ChangeMenuForegroundColor $submenuname $i $colors(1) } 
	    default { }
	}
    }    
    
}

proc SetEditGeometrySubMenusState { menuname pos npoints nlines nsurfs nvols } {
    
    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Move point#C#menu"] [_ "Polyline#C#menu"]  \
	    [_ "Divide#C#menu"] [_ "Lines operations#C#menu"] [_ "Swap arc#C#menu"] \
	    [_ "Uncollapse#C#menu"] [_ "Edit NURBS#C#menu"] [_ "Hole NURBS surface#C#menu"] \
	    [_ "SurfMesh#C#menu"] [_ "Convert to NURBS#C#menu"] [_ "Simplify NURBS#C#menu"] \
	    [_ "Collapse#C#menu"] [_ "Intersection#C#menu"] [_ "Surface boolean op.#C#menu"]\
	    [_ "Volume boolean op.#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"   
    if { $nvols } {
	set colors(3) "black"
	set colors(2) "black"  
	set colors(1) "black"  
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(2) "black"  
	    set colors(1) "black"   
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(1) "black"   
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	    }
	}
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	switch $option {
	    0 { ChangeMenuForegroundColor $submenuname $i $colors(0) } ; #points
	    1 { ChangeMenuForegroundColor $submenuname $i $colors(1) } ; #lines
	    2 { SetDivideSubMenusState $submenuname $i $nlines $nsurfs $nvols } 
	    3 { ChangeMenuForegroundColor $submenuname $i $colors(1) } ; #lines
	    4 { ChangeMenuForegroundColor $submenuname $i $colors(1) } ; #lines
	    5 { SetUnCollapseSubMenusState $submenuname $i $nlines $nsurfs } 
	    6 { SetEditNurbSubMenusState $submenuname $i $nlines $nsurfs } 
	    7 { ChangeMenuForegroundColor $submenuname $i $colors(2) } ; #surfaces
	    8 { ChangeMenuForegroundColor $submenuname $i $colors(2) } ; #surfaces
	    9 { SetLineSurfaceSubMenusState $submenuname $i $nlines $nsurfs } 
	    10 { SetSimplifySubMenusState $submenuname $i $nlines $nsurfs }
	    11 { SetCollapseGeomMenusState $submenuname $i $npoints $nlines $nsurfs $nvols} 
	    12 { 
		if { !$nlines } {
		    ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu) 
		} else {
		    $submenuname entryconfigure $i -foreground "black"
		    set subsubmenuname "$submenuname.$i"
		    set optionstochange2 [list [_ "Line-line#C#menu"] [_ "Multiple lines#C#menu"] [_ "Surface-2 points#C#menu"] \
		            [_ "Surface-lines#C#menu"] [_ "Surface-surface#C#menu"] [_ "Multiple surfaces#C#menu"]]
		    set nj [$subsubmenuname index end]
		    if { $nj == "none" } { set nj -1 }
		    for {set j 0} {$j <= $nj} {incr j} {
		        if { [$subsubmenuname type $j] != {command} && [$subsubmenuname type $j] != {cascade} } { continue }
		        set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		        if { $option == -1 } { continue }
		        if { $option == 0 || $option == 1 } { 
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(1) 
		        } elseif { $option == 2 || $option == 3 || $option == 4 || $option == 5  } {
		            ChangeMenuForegroundColor $subsubmenuname $j $colors(2) 
		        } 
		    }
		}
	    }
	    13 { ChangeMenuForegroundColor $submenuname $i $colors(2) } ; #surfaces
	    14 { ChangeMenuForegroundColor $submenuname $i $colors(3) } ; #volumes
	    default { }
	}
    }
}

proc SetCollapseGeomMenusState { menuname pos npoints nlines nsurfs nvols } {
    
    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [ list [_ "Model#C#menu"] [_ "Points#C#menu"] [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"] ]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    set colors(1) "black"
    
    if { $nvols } {
	set colors(4) "black"
	set colors(3) "black"  
	set colors(2) "black"  
    } else {
	set colors(4) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(3) "black"  
	    set colors(2) "black"   
	} else {
	    set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(2) "black"   
	    } else {
		set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    }
	}
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }    
}

proc SetDivideSubMenusState {  menuname pos nlines nsurfs nvols } {
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"

    set optionstochange [list [_ "Lines#C#menu"]  [_ "Polylines#C#menu"] \
	    [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"]]
	
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    if { $nvols } {
	set colors(0) "black"  
	set colors(1) "black"  
	set colors(2) "black"   
	set colors(3) "black"   
    } elseif { $nsurfs } {
	set colors(0) "black"  
	set colors(1) "black"  
	set colors(2) "black"  
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu) 
    } elseif { $nlines } {
	set colors(0) "black"  
	set colors(1) "black"  
	set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu) 
    } else {
	set colors(0) $::GidPriv(Color,DisabledForegroundMenu)
	set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
    }
    for {set i 0} {$i <= [$submenuname index end]} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue
	ChangeMenuForegroundColor $submenuname $i $colors($option) ; #surfaces
    } 
}


proc SetSimplifySubMenusState { menuname pos nlines nsurfs } {
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Model#C#menu"] [_ "Line#C#menu"] [_ "Surface#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
  
    if { $nsurfs } {
	set colors(0) "black"  
	set colors(1) "black"  
	set colors(2) "black"   
    } elseif { $nlines } {
	set colors(0) "black"  
	set colors(1) "black"  
	set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
    } else {
	set colors(0) $::GidPriv(Color,DisabledForegroundMenu)
	set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }
}

proc SetLineSurfaceSubMenusState {  menuname pos nlines nsurfs } {
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Line#C#menu"] [_ "Surface#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"   
    if { $nsurfs } {
	set colors(1) "black"   
    } else {
	set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }
}

proc SetEditNurbSubMenusState {  menuname pos nlines nsurfs } {
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Line#C#menu"]... [_ "Surface#C#menu"]...]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"   
    if { $nsurfs } {
	set colors(1) "black"   
    } else {
	set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }
}

proc SetDeleteGeometrySubMenusState {  menuname pos npoints nlines nsurfs nvols } {
    
    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Points#C#menu"] [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"] [_ "All types#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    set colors(4) "black"
    
    if { $nvols } {
	set colors(3) "black"
	set colors(2) "black"  
	set colors(1) "black"  
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(2) "black"  
	    set colors(1) "black"   
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(1) "black"   
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	    }
	}
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }    
}

proc SetUtilitiesMenuState { } {
    set menu_name "Utilities#C#menu"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return
   
    set menuname [GetMenuName $i]
 
    set npoints [GiD_Info Geometry NumPoints]
    set nlines [GiD_Info Geometry NumLines]
    set nsurfs [GiD_Info Geometry NumSurfaces]
    set nvols [GiD_Info Geometry NumVolumes]
    
    set optionstochange [list [list "Copy#C#menu"] \
	    [list "Move#C#menu"] [list "Renumber#C#menu"] [list "Id#C#menu"] \
	    [list "Distance#C#menu"] [list "Repair#C#menu"]]

    #Aquests no valen perquè hi ha menús que es tracten a part.
    #set optionstochange [list [_ "List"] [_ "Copy"]...\
	    [_ "Move#C#menu"]... [_ "Renumber#C#menu"] [_ "Id#C#menu"] [_ "Signal#C#menu"]\
	    [_ "Distance#C#menu"] [_ "Draw normals#C#menu"] [_ "Draw higher entities#C#menu"]\
	    [_ "Repair#C#menu"] [_ "Collapse#C#menu"] [_ "Uncollapse#C#menu"]]

    if { [GiD_Info Project ViewMode] == "MESHUSE" } {
	set color "black"
	foreach option_name $optionstochange {
	    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	    if { $position != -1 } {
		ChangeMenuForegroundColor $menuname $position $color
	    }
	}        
	return
    }
    
    if { !$npoints } {
	set color $::GidPriv(Color,DisabledForegroundMenu)        
	foreach option_name $optionstochange {
	    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	    if { $position != -1 } {
		ChangeMenuForegroundColor $menuname $position $color
	    }
	}        
    } else {     
	set color "black"
	foreach option_name $optionstochange {
	    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	    if { $position != -1 } {
		ChangeMenuForegroundColor $menuname $position $color
	    }
	} 
    }

}

proc SetUnCollapseSubMenusState { menuname pos nlines nsurfs } {
     
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] ]
    
    set submenuname $menuname.$pos
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    if { $nsurfs } {
	set colors(1) "black" 
    } else {
	set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
    }
	     
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }
    
}

proc SetDataMenuState { } {
    
    if { [lindex [GiD_Info project] 0] != "UNKNOWN" } return

    set menu_name "Data#C#menu"    
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return
    
    set menuname [GetMenuName $i]
    
    if { ![winfo exists $menuname] } return
   
    if { [lindex [GiD_Info project] 0] == "UNKNOWN" } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
    } else {
	set color black
    }
  
    set optionstochange [list [_ "Conditions#C#menu"] [_ "Materials#C#menu"] [_ "Interval Data#C#menu"] [_ "Problem Data#C#menu"]]
    set ni [$menuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$menuname type $i] == {command} } {
	    if { [lsearch $optionstochange [$menuname entrycget $i -label]] != -1 } {
		ChangeMenuForegroundColor $menuname $i $color
	    }
	}
    }
    
    if { ![.central.s info unitssystems] || ![ModUnitWin::Init .gid.dataunits] } {
	set optionstochange [list [_ "Data units#C#menu"]]
	set ni [$menuname index end]
	if { $ni == "none" } { set ni -1 }
	for {set i 0} {$i <= $ni} {incr i} {
	    if { [$menuname type $i] == {command} } {
		if { [lsearch $optionstochange [$menuname entrycget $i -label]] != -1 } {
		    ChangeMenuForegroundColor $menuname $i $color
		}
	    }
	}
    }
}

proc SetMeshMenuState { } {
    set menu_name "Mesh#C#menu"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return
   
    set menuname [GetMenuName $i]
	
    set npoints [GiD_Info Geometry NumPoints]
    set nlines [GiD_Info Geometry NumLines]
    set nsurfs [GiD_Info Geometry NumSurfaces]
    set nvols [GiD_Info Geometry NumVolumes]
	
    if { !$npoints } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
	
	if { $::BOUNDARY_LAYER_MESH } {
	    set optionstochange [list [list "Unstructured#C#menu"] [list "Structured#C#menu"]\
		    [list "SemiStructured#C#menu"] [list "Cartesian#C#menu"] \
		    [list "Element type#C#menu"] [list "Reset mesh data#C#menu"] \
		    [list "Mesh criteria#C#menu"] [list "Draw#C#menu"] [list "Generate mesh#C#menu"] \
		    [list "Show errors#C#menu"] [list "Boundary layer#C#menu"]]
	} else {
	    set optionstochange [list [list "Unstructured#C#menu"] [list "Structured#C#menu"]\
		    [list "SemiStructured#C#menu"] [list "Cartesian#C#menu"] \
		    [list "Element type#C#menu"] [list "Reset mesh data#C#menu"] \
		    [list "Mesh criteria#C#menu"] [list "Draw#C#menu"] [list "Generate mesh#C#menu"] \
		    [list "Show errors#C#menu"] ]
	}
	
	foreach option_name $optionstochange {
	    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	    if { $position != -1 } {
		ChangeMenuForegroundColor $menuname $position $color
	    }
	}        
    } else {
	set option_name [list "Unstructured#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetUnstructuredSubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
	}
	
	set option_name [list "Structured#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetStructuredSubMenusState $menuname $position $nlines $nsurfs $nvols
	}
	
	set option_name [list "SemiStructured#C#menu"]
	if { !$nvols } {
	    set color $::GidPriv(Color,DisabledForegroundMenu)
	} else {
	    set color "black" 
	}
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}

	set option_name [list "Cartesian#C#menu"]       
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetCartesianSubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
	}
	
	if { $::BOUNDARY_LAYER_MESH } {
	    set option_name [list "Boundary layer#C#menu"]
	    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	    if { $position != -1 } {
		SetBoundaryLayerSubMenusState $menuname $position $nsurfs $nvols
	    }
	}

	set option_name [list "Element type#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetElementTypeSubMenusState $menuname $position $nlines $nsurfs $nvols
	}
	
	set color "black"
	set option_name [list "Reset mesh data#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Mesh criteria#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetMeshCriteriaSubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
	}
	
	set option_name [list "Draw#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    SetDrawSubMenusState $menuname $position $npoints $nlines $nsurfs $nvols
	}
	
	set color "black"
	set option_name [list "Generate mesh#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
    }
    
    set optionstochange [list [list "Erase mesh#C#menu"] [list "Edit mesh#C#menu"] \
	    [list "View mesh boundary#C#menu"] [list "Create boundary mesh#C#menu"] \
	    [list "Mesh quality#C#menu"]]
    if { [lindex [GiD_Info Mesh] 0] == 0 } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
    } else {
	set color "black"
    }
    foreach option_name $optionstochange {
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
    }  

    set option_name [list "Show errors#C#menu"]
    if { ![info exists ::MeshErrorsPriv] ||
    ($::MeshErrorsPriv(ErrSurf,NumErr) == 0 && $::MeshErrorsPriv(ErrVol,NumErr) == 0) } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
    } else {
	set color "black" 
    }
    set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
    if { $position != -1 } {
	ChangeMenuForegroundColor $menuname $position $color
    }
    
}

proc SetCartesianSubMenusState {  menuname pos npoints nlines nsurfs nvols } {
    
    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Points#C#menu"] [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
	
    if { $nvols } {
	set colors(3) "black"
	set colors(2) "black"  
	set colors(1) "black"  
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(2) "black"  
	    set colors(1) "black"   
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(1) "black"   
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	    }
	}
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } continue 
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }    
}

proc SetUnstructuredSubMenusState { menuname pos npoints nlines nsurfs nvols } {
      
    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
      
    set optionstochange [list [_ "Assign sizes on points#C#menu"] [_ "Assign sizes on lines#C#menu"] \
	    [_ "Assign sizes on surfaces#C#menu"] [_ "Assign sizes on volumes#C#menu"] \
	    [_ "Sizes by chordal error#C#menu"]... [_ "Correct sizes#C#menu"]... \
	    [_ "Sizes by background mesh#C#menu"]... [_ "Assign entities#C#menu"]]
    
    set optionstochange2 [list [_ "Points#C#menu"] [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"] ]

    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    if { $nvols } {
	set colors(3) "black" 
	set colors(2) "black" 
	set colors(1) "black" 
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(2) "black" 
	    set colors(1) "black" 
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(1) "black" 
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	    }
	}
    }
    set colors(4) "black"
    set colors(5) "black"
    set colors(6) "black"
       
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	if { $option < 7 } {
	    ChangeMenuForegroundColor $submenuname $i $colors($option)
	} else {
	    if { !$nlines && !$nsurfs && !$nvols } {
		ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu)
	    } else {
		$submenuname entryconfigure $i -foreground "black"
		set subsubmenuname "$submenuname.$i"
		set nj [$subsubmenuname index end]
		if { $nj == "none" } { set nj -1 }
		for {set j 0} {$j <= $nj} {incr j} {
		    if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		    set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		    if { $option == -1 } { continue }
		    ChangeMenuForegroundColor $subsubmenuname $j $colors($option)
		} 
	    }
	}
    }
    
}

proc SetBoundaryLayerSubMenusState {  menuname pos nsurfs nvols } {

  if { !$nsurfs } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }

    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "2 dimensional#C#menu"] [_ "3 dimensional#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    if { $nvols } {
	set colors(1) "black"  
    } else {
	set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
    }
	  
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }

}

proc SetStructuredSubMenusState {  menuname pos nlines nsurfs nvols } {
    
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"]]
    
    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    if { $nvols } {
	set colors(2) "black"
	set colors(1) "black"  
    } else {
	set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(1) "black" 
	} else {
	    set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	}
    }
	  
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	ChangeMenuForegroundColor $submenuname $i $colors($option)
    }
    
}

proc SetElementTypeSubMenusState { menuname pos nlines nsurfs nvols } {
    
    if { !$nlines } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Default#C#menu"] [_ "Linear#C#menu"] [_ "Triangle#C#menu"] \
	    [_ "Quadrilateral#C#menu"] [_ "Tetrahedra#C#menu"] [_ "Hexahedra#C#menu"] \
	    [_ "Prism#C#menu"] [_ "Only points#C#menu"] [_ "Sphere#C#menu"] [_ "Circle#C#menu"] ]
    set optionstochange2 [list [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"] ]
    
    set submenuname $menuname.$pos
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    if { $nvols } {
	set colors(2) "black"
	set colors(1) "black"  
    } else {
	set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(1) "black" 
	} else {
	    set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	}
    }
    
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	if { $option == 0} { 
	    #Default
	    $submenuname entryconfigure $i -foreground $colors(1)
	    set subsubmenuname "$submenuname.$i"
	    set nj [$subsubmenuname index end]
	    if { $nj == "none" } { set nj -1 }
	    for {set j 0} {$j <= $nj} {incr j} {
		if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		if { $option == -1 } { continue }
		ChangeMenuForegroundColor $subsubmenuname $j $colors($option)
	    } 
	} elseif { $option == 1} { 
	    #Linear
	    if { !$nsurfs } {
		ChangeMenuForegroundColor $submenuname $i $colors(1)
		continue
	    } 
	    $submenuname entryconfigure $i -foreground "black"
	    set subsubmenuname "$submenuname.$i"
	    set nj [$subsubmenuname index end]
	    if { $nj == "none" } { set nj -1 }
	    for {set j 0} {$j <= $nj} {incr j} {
		if { [$subsubmenuname type $i] != {command} && [$subsubmenuname type $i] != {cascade} } { continue }
		set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
		if { $option == -1 } { continue }
		ChangeMenuForegroundColor $subsubmenuname $j $colors($option)
	    } 
	} elseif { $option == 2 || $option == 3 || $option == 9} { 
	    #surface elements
	    ChangeMenuForegroundColor $submenuname $i $colors(1)
	} else { 
	    #volume elements
	    ChangeMenuForegroundColor $submenuname $i $colors(2)
	}
    }
}

proc SetMeshCriteriaSubMenusState { menuname pos npoints nlines nsurfs nvols } {

    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Default#C#menu"] [_ "Mesh#C#menu,verb"] [_ "No mesh#C#menu,verb"] [_ "Default mesh#C#menu,verb" ] \
	    [_ "Skip#C#menu"] [_ "No skip#C#menu"] [_ "Automatic skip#C#menu"] [_ "Skip by...#C#menu"] [_ "Force points to#C#menu"] \
	    [_ "Duplicate#C#menu"] [_ "No Duplicate#C#menu"] ]
    set optionstochange2 [list [_ "Points#C#menu"] [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"] \
	    [_ "Surface mesh#C#menu"] [_ "Volume mesh#C#menu"] ]

    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    set colors(0) "black"
    if { $nvols } {
	set colors(3) "black" 
	set colors(2) "black" 
	set colors(1) "black" 
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(2) "black" 
	    set colors(1) "black" 
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(1) "black" 
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
	    }
	}
    }
    set colors(4) $colors(2)
    set colors(5) $colors(3)

    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	if  { $option ==1 || $option ==2 || $option ==3 || $option ==9 || $option ==10 } {
	    if { !$nlines } {
		ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu)
		continue
	    }
	} 
	if  { $option ==7 } {#this is not cascade menu
	    ChangeMenuForegroundColor $submenuname $i $colors(2)
	    continue
	} 
	if  { $option ==8 } {
	    if { !$nsurfs } {
		ChangeMenuForegroundColor $submenuname $i $::GidPriv(Color,DisabledForegroundMenu)
		continue
	    }
	} 
	$submenuname entryconfigure $i -foreground "black"
	set subsubmenuname "$submenuname.$i"
	set nj [$subsubmenuname index end]
	if { $nj == "none" } { set nj -1 }
	for {set j 0} {$j <= $nj} {incr j} {
	    if { [$subsubmenuname type $j] != {command} && [$subsubmenuname type $j] != {cascade} } { continue }
	    set option [lsearch $optionstochange2 [$subsubmenuname entrycget $j -label]]
	    if { $option == -1 } { continue }
	    ChangeMenuForegroundColor $subsubmenuname $j $colors($option)
	} 
    }
    
}

proc SetDrawSubMenusState { menuname pos npoints nlines nsurfs nvols} {

    if { !$npoints } { 
	ChangeMenuForegroundColor $menuname $pos $::GidPriv(Color,DisabledForegroundMenu)
	return
    }
    
    $menuname entryconfigure $pos -foreground "black"
    
    set optionstochange [list [_ "Sizes#C#menu"] [_ "Element type#C#menu"] [_ "Skip entities (Rjump)#C#menu"] \
	[_ "Duplicate#C#menu"] [_ "Mesh / No mesh#C#menu,verb"] [_ "Structured type#C#menu"]]
    set optionstochange2 [list [_ "Lines#C#menu"] [_ "Surfaces#C#menu"] [_ "Volumes#C#menu"] [_ "All types#C#menu"]]

    set submenuname "$menuname.$pos"
    if { [$submenuname index end] == "none" } { return }
    
    
    if { $nvols } {
	set colors(3) "black" 
	set colors(2) "black" 
	set colors(1) "black" 
	set colors(0) "black"
    } else {
	set colors(3) $::GidPriv(Color,DisabledForegroundMenu)
	if { $nsurfs } {
	    set colors(2) "black" 
	    set colors(1) "black" 
	    set colors(0) "black" 
	} else {
	    set colors(2) $::GidPriv(Color,DisabledForegroundMenu)
	    if { $nlines } {
		set colors(1) "black" 
		set colors(0) "black" 
	    } else {
		set colors(1) $::GidPriv(Color,DisabledForegroundMenu)
		if { $npoints } {
		    set colors(0) "black" 
		} else {
		    set colors(0) $::GidPriv(Color,DisabledForegroundMenu) 
		}
	    }
	}
    }

    #set submenuname "$menuname.$pos"
    #if { [$submenuname index end] == "none" } { return }
	
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	if { [$submenuname type $i] != {command} && [$submenuname type $i] != {cascade} } { continue }
	set option [lsearch $optionstochange [$submenuname entrycget $i -label]]
	if { $option == -1 } { continue }
	if { $option == 0 } {
	    #option == 0 : Sizes submenu is treated in DrawSizeMenu            
	    $submenuname entryconfigure $i -foreground "black"
	} elseif { $option == 1 } {
	    #option == 1 : Element Type submenu is treated in DrawElementTypeMenu
	    $submenuname entryconfigure $i -foreground $colors(0)
	} elseif { $option == 2 } {
	    #option == 2 :  Skip entities (Rjump)
	    ChangeMenuForegroundColor $submenuname $i $colors(0)
	} elseif { $option == 3 } {
	    #option == 3 :  Duplicate
	    ChangeMenuForegroundColor $submenuname $i $colors(1)
	} elseif { $option == 4 } {
	    #option == 4 :  Mesh/NoMesh
	    ChangeMenuForegroundColor $submenuname $i $colors(0)
	} elseif { $option == 5 } {
	    #option == 5 :  Structured Type
	    ChangeMenuForegroundColor $submenuname $i $colors(0)
	}
    } 
}

proc SetCalculateMenuState { } {
    set menu_name "Calculate#C#menu"
    set i [GiDMenu::_FindIndex $menu_name PRE]
    if { $i == -1 } return

    set menuname [GetMenuName $i]
	
    set problemtype 1
    if { [lindex [GiD_Info project] 0] == "UNKNOWN" } {
	set problemtype 0
    }
    
    if { !$problemtype } {
	set color $::GidPriv(Color,DisabledForegroundMenu)
	set option_name [list "Cancel process#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}

	set option_name [list "View process info#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Calculate window#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Calculate#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Calculate remote#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
    } else {
	set option_name [list "Cancel process#C#menu"]
	if { ![info exists RunProcInfo] } { 
	    set color $::GidPriv(Color,DisabledForegroundMenu)
	} else {
	    set color "black"
	}
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "View process info#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    set filenames [PWFIndOutputFilenameCurrent]          
	    if { $filenames == "" } {
		set color  $::GidPriv(Color,DisabledForegroundMenu)
	    } else {
		set color "black"
	    }
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Calculate window#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    set color "black"
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Calculate#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    if { [lindex [.central.s info Mesh] 0] == 0 } {
		set color  $::GidPriv(Color,DisabledForegroundMenu)
	    } else {
		set color "black"
	    }
	    ChangeMenuForegroundColor $menuname $position $color
	}
	
	set option_name [list "Calculate remote#C#menu"]
	set position [lindex [split [GiDMenu::_FindOptionIndex $i $option_name PRE] ,] end]
	if { $position != -1 } {
	    if { [lindex [.central.s info Mesh] 0] == 0 } {
		set color  $::GidPriv(Color,DisabledForegroundMenu)
	    } else {
		set color "black"
	    }
	    ChangeMenuForegroundColor $menuname $position $color
	}
    }
    
}

proc ChangeSubMenusColor { menuname pos color } {
    
    if { [$menuname type $pos] != {cascade} } { return }
    
    set submenuname "$menuname.$pos"
	
    set ni [$submenuname index end]
    if { $ni == "none" } { set ni -1 }
    for {set i 0} {$i <= $ni} {incr i} {
	ChangeMenuForegroundColor $submenuname $i $color
    }    
}

proc ChangeMenuForegroundColor { menuname pos color } {
    
    if {  [$menuname type $pos] != {command}  &&  [$menuname type $pos] != {cascade}  } { 
	return
    }
    
    $menuname entryconfigure $pos -foreground $color
    
    if { [$menuname type $pos] == {cascade} } { 
	ChangeSubMenusColor $menuname $pos $color
    }
}

proc OpenNotes { } {
    package require texteditor
    set modelname [GiD_Info Project ModelName]
    if  { $modelname == "UNNAMED" } {
	#WarnWin [_ "A project title is needed. Save project to get it"]
	set filename ""
    } else {
	set name [file tail $modelname]
	set filename [file join $modelname.gid $name.txt]
    }
    set w .gid.texteditor
    if { $filename != "" } {
	if { [file exists $filename] } {
	    TextEditor::Create $filename "" "" 1 $w utf-8
	} else {
	    TextEditor::Create "" "" "" 1 $w utf-8
	    set TextEditor::currentfile $filename ;#to save with this name
	    wm title $w $filename
	}
    } else {
	TextEditor::Create "" "" "" 1 $w utf-8
	TextEditor::ReadFromVariable
	set TextEditor::cansavetovariable 1 ;#to temporary store in a tcl variable
    }
}

proc RecursiveChangeState {w {what "restore"}} {
    if { ![winfo exists $w] } return
    foreach i [winfo children $w] {
	RecursiveChangeState $i $what
    }
    catch {
	if {$what eq "restore"} {
	    $w conf -state $::Gid_ChangeState($w)
	} else {
	    set ::Gid_ChangeState($w) [$w cget -state]
	    $w conf -state $what
	}
    }
}

proc FinishButton { w win text endaction disable {small 0} } {
  if { $small } {
    set w2 .gid.finish
    toplevel $w2
    wm transient $w2 .gid
    if { $::tcl_platform(platform) == "windows" } {
      AddAlwaysOnTopFlag $w2
    }
    wm title $w2 [wm title $w]
    scan [wm geometry $w] "%dx%d+%d+%d" width height x y
    if { ![regexp {[+]} $x] } { set x +$x}
    if { ![regexp {[+]} $y] } { set y +$y}
    wm geometry $w2 $x$y
    wm protocol $w2 WM_DELETE_WINDOW [list EndFinishButton $w $endaction $small]
    wm minsize $w2 50 25
    
    frame $w2.finf -relief ridge -bd 2
    button $w2.finf.f -text [_ "Finish"] -width 6 -command [list EndFinishButton $w $endaction $small]
    label $w2.finf.l -text $text
    grid $w2.finf.f $w2.finf.l -sticky ew -padx 2
    #grid configure $w2.finf.f -sticky w
    grid columnconf $w2.finf 1 -weight 1
    grid $w2.finf -sticky nsew -columnspan 2
    grid columnconf $w2 0 -weight 1
    grid rowconfig $w2 0 -weight 1
    wm withdraw $w
  } else {
    set w2 $w
    frame $w2.finf -relief ridge -bd 2
    button $w2.finf.f -text [_ "Finish"] -width 6 -command [list EndFinishButton $w $endaction $small]
    label $w2.finf.l -text $text
    pack $w.finf.f $w.finf.l -side left
    place $w2.finf -in $win -x 0 -rely 1 -anchor sw -bordermode outside
    if { $disable == "disableall" } {
      RecursiveChangeState $w disabled
    }
    $w2.finf.f conf -state normal
  }
  focus $w2.finf.f
  bind $w2 <Escape> [list EndFinishButton $w $endaction $small]
  bind $w2 <Return> [list EndFinishButton $w $endaction $small]
  delayedop changefunc [list EndFinishButton $w $endaction $small]
}

proc EndFinishButton { w endaction small } {
  global GidPriv
  
  set found [delayedop -cancel changefunc "$w.finf.f invoke"]
  .central.s process escape escape
  
  if { $small } {
    set w2 .gid.finish
    if { [winfo exists $w2] } { 
      scan [wm geometry $w2] "%dx%d+%d+%d" width height x y
      if { ![regexp {[+]} $x] } { set x +$x}
      if { ![regexp {[+]} $y] } { set y +$y}
      destroy $w2
      wm geometry $w $x$y
    }
    wm deiconify $w
  } else {
    destroy $w.finf
    RecursiveChangeState $w
    array unset ::Gid_ChangeState
  }  
  if { $endaction != "" } { eval $endaction }
}


#to centralize this name, because change bethween platforms, and also if menu is outside
#i index 0 ... n
proc GetMenuName { i } {
    if { $::GidPriv(MenuType) == "Win" } {
	if { [winfo toplevel .gid.menu] == ".gid" } {
	    set menuname .mm1.button$i
	} else {
	    set menuname .gid.menu.mm1.button$i
	}
    } else {
	set menuname .gid.menu.button$i.m.m
    }
    return $menuname
}

proc WatchBitmaps {name index op} {
    if {$index eq "PrePostBitmapsWindowGeom"}  {
	WarnWinText $::GidPriv($index)
    }
}

#define this functions here instead Layers.tcl because can be used before source Layers.tcl
proc CreateLayerCombo { w } {
    global GidPriv

    entry $w.e -background white
    #-disabledbackground white -state disabled
    button $w.b -image [GetImage arrow_down_vsmall.gif] -command "open_layers_as_menu .gid $w"
    grid $w.e $w.b -sticky ew
    grid configure $w.b -sticky e
    grid columnconfigure $w.b 0 -weight 1

    bind $w.e <ButtonPress-1> "[list $w.b invoke]; break"
    bind $w.e <ButtonRelease-1> "break"
    bind $w.e <KeyPress> "break"
    bind $w.e <Key-Down> [list $w.b invoke]

    set GidPriv(ComboLayers,entry) $w.e
    $w.e insert end [GiD_Info Project LayerToUse]
    return
}

proc open_layers_as_menu { parent open_over } {
    global GidPriv
   
    set w $parent.m
    if { [winfo exists $w] } {
	destroy $w
	return
    }
    set w [toplevel $parent.m -bd 1 -relief raised]
    
    wm withdraw $w
    wm overrideredirect $w 1
    set orient "horizontal"
    if { $orient == "vertical" } {
	set x [expr {[winfo rootx $open_over]+[winfo width $open_over]}]
	set y [expr {[winfo rooty $open_over]}]
    } else {
	set x [expr {[winfo rootx $open_over]}]
	set y [expr {[winfo rooty $open_over]+[winfo height $open_over]}]
    }
    set width 250
    #set height 210
    set nrows [llength [GiD_Info Layers]]
    if { $nrows < 1 } {
	set nrows 1
    } elseif { $nrows > 20 } {
	set nrows 20
    }
    set height [expr {$nrows*18+22}]
    wm geometry $w ${width}x${height}+${x}+$y
    
    
    frame $w.gg
    set tbl [LayersCreateListWin $w.gg]
    set GidPriv(ComboLayers,table) $tbl
    FillTableInfoLayers $tbl
    focus $tbl
    SetLayerAsSelection $tbl [GiD_Info Project LayerToUse]
    bind [$tbl bodytag] <Double-Button-1> "+destroy $w"
    #bind [$tbl bodytag] <Return> "DoubleClickTableList %x %y %W $tbl ; destroy $w"
    bind $w <Return> "SetLayersTo TOUSE $tbl ; destroy $w"
    
    bind $w <Destroy> [list +DestroyComboLayers %W $w]

    grid $w.gg - -sticky nsew -pady 2
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1
    
    tkTabToWindow $w.gg
    grab $w.gg
    bind $w <Escape> [list destroy $w]
    bind $w <1> [list _layers_destroy_if_outside $w %X %Y]
    
    update
    if { [winfo exists $w] } {
	wm deiconify $w 
    }
}

proc DestroyComboLayers { W w } {
    if { $W == $w } {
	unset -nocomplain ::GidPriv(ComboLayers,table)
    }
}

proc _layers_destroy_if_outside { w x y } {    
    if { ![winfo exists $w] } { return }
    if { $x < [winfo rootx $w] || $x > [winfo rootx $w]+[winfo width $w] || \
	$y < [winfo rooty $w] || $y > [winfo rooty $w]+[winfo height $w] } {
	open_layers_as_menu [winfo parent $w] [winfo parent $w]
    }
}

GidTkBackCompatibility ;#repeat againg, else "tkGenerateMenuSelect" dissapear!!

# trace add variable ::GidPriv write WatchBitmaps